Package: maxentcpp
Type: Package
Title: Maximum Entropy Species Distribution Modeling ('C++'
        Implementation)
Version: 1.0.0
Authors@R: c(
    person("Angel", "Robles", role = c("aut", "cre"),
           email = "a.l.robles.fernandez@gmail.com"))
Description: 'C++' implementation of Maximum Entropy (Maxent) species distribution
    modeling with R bindings via 'Rcpp'. Provides a high-performance
    reimplementation of the Maxent algorithm for modeling species geographic
    distributions from occurrence data and environmental variables, following
    Phillips et al. (2006) <doi:10.1016/j.ecolmodel.2005.03.026>. Supports
    linear, quadratic, product, hinge, and threshold feature transformations,
    spatial projection in raw, logistic, and cloglog scales, and model
    diagnostics including Area Under the ROC Curve (AUC), variable importance,
    response curves, and Multivariate Environmental Similarity Surfaces (MESS)
    maps.
URL: https://github.com/alrobles/maxentcpp,
        https://alrobles.github.io/maxentcpp/
BugReports: https://github.com/alrobles/maxentcpp/issues
License: MIT + file LICENSE
Depends: R (>= 3.5.0)
Imports: Rcpp (>= 1.0.0), methods
LinkingTo: Rcpp, RcppEigen
Suggests: testthat (>= 3.0.0), knitr, rmarkdown, terra (>= 1.7-46),
        raster, sf, png, rJava, maxnet
SystemRequirements: C++17
Encoding: UTF-8
LazyData: true
Config/testthat/edition: 3
Config/roxygen2/version: 8.0.0
RoxygenNote: 7.3.3
NeedsCompilation: yes
Packaged: 2026-05-13 03:58:32 UTC; alrob
Author: Angel Robles [aut, cre]
Maintainer: Angel Robles <a.l.robles.fernandez@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-18 18:30:11 UTC
Built: R 4.6.0; aarch64-apple-darwin23; 2026-05-18 19:10:49 UTC; unix
Archs: maxentcpp.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method(print,maxent_sample)
export(clamp_grids)
export(compute_mess)
export(compute_mess_range)
export(compute_percent_contribution)
export(compute_permutation_importance)
export(compute_response_curve)
export(compute_response_curve_fixed)
export(compute_variable_ranges)
export(coords_to_rowcol)
export(create_grid_dimension)
export(create_grid_float)
export(create_hinge_feature)
export(create_layer)
export(create_linear_feature)
export(create_product_feature)
export(create_quadratic_feature)
export(create_sample)
export(create_threshold_feature)
export(csv_close)
export(csv_headers)
export(csv_next_record)
export(csv_open)
export(csv_read_double_column)
export(csv_writer_close)
export(csv_writer_open)
export(csv_writer_print)
export(csv_writer_print_double)
export(csv_writer_println)
export(eval_auc)
export(eval_correlation)
export(eval_logloss)
export(eval_misclassification)
export(eval_model)
export(eval_square_error)
export(extract_predictions_raw)
export(feature_eval)
export(feature_get_info)
export(generate_features)
export(get_grid_dimension_info)
export(get_layer_info)
export(get_sample_info)
export(grid_float_from_matrix)
export(grid_float_info)
export(grid_float_to_matrix)
export(grid_from_matrix)
export(grid_get_value)
export(grid_get_values_batch)
export(grid_has_data)
export(grid_read_asc)
export(grid_read_file)
export(grid_set_value)
export(grid_to_matrix)
export(grid_write_asc)
export(layer_name_from_path)
export(maxent_append_results_csv)
export(maxent_auc)
export(maxent_background_indices)
export(maxent_clamp)
export(maxent_color_ramp)
export(maxent_correlation)
export(maxent_csv_close)
export(maxent_csv_headers)
export(maxent_csv_next)
export(maxent_csv_open)
export(maxent_csv_read_column)
export(maxent_csv_write)
export(maxent_csv_write_close)
export(maxent_csv_write_num)
export(maxent_csv_write_open)
export(maxent_csv_write_row)
export(maxent_dimension)
export(maxent_evaluate)
export(maxent_extract_lambdas)
export(maxent_extract_occurrence_env_terra)
export(maxent_extract_predictions_cloglog)
export(maxent_extract_predictions_logistic)
export(maxent_extract_predictions_raw)
export(maxent_feature_eval)
export(maxent_feature_info)
export(maxent_featured_space)
export(maxent_featured_space_create)
export(maxent_featured_space_from_callback)
export(maxent_featured_space_from_rast)
export(maxent_featured_space_info)
export(maxent_fit)
export(maxent_generate_features)
export(maxent_get_entropy)
export(maxent_get_loss)
export(maxent_get_weights)
export(maxent_grid)
export(maxent_grid_from_matrix)
export(maxent_grid_from_terra)
export(maxent_grid_info)
export(maxent_grid_to_matrix)
export(maxent_grid_to_terra)
export(maxent_hinge_feature)
export(maxent_layer)
export(maxent_layer_info)
export(maxent_layer_name)
export(maxent_linear_feature)
export(maxent_load_lambdas)
export(maxent_logloss)
export(maxent_mess)
export(maxent_mess_range)
export(maxent_misclassification)
export(maxent_model_entropy)
export(maxent_model_loss)
export(maxent_model_weights)
export(maxent_percent_contribution)
export(maxent_permutation_importance)
export(maxent_plot_response_curves)
export(maxent_plot_variable_importance)
export(maxent_predict)
export(maxent_predict_model)
export(maxent_print_results)
export(maxent_product_feature)
export(maxent_project_cloglog)
export(maxent_project_logistic)
export(maxent_project_raw)
export(maxent_quadratic_feature)
export(maxent_raster_sample_indices)
export(maxent_read_asc)
export(maxent_read_grid)
export(maxent_read_lambdas)
export(maxent_read_occurrences)
export(maxent_response_curve)
export(maxent_response_curve_fixed)
export(maxent_run)
export(maxent_sample)
export(maxent_save_lambdas)
export(maxent_sequential_fit)
export(maxent_set_sample_expectations)
export(maxent_space_info)
export(maxent_square_error)
export(maxent_threshold_feature)
export(maxent_train)
export(maxent_train_terra)
export(maxent_variable_ranges)
export(maxent_write_asc)
export(maxent_write_lambdas)
export(maxent_write_omission_csv)
export(maxent_write_prediction_png)
export(maxent_write_sample_predictions)
export(project_cloglog)
export(project_logistic)
export(project_raw)
export(sample_get_feature)
export(sample_set_feature)
importFrom(Rcpp,sourceCpp)
importFrom(methods,slot)
importFrom(methods,slotNames)
useDynLib(maxentcpp, .registration = TRUE)

## maxentcpp Quick-Start Example
##
## Complete end-to-end species distribution modelling workflow using bundled
## example data (bio1 and bio12 cropped layers, Abeillia abeillei occurrence
## records from GBIF).
##
## Run from the package root:
##   Rscript inst/examples/quickstart.R
## or, after installing the package:
##   source(system.file("examples", "quickstart.R", package = "maxentcpp"))

library(maxentcpp)
library(terra)

## ---- 1. Load environmental layers ------------------------------------------

stack_path      <- system.file("extdata", "stack_1_12_crop.rds",
                               package = "maxentcpp")
example_rasters <- terra::unwrap(readRDS(stack_path))
bio1            <- example_rasters[[1]]
bio12           <- example_rasters[[2]]

# Convert to maxentcpp grids
g_bio1  <- maxent_grid_from_terra(bio1)
g_bio12 <- maxent_grid_from_terra(bio12)

cat("Grid info (bio1):\n")
print(maxent_grid_info(g_bio1))

## ---- 2. Prepare occurrence data --------------------------------------------

data(example_occ_df)        # columns: species, long, lat
occ_df <- example_occ_df

# Build a GridDimension that matches the environmental layers
info <- maxent_grid_info(g_bio1)
dim  <- maxent_dimension(nrows    = info$nrows,
                         ncols    = info$ncols,
                         xll      = info$xll,
                         yll      = info$yll,
                         cellsize = info$cellsize)

# Convert lon/lat to grid indices (note column names: "long" / "lat")
occ <- maxent_read_occurrences(occ_df, dim,
                               lon_col = "long",
                               lat_col = "lat")

cat("Number of presence points:", length(occ$indices), "\n")

## ---- 3. Background points --------------------------------------------------

bg <- maxent_background_indices(g_bio1, n = 10000, seed = 42)

cat("Number of background points:", length(bg$indices), "\n")

## ---- 4. Extract environmental values and build features --------------------

all_rows <- c(bg$rows, occ$rows)
all_cols <- c(bg$cols, occ$cols)
n_total  <- length(all_rows)

# Presence samples are appended at the end -> 0-based indices
sample_indices <- seq(length(bg$rows), n_total - 1L)

# grid_get_values_batch is an internal Rcpp helper (not in the public
# maxent_* surface).  Use maxentcpp::: to reach it from this example
# script.
bio1_vals  <- maxentcpp:::grid_get_values_batch(g_bio1,  all_rows, all_cols)
bio12_vals <- maxentcpp:::grid_get_values_batch(g_bio12, all_rows, all_cols)

env_data <- list(bio1 = bio1_vals, bio12 = bio12_vals)

features <- maxent_generate_features(env_data,
                                     types    = c("linear", "quadratic", "hinge"),
                                     n_hinges = 15)
cat("Generated", length(features), "features\n")

## ---- 5. Train the model ----------------------------------------------------

fs     <- maxent_featured_space(n_total, as.integer(sample_indices), features)
result <- maxent_fit(fs,
                     max_iter        = 500,
                     convergence     = 1e-5,
                     beta_multiplier = 1.0)

cat("Converged:", result$converged, "\n")
cat("Final loss:", result$loss,      "\n")
cat("Entropy:",    result$entropy,   "\n")
cat("Iterations:", result$iterations, "\n")

## Save lambdas to a temporary file
lambdas_file <- file.path(tempdir(), "quickstart_model.lambdas")
maxent_save_lambdas(fs, lambdas_file)
cat("Model saved to:", lambdas_file, "\n")

## ---- 6. Project the model --------------------------------------------------

pred_grid   <- maxent_project_cloglog(fs,
                                      list(g_bio1, g_bio12),
                                      c("bio1", "bio12"))
pred_raster <- maxent_grid_to_terra(pred_grid)

cat("Prediction raster extent:\n")
print(terra::ext(pred_raster))

# Save prediction map
out_png <- file.path(tempdir(), "quickstart_prediction.png")
png(out_png, width = 800, height = 600)
terra::plot(pred_raster,
            main = "Predicted Habitat Suitability (cloglog)",
            col  = hcl.colors(50, "YlOrRd", rev = TRUE))
dev.off()
cat("Prediction map saved to:", out_png, "\n")

## ---- 7. Evaluate the model -------------------------------------------------

pres_preds <- maxent_extract_predictions_raw(
    fs, list(g_bio1, g_bio12), c("bio1", "bio12"),
    occ$rows, occ$cols)

bg_preds <- maxent_extract_predictions_raw(
    fs, list(g_bio1, g_bio12), c("bio1", "bio12"),
    bg$rows, bg$cols)

eval_result <- maxent_evaluate(pres_preds, bg_preds)
cat("AUC:",       eval_result$auc,       "\n")
cat("Max Kappa:", eval_result$max_kappa, "\n")
cat("Log-loss:",  eval_result$logloss,   "\n")

## ---- 8. Variable importance ------------------------------------------------

contrib <- maxent_percent_contribution(fs, c("bio1", "bio12"))
cat("\nPercent contribution:\n")
print(contrib)

perm_imp <- maxent_permutation_importance(
    fs, list(g_bio1, g_bio12), c("bio1", "bio12"),
    occ$rows, occ$cols,
    bg$rows,  bg$cols,
    seed = 42)
cat("\nPermutation importance:\n")
print(perm_imp)

## ---- 9. Response curves ----------------------------------------------------

curve_bio1 <- maxent_response_curve(fs,
                                    list(g_bio1, g_bio12),
                                    c("bio1", "bio12"),
                                    var_index = 0L,
                                    n_steps   = 100L)

curve_bio12 <- maxent_response_curve(fs,
                                     list(g_bio1, g_bio12),
                                     c("bio1", "bio12"),
                                     var_index = 1L,
                                     n_steps   = 100L)

# Save response-curve plots
curves_png <- file.path(tempdir(), "quickstart_response_curves.png")
png(curves_png, width = 1000, height = 500)
par(mfrow = c(1, 2))
plot(curve_bio1$value,  curve_bio1$prediction,
     type = "l", lwd = 2,
     xlab = "Annual Mean Temperature (bio1)",
     ylab = "Cloglog Prediction",
     main = "Response Curve: bio1")
plot(curve_bio12$value, curve_bio12$prediction,
     type = "l", lwd = 2,
     xlab = "Annual Precipitation (bio12)",
     ylab = "Cloglog Prediction",
     main = "Response Curve: bio12")
dev.off()
cat("Response curves saved to:", curves_png, "\n")

## ---- 10. MESS analysis -----------------------------------------------------

ref_vals <- list(
    bio1_vals[seq_len(length(bg$rows))],
    bio12_vals[seq_len(length(bg$rows))]
)

mess_result <- maxent_mess(list(g_bio1, g_bio12),
                           ref_vals,
                           c("bio1", "bio12"))
mess_raster <- maxent_grid_to_terra(mess_result$mess_grid)

mess_png <- file.path(tempdir(), "quickstart_mess.png")
png(mess_png, width = 800, height = 600)
terra::plot(mess_raster,
            main = "MESS: Negative = Novel Environment",
            col  = hcl.colors(50, "RdYlGn"))
dev.off()
cat("MESS map saved to:", mess_png, "\n")

## ---- 11. Clamping ----------------------------------------------------------

ranges  <- maxent_variable_ranges(list(g_bio1, g_bio12))
clamped <- maxent_clamp(list(g_bio1, g_bio12),
                        ranges$min, ranges$max)

pred_clamped <- maxent_project_cloglog(fs,
                                       clamped$clamped_grids,
                                       c("bio1", "bio12"))

cat("\nClamped prediction range:",
    range(terra::values(maxent_grid_to_terra(pred_clamped)), na.rm = TRUE),
    "\n")

cat("\nQuick-start workflow complete.\n")

#' Create a FeaturedSpace Object
#'
#' Constructs a MaxEnt FeaturedSpace from background point data, occurrence
#' sample indices, and a list of pre-built Feature objects.
#'
#' @param num_points     Integer: total number of background points.
#' @param sample_indices Integer vector: 0-based indices (into the background
#'   array) of the occurrence sample locations.
#' @param features       List of external pointers to Feature objects, as
#'   returned by \code{maxent_linear_feature()}, \code{maxent_hinge_feature()},
#'   \code{maxent_generate_features()}, etc.
#' @param bias_weights   Optional numeric vector of per-point bias weights
#'   (length \code{num_points}).  When supplied, background density is computed
#'   as \code{bias[i] * exp(lp[i] - lpn)} instead of the standard
#'   \code{exp(lp[i] - lpn)}.  This mirrors Java Maxent's \code{biasFile}
#'   parameter.  Pass \code{NULL} (default) for uniform (unbiased) background.
#' @return External pointer to a FeaturedSpace C++ object.
#' @export
#' @examples
#' n   <- 100L
#' idx <- 90:99  # 0-based sample indices
#' env <- seq(0, 1, length.out = n)
#' f   <- maxent_linear_feature(env, "env1")
#' fs  <- maxent_featured_space(n, idx, list(f))
maxent_featured_space <- function(num_points, sample_indices, features,
                                  bias_weights = NULL) {
    if (!is.list(features)) {
        stop("features must be a list of Feature external pointers")
    }
    bw <- if (is.null(bias_weights)) numeric(0) else as.numeric(bias_weights)
    maxent_featured_space_create(
        as.integer(num_points),
        as.integer(sample_indices),
        features,
        bw
    )
}

#' Train a MaxEnt Model
#'
#' Runs the sequential coordinate-ascent MaxEnt optimization on a FeaturedSpace.
#'
#' @param featured_space  External pointer to a FeaturedSpace object (from
#'   \code{maxent_featured_space()}).
#' @param max_iter        Maximum number of training iterations (default 500).
#' @param convergence     Convergence threshold: stop when the per-20-iteration
#'   loss improvement is below this value (default 1e-5).
#' @param beta_multiplier Regularization multiplier (default 1.0). Higher
#'   values increase regularization strength.
#' @param min_deviation   Minimum sample deviation floor used in regularization
#'   (default 0.001).
#' @return Named list with:
#'   \describe{
#'     \item{loss}{Final regularized loss (scalar).}
#'     \item{entropy}{Shannon entropy of the trained distribution.}
#'     \item{iterations}{Number of training iterations completed.}
#'     \item{converged}{Logical: whether the convergence threshold was reached.}
#'     \item{lambdas}{Numeric vector of final lambda (weight) values.}
#'   }
#' @export
#' @examples
#' n   <- 50L
#' idx <- c(5L, 15L, 25L, 35L, 45L)
#' env <- list(bio1 = runif(n), bio12 = runif(n))
#' feats <- maxent_generate_features(env, types = "linear")
#' fs  <- maxent_featured_space(n, idx, feats)
#' result <- maxent_fit(fs, max_iter = 100, convergence = 1e-5)
#' result$loss
maxent_fit <- function(featured_space,
                       max_iter        = 500L,
                       convergence     = 1e-5,
                       beta_multiplier = 1.0,
                       min_deviation   = 0.001) {
    maxent_train(
        featured_space,
        as.integer(max_iter),
        as.double(convergence),
        as.double(beta_multiplier),
        as.double(min_deviation)
    )
}

#' Predict with a Trained MaxEnt Model
#'
#' Computes raw Gibbs distribution scores for new environmental data, using
#' the feature lambdas stored in the trained FeaturedSpace.
#'
#' @param featured_space External pointer to a trained FeaturedSpace object.
#' @param newdata        Numeric matrix: one row per new point, one column per
#'   feature.  Column values must be the \emph{already-evaluated} feature
#'   values (e.g., from running \code{maxent_feature_eval()} for each feature
#'   and each point).
#' @return Numeric vector of raw (unnormalized) prediction scores.
#' @export
#' @examples
#' \donttest{
#' # After training, predict on 5 new points with 2 features each
#' newdata <- matrix(runif(10), nrow = 5, ncol = 2)
#' preds <- maxent_predict_model(fs, newdata)
#' }
maxent_predict_model <- function(featured_space, newdata) {
    newdata <- as.matrix(newdata)
    maxent_predict(featured_space, newdata)
}

#' Get Model Distribution Weights
#'
#' Returns the normalized probability weights of the current MaxEnt
#' distribution over the background points.
#'
#' @param featured_space External pointer to a FeaturedSpace object.
#' @return Numeric vector of weights that sum to 1.
#' @export
maxent_model_weights <- function(featured_space) {
    maxent_get_weights(featured_space)
}

#' Get Model Entropy
#'
#' Returns the Shannon entropy of the current MaxEnt distribution.
#'
#' @param featured_space External pointer to a FeaturedSpace object.
#' @return Scalar: Shannon entropy (non-negative).
#' @export
maxent_model_entropy <- function(featured_space) {
    maxent_get_entropy(featured_space)
}

#' Get Model Loss
#'
#' Returns the current (negative log-likelihood) loss of the model.
#'
#' @param featured_space External pointer to a FeaturedSpace object.
#' @return Scalar: current loss value.
#' @export
maxent_model_loss <- function(featured_space) {
    maxent_get_loss(featured_space)
}

#' Save Model Lambdas to File
#'
#' Writes the trained model coefficients (lambdas) to a CSV file in
#' the standard Maxent .lambdas format.
#'
#' @param featured_space External pointer to a trained FeaturedSpace object.
#' @param file           Character: path to the output file.
#' @return Invisibly returns the output file path.
#' @export
#' @examples
#' \donttest{
#' maxent_save_lambdas(fs, tempfile(fileext = ".lambdas"))
#' }
maxent_save_lambdas <- function(featured_space, file) {
    maxent_write_lambdas(featured_space, as.character(file))
    invisible(file)
}

#' Load Model Lambdas from File
#'
#' Reads model coefficients from a .lambdas file and applies them to
#' a FeaturedSpace that was created with the same features.
#'
#' @param featured_space External pointer to a FeaturedSpace object (must have
#'   features with the same names as those in the file).
#' @param file           Character: path to the lambdas file.
#' @return Invisibly returns the FeaturedSpace external pointer (updated
#'   in-place).
#' @export
#' @examples
#' \donttest{
#' maxent_load_lambdas(fs, tempfile(fileext = ".lambdas"))
#' }
maxent_load_lambdas <- function(featured_space, file) {
    maxent_read_lambdas(featured_space, as.character(file))
    invisible(featured_space)
}

#' Get FeaturedSpace Information
#'
#' Returns basic metadata about a FeaturedSpace object.
#'
#' @param featured_space External pointer to a FeaturedSpace object.
#' @return Named list with: \code{num_points}, \code{num_samples},
#'   \code{num_features}, \code{density_normalizer},
#'   \code{linear_predictor_normalizer}.
#' @export
maxent_space_info <- function(featured_space) {
    maxent_featured_space_info(featured_space)
}

#' Extract Lambda Values from a Maxent Run Result
#'
#' Retrieves the trained lambda (weight) coefficients from the output of
#' \code{\link{maxent_run}}, returning them as a named numeric vector.
#' Each element is the lambda value for the corresponding feature, and the
#' name of each element is the feature name (e.g. \code{"bio1"},
#' \code{"bio1^2"}, \code{"bio1'1"}).
#'
#' The function reads lambdas directly from the trained model object, so
#' it also works correctly after reloading coefficients with
#' \code{\link{maxent_load_lambdas}}.
#'
#' @param run_result A named list as returned by \code{\link{maxent_run}}.
#'   Must contain an element \code{model} holding the trained FeaturedSpace
#'   external pointer.
#' @return A named numeric vector of lambda values, one per feature.  Names
#'   are the feature names as stored in the model.
#' @export
#' @examples
#' \donttest{
#' stack_path      <- system.file("extdata", "stack_1_12_crop.rds",
#'                                package = "maxentcpp")
#' example_rasters <- terra::unwrap(readRDS(stack_path))
#' grids <- list(
#'   bio1  = maxent_grid_from_terra(example_rasters[[1]]),
#'   bio12 = maxent_grid_from_terra(example_rasters[[2]])
#' )
#' data(example_occ_df)
#'
#' result <- maxent_run(
#'   species    = "Abeillia_abeillei",
#'   env_grids  = grids,
#'   occ_df     = example_occ_df,
#'   output_dir = tempdir(),
#'   lon_col    = "long",
#'   lat_col    = "lat")
#'
#' lambdas <- maxent_extract_lambdas(result)
#' print(lambdas)
#' }
maxent_extract_lambdas <- function(run_result) {
    if (!is.list(run_result) || !"model" %in% names(run_result)) {
        stop("'run_result' must be the list returned by maxent_run()")
    }

    feature_ptrs <- attr(run_result$model, "feature_ptrs")
    if (is.null(feature_ptrs)) {
        stop("'run_result$model' has no 'feature_ptrs' attribute; ",
             "was it created by maxent_run() or maxent_featured_space()?")
    }

    n          <- length(feature_ptrs)
    feat_names <- character(n)
    lambdas    <- numeric(n)

    for (i in seq_len(n)) {
        info          <- feature_get_info(feature_ptrs[[i]])
        feat_names[i] <- info$name
        lambdas[i]    <- info$lambda
    }

    names(lambdas) <- feat_names
    lambdas
}

test_that("FeaturedSpace can be created from features", {
    skip_if_not_installed("maxentcpp")

    n   <- 100L
    env <- seq(0, 1, length.out = n)
    f   <- maxent_linear_feature(env, "env1")

    # Sample indices (0-based): last 10 points
    idx <- as.integer(90:99)

    fs <- maxent_featured_space(n, idx, list(f))
    expect_true(!is.null(fs))
})

test_that("FeaturedSpace retains feature pointers after garbage collection", {
    skip_if_not_installed("maxentcpp")

    n <- 50L
    idx <- as.integer(40:49)

    fs <- local({
        env <- seq(0, 1, length.out = n)
        maxent_featured_space(n, idx, list(maxent_linear_feature(env, "env1")))
    })

    invisible(gc())

    w <- maxent_model_weights(fs)
    expect_equal(length(w), n)
    expect_equal(sum(w), 1.0, tolerance = 1e-12)
})

test_that("Initial distribution is uniform (all lambdas = 0)", {
    skip_if_not_installed("maxentcpp")

    n   <- 100L
    env <- seq(0, 1, length.out = n)
    f   <- maxent_linear_feature(env, "env1")
    idx <- as.integer(90:99)

    fs <- maxent_featured_space(n, idx, list(f))
    w  <- maxent_model_weights(fs)

    expect_equal(length(w), n)
    expect_equal(sum(w), 1.0, tolerance = 1e-12)
    # Each weight should be 1/n for uniform distribution
    expect_equal(w[1], 1.0 / n, tolerance = 1e-9)
    expect_equal(w[50], 1.0 / n, tolerance = 1e-9)
})

test_that("Model weights sum to ~1.0", {
    skip_if_not_installed("maxentcpp")

    n   <- 50L
    env1 <- runif(n)
    env2 <- runif(n)
    f1   <- maxent_linear_feature(env1, "v1")
    f2   <- maxent_linear_feature(env2, "v2")
    idx  <- as.integer(40:49)

    fs <- maxent_featured_space(n, idx, list(f1, f2))

    # Even before training, weights should sum to 1
    w <- maxent_model_weights(fs)
    expect_equal(sum(w), 1.0, tolerance = 1e-12)
})

test_that("Entropy is non-negative and bounded by log(n)", {
    skip_if_not_installed("maxentcpp")

    n   <- 100L
    env <- seq(0, 1, length.out = n)
    f   <- maxent_linear_feature(env, "env1")
    idx <- as.integer(90:99)

    fs <- maxent_featured_space(n, idx, list(f))
    H  <- maxent_model_entropy(fs)

    expect_true(H >= 0)
    expect_true(H <= log(n) + 1e-9)

    # Uniform distribution has max entropy = log(n)
    expect_equal(H, log(n), tolerance = 1e-9)
})

test_that("Training reduces loss", {
    skip_if_not_installed("maxentcpp")

    set.seed(42)
    n    <- 100L
    env  <- seq(0, 1, length.out = n)
    f    <- maxent_linear_feature(env, "env1")
    idx  <- as.integer(90:99)

    # Record initial loss (all lambdas = 0)
    fs0 <- maxent_featured_space(n, idx, list(f))
    # Set sample expectations to get meaningful loss
    maxent_set_sample_expectations(fs0)
    initial_loss <- maxent_model_loss(fs0)

    # Train a fresh model
    f2  <- maxent_linear_feature(env, "env1")
    fs  <- maxent_featured_space(n, idx, list(f2))
    res <- maxent_fit(fs, max_iter = 300L, convergence = 1e-5)

    expect_true(res$loss <= initial_loss + 1e-9)
    expect_true(res$iterations > 0L)
    expect_equal(length(res$lambdas), 1L)
    expect_true(res$entropy >= 0)
    expect_true(is.logical(res$converged))
})

test_that("Loss is finite after training", {
    skip_if_not_installed("maxentcpp")

    n   <- 80L
    env <- seq(0, 1, length.out = n)
    f   <- maxent_linear_feature(env, "env1")
    idx <- as.integer(70:79)

    fs  <- maxent_featured_space(n, idx, list(f))
    res <- maxent_fit(fs, max_iter = 100L)

    expect_true(is.finite(res$loss))
    expect_true(is.finite(res$entropy))
})

test_that("Prediction returns a numeric vector", {
    skip_if_not_installed("maxentcpp")

    n   <- 50L
    env <- seq(0, 1, length.out = n)
    f   <- maxent_linear_feature(env, "env1")
    idx <- as.integer(40:49)

    fs <- maxent_featured_space(n, idx, list(f))
    maxent_fit(fs, max_iter = 100L)

    # Predict on 5 new points (feature values already evaluated)
    newdata <- matrix(seq(0.1, 0.9, length.out = 5), ncol = 1)
    preds   <- maxent_predict_model(fs, newdata)

    expect_equal(length(preds), 5L)
    expect_true(all(is.finite(preds)))
    expect_true(all(preds >= 0))
})

test_that("Lambda save/load round-trip preserves values", {
    skip_if_not_installed("maxentcpp")

    n    <- 80L
    env1 <- seq(0, 1, length.out = n)
    env2 <- rev(env1)
    f1   <- maxent_linear_feature(env1, "v1")
    f2   <- maxent_linear_feature(env2, "v2")
    idx  <- as.integer(70:79)

    # Train
    fs   <- maxent_featured_space(n, idx, list(f1, f2))
    res1 <- maxent_fit(fs, max_iter = 200L)

    # Save
    tmpfile <- tempfile(fileext = ".lambdas")
    on.exit(unlink(tmpfile))
    maxent_save_lambdas(fs, tmpfile)

    expect_true(file.exists(tmpfile))

    # Load into a fresh space
    f1b <- maxent_linear_feature(env1, "v1")
    f2b <- maxent_linear_feature(env2, "v2")
    fs2 <- maxent_featured_space(n, idx, list(f1b, f2b))
    maxent_load_lambdas(fs2, tmpfile)

    # Weights should match
    w1 <- maxent_model_weights(fs)
    w2 <- maxent_model_weights(fs2)
    expect_equal(w1, w2, tolerance = 1e-6)

    # Density normalizer should match
    info1 <- maxent_space_info(fs)
    info2 <- maxent_space_info(fs2)
    expect_equal(info1$density_normalizer, info2$density_normalizer,
                 tolerance = 1e-6)
})

test_that("maxent_space_info returns correct structure", {
    skip_if_not_installed("maxentcpp")

    n   <- 60L
    env <- seq(0, 1, length.out = n)
    f   <- maxent_linear_feature(env, "env1")
    idx <- as.integer(50:59)

    fs   <- maxent_featured_space(n, idx, list(f))
    info <- maxent_space_info(fs)

    expect_equal(info$num_points,   n)
    expect_equal(info$num_samples,  10L)
    expect_equal(info$num_features, 1L)
    expect_true(is.numeric(info$density_normalizer))
    expect_true(is.numeric(info$linear_predictor_normalizer))
})

test_that("Multi-feature training works", {
    skip_if_not_installed("maxentcpp")

    set.seed(123)
    n    <- 100L
    env1 <- seq(0, 1, length.out = n)
    env2 <- rev(env1)
    idx  <- as.integer(80:99)

    f1 <- maxent_linear_feature(env1, "v1")
    f2 <- maxent_linear_feature(env2, "v2")
    fs <- maxent_featured_space(n, idx, list(f1, f2))

    res <- maxent_fit(fs, max_iter = 300L)

    expect_equal(length(res$lambdas), 2L)
    expect_true(is.finite(res$loss))

    w <- maxent_model_weights(fs)
    expect_equal(sum(w), 1.0, tolerance = 1e-12)
})

test_that("maxent_extract_lambdas returns a named numeric vector", {
    skip_if_not_installed("maxentcpp")

    set.seed(42)
    n    <- 100L
    env1 <- seq(0, 1, length.out = n)
    env2 <- rev(env1)
    idx  <- as.integer(80:99)

    f1 <- maxent_linear_feature(env1, "bio1")
    f2 <- maxent_linear_feature(env2, "bio12")
    fs <- maxent_featured_space(n, idx, list(f1, f2))

    fit  <- maxent_fit(fs, max_iter = 300L)

    # Wrap into a minimal run_result list (same structure as maxent_run())
    run_result <- list(model = fs, fit_result = fit)

    lambdas <- maxent_extract_lambdas(run_result)

    expect_true(is.numeric(lambdas))
    expect_equal(length(lambdas), 2L)
    expect_equal(names(lambdas), c("bio1", "bio12"))
    expect_equal(unname(lambdas), fit$lambdas)
})

test_that("maxent_extract_lambdas errors on invalid input", {
    skip_if_not_installed("maxentcpp")

    expect_error(maxent_extract_lambdas(list()), "'run_result' must be")
    expect_error(maxent_extract_lambdas(42),     "'run_result' must be")
})

# =========================================================================
# Bias weights tests
# =========================================================================

test_that("Uniform bias weights match no-bias result", {
    skip_if_not_installed("maxentcpp")

    n   <- 100L
    env <- seq(0, 1, length.out = n)
    idx <- as.integer(90:99)

    f1 <- maxent_linear_feature(env, "env1")
    fs_no_bias <- maxent_featured_space(n, idx, list(f1))

    f2 <- maxent_linear_feature(env, "env1")
    bias <- rep(1.0, n)
    fs_biased <- maxent_featured_space(n, idx, list(f2), bias_weights = bias)

    w1 <- maxent_model_weights(fs_no_bias)
    w2 <- maxent_model_weights(fs_biased)
    expect_equal(w1, w2, tolerance = 1e-12)

    info1 <- maxent_space_info(fs_no_bias)
    info2 <- maxent_space_info(fs_biased)
    expect_false(info1$has_bias)
    expect_true(info2$has_bias)
})

test_that("Bias weights change density distribution", {
    skip_if_not_installed("maxentcpp")

    n   <- 100L
    env <- seq(0, 1, length.out = n)
    idx <- as.integer(90:99)

    # Heavy bias on first half
    bias <- rep(1.0, n)
    bias[1:50] <- 10.0

    f1 <- maxent_linear_feature(env, "env1")
    fs_no_bias <- maxent_featured_space(n, idx, list(f1))

    f2 <- maxent_linear_feature(env, "env1")
    fs_biased <- maxent_featured_space(n, idx, list(f2), bias_weights = bias)

    w_no <- maxent_model_weights(fs_no_bias)
    w_bi <- maxent_model_weights(fs_biased)

    # First half should have higher weight with bias
    expect_true(sum(w_bi[1:50]) > sum(w_no[1:50]))
    # Weights still sum to 1
    expect_equal(sum(w_bi), 1.0, tolerance = 1e-12)
})

test_that("Bias weights affect training results", {
    skip_if_not_installed("maxentcpp")

    n   <- 100L
    env <- seq(0, 1, length.out = n)
    idx <- as.integer(90:99)

    bias <- seq(1.0, 10.0, length.out = n)

    f1 <- maxent_linear_feature(env, "env1")
    fs_no_bias <- maxent_featured_space(n, idx, list(f1))
    res1 <- maxent_fit(fs_no_bias, max_iter = 200L)

    f2 <- maxent_linear_feature(env, "env1")
    fs_biased <- maxent_featured_space(n, idx, list(f2), bias_weights = bias)
    res2 <- maxent_fit(fs_biased, max_iter = 200L)

    # Lambdas should differ
    expect_false(isTRUE(all.equal(res1$lambdas, res2$lambdas, tolerance = 1e-6)))

    # Both should produce valid results
    expect_true(is.finite(res1$loss))
    expect_true(is.finite(res2$loss))
})

test_that("NULL bias_weights produces unbiased model", {
    skip_if_not_installed("maxentcpp")

    n   <- 50L
    env <- seq(0, 1, length.out = n)
    idx <- as.integer(40:49)

    f <- maxent_linear_feature(env, "env1")
    fs <- maxent_featured_space(n, idx, list(f), bias_weights = NULL)

    info <- maxent_space_info(fs)
    expect_false(info$has_bias)
    w <- maxent_model_weights(fs)
    expect_equal(w[1], 1.0 / n, tolerance = 1e-9)
})

## Helpers reuse the same minimal model builder as test-maxent-outputs.R
.make_plot_test_model <- function() {
    skip_if_not_installed("maxentcpp")
    dim_obj <- maxent_dimension(nrows = 4, ncols = 4, xll = 0, yll = 0,
                                cellsize = 1)
    g0 <- maxent_grid(dim_obj, "env0")
    g1 <- maxent_grid(dim_obj, "env1")
    idx <- 0L
    for (r in 0:3) {
        for (cc in 0:3) {
            grid_set_value(g0, r, cc, idx / 15.0)
            grid_set_value(g1, r, cc, 1.0 - idx / 15.0)
            idx <- idx + 1L
        }
    }
    vals0 <- sapply(0:15, function(i) i / 15.0)
    vals1 <- 1.0 - vals0
    features <- maxent_generate_features(
        list(env0 = vals0, env1 = vals1),
        types = "linear", n_thresholds = 0, n_hinges = 0)
    pres_idx <- 12:15
    model    <- maxent_featured_space(16L, as.integer(pres_idx), features)
    maxent_fit(model, max_iter = 200, convergence = 1e-4,
               beta_multiplier = 1.0)
    pres_rows <- as.integer(pres_idx %/% 4L)
    pres_cols <- as.integer(pres_idx  %%  4L)
    list(model = model, g0 = g0, g1 = g1,
         pres_rows = pres_rows, pres_cols = pres_cols)
}

## ---- Response Curve PNGs ----------------------------------------------------

test_that("maxent_plot_response_curves creates full and thumbnail PNGs", {
    skip_if_not_installed("maxentcpp")
    m       <- .make_plot_test_model()
    out_dir <- file.path(tempdir(), "test_rc_plots")
    on.exit(unlink(out_dir, recursive = TRUE), add = TRUE)

    file_list <- maxent_plot_response_curves(
        m$model, list(m$g0, m$g1), c("env0", "env1"),
        output_dir = out_dir,
        species    = "TestSp",
        thumbnail  = TRUE)

    expect_true(length(file_list) >= 4L)  # 2 full + 2 thumbs

    for (p in unlist(file_list)) {
        expect_true(file.exists(p))
        expect_gt(file.size(p), 0L)
    }
})

test_that("maxent_plot_response_curves thumbnail dimensions are ~210x140", {
    skip_if_not_installed("maxentcpp")
    skip_if_not_installed("png")

    m       <- .make_plot_test_model()
    out_dir <- file.path(tempdir(), "test_rc_thumb_dims")
    on.exit(unlink(out_dir, recursive = TRUE), add = TRUE)

    file_list <- maxent_plot_response_curves(
        m$model, list(m$g0, m$g1), c("env0", "env1"),
        output_dir = out_dir,
        species    = "ThumbSp",
        thumbnail  = TRUE)

    thumb_files <- file_list[grepl("_thumb$", names(file_list))]
    if (length(thumb_files) > 0) {
        img <- png::readPNG(thumb_files[[1]])
        # dim is [height, width, channels]
        expect_equal(dim(img)[1], 140L)
        expect_equal(dim(img)[2], 210L)
    }
})

test_that("maxent_plot_response_curves write_dat creates .dat files", {
    skip_if_not_installed("maxentcpp")
    m       <- .make_plot_test_model()
    out_dir <- file.path(tempdir(), "test_rc_dat")
    on.exit(unlink(out_dir, recursive = TRUE), add = TRUE)

    file_list <- maxent_plot_response_curves(
        m$model, list(m$g0, m$g1), c("env0", "env1"),
        output_dir = out_dir,
        species    = "DatSp",
        thumbnail  = FALSE,
        write_dat  = TRUE)

    dat_files <- file_list[grepl("_dat$", names(file_list))]
    expect_true(length(dat_files) > 0)
    for (p in unlist(dat_files)) {
        expect_true(file.exists(p))
    }
})

test_that("maxent_plot_response_curves var_indices subset works", {
    skip_if_not_installed("maxentcpp")
    m       <- .make_plot_test_model()
    out_dir <- file.path(tempdir(), "test_rc_subset")
    on.exit(unlink(out_dir, recursive = TRUE), add = TRUE)

    file_list <- maxent_plot_response_curves(
        m$model, list(m$g0, m$g1), c("env0", "env1"),
        output_dir  = out_dir,
        species     = "SubSp",
        var_indices = 0L,   # only env0
        thumbnail   = FALSE)

    full_files <- file_list[grepl("_full$", names(file_list))]
    expect_equal(length(full_files), 1L)
})

## ---- Variable Importance Plot -----------------------------------------------

test_that("maxent_plot_variable_importance creates a valid PNG", {
    skip_if_not_installed("maxentcpp")
    m       <- .make_plot_test_model()
    contrib <- maxent_percent_contribution(m$model, c("env0", "env1"))
    perm_ip <- maxent_permutation_importance(
        m$model, list(m$g0, m$g1), c("env0", "env1"),
        m$pres_rows, m$pres_cols,
        c(0L, 1L, 2L), c(0L, 1L, 2L))

    out_dir <- file.path(tempdir(), "test_varimp")
    on.exit(unlink(out_dir, recursive = TRUE), add = TRUE)

    png_path <- maxent_plot_variable_importance(contrib, perm_ip,
                                                species    = "VISp",
                                                output_dir = out_dir)
    expect_true(file.exists(png_path))
    expect_gt(file.size(png_path), 0L)
})

# Phase E: Extend Fidelity to All Feature Types
#
# This file verifies end-to-end fidelity (training, prediction, clamping,
# and lambda round-trip) for every feature type supported by maxentcpp:
#   linear, quadratic, product, threshold, hinge (forward), reverse_hinge.
#
# The tests in prior phases (A-D) established fidelity for linear,
# linear+quadratic, and linear+quadratic+product features via trajectory
# comparison against Java golden values.  This file extends that coverage
# to threshold and hinge features and verifies key invariants hold for
# every feature type under end-to-end training and projection.
#
# Groups:
#   J — Per-type training convergence & prediction invariants
#   K — Feature clamping during projection (eval_from_env path)
#   L — Lambda save/load round-trip for multi-type models
#   M — Combined all-feature-type model fidelity

# ===========================================================================
# Helper: build a model with the given feature types on the asym fixture
# ===========================================================================

.build_asym_model <- function(types = "linear",
                              n_thresholds = 10L,
                              n_hinges = 10L,
                              max_iter = 500L,
                              beta_multiplier = 1.0) {
    d <- asym_fixture()

    feat <- maxent_generate_features(
        list(bio1 = d$bio1, bio2 = d$bio2),
        types        = types,
        n_thresholds = as.integer(n_thresholds),
        n_hinges     = as.integer(n_hinges)
    )
    fs <- maxent_featured_space(d$n, d$sample_idx, feat)
    res <- maxent_fit(fs,
                      max_iter        = as.integer(max_iter),
                      convergence     = 1e-5,
                      beta_multiplier = beta_multiplier)
    list(model = fs, fit = res, features = feat, data = d)
}

# ===========================================================================
# Group J — Per-type training convergence & prediction invariants
# ===========================================================================

# --- J.1  Quadratic-only training -------------------------------------------

test_that("Group J: quadratic-only model trains and converges", {
    skip_if_not_installed("maxentcpp")

    m <- .build_asym_model(types = c("linear", "quadratic"),
                           n_thresholds = 0L, n_hinges = 0L)

    expect_true(m$fit$iterations > 0L)
    expect_true(is.finite(m$fit$loss))
    expect_true(is.finite(m$fit$entropy))
    expect_true(m$fit$entropy >= 0)
    expect_equal(length(m$fit$lambdas), length(m$features))
})

test_that("Group J: quadratic-only model predictions are finite and non-negative", {
    skip_if_not_installed("maxentcpp")

    m  <- .build_asym_model(types = c("linear", "quadratic"),
                            n_thresholds = 0L, n_hinges = 0L)
    nf <- length(m$features)
    n  <- m$data$n

    fmat <- matrix(0.0, nrow = n, ncol = nf)
    for (j in seq_len(nf))
        for (i in seq_len(n))
            fmat[i, j] <- maxent_feature_eval(m$features[[j]], i)

    preds <- maxent_predict_model(m$model, fmat)

    expect_equal(length(preds), n)
    expect_true(all(is.finite(preds)))
    expect_true(all(preds >= 0))
})

# --- J.2  Threshold-only training -------------------------------------------

test_that("Group J: threshold model trains and converges", {
    skip_if_not_installed("maxentcpp")

    m <- .build_asym_model(types = c("linear", "threshold"),
                           n_thresholds = 5L, n_hinges = 0L)

    expect_true(m$fit$iterations > 0L)
    expect_true(is.finite(m$fit$loss))
    expect_true(is.finite(m$fit$entropy))
    expect_true(m$fit$entropy >= 0)
    expect_equal(length(m$fit$lambdas), length(m$features))
})

test_that("Group J: threshold model predictions are finite and non-negative", {
    skip_if_not_installed("maxentcpp")

    m  <- .build_asym_model(types = c("linear", "threshold"),
                            n_thresholds = 5L, n_hinges = 0L)
    nf <- length(m$features)
    n  <- m$data$n

    fmat <- matrix(0.0, nrow = n, ncol = nf)
    for (j in seq_len(nf))
        for (i in seq_len(n))
            fmat[i, j] <- maxent_feature_eval(m$features[[j]], i)

    preds <- maxent_predict_model(m$model, fmat)

    expect_equal(length(preds), n)
    expect_true(all(is.finite(preds)))
    expect_true(all(preds >= 0))
})

# --- J.3  Hinge-only training -----------------------------------------------

test_that("Group J: hinge model trains and converges", {
    skip_if_not_installed("maxentcpp")

    m <- .build_asym_model(types = c("linear", "hinge"),
                           n_thresholds = 0L, n_hinges = 5L)

    expect_true(m$fit$iterations > 0L)
    expect_true(is.finite(m$fit$loss))
    expect_true(is.finite(m$fit$entropy))
    expect_true(m$fit$entropy >= 0)
    expect_equal(length(m$fit$lambdas), length(m$features))

    types <- sapply(m$features, function(f) maxent_feature_info(f)$type)
    expect_true("hinge"         %in% types)
    expect_true("reverse_hinge" %in% types)
})

test_that("Group J: hinge model predictions are finite and non-negative", {
    skip_if_not_installed("maxentcpp")

    m  <- .build_asym_model(types = c("linear", "hinge"),
                            n_thresholds = 0L, n_hinges = 5L)
    nf <- length(m$features)
    n  <- m$data$n

    fmat <- matrix(0.0, nrow = n, ncol = nf)
    for (j in seq_len(nf))
        for (i in seq_len(n))
            fmat[i, j] <- maxent_feature_eval(m$features[[j]], i)

    preds <- maxent_predict_model(m$model, fmat)

    expect_equal(length(preds), n)
    expect_true(all(is.finite(preds)))
    expect_true(all(preds >= 0))
})

# --- J.4  Product-only training ---------------------------------------------

test_that("Group J: product model trains and converges", {
    skip_if_not_installed("maxentcpp")

    m <- .build_asym_model(types = c("linear", "product"),
                           n_thresholds = 0L, n_hinges = 0L)

    expect_true(m$fit$iterations > 0L)
    expect_true(is.finite(m$fit$loss))
    expect_true(is.finite(m$fit$entropy))
    expect_true(m$fit$entropy >= 0)

    types <- sapply(m$features, function(f) maxent_feature_info(f)$type)
    expect_true("product" %in% types)
})

test_that("Group J: product model predictions are finite and non-negative", {
    skip_if_not_installed("maxentcpp")

    m  <- .build_asym_model(types = c("linear", "product"),
                            n_thresholds = 0L, n_hinges = 0L)
    nf <- length(m$features)
    n  <- m$data$n

    fmat <- matrix(0.0, nrow = n, ncol = nf)
    for (j in seq_len(nf))
        for (i in seq_len(n))
            fmat[i, j] <- maxent_feature_eval(m$features[[j]], i)

    preds <- maxent_predict_model(m$model, fmat)

    expect_equal(length(preds), n)
    expect_true(all(is.finite(preds)))
    expect_true(all(preds >= 0))
})

# ===========================================================================
# Group K — Feature clamping during projection (eval_from_env path)
# ===========================================================================

# --- K.1  Quadratic feature clamping ----------------------------------------

test_that("Group K: quadratic model clamping — out-of-range matches boundary", {
    skip_if_not_installed("maxentcpp")

    m <- .build_asym_model(types = c("linear", "quadratic"),
                           n_thresholds = 0L, n_hinges = 0L)

    dim_1 <- maxent_dimension(nrows = 1L, ncols = 1L,
                              xll = 0, yll = 0, cellsize = 1)

    # Grid at boundary minimum (both vars = 0, the pre-scaled minimum)
    g_min1 <- maxent_grid(dim_1, "bio1")
    g_min2 <- maxent_grid(dim_1, "bio2")
    grid_set_value(g_min1, 0L, 0L, 0.0)
    grid_set_value(g_min2, 0L, 0L, 0.0)

    # Grid far below training range
    g_low1 <- maxent_grid(dim_1, "bio1")
    g_low2 <- maxent_grid(dim_1, "bio2")
    grid_set_value(g_low1, 0L, 0L, -100.0)
    grid_set_value(g_low2, 0L, 0L, -100.0)

    pred_min <- maxent_grid_to_matrix(
        maxent_project_raw(m$model, list(g_min1, g_min2), c("bio1", "bio2")))[1, 1]
    pred_low <- maxent_grid_to_matrix(
        maxent_project_raw(m$model, list(g_low1, g_low2), c("bio1", "bio2")))[1, 1]

    expect_equal(pred_low, pred_min, tolerance = 1e-9,
                 label = "below-range quadratic prediction equals min boundary")
})

test_that("Group K: quadratic model clamping — above-range matches max boundary", {
    skip_if_not_installed("maxentcpp")

    m <- .build_asym_model(types = c("linear", "quadratic"),
                           n_thresholds = 0L, n_hinges = 0L)

    dim_1 <- maxent_dimension(nrows = 1L, ncols = 1L,
                              xll = 0, yll = 0, cellsize = 1)

    # Grid at boundary maximum (both vars = 1.0, the pre-scaled maximum)
    g_max1 <- maxent_grid(dim_1, "bio1")
    g_max2 <- maxent_grid(dim_1, "bio2")
    grid_set_value(g_max1, 0L, 0L, 1.0)
    grid_set_value(g_max2, 0L, 0L, 1.0)

    # Grid far above training range
    g_hi1 <- maxent_grid(dim_1, "bio1")
    g_hi2 <- maxent_grid(dim_1, "bio2")
    grid_set_value(g_hi1, 0L, 0L, 100.0)
    grid_set_value(g_hi2, 0L, 0L, 100.0)

    pred_max <- maxent_grid_to_matrix(
        maxent_project_raw(m$model, list(g_max1, g_max2), c("bio1", "bio2")))[1, 1]
    pred_hi  <- maxent_grid_to_matrix(
        maxent_project_raw(m$model, list(g_hi1, g_hi2), c("bio1", "bio2")))[1, 1]

    expect_equal(pred_hi, pred_max, tolerance = 1e-9,
                 label = "above-range quadratic prediction equals max boundary")
})

# --- K.2  Product feature clamping ------------------------------------------

test_that("Group K: product model clamping — out-of-range matches boundary", {
    skip_if_not_installed("maxentcpp")

    m <- .build_asym_model(types = c("linear", "product"),
                           n_thresholds = 0L, n_hinges = 0L)

    dim_1 <- maxent_dimension(nrows = 1L, ncols = 1L,
                              xll = 0, yll = 0, cellsize = 1)

    # Boundary minimum
    g_min1 <- maxent_grid(dim_1, "bio1")
    g_min2 <- maxent_grid(dim_1, "bio2")
    grid_set_value(g_min1, 0L, 0L, 0.0)
    grid_set_value(g_min2, 0L, 0L, 0.0)

    # Far below range
    g_low1 <- maxent_grid(dim_1, "bio1")
    g_low2 <- maxent_grid(dim_1, "bio2")
    grid_set_value(g_low1, 0L, 0L, -100.0)
    grid_set_value(g_low2, 0L, 0L, -100.0)

    pred_min <- maxent_grid_to_matrix(
        maxent_project_raw(m$model, list(g_min1, g_min2), c("bio1", "bio2")))[1, 1]
    pred_low <- maxent_grid_to_matrix(
        maxent_project_raw(m$model, list(g_low1, g_low2), c("bio1", "bio2")))[1, 1]

    expect_equal(pred_low, pred_min, tolerance = 1e-9,
                 label = "below-range product prediction equals min boundary")
})

# --- K.3  Hinge feature clamping --------------------------------------------

test_that("Group K: hinge model clamping — out-of-range matches boundary", {
    skip_if_not_installed("maxentcpp")

    m <- .build_asym_model(types = c("linear", "hinge"),
                           n_thresholds = 0L, n_hinges = 3L)

    dim_1 <- maxent_dimension(nrows = 1L, ncols = 1L,
                              xll = 0, yll = 0, cellsize = 1)

    # Boundary minimum
    g_min1 <- maxent_grid(dim_1, "bio1")
    g_min2 <- maxent_grid(dim_1, "bio2")
    grid_set_value(g_min1, 0L, 0L, 0.0)
    grid_set_value(g_min2, 0L, 0L, 0.0)

    # Far below range
    g_low1 <- maxent_grid(dim_1, "bio1")
    g_low2 <- maxent_grid(dim_1, "bio2")
    grid_set_value(g_low1, 0L, 0L, -100.0)
    grid_set_value(g_low2, 0L, 0L, -100.0)

    pred_min <- maxent_grid_to_matrix(
        maxent_project_raw(m$model, list(g_min1, g_min2), c("bio1", "bio2")))[1, 1]
    pred_low <- maxent_grid_to_matrix(
        maxent_project_raw(m$model, list(g_low1, g_low2), c("bio1", "bio2")))[1, 1]

    expect_equal(pred_low, pred_min, tolerance = 1e-9,
                 label = "below-range hinge prediction equals min boundary")
})

test_that("Group K: hinge model clamping — above-range matches max boundary", {
    skip_if_not_installed("maxentcpp")

    m <- .build_asym_model(types = c("linear", "hinge"),
                           n_thresholds = 0L, n_hinges = 3L)

    dim_1 <- maxent_dimension(nrows = 1L, ncols = 1L,
                              xll = 0, yll = 0, cellsize = 1)

    # Boundary maximum
    g_max1 <- maxent_grid(dim_1, "bio1")
    g_max2 <- maxent_grid(dim_1, "bio2")
    grid_set_value(g_max1, 0L, 0L, 1.0)
    grid_set_value(g_max2, 0L, 0L, 1.0)

    # Far above range
    g_hi1 <- maxent_grid(dim_1, "bio1")
    g_hi2 <- maxent_grid(dim_1, "bio2")
    grid_set_value(g_hi1, 0L, 0L, 100.0)
    grid_set_value(g_hi2, 0L, 0L, 100.0)

    pred_max <- maxent_grid_to_matrix(
        maxent_project_raw(m$model, list(g_max1, g_max2), c("bio1", "bio2")))[1, 1]
    pred_hi  <- maxent_grid_to_matrix(
        maxent_project_raw(m$model, list(g_hi1, g_hi2), c("bio1", "bio2")))[1, 1]

    expect_equal(pred_hi, pred_max, tolerance = 1e-9,
                 label = "above-range hinge prediction equals max boundary")
})

# --- K.4  Threshold feature clamping ----------------------------------------

test_that("Group K: threshold model clamping — extreme values stay in [0,1] cloglog", {
    skip_if_not_installed("maxentcpp")

    m <- .build_asym_model(types = c("linear", "threshold"),
                           n_thresholds = 5L, n_hinges = 0L)

    dim_1 <- maxent_dimension(nrows = 1L, ncols = 2L,
                              xll = 0, yll = 0, cellsize = 1)

    g1 <- maxent_grid(dim_1, "bio1")
    g2 <- maxent_grid(dim_1, "bio2")
    grid_set_value(g1, 0L, 0L, -1e6)
    grid_set_value(g1, 0L, 1L,  1e6)
    grid_set_value(g2, 0L, 0L, -1e6)
    grid_set_value(g2, 0L, 1L,  1e6)

    cloglog_grid <- maxent_project_cloglog(m$model, list(g1, g2), c("bio1", "bio2"))
    cloglog_mat  <- maxent_grid_to_matrix(cloglog_grid)

    expect_true(all(cloglog_mat >= 0.0 & cloglog_mat <= 1.0))
})

# --- K.5  All-type model cloglog bounds -------------------------------------

test_that("Group K: all-type model cloglog stays in [0,1] for extreme env", {
    skip_if_not_installed("maxentcpp")

    m <- .build_asym_model(
        types = c("linear", "quadratic", "product", "threshold", "hinge"),
        n_thresholds = 3L, n_hinges = 3L)

    dim_1 <- maxent_dimension(nrows = 1L, ncols = 2L,
                              xll = 0, yll = 0, cellsize = 1)

    g1 <- maxent_grid(dim_1, "bio1")
    g2 <- maxent_grid(dim_1, "bio2")
    grid_set_value(g1, 0L, 0L, -1e6)
    grid_set_value(g1, 0L, 1L,  1e6)
    grid_set_value(g2, 0L, 0L, -1e6)
    grid_set_value(g2, 0L, 1L,  1e6)

    cloglog_grid <- maxent_project_cloglog(m$model, list(g1, g2), c("bio1", "bio2"))
    cloglog_mat  <- maxent_grid_to_matrix(cloglog_grid)

    expect_true(all(cloglog_mat >= 0.0 & cloglog_mat <= 1.0))
})

# ===========================================================================
# Group L — Lambda save/load round-trip for multi-type models
# ===========================================================================

test_that("Group L: L+Q lambda round-trip preserves lambda values", {
    skip_if_not_installed("maxentcpp")

    m <- .build_asym_model(types = c("linear", "quadratic"),
                           n_thresholds = 0L, n_hinges = 0L)
    d <- m$data

    lambda_file <- tempfile(fileext = ".csv")
    on.exit(unlink(lambda_file), add = TRUE)
    maxent_save_lambdas(m$model, lambda_file)

    feat2 <- maxent_generate_features(
        list(bio1 = d$bio1, bio2 = d$bio2),
        types        = c("linear", "quadratic"),
        n_thresholds = 0L, n_hinges = 0L
    )
    fs2 <- maxent_featured_space(d$n, d$sample_idx, feat2)
    maxent_load_lambdas(fs2, lambda_file)

    # Lambda values should be exactly preserved
    info1 <- maxent_space_info(m$model)
    info2 <- maxent_space_info(fs2)
    expect_equal(info1$num_features, info2$num_features)

    # Model weights should sum to ~1
    w <- maxent_model_weights(fs2)
    expect_equal(sum(w), 1.0, tolerance = 1e-12)

    # Entropy should be restored
    H2 <- maxent_model_entropy(fs2)
    expect_true(is.finite(H2))
    expect_true(H2 >= 0)
})

test_that("Group L: L+Q+P lambda round-trip preserves lambda values", {
    skip_if_not_installed("maxentcpp")

    m <- .build_asym_model(types = c("linear", "quadratic", "product"),
                           n_thresholds = 0L, n_hinges = 0L)
    d <- m$data

    lambda_file <- tempfile(fileext = ".csv")
    on.exit(unlink(lambda_file), add = TRUE)
    maxent_save_lambdas(m$model, lambda_file)

    feat2 <- maxent_generate_features(
        list(bio1 = d$bio1, bio2 = d$bio2),
        types        = c("linear", "quadratic", "product"),
        n_thresholds = 0L, n_hinges = 0L
    )
    fs2 <- maxent_featured_space(d$n, d$sample_idx, feat2)
    maxent_load_lambdas(fs2, lambda_file)

    # Lambda values should be preserved
    info1 <- maxent_space_info(m$model)
    info2 <- maxent_space_info(fs2)
    expect_equal(info1$num_features, info2$num_features)

    # Model weights should sum to ~1
    w <- maxent_model_weights(fs2)
    expect_equal(sum(w), 1.0, tolerance = 1e-12)

    # Entropy should be restored
    H2 <- maxent_model_entropy(fs2)
    expect_true(is.finite(H2))
    expect_true(H2 >= 0)
})

test_that("Group L: L+T+H lambda round-trip preserves entropy and normalizers", {
    skip_if_not_installed("maxentcpp")

    m <- .build_asym_model(types = c("linear", "threshold", "hinge"),
                           n_thresholds = 3L, n_hinges = 3L)
    d <- m$data

    H_trained   <- maxent_model_entropy(m$model)
    info_trained <- maxent_space_info(m$model)

    lambda_file <- tempfile(fileext = ".csv")
    on.exit(unlink(lambda_file), add = TRUE)
    maxent_save_lambdas(m$model, lambda_file)

    feat2 <- maxent_generate_features(
        list(bio1 = d$bio1, bio2 = d$bio2),
        types        = c("linear", "threshold", "hinge"),
        n_thresholds = 3L, n_hinges = 3L
    )
    fs2 <- maxent_featured_space(d$n, d$sample_idx, feat2)
    maxent_load_lambdas(fs2, lambda_file)

    H2    <- maxent_model_entropy(fs2)
    info2 <- maxent_space_info(fs2)

    expect_equal(H2, H_trained, tolerance = 1e-9)
    # Note: density_normalizer and linear_predictor_normalizer may differ after
    # lambda round-trip because increase_lambda() only increases lpn (never
    # decreases), while read_lambdas() recomputes it as the actual max.
    # We verify that entropy is preserved (above) and that normalizers are
    # at least finite and positive.
    expect_true(is.finite(info2$density_normalizer))
    expect_true(info2$density_normalizer > 0)
    expect_true(is.finite(info2$linear_predictor_normalizer))
    expect_true(info2$linear_predictor_normalizer > 0)
})

test_that("Group L: all-type lambda round-trip preserves lambda values and validity", {
    skip_if_not_installed("maxentcpp")

    m <- .build_asym_model(
        types = c("linear", "quadratic", "product", "threshold", "hinge"),
        n_thresholds = 3L, n_hinges = 3L)
    d <- m$data

    lambda_file <- tempfile(fileext = ".csv")
    on.exit(unlink(lambda_file), add = TRUE)
    maxent_save_lambdas(m$model, lambda_file)

    feat2 <- maxent_generate_features(
        list(bio1 = d$bio1, bio2 = d$bio2),
        types = c("linear", "quadratic", "product", "threshold", "hinge"),
        n_thresholds = 3L, n_hinges = 3L
    )
    fs2 <- maxent_featured_space(d$n, d$sample_idx, feat2)
    maxent_load_lambdas(fs2, lambda_file)

    # Lambda values should be preserved
    info1 <- maxent_space_info(m$model)
    info2 <- maxent_space_info(fs2)
    expect_equal(info1$num_features, info2$num_features)

    # Model weights should sum to ~1
    w <- maxent_model_weights(fs2)
    expect_equal(sum(w), 1.0, tolerance = 1e-12)

    # Entropy should be restored and valid
    H2 <- maxent_model_entropy(fs2)
    expect_true(is.finite(H2))
    expect_true(H2 >= 0)

    # Predictions should be valid (finite, in bounds for cloglog)
    dim_1 <- maxent_dimension(nrows = 1L, ncols = 3L,
                              xll = 0, yll = 0, cellsize = 0.33)
    g1 <- maxent_grid(dim_1, "bio1")
    g2 <- maxent_grid(dim_1, "bio2")
    for (cc in 0:2) {
        grid_set_value(g1, 0L, as.integer(cc), cc * 0.5)
        grid_set_value(g2, 0L, as.integer(cc), cc * 0.5)
    }

    cloglog_grid <- maxent_project_cloglog(fs2, list(g1, g2), c("bio1", "bio2"))
    cloglog_mat  <- maxent_grid_to_matrix(cloglog_grid)

    expect_true(all(is.finite(cloglog_mat)))
    expect_true(all(cloglog_mat >= 0.0 & cloglog_mat <= 1.0))
})

# ===========================================================================
# Group M — Combined all-feature-type model fidelity
# ===========================================================================

test_that("Group M: all-type model trains with correct feature count", {
    skip_if_not_installed("maxentcpp")

    m <- .build_asym_model(
        types = c("linear", "quadratic", "product", "threshold", "hinge"),
        n_thresholds = 3L, n_hinges = 2L)

    # 2 variables:
    # Per variable: linear(1) + quadratic(1) + threshold(3) + hinge_fwd(2) + hinge_rev(2) = 9
    # Product pairs: 1
    # Total: 9*2 + 1 = 19
    expect_equal(length(m$features), 19L)
    expect_equal(length(m$fit$lambdas), 19L)
})

test_that("Group M: all-type model has all feature types present", {
    skip_if_not_installed("maxentcpp")

    m <- .build_asym_model(
        types = c("linear", "quadratic", "product", "threshold", "hinge"),
        n_thresholds = 3L, n_hinges = 2L)

    types <- sapply(m$features, function(f) maxent_feature_info(f)$type)
    expect_true("linear"        %in% types)
    expect_true("quadratic"     %in% types)
    expect_true("product"       %in% types)
    expect_true("threshold"     %in% types)
    expect_true("hinge"         %in% types)
    expect_true("reverse_hinge" %in% types)
})

test_that("Group M: all-type model converges with finite loss and entropy", {
    skip_if_not_installed("maxentcpp")

    m <- .build_asym_model(
        types = c("linear", "quadratic", "product", "threshold", "hinge"),
        n_thresholds = 3L, n_hinges = 2L)

    expect_true(m$fit$iterations > 0L)
    expect_true(is.finite(m$fit$loss))
    expect_true(is.finite(m$fit$entropy))
    expect_true(m$fit$entropy >= 0)
})

test_that("Group M: all-type model weights sum to ~1", {
    skip_if_not_installed("maxentcpp")

    m <- .build_asym_model(
        types = c("linear", "quadratic", "product", "threshold", "hinge"),
        n_thresholds = 3L, n_hinges = 2L)

    w <- maxent_model_weights(m$model)
    expect_equal(sum(w), 1.0, tolerance = 1e-12)
})

test_that("Group M: all-type model raw_java bounded by [0,1] in-range", {
    skip_if_not_installed("maxentcpp")

    m <- .build_asym_model(
        types = c("linear", "quadratic", "product", "threshold", "hinge"),
        n_thresholds = 3L, n_hinges = 2L)

    dim_1 <- maxent_dimension(nrows = 1L, ncols = 5L,
                              xll = 0, yll = 0, cellsize = 0.2)
    g1 <- maxent_grid(dim_1, "bio1")
    g2 <- maxent_grid(dim_1, "bio2")
    for (cc in 0:4) {
        grid_set_value(g1, 0L, as.integer(cc), cc * 0.25)
        grid_set_value(g2, 0L, as.integer(cc), cc * 0.25)
    }

    raw_grid <- maxent_project_raw(m$model, list(g1, g2), c("bio1", "bio2"))
    raw_mat  <- maxent_grid_to_matrix(raw_grid)

    expect_true(all(raw_mat >= 0.0))
    expect_true(all(raw_mat <= 1.0))
})

test_that("Group M: all-type model training loss <= initial uniform loss", {
    skip_if_not_installed("maxentcpp")

    d <- asym_fixture()

    # Build untrained model to get initial loss
    feat0 <- maxent_generate_features(
        list(bio1 = d$bio1, bio2 = d$bio2),
        types = c("linear", "quadratic", "product", "threshold", "hinge"),
        n_thresholds = 3L, n_hinges = 2L
    )
    fs0 <- maxent_featured_space(d$n, d$sample_idx, feat0)
    maxent_set_sample_expectations(fs0)
    initial_loss <- maxent_model_loss(fs0)

    # Now train
    m <- .build_asym_model(
        types = c("linear", "quadratic", "product", "threshold", "hinge"),
        n_thresholds = 3L, n_hinges = 2L)

    expect_true(m$fit$loss <= initial_loss + 1e-9,
                label = "trained loss <= initial uniform loss")
})

test_that("Group M: sequential_fit on all-type model matches maxent_fit loss", {
    skip_if_not_installed("maxentcpp")

    d <- asym_fixture()

    # Build and train with maxent_fit (simple optimizer)
    feat1 <- maxent_generate_features(
        list(bio1 = d$bio1, bio2 = d$bio2),
        types        = c("linear", "quadratic", "product"),
        n_thresholds = 0L, n_hinges = 0L
    )
    fs1 <- maxent_featured_space(d$n, d$sample_idx, feat1)
    res1 <- maxent_fit(fs1, max_iter = 500L, convergence = 1e-5,
                       beta_multiplier = 1.0)

    # Build and train with maxent_sequential_fit (full optimizer)
    feat2 <- maxent_generate_features(
        list(bio1 = d$bio1, bio2 = d$bio2),
        types        = c("linear", "quadratic", "product"),
        n_thresholds = 0L, n_hinges = 0L
    )
    fs2 <- maxent_featured_space(d$n, d$sample_idx, feat2)
    res2 <- maxent_sequential_fit(fs2, max_iter = 500L, convergence = 1e-5,
                                  beta_multiplier = 1.0)

    # Both should achieve similar converged loss.  The two optimizers
    # (goodAlpha vs full sequential) follow different paths and may converge
    # to slightly different solutions, especially with product features.
    expect_equal(res1$loss, res2$loss, tolerance = 0.5,
                 label = "maxent_fit loss",
                 expected.label = "sequential_fit loss")
})

## ---- Console Results Print --------------------------------------------------

.make_print_test_model <- function() {
    skip_if_not_installed("maxentcpp")
    dim_obj <- maxent_dimension(nrows = 4, ncols = 4, xll = 0, yll = 0,
                                cellsize = 1)
    g0 <- maxent_grid(dim_obj, "env0")
    g1 <- maxent_grid(dim_obj, "env1")
    idx <- 0L
    for (r in 0:3) {
        for (cc in 0:3) {
            grid_set_value(g0, r, cc, idx / 15.0)
            grid_set_value(g1, r, cc, 1.0 - idx / 15.0)
            idx <- idx + 1L
        }
    }
    vals0 <- sapply(0:15, function(i) i / 15.0)
    vals1 <- 1.0 - vals0
    features <- maxent_generate_features(
        list(env0 = vals0, env1 = vals1),
        types = "linear", n_thresholds = 0, n_hinges = 0)
    pres_idx <- 12:15
    model <- maxent_featured_space(16L, as.integer(pres_idx), features)
    fit_r <- maxent_fit(model, max_iter = 200, convergence = 1e-4,
                        beta_multiplier = 1.0)
    pres_rows <- as.integer(pres_idx %/% 4L)
    pres_cols <- as.integer(pres_idx  %%  4L)
    bg_rows   <- as.integer(c(0L, 1L, 2L))
    bg_cols   <- as.integer(c(0L, 1L, 2L))
    list(model = model, fit_result = fit_r,
         g0 = g0, g1 = g1,
         pres_rows = pres_rows, pres_cols = pres_cols,
         bg_rows = bg_rows, bg_cols = bg_cols)
}

test_that("maxent_print_results returns invisibly with correct structure", {
    skip_if_not_installed("maxentcpp")
    m <- .make_print_test_model()

    pres_preds <- maxent_extract_predictions_raw(
        m$model, list(m$g0, m$g1), c("env0", "env1"),
        m$pres_rows, m$pres_cols)
    bg_preds <- maxent_extract_predictions_raw(
        m$model, list(m$g0, m$g1), c("env0", "env1"),
        m$bg_rows, m$bg_cols)
    eval_r   <- maxent_evaluate(pres_preds, bg_preds)
    contrib  <- maxent_percent_contribution(m$model, c("env0", "env1"))
    perm_imp <- maxent_permutation_importance(
        m$model, list(m$g0, m$g1), c("env0", "env1"),
        m$pres_rows, m$pres_cols,
        m$bg_rows, m$bg_cols)

    result <- maxent_print_results(
        species          = "TestSp",
        eval_result      = eval_r,
        contributions_df = contrib,
        perm_imp_df      = perm_imp,
        n_presence       = length(m$pres_rows),
        n_background     = length(m$bg_rows),
        fit_result       = m$fit_result)

    expect_type(result, "list")
    expect_equal(result$species, "TestSp")
    expect_equal(result$n_presence, length(m$pres_rows))
    expect_equal(result$n_background, length(m$bg_rows))
    expect_equal(result$n_test, 0L)
    expect_true(is.numeric(result$training_auc))
    expect_true(result$training_auc >= 0 && result$training_auc <= 1)
})

test_that("maxent_print_results outputs expected lines to console", {
    skip_if_not_installed("maxentcpp")
    m <- .make_print_test_model()

    pres_preds <- maxent_extract_predictions_raw(
        m$model, list(m$g0, m$g1), c("env0", "env1"),
        m$pres_rows, m$pres_cols)
    bg_preds <- maxent_extract_predictions_raw(
        m$model, list(m$g0, m$g1), c("env0", "env1"),
        m$bg_rows, m$bg_cols)
    eval_r   <- maxent_evaluate(pres_preds, bg_preds)
    contrib  <- maxent_percent_contribution(m$model, c("env0", "env1"))
    perm_imp <- maxent_permutation_importance(
        m$model, list(m$g0, m$g1), c("env0", "env1"),
        m$pres_rows, m$pres_cols,
        m$bg_rows, m$bg_cols)

    output <- capture.output(
        maxent_print_results(
            species          = "TestSp",
            eval_result      = eval_r,
            contributions_df = contrib,
            perm_imp_df      = perm_imp,
            n_presence       = length(m$pres_rows),
            n_background     = length(m$bg_rows),
            fit_result       = m$fit_result))

    combined <- paste(output, collapse = "\n")
    expect_true(grepl("MaxEnt",             combined))
    expect_true(grepl("TestSp",             combined))
    expect_true(grepl("n presence",         combined))
    expect_true(grepl("n background",       combined))
    expect_true(grepl("Training statistics", combined))
    expect_true(grepl("AUC",               combined))
    expect_true(grepl("Variable contributions", combined))
    expect_true(grepl("env0",              combined))
    expect_true(grepl("env1",              combined))
})

test_that("maxent_print_results reports test AUC when test_eval_result provided", {
    skip_if_not_installed("maxentcpp")
    m <- .make_print_test_model()

    pres_preds <- maxent_extract_predictions_raw(
        m$model, list(m$g0, m$g1), c("env0", "env1"),
        m$pres_rows, m$pres_cols)
    bg_preds <- maxent_extract_predictions_raw(
        m$model, list(m$g0, m$g1), c("env0", "env1"),
        m$bg_rows, m$bg_cols)
    eval_r      <- maxent_evaluate(pres_preds, bg_preds)
    test_eval_r <- maxent_evaluate(pres_preds, bg_preds)
    contrib  <- maxent_percent_contribution(m$model, c("env0", "env1"))
    perm_imp <- maxent_permutation_importance(
        m$model, list(m$g0, m$g1), c("env0", "env1"),
        m$pres_rows, m$pres_cols,
        m$bg_rows, m$bg_cols)

    output <- capture.output(
        result <- maxent_print_results(
            species          = "TestSp",
            eval_result      = eval_r,
            contributions_df = contrib,
            perm_imp_df      = perm_imp,
            n_presence       = length(m$pres_rows),
            n_background     = length(m$bg_rows),
            test_eval_result = test_eval_r,
            n_test           = 4L,
            fit_result       = m$fit_result))

    combined <- paste(output, collapse = "\n")
    expect_true(grepl("n test",           combined))
    expect_true(grepl("Test statistics",  combined))
    expect_equal(result$n_test, 4L)
    expect_true(is.numeric(result$test_auc))
})

# Helper functions for Java numerical equivalency tests.
#
# These helpers are silently skipped when rJava is not installed or the
# JAR has not been compiled.  All other testthat test files are unaffected.

# ---------------------------------------------------------------------------
# JAR location and JVM initialisation
# ---------------------------------------------------------------------------

.java_jar_path <- function() {
    system.file("java", "maxent_mini.jar", package = "maxentcpp")
}

# Called once (idempotent) to start the JVM and load the JAR.
.init_jvm <- local({
    done <- FALSE
    function() {
        if (done) return(invisible(NULL))
        if (!requireNamespace("rJava", quietly = TRUE))
            stop("rJava not available")
        jar <- .java_jar_path()
        if (!file.exists(jar))
            stop("maxent_mini.jar not found; run `cd java && ./build.sh`")
        rJava::.jinit()
        rJava::.jaddClassPath(jar)
        done <<- TRUE
        invisible(NULL)
    }
})

# Skip helpers for Groups B / C / D.
skip_if_no_java <- function() {
    skip_if_not_installed("rJava")
    skip_if(
        !file.exists(.java_jar_path()),
        "maxent_mini.jar not compiled; run `cd java && ./build.sh`"
    )
    # Initialise JVM and verify MaxentJavaRunner is loadable (guards
    # against Java version mismatch: jar compiled with Java 17 but
    # runtime may be older).
    tryCatch({
        .init_jvm()
        rJava::.jfindClass("MaxentJavaRunner")
    }, error = function(e) {
        skip(paste("Java runtime cannot load MaxentJavaRunner:",
                   conditionMessage(e)))
    })
}

# ---------------------------------------------------------------------------
# Canonical datasets
# ---------------------------------------------------------------------------

# Tiny 1-variable dataset (9 background + presence flags).
# Matches make_tiny_model() used in other test files.
tiny_1var_data <- function() {
    vals  <- seq(0, 1, length.out = 9L)
    flags <- c(0L, 0L, 1L, 0L, 0L, 1L, 0L, 0L, 1L)   # 1 at 1-based R indices 3,6,9 (0-based: 2,5,8)
    list(
        env    = vals,
        flags  = flags,
        n      = 9L,
        idx0   = c(2L, 5L, 8L)          # 0-based sample indices
    )
}

# Two-variable dataset from inst/extdata/test_data.csv (70 points).
# Returns NULL (and emits a skip) when the file is not found.
twovar_data <- function() {
    csv_path <- system.file("extdata", "test_data.csv", package = "maxentcpp")
    if (!nzchar(csv_path) || !file.exists(csv_path))
        testthat::skip("test_data.csv not found in inst/extdata/")

    df  <- read.csv(csv_path)
    idx <- which(df$species == 1L) - 1L   # 0-based presence indices

    list(
        bio1   = df$bio1,
        bio12  = df$bio12,
        flags  = as.integer(df$species),
        n      = nrow(df),
        idx0   = as.integer(idx)
    )
}

# ---------------------------------------------------------------------------
# C++ model builders
# ---------------------------------------------------------------------------

# Build and train a C++ model using bio1 from test_data.csv.
build_cpp_model_1var <- function() {
    d <- twovar_data()
    feat <- maxent_generate_features(
        list(bio1 = d$bio1),
        types        = "linear",
        n_thresholds = 0L,
        n_hinges     = 0L
    )
    model      <- maxent_featured_space(d$n, d$idx0, feat)
    fit_result <- maxent_fit(model,
                             max_iter        = 500L,
                             convergence     = 1e-5,
                             beta_multiplier = 1.0,
                             min_deviation   = 0.001)
    list(model = model, fit = fit_result, features = feat, data = d)
}

# Build and train a C++ model using bio1 + bio12 from test_data.csv.
build_cpp_model_2var <- function() {
    d <- twovar_data()
    feat <- maxent_generate_features(
        list(bio1 = d$bio1, bio12 = d$bio12),
        types        = "linear",
        n_thresholds = 0L,
        n_hinges     = 0L
    )
    model      <- maxent_featured_space(d$n, d$idx0, feat)
    fit_result <- maxent_fit(model,
                             max_iter        = 500L,
                             convergence     = 1e-5,
                             beta_multiplier = 1.0,
                             min_deviation   = 0.001)
    list(model = model, fit = fit_result, features = feat, data = d)
}

# Compute Java-compatible raw scores for all training points of a C++ model.
# Returns a numeric vector of length n, each value in [0, 1].
cpp_raw_scores <- function(cpp_result) {
    model    <- cpp_result$model
    features <- cpp_result$features
    n        <- cpp_result$data$n
    info     <- maxent_space_info(model)

    # Build pre-evaluated feature matrix (n × num_features)
    nf   <- length(features)
    fmat <- matrix(0.0, nrow = n, ncol = nf)
    for (j in seq_len(nf)) {
        for (i in seq_len(n)) {
            fmat[i, j] <- maxent_feature_eval(features[[j]], i)
        }
    }

    unnorm <- maxent_predict_model(model, fmat)           # exp(lp - lpNorm)
    raw    <- unnorm / info$density_normalizer            # divide by Z
    pmin(raw, 1.0)                                        # cap at 1
}

# ---------------------------------------------------------------------------
# Java model builders
# ---------------------------------------------------------------------------

# Train a 1-variable Java model on bio1 from test_data.csv.
# Returns a named list mirroring the C++ model's information.
build_java_model_1var <- function() {
    .init_jvm()
    d <- twovar_data()

    j_out <- rJava::.jcall(
        "MaxentJavaRunner", "[D", "trainLinear1Var",
        rJava::.jarray(as.double(d$bio1)),
        rJava::.jarray(as.integer(d$flags)),
        500L, 1e-5, 1.0, 0.001
    )
    result <- if (is.numeric(j_out)) j_out else rJava::.jevalArray(j_out)

    n <- d$n
    list(
        lambda             = result[1],
        entropy            = result[2],
        density_normalizer = result[3],
        lp_normalizer      = result[4],
        loss               = result[5],
        raw_scores         = result[seq(6, 5 + n)]
    )
}

# Train a 2-variable Java model on bio1 + bio12 from test_data.csv.
build_java_model_2var <- function() {
    .init_jvm()
    d <- twovar_data()

    j_out <- rJava::.jcall(
        "MaxentJavaRunner", "[D", "trainLinear2Var",
        rJava::.jarray(as.double(d$bio1)),
        rJava::.jarray(as.double(d$bio12)),
        rJava::.jarray(as.integer(d$flags)),
        500L, 1e-5, 1.0, 0.001
    )
    result <- if (is.numeric(j_out)) j_out else rJava::.jevalArray(j_out)

    n <- d$n
    list(
        lambda1            = result[1],
        lambda2            = result[2],
        entropy            = result[3],
        density_normalizer = result[4],
        lp_normalizer      = result[5],
        loss               = result[6],
        raw_scores         = result[seq(7, 6 + n)]
    )
}

# ---------------------------------------------------------------------------
# Java trajectory builders (Phase B)
# ---------------------------------------------------------------------------

# Parse the flat double[] returned by trainLinear*VarTrajectory into a
# data.frame with columns (iteration, loss, entropy, lambda_0, [lambda_1, ...]).
#
# The return layout (from Sequential.runWithTrajectory) is:
#   result[1]                     = K (number of captured checkpoints)
#   For each k = 1..K, row width = 3 + nf:
#     result[1 + (k-1)*(3+nf) + 1] = 1-based iteration
#     result[1 + (k-1)*(3+nf) + 2] = loss
#     result[1 + (k-1)*(3+nf) + 3] = entropy
#     result[1 + (k-1)*(3+nf) + 4 .. 3+nf] = lambda values (one per feature)
.parse_trajectory <- function(result, nf) {
    raw <- if (is.numeric(result)) result else rJava::.jevalArray(result)
    K   <- as.integer(raw[1])
    if (K == 0L) {
        return(data.frame(
            iteration = integer(0), loss = double(0), entropy = double(0)))
    }
    rw  <- 3L + nf
    df  <- data.frame(
        iteration = integer(K),
        loss      = double(K),
        entropy   = double(K)
    )
    lam_cols <- setNames(
        replicate(nf, double(K), simplify = FALSE),
        paste0("lambda_", seq_len(nf) - 1L)
    )
    for (k in seq_len(K)) {
        base <- 1L + (k - 1L) * rw    # 1-based R index into raw[]
        df$iteration[k] <- as.integer(raw[base + 1L])
        df$loss[k]      <- raw[base + 2L]
        df$entropy[k]   <- raw[base + 3L]
        for (j in seq_len(nf)) {
            lam_cols[[j]][k] <- raw[base + 3L + j]
        }
    }
    cbind(df, as.data.frame(lam_cols))
}

# Capture the simple-Java-Sequential trajectory for a 1-variable model on
# bio1 from test_data.csv.  Returns a data.frame with columns:
#   iteration, loss, entropy, lambda_0
#
# convergence=0.0 is used so that the optimizer runs to max_iter
# deterministically, ensuring all checkpoint rows are captured.
build_java_trajectory_1var <- function(
        checkpoints = c(1L, 2L, 3L, 5L, 10L, 20L, 50L, 100L, 200L, 500L)) {
    .init_jvm()
    d <- twovar_data()

    j_out <- rJava::.jcall(
        "MaxentJavaRunner", "[D", "trainLinear1VarTrajectory",
        rJava::.jarray(as.double(d$bio1)),
        rJava::.jarray(as.integer(d$flags)),
        500L, 0.0, 1.0, 0.001,
        rJava::.jarray(as.integer(checkpoints))
    )
    .parse_trajectory(j_out, nf = 1L)
}

# Capture the simple-Java-Sequential trajectory for a 2-variable model on
# bio1 + bio12 from test_data.csv.  Returns a data.frame with columns:
#   iteration, loss, entropy, lambda_0, lambda_1
#
# convergence=0.0 is used so that the optimizer runs to max_iter
# deterministically, ensuring all checkpoint rows are captured.
build_java_trajectory_2var <- function(
        checkpoints = c(1L, 2L, 3L, 5L, 10L, 20L, 50L, 100L, 200L, 500L)) {
    .init_jvm()
    d <- twovar_data()

    j_out <- rJava::.jcall(
        "MaxentJavaRunner", "[D", "trainLinear2VarTrajectory",
        rJava::.jarray(as.double(d$bio1)),
        rJava::.jarray(as.double(d$bio12)),
        rJava::.jarray(as.integer(d$flags)),
        500L, 0.0, 1.0, 0.001,
        rJava::.jarray(as.integer(checkpoints))
    )
    .parse_trajectory(j_out, nf = 2L)
}

# ---------------------------------------------------------------------------
# Asymmetric fixture (Phase B)
# ---------------------------------------------------------------------------

# The 10x10 asymmetric fixture used by the C++ test_sequential.cpp golden.
# bio1[r][c] = 10 + c + 2*r  (scaled to [0,1])
# bio2[r][c] = 100 + 50*c + 10*r  (scaled to [0,1])
# 10 occurrences clustered in the bottom-right quadrant.
asym_fixture <- function() {
    nrows <- 10L; ncols <- 10L
    n     <- nrows * ncols

    # Build row-major (C-order) vectors to match the C++ FeaturedSpace
    # layout: index = row * ncols + col.
    bio1_raw <- numeric(n)
    bio2_raw <- numeric(n)
    for (r in seq_len(nrows) - 1L) {
        for (c in seq_len(ncols) - 1L) {
            idx <- r * ncols + c + 1L  # 1-based R index
            bio1_raw[idx] <- 10 + c + 2 * r          # range 10..37
            bio2_raw[idx] <- 100 + 50 * c + 10 * r   # range 100..640
        }
    }

    prescale <- function(v) {
        mn <- min(v); mx <- max(v)
        if (mx == mn) v else (v - mn) / (mx - mn)
    }
    bio1 <- prescale(bio1_raw)
    bio2 <- prescale(bio2_raw)

    # 0-based sample indices (row-major: idx = row * ncols + col)
    sample_idx <- c(
        6L * ncols + 8L,   # lon=8.5 lat=3.5
        7L * ncols + 7L,   # lon=7.5 lat=2.5
        7L * ncols + 8L,   # lon=8.5 lat=2.5
        7L * ncols + 9L,   # lon=9.5 lat=2.5
        8L * ncols + 7L,   # lon=7.5 lat=1.5
        8L * ncols + 8L,   # lon=8.5 lat=1.5
        8L * ncols + 9L,   # lon=9.5 lat=1.5
        9L * ncols + 7L,   # lon=7.5 lat=0.5
        9L * ncols + 8L,   # lon=8.5 lat=0.5
        9L * ncols + 9L    # lon=9.5 lat=0.5
    )

    list(
        bio1       = bio1,
        bio2       = bio2,
        n          = nrows * ncols,
        sample_idx = sample_idx
    )
}

# Committed Phase D golden: trajectory of the full Java Sequential optimizer
# on the asymmetric fixture with linear + quadratic features (L+Q, 4 features).
# Values are from inst/extdata/golden/asym/trajectory_java_lq.csv.
# beta_multiplier = 0.8 (Runner.java schedule for n=10, quadratic=true).
asym_java_golden_lq <- function() {
    path <- system.file("extdata", "golden", "asym", "trajectory_java_lq.csv",
                        package = "maxentcpp")
    if (nzchar(path) && file.exists(path)) {
        return(read.csv(path))
    }
    # Fallback: inline values identical to kJavaGolden in test_sequential_lq.cpp.
    data.frame(
        iteration = c(1L, 2L, 3L, 5L, 10L, 20L, 50L, 100L, 200L, 500L),
        loss      = c(4.5421182033281940, 4.4821067643752470, 4.4247504414752160,
                      4.3158369946915425, 4.0820758340185614, 3.2325401364872697,
                      3.1617709439417760, 3.1197924344039323, 3.1197924344039314,
                      3.1197924344039314),
        entropy   = c(4.6042914838480070, 4.6027068726461810, 4.5998290427816150,
                      4.5903933898033730, 4.5483808348650310, 3.1735804811921800,
                      3.1469676964979194, 3.1197924344050634, 3.1197924344039296,
                      3.1197924344039296),
        lambda_0  = c(0.0, 0.0, 0.0, 0.0, 0.0, 0.1840686396557502,
                      1.6527633458811182, 6.0624645573901740, 6.0624645574787560,
                      6.0624645574787560),
        lambda_1  = c(0.0, 0.1440826808525652, 0.2848388206506464,
                      0.4199891999959157, 0.7973445741951862, 3.8763474799751347,
                      2.6166075278309027, 0.0, 0.0, 0.0),
        lambda_2  = c(0.0, 0.0, 0.0, 0.0, 0.0, 0.1918902211537277,
                      4.4453724246267360, 4.1317372609637160, 4.1317372608782830,
                      4.1317372608782830),
        lambda_3  = c(0.1485486498637492, 0.1485486498637492, 0.1485486498637492,
                      0.2883496256734949, 0.5470248844025938, 2.4854448484169174,
                      0.0, 0.0, 0.0, 0.0)
    )
}

# Committed Phase D golden: trajectory of the full Java Sequential optimizer
# on the asymmetric fixture with linear + quadratic + product features
# (L+Q+P, 5 features).
# Values are from inst/extdata/golden/asym/trajectory_java_lqp.csv.
# beta_multiplier = 1.6 (Runner.java schedule for n=10, product=true).
asym_java_golden_lqp <- function() {
    path <- system.file("extdata", "golden", "asym", "trajectory_java_lqp.csv",
                        package = "maxentcpp")
    if (nzchar(path) && file.exists(path)) {
        return(read.csv(path))
    }
    # Fallback: inline values identical to kJavaGolden in test_sequential_lqp.cpp.
    data.frame(
        iteration = c(1L, 2L, 3L, 5L, 10L, 20L, 50L, 100L, 200L, 500L),
        loss      = c(4.5445471190472690, 4.4864626474063850, 4.4308592054085090,
                      4.3268492682186140, 4.1056844205583840, 3.3687650148865110,
                      3.3161976479591400, 3.3156340906479533, 3.3156340906479533,
                      3.3156340906479533),
        entropy   = c(4.6045625322128485, 4.6027457807404190, 4.5997345788713020,
                      4.5902214481545150, 4.5477752860752190, 3.3074182965897080,
                      3.3152035092852740, 3.3156340904802220, 3.3156340904802220,
                      3.3156340904802220),
        lambda_0  = c(0.0, 0.0, 0.0, 0.0, 0.0, 0.2871495803078118,
                      5.6095976837979840, 5.3733744303312730, 5.3733744303312730,
                      5.3733744303312730),
        lambda_1  = c(0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0),
        lambda_2  = c(0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 3.4527918072852968,
                      3.6925604260961236, 3.6925604260961236, 3.6925604260961236),
        lambda_3  = c(0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0),
        lambda_4  = c(0.1445946125260754, 0.2859648790693891, 0.4241384982541208,
                      0.6910226500687243, 1.3048083334902638, 5.7652180937050250,
                      0.0, 0.0, 0.0, 0.0)
    )
}

# Capture the simplified-Java-Sequential trajectory for the asymmetric
# 10x10 fixture with linear + quadratic features (L+Q, 4 features).
# Returns a data.frame with columns:
#   iteration, loss, entropy, lambda_0, lambda_1, lambda_2, lambda_3
#
# Note: the simplified Java optimizer (goodAlpha / reduceAlpha loop) produces
# a different trajectory from the full Java Sequential (density/Sequential.java).
# These trajectory values are only comparable with themselves, not with the
# golden CSV files produced by the full optimizer.
build_java_trajectory_asym_lq <- function(
        checkpoints = c(1L, 2L, 3L, 5L, 10L, 20L, 50L, 100L, 200L, 500L)) {
    .init_jvm()

    j_out <- rJava::.jcall(
        "MaxentJavaRunner", "[D", "trainAsymLQTrajectory",
        rJava::.jarray(as.integer(checkpoints))
    )
    .parse_trajectory(j_out, nf = 4L)
}

# Capture the simplified-Java-Sequential trajectory for the asymmetric
# 10x10 fixture with linear + quadratic + product features (L+Q+P, 5 features).
# Returns a data.frame with columns:
#   iteration, loss, entropy, lambda_0, lambda_1, lambda_2, lambda_3, lambda_4
#
# Note: the simplified Java optimizer (goodAlpha / reduceAlpha loop) produces
# a different trajectory from the full Java Sequential (density/Sequential.java).
# These trajectory values are only comparable with themselves, not with the
# golden CSV files produced by the full optimizer.
build_java_trajectory_asym_lqp <- function(
        checkpoints = c(1L, 2L, 3L, 5L, 10L, 20L, 50L, 100L, 200L, 500L)) {
    .init_jvm()

    j_out <- rJava::.jcall(
        "MaxentJavaRunner", "[D", "trainAsymLQPTrajectory",
        rJava::.jarray(as.integer(checkpoints))
    )
    .parse_trajectory(j_out, nf = 5L)
}

# Committed Phase B golden: trajectory of the full Java Sequential optimizer
# on the asymmetric fixture.  Values are embedded from
# inst/extdata/golden/asym/trajectory_java.csv and the C++ test
# test_sequential.cpp (kJavaGolden array).
asym_java_golden <- function() {
    path <- system.file("extdata", "golden", "asym", "trajectory_java.csv",
                        package = "maxentcpp")
    if (nzchar(path) && file.exists(path)) {
        return(read.csv(path))
    }
    # Fallback: inline values identical to kJavaGolden in test_sequential.cpp.
    data.frame(
        iteration = c(  1L,   2L,   3L,   5L,   10L,    20L,     50L,    100L,    200L,    500L),
        loss      = c(4.5534670404943647, 4.5044834711287370, 4.4580810114815135,
                      4.3720945694427646, 4.1835670471368118, 3.2037850143863942,
                      3.1709389711531495, 3.1709389280031175, 3.1709389280031170,
                      3.1709389280031170),
        entropy   = c(4.6043911168777019, 4.6021307223743895, 4.5985044305670746,
                      4.5902747732142490, 4.5581876611341912, 3.4675596532919681,
                      3.1705613150634631, 3.1709389276298126, 3.1709389281312537,
                      3.1709389281312537),
        lambda_0  = c(0.0000000000000000, 0.0000000000000000, 0.0000000000000000,
                      0.13045194323897974, 0.50462215463779610, 4.1326468050985640,
                      5.8805050188416480, 5.8795313549995200, 5.8795313516019400,
                      5.8795313516019400),
        lambda_1  = c(0.14557239069988340, 0.28771862438045490, 0.42652212318552490,
                      0.56206675272495440, 0.81416830898974410, 4.1203851942018460,
                      4.0147649660976490, 4.0135187975142200, 4.0135187980476040,
                      4.0135187980476040)
    )
}

# Tests for Java Maxent numerical equivalency.
#
# Compares outputs of the maxentcpp C++ implementation with a minimal
# Java reimplementation of the core Maxent algorithm (linear features,
# raw output).
#
# Java source files live in java/ at the repository root.
# The compiled JAR (inst/java/maxent_mini.jar) is required for
# Groups B, C, and D.  Build it with:
#   cd java && ./build.sh
#
# Source classes tested:
#   Group A — density/LinearFeature.java    (feature evaluation formula)
#   Group B — density/FeaturedSpace.java +  (training invariants)
#             density/Sequential.java
#   Group C — density/FeaturedSpace.java    (raw output formula)
#   Group D — end-to-end on test_data.csv   (bio1 + bio12 linear model)

# ===========================================================================
# Group A — Linear feature normalization
# (No Java runtime required; tests the formula against the C++ implementation)
# ===========================================================================

test_that("Linear feature eval matches (x - min)/(max - min) for in-range values", {
    skip_if_not_installed("maxentcpp")

    # Use bio1 column from test_data.csv if available, otherwise a small vector
    csv_path <- system.file("extdata", "test_data.csv", package = "maxentcpp")
    if (nzchar(csv_path) && file.exists(csv_path)) {
        vals <- read.csv(csv_path)$bio1
    } else {
        vals <- c(56.6, 104.7, 181.1, 248.2, 299.4)
    }

    mn <- min(vals)
    mx <- max(vals)
    f  <- maxent_linear_feature(vals, "bio1", min_val = mn, max_val = mx)

    expected <- (vals - mn) / (mx - mn)
    actual   <- sapply(seq_along(vals), function(i) maxent_feature_eval(f, i))

    expect_equal(actual, expected, tolerance = 1e-12)
})

test_that("Linear feature eval clamps: below min -> 0, above max -> 1 (projection)", {
    skip_if_not_installed("maxentcpp")

    # Use projection (eval_from_env) to test clamping:
    # build a tiny 1-point grid with an out-of-range value and compare
    # to the boundary value.
    vals <- seq(10.0, 100.0, length.out = 5)
    features <- maxent_generate_features(
        list(bio1 = vals), types = "linear",
        n_thresholds = 0L, n_hinges = 0L
    )
    model <- maxent_featured_space(5L, c(3L, 4L), features)
    maxent_fit(model, max_iter = 100L, convergence = 1e-5)

    # Build single-cell grids at boundary values
    dim_1 <- maxent_dimension(nrows = 1L, ncols = 1L,
                              xll = 0, yll = 0, cellsize = 1)
    g_min <- maxent_grid(dim_1, "bio1")
    g_max <- maxent_grid(dim_1, "bio1")
    g_low <- maxent_grid(dim_1, "bio1")   # below training min
    g_hi  <- maxent_grid(dim_1, "bio1")   # above training max

    grid_set_value(g_min, 0L, 0L, min(vals))
    grid_set_value(g_max, 0L, 0L, max(vals))
    grid_set_value(g_low, 0L, 0L, min(vals) - 1000)
    grid_set_value(g_hi,  0L, 0L, max(vals) + 1000)

    pred_min <- maxent_grid_to_matrix(
        maxent_project_raw(model, list(g_min), c("bio1")))[1, 1]
    pred_max <- maxent_grid_to_matrix(
        maxent_project_raw(model, list(g_max), c("bio1")))[1, 1]
    pred_low <- maxent_grid_to_matrix(
        maxent_project_raw(model, list(g_low), c("bio1")))[1, 1]
    pred_hi  <- maxent_grid_to_matrix(
        maxent_project_raw(model, list(g_hi),  c("bio1")))[1, 1]

    expect_equal(pred_low, pred_min, tolerance = 1e-9,
                 label = "below-range prediction equals boundary min prediction")
    expect_equal(pred_hi,  pred_max, tolerance = 1e-9,
                 label = "above-range prediction equals boundary max prediction")
})

test_that("Linear feature formula: min == max returns 0", {
    skip_if_not_installed("maxentcpp")

    vals <- c(5.0, 5.0, 5.0)
    f    <- maxent_linear_feature(vals, "constant", min_val = 5.0, max_val = 5.0)
    expect_equal(maxent_feature_eval(f, 1), 0.0)
    expect_equal(maxent_feature_eval(f, 2), 0.0)
    expect_equal(maxent_feature_eval(f, 3), 0.0)
})

# ===========================================================================
# Group B — Training invariants (C++ vs Java)
# Tolerance 1e-6 across all invariants.
# ===========================================================================

test_that("Entropy matches Java after training (1-variable model)", {
    skip_if_not_installed("maxentcpp")
    skip_if_no_java()

    cpp  <- build_cpp_model_1var()
    java <- build_java_model_1var()

    expect_equal(cpp$fit$entropy, java$entropy,
                 tolerance = 1e-6,
                 label = "C++ entropy", expected.label = "Java entropy")
})

test_that("density_normalizer matches Java after training (1-variable model)", {
    skip_if_not_installed("maxentcpp")
    skip_if_no_java()

    cpp  <- build_cpp_model_1var()
    java <- build_java_model_1var()

    cpp_dn <- maxent_space_info(cpp$model)$density_normalizer
    expect_equal(cpp_dn, java$density_normalizer,
                 tolerance = 1e-6,
                 label = "C++ density_normalizer",
                 expected.label = "Java density_normalizer")
})

test_that("linear_predictor_normalizer matches Java after training (1-variable model)", {
    skip_if_not_installed("maxentcpp")
    skip_if_no_java()

    cpp  <- build_cpp_model_1var()
    java <- build_java_model_1var()

    cpp_lpn <- maxent_space_info(cpp$model)$linear_predictor_normalizer
    expect_equal(cpp_lpn, java$lp_normalizer,
                 tolerance = 1e-6,
                 label = "C++ lp_normalizer",
                 expected.label = "Java lp_normalizer")
})

test_that("Lambda value matches Java for bio1 linear feature (1-variable model)", {
    skip_if_not_installed("maxentcpp")
    skip_if_no_java()

    cpp  <- build_cpp_model_1var()
    java <- build_java_model_1var()

    expect_equal(cpp$fit$lambdas[[1]], java$lambda,
                 tolerance = 1e-6,
                 label = "C++ lambda", expected.label = "Java lambda")
})

test_that("Loss matches Java after training (1-variable model)", {
    skip_if_not_installed("maxentcpp")
    skip_if_no_java()

    cpp  <- build_cpp_model_1var()
    java <- build_java_model_1var()

    expect_equal(cpp$fit$loss, java$loss,
                 tolerance = 1e-6,
                 label = "C++ loss", expected.label = "Java loss")
})

# ===========================================================================
# Group C — Raw output formula (C++ vs Java)
# ===========================================================================

test_that("Raw score matches Java for all background training points (1-variable)", {
    skip_if_not_installed("maxentcpp")
    skip_if_no_java()

    cpp  <- build_cpp_model_1var()
    java <- build_java_model_1var()

    cpp_raw  <- cpp_raw_scores(cpp)
    java_raw <- java$raw_scores

    expect_equal(cpp_raw, java_raw,
                 tolerance = 1e-9,
                 label = "C++ raw scores",
                 expected.label = "Java raw scores")
})

test_that("Raw scores are capped at 1.0 by both C++ and Java (1-variable)", {
    skip_if_not_installed("maxentcpp")
    skip_if_no_java()

    cpp  <- build_cpp_model_1var()
    java <- build_java_model_1var()

    cpp_raw  <- cpp_raw_scores(cpp)
    java_raw <- java$raw_scores

    expect_true(all(cpp_raw  <= 1.0), label = "C++ raw scores <= 1")
    expect_true(all(java_raw <= 1.0), label = "Java raw scores <= 1")
    expect_true(all(cpp_raw  >= 0.0), label = "C++ raw scores >= 0")
    expect_true(all(java_raw >= 0.0), label = "Java raw scores >= 0")
})

test_that("Raw scores for presence points match Java (1-variable model)", {
    skip_if_not_installed("maxentcpp")
    skip_if_no_java()

    cpp  <- build_cpp_model_1var()
    java <- build_java_model_1var()

    d <- twovar_data()
    pres_idx <- which(d$flags == 1L)    # 1-based R indices into all-points vector

    cpp_raw  <- cpp_raw_scores(cpp)[pres_idx]
    java_raw <- java$raw_scores[pres_idx]

    expect_equal(cpp_raw, java_raw,
                 tolerance = 1e-9,
                 label = "C++ presence raw scores",
                 expected.label = "Java presence raw scores")
})

# ===========================================================================
# Group D — End-to-end on test_data.csv (bio1 + bio12 linear model)
# ===========================================================================

test_that("Entropy matches Java after training (bio1 + bio12 linear model)", {
    skip_if_not_installed("maxentcpp")
    skip_if_no_java()

    cpp  <- build_cpp_model_2var()
    java <- build_java_model_2var()

    expect_equal(cpp$fit$entropy, java$entropy,
                 tolerance = 1e-6,
                 label = "C++ entropy", expected.label = "Java entropy")
})

test_that("Lambdas match Java for bio1 and bio12 features", {
    skip_if_not_installed("maxentcpp")
    skip_if_no_java()

    cpp  <- build_cpp_model_2var()
    java <- build_java_model_2var()

    expect_equal(cpp$fit$lambdas[[1]], java$lambda1,
                 tolerance = 1e-6,
                 label = "C++ lambda bio1", expected.label = "Java lambda bio1")
    expect_equal(cpp$fit$lambdas[[2]], java$lambda2,
                 tolerance = 1e-6,
                 label = "C++ lambda bio12", expected.label = "Java lambda bio12")
})

test_that("density_normalizer matches Java (bio1 + bio12 linear model)", {
    skip_if_not_installed("maxentcpp")
    skip_if_no_java()

    cpp  <- build_cpp_model_2var()
    java <- build_java_model_2var()

    cpp_dn <- maxent_space_info(cpp$model)$density_normalizer
    expect_equal(cpp_dn, java$density_normalizer,
                 tolerance = 1e-6,
                 label = "C++ density_normalizer",
                 expected.label = "Java density_normalizer")
})

test_that("linear_predictor_normalizer matches Java (bio1 + bio12 linear model)", {
    skip_if_not_installed("maxentcpp")
    skip_if_no_java()

    cpp  <- build_cpp_model_2var()
    java <- build_java_model_2var()

    cpp_lpn <- maxent_space_info(cpp$model)$linear_predictor_normalizer
    expect_equal(cpp_lpn, java$lp_normalizer,
                 tolerance = 1e-6,
                 label = "C++ lp_normalizer",
                 expected.label = "Java lp_normalizer")
})

test_that("Loss matches Java (bio1 + bio12 linear model)", {
    skip_if_not_installed("maxentcpp")
    skip_if_no_java()

    cpp  <- build_cpp_model_2var()
    java <- build_java_model_2var()

    expect_equal(cpp$fit$loss, java$loss,
                 tolerance = 1e-6,
                 label = "C++ loss", expected.label = "Java loss")
})

test_that("Raw scores at all training points match Java (bio1 + bio12 linear model)", {
    skip_if_not_installed("maxentcpp")
    skip_if_no_java()

    cpp  <- build_cpp_model_2var()
    java <- build_java_model_2var()

    cpp_raw  <- cpp_raw_scores(cpp)
    java_raw <- java$raw_scores

    expect_equal(cpp_raw, java_raw,
                 tolerance = 1e-9,
                 label = "C++ raw scores",
                 expected.label = "Java raw scores")
})

# Phase B/D: Quantify Optimizer Gap — optimizer trajectory comparison tests.
#
# This file implements the "side-by-side diagnostics" called for in
# issue #38 (Phase B of the fidelity roadmap, issue #36) and the 1e-6 parity
# gate for Phase D (issue #40).
#
# Groups of tests provided:
#
# Group E — Full C++ Sequential trajectory vs. full Java golden (linear)
#   Uses maxent_sequential_fit() (the Phase C full port of density/Sequential.java)
#   on the 10×10 asymmetric fixture and asserts that every checkpoint in
#   {1,2,3,5,10,20,50,100,200,500} matches the committed Java golden to 1e-6.
#
# Group F — Simple Java Sequential trajectory self-consistency
#   (Requires rJava)  Runs the simplified goodAlpha / reduceAlpha loop in the
#   Java oracle (MaxentJavaRunner::trainLinear2VarTrajectory) on the bio1+bio12
#   dataset and verifies the trajectory is internally consistent: loss must be
#   monotonically non-increasing, entropy must be non-negative, and lambdas
#   must be finite at every checkpoint.
#
# Group G — Phase D: C++ Sequential trajectory vs. committed golden (L+Q)
#   Verifies that the full C++ Sequential optimizer matches the committed
#   Phase D golden (trajectory_java_lq.csv) on the asym fixture with
#   linear + quadratic features to 1e-6.
#
# Group H — Phase D: C++ Sequential trajectory vs. committed golden (L+Q+P)
#   Verifies that the full C++ Sequential optimizer matches the committed
#   Phase D golden (trajectory_java_lqp.csv) on the asym fixture with
#   linear + quadratic + product features to 1e-6.
#
# Group I — Phase D: simplified Java L+Q and L+Q+P trajectory self-consistency
#   (Requires rJava)  Runs the simplified Java optimizer on the asym fixture
#   with L+Q and L+Q+P features and verifies internal consistency.
#
# Together these groups establish the Phase D fidelity baseline.

# ===========================================================================
# Helper: build and run C++ sequential_fit on the asymmetric fixture
# ===========================================================================

.run_cpp_asym_trajectory <- function() {
    d <- asym_fixture()

    feat <- maxent_generate_features(
        list(bio1 = d$bio1, bio2 = d$bio2),
        types        = "linear",
        n_thresholds = 0L,
        n_hinges     = 0L
    )
    fs <- maxent_featured_space(d$n, d$sample_idx, feat)

    maxent_sequential_fit(
        fs,
        max_iter                  = 500L,
        convergence               = 0.0,
        disable_convergence_test  = TRUE,
        trajectory_iterations     = c(1L, 2L, 3L, 5L, 10L, 20L, 50L, 100L, 200L, 500L)
    )$trajectory
}

# ===========================================================================
# Group E — Full C++ Sequential trajectory vs. committed Java golden
# ===========================================================================

test_that("Group E: C++ trajectory row count matches Java golden (asym fixture)", {
    skip_if_not_installed("maxentcpp")

    traj   <- .run_cpp_asym_trajectory()
    golden <- asym_java_golden()

    expect_equal(nrow(traj), nrow(golden),
                 label = "number of captured checkpoints")
})

test_that("Group E: C++ trajectory iteration labels match Java golden (asym fixture)", {
    skip_if_not_installed("maxentcpp")

    traj   <- .run_cpp_asym_trajectory()
    golden <- asym_java_golden()

    expect_equal(traj$iteration, golden$iteration,
                 label = "checkpoint iteration numbers")
})

test_that("Group E: C++ loss trajectory matches Java golden to 1e-6 (asym fixture)", {
    skip_if_not_installed("maxentcpp")

    traj   <- .run_cpp_asym_trajectory()
    golden <- asym_java_golden()

    expect_equal(traj$loss, golden$loss,
                 tolerance = 1e-6,
                 label = "C++ loss trajectory",
                 expected.label = "Java golden loss")
})

test_that("Group E: C++ entropy trajectory matches Java golden to 1e-6 (asym fixture)", {
    skip_if_not_installed("maxentcpp")

    traj   <- .run_cpp_asym_trajectory()
    golden <- asym_java_golden()

    expect_equal(traj$entropy, golden$entropy,
                 tolerance = 1e-6,
                 label = "C++ entropy trajectory",
                 expected.label = "Java golden entropy")
})

test_that("Group E: C++ lambda_0 trajectory matches Java golden to 1e-6 (asym fixture)", {
    skip_if_not_installed("maxentcpp")

    traj   <- .run_cpp_asym_trajectory()
    golden <- asym_java_golden()

    expect_equal(traj$lambda_0, golden$lambda_0,
                 tolerance = 1e-6,
                 label = "C++ lambda_0 trajectory",
                 expected.label = "Java golden lambda_0")
})

test_that("Group E: C++ lambda_1 trajectory matches Java golden to 1e-6 (asym fixture)", {
    skip_if_not_installed("maxentcpp")

    traj   <- .run_cpp_asym_trajectory()
    golden <- asym_java_golden()

    expect_equal(traj$lambda_1, golden$lambda_1,
                 tolerance = 1e-6,
                 label = "C++ lambda_1 trajectory",
                 expected.label = "Java golden lambda_1")
})

test_that("Group E: worst ||delta_lambda||_inf < 1e-6 across all checkpoints (asym fixture)", {
    skip_if_not_installed("maxentcpp")

    traj   <- .run_cpp_asym_trajectory()
    golden <- asym_java_golden()

    worst <- max(
        abs(traj$lambda_0 - golden$lambda_0),
        abs(traj$lambda_1 - golden$lambda_1)
    )
    expect_lt(worst, 1e-6,
              label = "worst ||delta_lambda||_inf across all checkpoints")
})

# ===========================================================================
# Group F — Simple Java Sequential trajectory self-consistency
# (rJava required)
# ===========================================================================

test_that("Group F: simple Java trajectory has correct number of checkpoints (2-var)", {
    skip_if_not_installed("maxentcpp")
    skip_if_no_java()

    checkpoints <- c(1L, 2L, 3L, 5L, 10L, 20L, 50L, 100L, 200L, 500L)
    traj <- build_java_trajectory_2var(checkpoints)

    # The Java optimizer may converge early (before max_iter) so it may
    # not capture all requested checkpoints.  We only require that the
    # returned checkpoints are a prefix of the requested ones.
    expect_true(nrow(traj) >= 1L && nrow(traj) <= length(checkpoints),
                label = "number of Java trajectory rows is in valid range")
    expect_equal(traj$iteration, checkpoints[seq_len(nrow(traj))],
                 label = "Java trajectory iteration labels")
})

test_that("Group F: simple Java trajectory loss is non-increasing at checkpoints (2-var)", {
    skip_if_not_installed("maxentcpp")
    skip_if_no_java()

    traj <- build_java_trajectory_2var()

    # Allow a tiny numerical tolerance for non-strict monotonicity.
    diffs <- diff(traj$loss)
    expect_true(all(diffs <= 1e-9),
                label = "Java trajectory loss is non-increasing at every checkpoint")
})

test_that("Group F: simple Java trajectory entropy is non-negative at all checkpoints (2-var)", {
    skip_if_not_installed("maxentcpp")
    skip_if_no_java()

    traj <- build_java_trajectory_2var()

    expect_true(all(traj$entropy >= 0),
                label = "Java trajectory entropy >= 0 at every checkpoint")
})

test_that("Group F: simple Java trajectory lambdas are finite at all checkpoints (2-var)", {
    skip_if_not_installed("maxentcpp")
    skip_if_no_java()

    traj <- build_java_trajectory_2var()

    expect_true(all(is.finite(traj$lambda_0)),
                label = "Java trajectory lambda_0 finite at every checkpoint")
    expect_true(all(is.finite(traj$lambda_1)),
                label = "Java trajectory lambda_1 finite at every checkpoint")
})

test_that("Group F: simple Java trajectory matches expected checkpoint values (2-var)", {
    skip_if_not_installed("maxentcpp")
    skip_if_no_java()

    checkpoints <- c(1L, 10L, 100L, 500L)
    traj <- build_java_trajectory_2var(checkpoints)

    cpp <- build_cpp_model_2var()

    # At convergence (checkpoint 500 or final), simple Java lambda should agree
    # with C++ maxent_fit() (goodAlpha loop) to within 1e-5 tolerance.
    final_row <- traj[traj$iteration == max(traj$iteration), ]
    expect_equal(final_row$lambda_0[1], cpp$fit$lambdas[[1]],
                 tolerance = 1e-5,
                 label = "simple Java final lambda_0",
                 expected.label = "C++ maxent_fit final lambda_0 (same algorithm)")
    expect_equal(final_row$lambda_1[1], cpp$fit$lambdas[[2]],
                 tolerance = 1e-5,
                 label = "simple Java final lambda_1",
                 expected.label = "C++ maxent_fit final lambda_1 (same algorithm)")
})

# ===========================================================================
# Helpers: build and run C++ sequential_fit on the asym fixture with L+Q / L+Q+P
# ===========================================================================

.run_cpp_asym_lq_trajectory <- function() {
    d <- asym_fixture()

    feat <- maxent_generate_features(
        list(bio1 = d$bio1, bio2 = d$bio2),
        types        = c("linear", "quadratic"),
        n_thresholds = 0L,
        n_hinges     = 0L
    )
    fs <- maxent_featured_space(d$n, d$sample_idx, feat)

    maxent_sequential_fit(
        fs,
        max_iter                  = 500L,
        convergence               = 0.0,
        beta_multiplier           = 0.8,
        min_deviation             = 0.001,
        disable_convergence_test  = TRUE,
        trajectory_iterations     = c(1L, 2L, 3L, 5L, 10L, 20L, 50L, 100L, 200L, 500L)
    )$trajectory
}

.run_cpp_asym_lqp_trajectory <- function() {
    d <- asym_fixture()

    feat <- maxent_generate_features(
        list(bio1 = d$bio1, bio2 = d$bio2),
        types        = c("linear", "quadratic", "product"),
        n_thresholds = 0L,
        n_hinges     = 0L
    )
    fs <- maxent_featured_space(d$n, d$sample_idx, feat)

    maxent_sequential_fit(
        fs,
        max_iter                  = 500L,
        convergence               = 0.0,
        beta_multiplier           = 1.6,
        min_deviation             = 0.001,
        disable_convergence_test  = TRUE,
        trajectory_iterations     = c(1L, 2L, 3L, 5L, 10L, 20L, 50L, 100L, 200L, 500L)
    )$trajectory
}

# ===========================================================================
# Group G — Phase D: C++ Sequential L+Q trajectory vs. committed golden
# ===========================================================================

test_that("Group G: C++ L+Q trajectory row count matches golden (asym fixture)", {
    skip_if_not_installed("maxentcpp")

    traj   <- .run_cpp_asym_lq_trajectory()
    golden <- asym_java_golden_lq()

    expect_equal(nrow(traj), nrow(golden),
                 label = "number of captured L+Q checkpoints")
})

test_that("Group G: C++ L+Q iteration labels match golden (asym fixture)", {
    skip_if_not_installed("maxentcpp")

    traj   <- .run_cpp_asym_lq_trajectory()
    golden <- asym_java_golden_lq()

    expect_equal(traj$iteration, golden$iteration,
                 label = "L+Q checkpoint iteration numbers")
})

test_that("Group G: C++ L+Q loss trajectory matches golden to 1e-6 (asym fixture)", {
    skip_if_not_installed("maxentcpp")

    traj   <- .run_cpp_asym_lq_trajectory()
    golden <- asym_java_golden_lq()

    expect_equal(traj$loss, golden$loss,
                 tolerance = 1e-6,
                 label = "C++ L+Q loss trajectory",
                 expected.label = "Phase D golden loss (L+Q)")
})

test_that("Group G: C++ L+Q entropy trajectory matches golden to 1e-6 (asym fixture)", {
    skip_if_not_installed("maxentcpp")

    traj   <- .run_cpp_asym_lq_trajectory()
    golden <- asym_java_golden_lq()

    expect_equal(traj$entropy, golden$entropy,
                 tolerance = 1e-6,
                 label = "C++ L+Q entropy trajectory",
                 expected.label = "Phase D golden entropy (L+Q)")
})

test_that("Group G: C++ L+Q lambda trajectories match golden to 1e-6 (asym fixture)", {
    skip_if_not_installed("maxentcpp")

    traj   <- .run_cpp_asym_lq_trajectory()
    golden <- asym_java_golden_lq()

    worst <- max(
        abs(traj$lambda_0 - golden$lambda_0),
        abs(traj$lambda_1 - golden$lambda_1),
        abs(traj$lambda_2 - golden$lambda_2),
        abs(traj$lambda_3 - golden$lambda_3)
    )
    expect_lt(worst, 1e-6,
              label = "worst ||delta_lambda||_inf across all L+Q checkpoints")
})

# ===========================================================================
# Group H — Phase D: C++ Sequential L+Q+P trajectory vs. committed golden
# ===========================================================================

test_that("Group H: C++ L+Q+P trajectory row count matches golden (asym fixture)", {
    skip_if_not_installed("maxentcpp")

    traj   <- .run_cpp_asym_lqp_trajectory()
    golden <- asym_java_golden_lqp()

    expect_equal(nrow(traj), nrow(golden),
                 label = "number of captured L+Q+P checkpoints")
})

test_that("Group H: C++ L+Q+P iteration labels match golden (asym fixture)", {
    skip_if_not_installed("maxentcpp")

    traj   <- .run_cpp_asym_lqp_trajectory()
    golden <- asym_java_golden_lqp()

    expect_equal(traj$iteration, golden$iteration,
                 label = "L+Q+P checkpoint iteration numbers")
})

test_that("Group H: C++ L+Q+P loss trajectory matches golden to 1e-6 (asym fixture)", {
    skip_if_not_installed("maxentcpp")

    traj   <- .run_cpp_asym_lqp_trajectory()
    golden <- asym_java_golden_lqp()

    expect_equal(traj$loss, golden$loss,
                 tolerance = 1e-6,
                 label = "C++ L+Q+P loss trajectory",
                 expected.label = "Phase D golden loss (L+Q+P)")
})

test_that("Group H: C++ L+Q+P entropy trajectory matches golden to 1e-6 (asym fixture)", {
    skip_if_not_installed("maxentcpp")

    traj   <- .run_cpp_asym_lqp_trajectory()
    golden <- asym_java_golden_lqp()

    expect_equal(traj$entropy, golden$entropy,
                 tolerance = 1e-6,
                 label = "C++ L+Q+P entropy trajectory",
                 expected.label = "Phase D golden entropy (L+Q+P)")
})

test_that("Group H: C++ L+Q+P lambda trajectories match golden to 1e-6 (asym fixture)", {
    skip_if_not_installed("maxentcpp")

    traj   <- .run_cpp_asym_lqp_trajectory()
    golden <- asym_java_golden_lqp()

    worst <- max(
        abs(traj$lambda_0 - golden$lambda_0),
        abs(traj$lambda_1 - golden$lambda_1),
        abs(traj$lambda_2 - golden$lambda_2),
        abs(traj$lambda_3 - golden$lambda_3),
        abs(traj$lambda_4 - golden$lambda_4)
    )
    expect_lt(worst, 1e-6,
              label = "worst ||delta_lambda||_inf across all L+Q+P checkpoints")
})

# ===========================================================================
# Group I — Phase D: simplified Java L+Q / L+Q+P trajectory self-consistency
# (Requires rJava)
# ===========================================================================

test_that("Group I: simplified Java L+Q trajectory has correct checkpoints", {
    skip_if_not_installed("maxentcpp")
    skip_if_no_java()

    checkpoints <- c(1L, 2L, 3L, 5L, 10L, 20L, 50L, 100L, 200L, 500L)
    traj <- build_java_trajectory_asym_lq(checkpoints)

    expect_equal(nrow(traj), length(checkpoints),
                 label = "number of Java L+Q trajectory rows")
    expect_equal(traj$iteration, checkpoints,
                 label = "Java L+Q trajectory iteration labels")
})

test_that("Group I: simplified Java L+Q trajectory loss is non-increasing", {
    skip_if_not_installed("maxentcpp")
    skip_if_no_java()

    traj <- build_java_trajectory_asym_lq()

    diffs <- diff(traj$loss)
    expect_true(all(diffs <= 1e-9),
                label = "simplified Java L+Q loss non-increasing at every checkpoint")
})

test_that("Group I: simplified Java L+Q trajectory entropy is non-negative", {
    skip_if_not_installed("maxentcpp")
    skip_if_no_java()

    traj <- build_java_trajectory_asym_lq()

    expect_true(all(traj$entropy >= 0),
                label = "simplified Java L+Q entropy >= 0 at every checkpoint")
})

test_that("Group I: simplified Java L+Q trajectory lambdas are finite", {
    skip_if_not_installed("maxentcpp")
    skip_if_no_java()

    traj <- build_java_trajectory_asym_lq()

    expect_true(all(is.finite(traj$lambda_0)), label = "Java L+Q lambda_0 finite")
    expect_true(all(is.finite(traj$lambda_1)), label = "Java L+Q lambda_1 finite")
    expect_true(all(is.finite(traj$lambda_2)), label = "Java L+Q lambda_2 finite")
    expect_true(all(is.finite(traj$lambda_3)), label = "Java L+Q lambda_3 finite")
})

test_that("Group I: simplified Java L+Q+P trajectory has correct checkpoints", {
    skip_if_not_installed("maxentcpp")
    skip_if_no_java()

    checkpoints <- c(1L, 2L, 3L, 5L, 10L, 20L, 50L, 100L, 200L, 500L)
    traj <- build_java_trajectory_asym_lqp(checkpoints)

    expect_equal(nrow(traj), length(checkpoints),
                 label = "number of Java L+Q+P trajectory rows")
    expect_equal(traj$iteration, checkpoints,
                 label = "Java L+Q+P trajectory iteration labels")
})

test_that("Group I: simplified Java L+Q+P trajectory loss is non-increasing", {
    skip_if_not_installed("maxentcpp")
    skip_if_no_java()

    traj <- build_java_trajectory_asym_lqp()

    diffs <- diff(traj$loss)
    expect_true(all(diffs <= 1e-9),
                label = "simplified Java L+Q+P loss non-increasing at every checkpoint")
})

test_that("Group I: simplified Java L+Q+P trajectory entropy is non-negative", {
    skip_if_not_installed("maxentcpp")
    skip_if_no_java()

    traj <- build_java_trajectory_asym_lqp()

    expect_true(all(traj$entropy >= 0),
                label = "simplified Java L+Q+P entropy >= 0 at every checkpoint")
})

test_that("Group I: simplified Java L+Q+P trajectory lambdas are finite", {
    skip_if_not_installed("maxentcpp")
    skip_if_no_java()

    traj <- build_java_trajectory_asym_lqp()

    expect_true(all(is.finite(traj$lambda_0)), label = "Java L+Q+P lambda_0 finite")
    expect_true(all(is.finite(traj$lambda_1)), label = "Java L+Q+P lambda_1 finite")
    expect_true(all(is.finite(traj$lambda_2)), label = "Java L+Q+P lambda_2 finite")
    expect_true(all(is.finite(traj$lambda_3)), label = "Java L+Q+P lambda_3 finite")
    expect_true(all(is.finite(traj$lambda_4)), label = "Java L+Q+P lambda_4 finite")
})

#' Run a Complete Maxent Species Distribution Modelling Workflow
#'
#' Provides a single high-level entry point that mirrors running the Java
#' Maxent GUI with one click.  Starting from raw environmental grids and
#' occurrence records, the function:
#' \enumerate{
#'   \item Reads and validates occurrence records.
#'   \item Samples background points.
#'   \item Extracts environmental values and builds features.
#'   \item Trains the Maxent model.
#'   \item Projects the model onto the full landscape (cloglog output).
#'   \item Evaluates model performance (AUC).
#'   \item Computes percent contribution and permutation importance.
#'   \item Writes all standard output files (lambdas, prediction PNG, response
#'     curve PNGs, omission CSV, sample-predictions CSV, and
#'     \code{maxentResults.csv}).
#'   \item Prints a performance summary to the console.
#' }
#'
#' @param species       Character: species name (used in file names and the
#'   console report).
#' @param env_grids     Named list of external pointers to Grid<float> objects.
#'   The \emph{names} of the list are used as the environmental variable names.
#' @param occ_df        A \code{data.frame} (or CSV path) with occurrence
#'   records.
#' @param output_dir    Character: directory where all output files are written
#'   (created if it does not exist).
#' @param lon_col       Character: longitude column name in \code{occ_df}
#'   (default \code{"long"}).
#' @param lat_col       Character: latitude column name in \code{occ_df}
#'   (default \code{"lat"}).
#' @param n_background  Integer: number of background points (default 10000).
#' @param types         Character vector of feature types passed to
#'   \code{\link{maxent_generate_features}} (default
#'   \code{c("linear", "quadratic", "hinge")}).
#' @param n_hinges      Integer: number of hinge knots (default 15).
#' @param max_iter      Integer: maximum training iterations (default 500).
#' @param seed          Integer: random seed for background sampling
#'   (default 42).
#' @param bias_weights  Optional numeric vector of per-background-point bias
#'   weights.  Must have length equal to \code{n_background} (the background
#'   points only, not including occurrence points).  When supplied, the
#'   background density is weighted by \code{bias[i] * exp(lp[i] - lpn)},
#'   mirroring Java Maxent's \code{biasFile}.  Pass \code{NULL} (default)
#'   for uniform (unbiased) background.
#' @param response_curves Logical: write response curve PNGs
#'   (default \code{TRUE}).
#' @param pictures      Logical: write prediction map PNG (default \code{TRUE}).
#' @return A named list with:
#'   \describe{
#'     \item{model}{Trained FeaturedSpace external pointer.}
#'     \item{fit_result}{List returned by \code{\link{maxent_fit}}.}
#'     \item{evaluation}{List returned by \code{\link{maxent_evaluate}}.}
#'     \item{contributions}{Data.frame of percent contributions.}
#'     \item{permutation_importance}{Data.frame of permutation importances.}
#'     \item{output_dir}{The output directory used.}
#'   }
#' @export
#' @examples
#' \donttest{
#' stack_path      <- system.file("extdata", "stack_1_12_crop.rds",
#'                                package = "maxentcpp")
#' example_rasters <- terra::unwrap(readRDS(stack_path))
#' grids <- list(
#'   bio1  = maxent_grid_from_terra(example_rasters[[1]]),
#'   bio12 = maxent_grid_from_terra(example_rasters[[2]])
#' )
#' data(example_occ_df)
#'
#' result <- maxent_run(
#'   species    = "Abeillia_abeillei",
#'   env_grids  = grids,
#'   occ_df     = example_occ_df,
#'   output_dir = tempdir(),
#'   lon_col    = "long",
#'   lat_col    = "lat")
#'
#' result$evaluation$auc
#' }
maxent_run <- function(species,
                       env_grids,
                       occ_df,
                       output_dir,
                       lon_col        = "long",
                       lat_col        = "lat",
                       n_background   = 10000L,
                       types          = c("linear", "quadratic", "hinge"),
                       n_hinges       = 15L,
                       max_iter       = 500L,
                       seed           = 42L,
                       bias_weights   = NULL,
                       response_curves = TRUE,
                       pictures        = TRUE) {

    if (!is.list(env_grids) || is.null(names(env_grids))) {
        stop("'env_grids' must be a named list of Grid objects")
    }
    feature_names <- names(env_grids)

    if (!dir.exists(output_dir)) dir.create(output_dir, recursive = TRUE)

    # --- 1. Read occurrences --------------------------------------------------
    ref_grid <- env_grids[[1]]
    info     <- maxent_grid_info(ref_grid)
    dim_obj  <- maxent_dimension(nrows    = info$nrows,
                                 ncols    = info$ncols,
                                 xll      = info$xll,
                                 yll      = info$yll,
                                 cellsize = info$cellsize)

    occ <- maxent_read_occurrences(occ_df, dim_obj,
                                   lon_col = lon_col,
                                   lat_col = lat_col)

    # --- 2. Background points -------------------------------------------------
    bg <- maxent_background_indices(ref_grid,
                                    n    = as.integer(n_background),
                                    seed = as.integer(seed))

    # --- 3. Extract environmental values and build features ------------------
    all_rows <- c(bg$rows, occ$rows)
    all_cols <- c(bg$cols, occ$cols)
    n_total  <- length(all_rows)

    sample_indices <- seq(length(bg$rows), n_total - 1L)

    int_rows <- as.integer(all_rows)
    int_cols <- as.integer(all_cols)

    env_vals <- lapply(env_grids, function(g) {
        grid_get_values_batch(g, int_rows, int_cols)
    })
    names(env_vals) <- feature_names

    features <- maxent_generate_features(env_vals,
                                         types    = as.character(types),
                                         n_hinges = as.integer(n_hinges))

    # --- 4. Train -------------------------------------------------------------
    # Extend per-background bias_weights to the full (bg + occ) array.
    # Background bias is user-supplied; occurrence points get weight 1.0.
    full_bias <- NULL
    if (!is.null(bias_weights)) {
        n_bg <- length(bg$rows)
        if (length(bias_weights) != n_bg) {
            stop(sprintf(
                "bias_weights must have length %d (background points only, excluding occurrences), got %d",
                n_bg, length(bias_weights)))
        }
        n_occ <- length(occ$rows)
        full_bias <- c(as.numeric(bias_weights), rep(1.0, n_occ))
    }
    fs         <- maxent_featured_space(n_total,
                                        as.integer(sample_indices),
                                        features,
                                        bias_weights = full_bias)
    fit_result <- maxent_fit(fs,
                             max_iter = as.integer(max_iter),
                             convergence = 1e-5)

    # --- 5. Evaluate ----------------------------------------------------------
    pres_preds <- maxent_extract_predictions_raw(fs, unname(env_grids),
                                                  feature_names,
                                                  occ$rows, occ$cols)
    bg_preds   <- maxent_extract_predictions_raw(fs, unname(env_grids),
                                                  feature_names,
                                                  bg$rows, bg$cols)
    eval_result <- maxent_evaluate(pres_preds, bg_preds)

    # --- 6. Diagnostics -------------------------------------------------------
    contrib  <- maxent_percent_contribution(fs, feature_names)
    perm_imp <- maxent_permutation_importance(
        fs, unname(env_grids), feature_names,
        occ$rows, occ$cols,
        bg$rows,  bg$cols,
        seed = as.integer(seed))

    # --- 7. Print performance summary to console -----------------------------
    maxent_print_results(
        species          = species,
        eval_result      = eval_result,
        contributions_df = contrib,
        perm_imp_df      = perm_imp,
        n_presence       = length(occ$rows),
        n_background     = length(bg$rows),
        fit_result       = fit_result)

    # --- 8. Append to maxentResults.csv --------------------------------------
    results_csv <- file.path(output_dir, "maxentResults.csv")
    maxent_append_results_csv(
        results_file     = results_csv,
        species          = species,
        n_training       = length(occ$rows),
        training_gain    = fit_result$loss,
        training_auc     = eval_result$auc,
        entropy          = fit_result$entropy,
        contributions_df = contrib,
        perm_imp_df      = perm_imp)

    list(
        model                  = fs,
        fit_result             = fit_result,
        evaluation             = eval_result,
        contributions          = contrib,
        permutation_importance = perm_imp,
        output_dir             = output_dir
    )
}

# ---- Stage 2: Prediction PNG writer -----------------------------------------

#' Write a Maxent Prediction Grid as a PNG Image
#'
#' Converts a \code{GridFloat} prediction grid to a colour PNG image using
#' the canonical Maxent colour ramp (red = high, blue = low).  Optionally
#' overlays presence and test-point locations, and renders a small legend.
#'
#' @param grid          External pointer to a GridFloat prediction grid (e.g.
#'   from \code{\link{maxent_project_cloglog}}).
#' @param filename      Character: path for the output PNG file.
#' @param presence_rows Integer vector of presence row indices (0-based) or
#'   \code{NULL} (default).
#' @param presence_cols Integer vector of presence column indices (0-based) or
#'   \code{NULL} (default).
#' @param test_rows     Integer vector of test-set row indices (0-based) or
#'   \code{NULL} (default).
#' @param test_cols     Integer vector of test-set column indices (0-based) or
#'   \code{NULL} (default).
#' @param mode          Colour mode passed to \code{\link{maxent_color_ramp}}:
#'   one of \code{"plain"} (default), \code{"log"}, \code{"blackandwhite"},
#'   or \code{"redandyellow"}.
#' @param legend        Logical: draw a colour-bar legend (default \code{TRUE}).
#' @param width         Integer: PNG width in pixels (default 800).
#' @param height        Integer: PNG height in pixels (default 600).
#' @return Invisibly returns \code{filename}.
#' @export
#' @examples
#' \donttest{
#' pred <- maxent_project_cloglog(model, list(g1, g2), c("bio1", "bio12"))
#' maxent_write_prediction_png(pred, tempfile(fileext = ".png"))
#' }
maxent_write_prediction_png <- function(grid, filename,
                                        presence_rows = NULL,
                                        presence_cols = NULL,
                                        test_rows     = NULL,
                                        test_cols     = NULL,
                                        mode          = "plain",
                                        legend        = TRUE,
                                        width         = 800L,
                                        height        = 600L) {
    mat <- maxent_grid_to_matrix(grid)
    if (is.null(mat)) stop("Could not extract matrix from grid")

    nrows <- nrow(mat)
    ncols <- ncol(mat)
    n_pal <- 1020L
    pal   <- maxent_color_ramp(n_pal, mode = mode)

    val_min <- min(mat, na.rm = TRUE)
    val_max <- max(mat, na.rm = TRUE)

    # Map values to palette indices (1..n_pal): high values → index 1 (red),
    # low values → index n_pal (blue).
    if (val_max > val_min) {
        idx <- round((val_max - mat) / (val_max - val_min) * (n_pal - 1L)) + 1L
    } else {
        idx <- matrix(1L, nrow = nrows, ncol = ncols)
    }
    idx[is.na(mat)] <- NA_integer_

    # Build RGB arrays (dims: nrows x ncols x 3, values in [0,1])
    hex_to_rgb01 <- function(hex) {
        m <- grDevices::col2rgb(hex) / 255
        m  # 3 x length(hex)
    }
    rgb_mat <- hex_to_rgb01(pal)   # 3 x n_pal

    red_layer   <- matrix(0.85, nrow = nrows, ncol = ncols)  # grey for NODATA
    green_layer <- matrix(0.85, nrow = nrows, ncol = ncols)
    blue_layer  <- matrix(0.85, nrow = nrows, ncol = ncols)

    valid <- !is.na(idx)
    idx_v <- idx[valid]

    red_layer[valid]   <- rgb_mat[1, idx_v]
    green_layer[valid] <- rgb_mat[2, idx_v]
    blue_layer[valid]  <- rgb_mat[3, idx_v]

    # Assemble raster array (rows from top → flip)
    img <- array(NA_real_, dim = c(nrows, ncols, 3))
    img[, , 1] <- red_layer[nrows:1, ]
    img[, , 2] <- green_layer[nrows:1, ]
    img[, , 3] <- blue_layer[nrows:1, ]

    grDevices::png(filename, width = as.integer(width),
                   height = as.integer(height))
    oldpar <- graphics::par(mar = c(0, 0, 0, 0))
    on.exit({graphics::par(oldpar); grDevices::dev.off()}, add = TRUE)
    graphics::plot.new()
    graphics::plot.window(xlim = c(0, ncols), ylim = c(0, nrows),
                          xaxs = "i", yaxs = "i")
    graphics::rasterImage(img, 0, 0, ncols, nrows)

    # Presence points (white dots)
    if (!is.null(presence_rows) && length(presence_rows) > 0) {
        px <- as.integer(presence_cols) + 0.5
        py <- as.integer(presence_rows) + 0.5
        graphics::points(px, py, pch = 21, bg = "white",
                         col = "black", cex = 0.8, lwd = 0.5)
    }

    # Test points (violet dots)
    if (!is.null(test_rows) && length(test_rows) > 0) {
        px <- as.integer(test_cols) + 0.5
        py <- as.integer(test_rows) + 0.5
        graphics::points(px, py, pch = 21, bg = "#EE82EE",
                         col = "black", cex = 0.8, lwd = 0.5)
    }

    # Legend strip (right side)
    if (legend) {
        leg_x <- ncols * 0.93
        leg_y0 <- nrows * 0.10
        leg_y1 <- nrows * 0.90
        leg_w  <- ncols * 0.04
        n_leg  <- 100L
        leg_pal <- maxent_color_ramp(n_leg, mode = mode)
        for (i in seq_len(n_leg)) {
            yb <- leg_y0 + (i - 1L) * (leg_y1 - leg_y0) / n_leg
            yt <- leg_y0 + i         * (leg_y1 - leg_y0) / n_leg
            graphics::rect(leg_x, yb, leg_x + leg_w, yt,
                           col = leg_pal[n_leg - i + 1L], border = NA)
        }
        graphics::rect(leg_x, leg_y0, leg_x + leg_w, leg_y1, border = "black",
                       lwd = 0.8)
        graphics::text(leg_x - leg_w * 0.3, leg_y1,
                       sprintf("%.2f", val_max), adj = 1, cex = 0.7)
        graphics::text(leg_x - leg_w * 0.3, leg_y0,
                       sprintf("%.2f", val_min), adj = 1, cex = 0.7)
    }

    invisible(filename)
}


# ---- Stage 5: Omission / Threshold CSV --------------------------------------

#' Write an Omission / Threshold CSV
#'
#' Computes cumulative predictions at all background + presence locations and
#' derives nine standard threshold statistics, writing the result to
#' \code{<output_dir>/<species>_omission.csv}.  The output format mirrors
#' \code{density/Runner.java::writeCumulativeIndex()}.
#'
#' @param model         External pointer to a trained FeaturedSpace object.
#' @param env_grids     List of external pointers to Grid<float> objects.
#' @param feature_names Character vector of environment variable names.
#' @param presence_rows Integer vector of presence row indices (0-based).
#' @param presence_cols Integer vector of presence column indices (0-based).
#' @param output_dir    Character: directory for output files.
#' @param species       Character: species name (used in the filename).
#' @param test_rows     Integer vector of test row indices (0-based) or
#'   \code{NULL} (default).
#' @param test_cols     Integer vector of test column indices (0-based) or
#'   \code{NULL} (default).
#' @return Invisibly returns the path to the written CSV file.
#' @export
#' @examples
#' \donttest{
#' maxent_write_omission_csv(model, list(g1, g2), c("bio1", "bio12"),
#'   pres_rows, pres_cols, output_dir = tempdir(), species = "Sp1")
#' }
maxent_write_omission_csv <- function(model, env_grids, feature_names,
                                      presence_rows, presence_cols,
                                      output_dir, species,
                                      test_rows = NULL, test_cols = NULL) {
    if (!dir.exists(output_dir)) dir.create(output_dir, recursive = TRUE)

    feature_names <- as.character(feature_names)
    presence_rows <- as.integer(presence_rows)
    presence_cols <- as.integer(presence_cols)

    # Model parameters for Java-compatible transforms
    H    <- maxent_model_entropy(model)
    expH <- exp(H)

    # Java-compatible cloglog from a raw_java score (as returned by
    # maxent_project_raw()/maxent_extract_predictions_raw(), which both
    # route through Projection::project_raw_java and are thus already
    # divided by densityNormalizer).  No further /dn division here.
    raw_java_to_cloglog_java <- function(raw_java) {
        1.0 - exp(-expH * raw_java)
    }

    # Project full grid for cumulative thresholds
    proj_grid <- maxent_project_raw(model, env_grids, feature_names)
    bg_mat    <- maxent_grid_to_matrix(proj_grid)
    all_raw   <- as.numeric(bg_mat[!is.na(bg_mat)])

    # Sort background to build cumulative index
    all_sorted <- sort(all_raw)
    n_bg       <- length(all_sorted)

    # Cumulative value for a given raw score = fraction of background <= score
    raw_to_cum <- function(raw_vals) {
        vapply(raw_vals, function(v) {
            mean(all_sorted <= v)
        }, numeric(1L))
    }

    pres_raw <- maxent_extract_predictions_raw(model, env_grids, feature_names,
                                           presence_rows, presence_cols)
    pres_cum <- raw_to_cum(pres_raw)

    has_test <- !is.null(test_rows) && length(test_rows) > 0
    if (has_test) {
        test_raw <- maxent_extract_predictions_raw(model, env_grids, feature_names,
                                               as.integer(test_rows),
                                               as.integer(test_cols))
        test_cum <- raw_to_cum(test_raw)
    }

    # ---- Threshold calculations ----------------------------------------------

    cum_thresholds <- seq(0, 1, length.out = n_bg)   # cumulative grid values

    # Helper: omission rate at cumulative threshold t
    omit_train <- function(t) mean(pres_cum < t)
    omit_test  <- function(t) {
        if (!has_test) return(NA_real_)
        mean(test_cum < t)
    }

    # Fixed cumulative thresholds 1, 5, 10
    fixed_thresholds <- c(0.01, 0.05, 0.10)
    fixed_names      <- c("Fixed cumulative value 1",
                          "Fixed cumulative value 5",
                          "Fixed cumulative value 10")

    # Minimum training presence
    min_train_cum <- min(pres_cum)

    # 10th percentile training presence
    pct10_cum <- stats::quantile(pres_cum, 0.10, names = FALSE)

    # Helper: find threshold maximising (sensitivity + specificity) or
    # minimising |sensitivity - (1 - specificity)|
    find_equal_ss <- function(preds_cum, label) {
        # sensitivity = TP/(TP+FN), specificity = TN/(TN+FP)
        # Over bg: all bg as negatives, preds_cum as positives
        ts <- sort(unique(c(preds_cum, cum_thresholds)))
        best_t <- ts[1]
        best_d <- Inf
        for (t in ts) {
            sens <- mean(preds_cum >= t)
            spec <- mean(all_sorted / max(all_sorted) < t)
            d <- abs(sens - spec)
            if (d < best_d) { best_d <- d; best_t <- t }
        }
        best_t
    }

    find_max_ss <- function(preds_cum) {
        ts <- sort(unique(c(preds_cum, cum_thresholds)))
        best_t <- ts[1]
        best_s <- -Inf
        for (t in ts) {
            sens <- mean(preds_cum >= t)
            spec <- mean(all_sorted / max(all_sorted) < t)
            s    <- sens + spec
            if (s > best_s) { best_s <- s; best_t <- t }
        }
        best_t
    }

    t_equal_train <- find_equal_ss(pres_cum)
    t_max_train   <- find_max_ss(pres_cum)

    # Build row table
    threshold_names <- c(
        fixed_names,
        "Minimum training presence",
        "10 percentile training presence",
        "Equal training sensitivity and specificity",
        "Maximum training sensitivity plus specificity"
    )
    threshold_values <- c(
        fixed_thresholds,
        min_train_cum,
        pct10_cum,
        t_equal_train,
        t_max_train
    )

    if (has_test) {
        t_equal_test <- find_equal_ss(test_cum)
        t_max_test   <- find_max_ss(test_cum)
        threshold_names  <- c(threshold_names,
                               "Equal test sensitivity and specificity",
                               "Maximum test sensitivity plus specificity")
        threshold_values <- c(threshold_values, t_equal_test, t_max_test)
    }

    n_thresh <- length(threshold_names)
    omission_train <- vapply(threshold_values, omit_train, numeric(1L))
    omission_test  <- vapply(threshold_values, omit_test,  numeric(1L))
    bg_predicted   <- vapply(threshold_values, function(t) {
        mean(all_sorted >= t * max(all_sorted))
    }, numeric(1L))

    # Compute cloglog threshold for each cumulative threshold
    # The raw value at cumulative fraction t is the t-th quantile of all_sorted
    cloglog_thresholds <- vapply(threshold_values, function(t) {
        raw_java_at_t <- stats::quantile(all_sorted, t, names = FALSE)
        raw_java_to_cloglog_java(raw_java_at_t)
    }, numeric(1L))

    out <- data.frame(
        Threshold                    = threshold_names,
        `Cumulative threshold`       = threshold_values,
        `Cloglog threshold`          = cloglog_thresholds,
        `Area`                       = bg_predicted,
        `Training omission`          = omission_train,
        `Test omission`              = omission_test,
        stringsAsFactors             = FALSE,
        check.names                  = FALSE
    )

    csv_path <- file.path(output_dir, paste0(species, "_omission.csv"))
    utils::write.csv(out, csv_path, row.names = FALSE)
    invisible(csv_path)
}


# ---- Stage 6: Sample Predictions CSV ----------------------------------------

#' Write Sample Predictions CSV
#'
#' Computes model predictions at presence (and optionally test) locations and
#' writes \code{<output_dir>/<species>_samplePredictions.csv}.
#'
#' @param model         External pointer to a trained FeaturedSpace object.
#' @param env_grids     List of external pointers to Grid<float> objects.
#' @param feature_names Character vector of environment variable names.
#' @param presence_rows Integer vector of presence row indices (0-based).
#' @param presence_cols Integer vector of presence column indices (0-based).
#' @param output_dir    Character: directory for the output file.
#' @param species       Character: species name (used in the filename).
#' @param test_rows     Integer vector of test row indices (0-based) or
#'   \code{NULL} (default).
#' @param test_cols     Integer vector of test column indices (0-based) or
#'   \code{NULL} (default).
#' @return Invisibly returns the path to the written CSV.
#' @export
#' @examples
#' \donttest{
#' maxent_write_sample_predictions(model, list(g1, g2), c("bio1", "bio12"),
#'   pres_rows, pres_cols, output_dir = tempdir(), species = "Sp1")
#' }
maxent_write_sample_predictions <- function(model, env_grids, feature_names,
                                            presence_rows, presence_cols,
                                            output_dir, species,
                                            test_rows = NULL,
                                            test_cols = NULL) {
    if (!dir.exists(output_dir)) dir.create(output_dir, recursive = TRUE)

    feature_names <- as.character(feature_names)
    presence_rows <- as.integer(presence_rows)
    presence_cols <- as.integer(presence_cols)

    pres_raw <- maxent_extract_predictions_raw(model, env_grids, feature_names,
                                           presence_rows, presence_cols)

    # Java-compatible cloglog: 1 - exp(-exp(H) * raw_java).  Note that
    # maxent_extract_predictions_raw() already returns raw_java (scores
    # are produced via predict_raw_java_from_env, which divides by
    # densityNormalizer), so the cloglog formula is applied directly.
    H    <- maxent_model_entropy(model)
    expH <- exp(H)

    raw_java_to_cloglog_java <- function(raw_java) {
        1.0 - exp(-expH * raw_java)
    }

    pres_cl  <- raw_java_to_cloglog_java(pres_raw)

    set_label <- rep("train", length(presence_rows))

    rows_all <- presence_rows
    cols_all <- presence_cols
    raw_all  <- pres_raw
    cl_all   <- pres_cl
    set_all  <- set_label

    if (!is.null(test_rows) && length(test_rows) > 0) {
        test_rows <- as.integer(test_rows)
        test_cols <- as.integer(test_cols)
        t_raw <- maxent_extract_predictions_raw(model, env_grids, feature_names,
                                            test_rows, test_cols)
        t_cl  <- raw_java_to_cloglog_java(t_raw)
        rows_all <- c(rows_all, test_rows)
        cols_all <- c(cols_all, test_cols)
        raw_all  <- c(raw_all, t_raw)
        cl_all   <- c(cl_all,  t_cl)
        set_all  <- c(set_all, rep("test", length(test_rows)))
    }

    out <- data.frame(
        species    = species,
        row        = rows_all,
        col        = cols_all,
        `set`      = set_all,
        raw        = raw_all,
        cloglog    = cl_all,
        stringsAsFactors = FALSE,
        check.names      = FALSE
    )

    csv_path <- file.path(output_dir,
                          paste0(species, "_samplePredictions.csv"))
    utils::write.csv(out, csv_path, row.names = FALSE)
    invisible(csv_path)
}


# ---- Stage 8: maxentResults.csv writer --------------------------------------

#' Append a Row to maxentResults.csv
#'
#' Appends one row of species-level summary statistics to a
#' \code{maxentResults.csv} file, creating the file (with a header) if it does
#' not yet exist.  Column names match the Java Maxent output for
#' cross-tool compatibility.
#'
#' @param results_file    Character: path to the results CSV file.
#' @param species         Character: species name.
#' @param n_training      Integer: number of training presence points.
#' @param n_test          Integer: number of test presence points (default 0).
#' @param training_gain   Numeric: regularized training gain.
#' @param training_auc    Numeric: training AUC.
#' @param test_gain       Numeric or \code{NA}: regularized test gain.
#' @param test_auc        Numeric or \code{NA}: test AUC.
#' @param entropy         Numeric: model entropy.
#' @param contributions_df A data.frame with columns \code{name} and
#'   \code{contribution} (from \code{\link{maxent_percent_contribution}}).
#' @param perm_imp_df     A data.frame with columns \code{name} and
#'   \code{permutation_importance} (from
#'   \code{\link{maxent_permutation_importance}}).
#' @return Invisibly returns \code{results_file}.
#' @export
#' @examples
#' \donttest{
#' maxent_append_results_csv(
#'   file.path(tempdir(), "maxentResults.csv"),
#'   species = "Sp1", n_training = 50L, training_gain = 1.23,
#'   training_auc = 0.95, entropy = 6.7,
#'   contributions_df = contrib, perm_imp_df = perm_imp)
#' }
maxent_append_results_csv <- function(results_file,
                                      species,
                                      n_training,
                                      n_test          = 0L,
                                      training_gain,
                                      training_auc,
                                      test_gain       = NA_real_,
                                      test_auc        = NA_real_,
                                      entropy,
                                      contributions_df,
                                      perm_imp_df) {
    # Base columns (Java column order)
    row <- data.frame(
        Species                    = as.character(species),
        `X..Training.samples`      = as.integer(n_training),
        `X..Test.samples`          = as.integer(n_test),
        `Regularized.training.gain`= as.numeric(training_gain),
        `Regularized.test.gain`    = as.numeric(test_gain),
        `Training.AUC`             = as.numeric(training_auc),
        `Test.AUC`                 = as.numeric(test_auc),
        Entropy                    = as.numeric(entropy),
        stringsAsFactors           = FALSE,
        check.names                = FALSE
    )

    # Contribution columns: "<name> contribution"
    if (!is.null(contributions_df) && is.data.frame(contributions_df)) {
        for (i in seq_len(nrow(contributions_df))) {
            nm <- paste0(contributions_df$name[i], " contribution")
            row[[nm]] <- as.numeric(contributions_df$contribution[i])
        }
    }

    # Permutation importance columns: "<name> permutation importance"
    if (!is.null(perm_imp_df) && is.data.frame(perm_imp_df)) {
        for (i in seq_len(nrow(perm_imp_df))) {
            nm <- paste0(perm_imp_df$name[i], " permutation importance")
            row[[nm]] <- as.numeric(perm_imp_df$permutation_importance[i])
        }
    }

    if (file.exists(results_file)) {
        existing <- utils::read.csv(results_file,
                                    stringsAsFactors = FALSE,
                                    check.names      = FALSE)
        # Align columns
        all_cols <- union(names(existing), names(row))
        for (col in setdiff(all_cols, names(existing))) existing[[col]] <- NA
        for (col in setdiff(all_cols, names(row)))     row[[col]]      <- NA
        combined <- rbind(existing[, all_cols], row[, all_cols])
        utils::write.csv(combined, results_file, row.names = FALSE)
    } else {
        utils::write.csv(row, results_file, row.names = FALSE)
    }

    invisible(results_file)
}

## Helpers to build a minimal trained model for output tests
.make_test_model <- function() {
    skip_if_not_installed("maxentcpp")
    dim_obj <- maxent_dimension(nrows = 4, ncols = 4, xll = 0, yll = 0,
                                cellsize = 1)
    g0 <- maxent_grid(dim_obj, "env0")
    g1 <- maxent_grid(dim_obj, "env1")
    idx <- 0L
    for (r in 0:3) {
        for (cc in 0:3) {
            grid_set_value(g0, r, cc, idx / 15.0)
            grid_set_value(g1, r, cc, 1.0 - idx / 15.0)
            idx <- idx + 1L
        }
    }

    n   <- 16L
    vals0 <- sapply(0:15, function(i) i / 15.0)
    vals1 <- 1.0 - vals0

    features <- maxent_generate_features(
        list(env0 = vals0, env1 = vals1),
        types = "linear", n_thresholds = 0, n_hinges = 0)

    # Background = first 12, presence = last 4
    pres_idx <- 12:15
    model    <- maxent_featured_space(n, as.integer(pres_idx), features)
    maxent_fit(model, max_iter = 200, convergence = 1e-4,
               beta_multiplier = 1.0)

    pres_rows <- as.integer(pres_idx %/% 4L)
    pres_cols <- as.integer(pres_idx  %%  4L)

    list(model = model, g0 = g0, g1 = g1,
         pres_rows = pres_rows, pres_cols = pres_cols)
}

## ---- Prediction PNG ---------------------------------------------------------

test_that("maxent_write_prediction_png creates a non-empty file", {
    skip_if_not_installed("maxentcpp")
    m <- .make_test_model()
    pred <- maxent_project_cloglog(m$model, list(m$g0, m$g1),
                                   c("env0", "env1"))
    out  <- file.path(tempdir(), "test_pred.png")
    maxent_write_prediction_png(pred, out)
    expect_true(file.exists(out))
    expect_gt(file.size(out), 0L)
    unlink(out)
})

test_that("maxent_write_prediction_png with presence points", {
    skip_if_not_installed("maxentcpp")
    m <- .make_test_model()
    pred <- maxent_project_cloglog(m$model, list(m$g0, m$g1),
                                   c("env0", "env1"))
    out  <- file.path(tempdir(), "test_pred_pres.png")
    maxent_write_prediction_png(pred, out,
                                presence_rows = m$pres_rows,
                                presence_cols = m$pres_cols)
    expect_true(file.exists(out))
    expect_gt(file.size(out), 0L)
    unlink(out)
})

test_that("maxent_write_prediction_png modes work", {
    skip_if_not_installed("maxentcpp")
    m <- .make_test_model()
    pred <- maxent_project_cloglog(m$model, list(m$g0, m$g1),
                                   c("env0", "env1"))
    for (mode in c("plain", "log", "blackandwhite", "redandyellow")) {
        out <- file.path(tempdir(), paste0("test_pred_", mode, ".png"))
        maxent_write_prediction_png(pred, out, mode = mode)
        expect_true(file.exists(out))
        expect_gt(file.size(out), 0L)
        unlink(out)
    }
})

## ---- Omission CSV -----------------------------------------------------------

test_that("maxent_write_omission_csv creates CSV with correct columns", {
    skip_if_not_installed("maxentcpp")
    m   <- .make_test_model()
    out_dir <- file.path(tempdir(), "maxent_test_omit")
    on.exit(unlink(out_dir, recursive = TRUE), add = TRUE)

    csv_path <- maxent_write_omission_csv(
        m$model, list(m$g0, m$g1), c("env0", "env1"),
        m$pres_rows, m$pres_cols,
        output_dir = out_dir, species = "TestSp")

    expect_true(file.exists(csv_path))
    df <- utils::read.csv(csv_path, check.names = FALSE)
    expect_true("Threshold" %in% names(df))
    expect_true("Training omission" %in% names(df))
    # At least 7 threshold rows (9 without test data for some configs)
    expect_gte(nrow(df), 7L)
})

## ---- Sample Predictions CSV -------------------------------------------------

test_that("maxent_write_sample_predictions creates CSV with correct shape", {
    skip_if_not_installed("maxentcpp")
    m   <- .make_test_model()
    out_dir <- file.path(tempdir(), "maxent_test_sp")
    on.exit(unlink(out_dir, recursive = TRUE), add = TRUE)

    csv_path <- maxent_write_sample_predictions(
        m$model, list(m$g0, m$g1), c("env0", "env1"),
        m$pres_rows, m$pres_cols,
        output_dir = out_dir, species = "TestSp")

    expect_true(file.exists(csv_path))
    df <- utils::read.csv(csv_path, check.names = FALSE)
    expect_equal(ncol(df), 6L)
    expect_equal(nrow(df), length(m$pres_rows))
    expect_true("species" %in% names(df))
    expect_true("cloglog" %in% names(df))
    expect_true(all(df$set == "train"))
})

test_that("maxent_write_sample_predictions includes test rows when provided", {
    skip_if_not_installed("maxentcpp")
    m   <- .make_test_model()
    out_dir <- file.path(tempdir(), "maxent_test_sp2")
    on.exit(unlink(out_dir, recursive = TRUE), add = TRUE)

    csv_path <- maxent_write_sample_predictions(
        m$model, list(m$g0, m$g1), c("env0", "env1"),
        m$pres_rows, m$pres_cols,
        output_dir = out_dir, species = "TestSp2",
        test_rows = c(0L, 1L), test_cols = c(0L, 1L))

    df <- utils::read.csv(csv_path, check.names = FALSE)
    expect_equal(nrow(df), length(m$pres_rows) + 2L)
    expect_true(any(df$set == "test"))
})

## ---- maxentResults.csv ------------------------------------------------------

test_that("maxent_append_results_csv creates file on first call", {
    skip_if_not_installed("maxentcpp")
    m   <- .make_test_model()
    contrib  <- maxent_percent_contribution(m$model, c("env0", "env1"))
    perm_imp <- maxent_permutation_importance(
        m$model, list(m$g0, m$g1), c("env0", "env1"),
        m$pres_rows, m$pres_cols,
        c(0L, 1L, 2L), c(0L, 1L, 2L))

    results_file <- file.path(tempdir(), "maxentResults_test.csv")
    on.exit(unlink(results_file), add = TRUE)

    maxent_append_results_csv(
        results_file  = results_file,
        species       = "TestSp",
        n_training    = 4L,
        training_gain = 1.23,
        training_auc  = 0.85,
        entropy       = 5.5,
        contributions_df = contrib,
        perm_imp_df      = perm_imp)

    expect_true(file.exists(results_file))
    df <- utils::read.csv(results_file, check.names = FALSE)
    expect_equal(nrow(df), 1L)
    expect_equal(df$Species, "TestSp")
})

test_that("maxent_append_results_csv appends second row without duplicating header", {
    skip_if_not_installed("maxentcpp")
    m   <- .make_test_model()
    contrib  <- maxent_percent_contribution(m$model, c("env0", "env1"))
    perm_imp <- maxent_permutation_importance(
        m$model, list(m$g0, m$g1), c("env0", "env1"),
        m$pres_rows, m$pres_cols,
        c(0L, 1L, 2L), c(0L, 1L, 2L))

    results_file <- file.path(tempdir(), "maxentResults_test2.csv")
    on.exit(unlink(results_file), add = TRUE)

    for (sp in c("Sp1", "Sp2")) {
        maxent_append_results_csv(
            results_file  = results_file,
            species       = sp,
            n_training    = 4L,
            training_gain = 1.0,
            training_auc  = 0.8,
            entropy       = 5.0,
            contributions_df = contrib,
            perm_imp_df      = perm_imp)
    }

    df <- utils::read.csv(results_file, check.names = FALSE)
    expect_equal(nrow(df), 2L)
    expect_equal(df$Species, c("Sp1", "Sp2"))
})

#' Print Maxent Model Results
#'
#' Prints a summary of Maxent model performance metrics to the console,
#' replicating the style of the \pkg{dismo} package's MaxEnt output.  The
#' report includes sample counts, training (and optionally test) evaluation
#' statistics, and a ranked variable-contributions table.
#'
#' @param species         Character: species name.
#' @param eval_result     Named list returned by \code{\link{maxent_evaluate}}
#'   (training predictions vs background).
#' @param contributions_df Data.frame with columns \code{name} and
#'   \code{contribution} (from \code{\link{maxent_percent_contribution}}).
#' @param perm_imp_df     Data.frame with columns \code{name} and
#'   \code{permutation_importance} (from
#'   \code{\link{maxent_permutation_importance}}).
#' @param n_presence      Integer: number of training presence records.
#' @param n_background    Integer: number of background records.
#' @param test_eval_result Named list returned by \code{\link{maxent_evaluate}}
#'   for test data, or \code{NULL} (default).
#' @param n_test          Integer: number of test presence records
#'   (default \code{0L}).
#' @param fit_result      Named list returned by \code{\link{maxent_fit}} or
#'   \code{NULL}.  Used to report regularized training gain and entropy.
#' @return Invisibly returns a named list with all reported metrics.
#' @export
#' @examples
#' \donttest{
#' maxent_print_results(
#'   species          = "Sp1",
#'   eval_result      = maxent_evaluate(pres_preds, bg_preds),
#'   contributions_df = contrib,
#'   perm_imp_df      = perm_imp,
#'   n_presence       = length(pres_rows),
#'   n_background     = length(bg_rows))
#' }
maxent_print_results <- function(species,
                                 eval_result,
                                 contributions_df,
                                 perm_imp_df,
                                 n_presence,
                                 n_background,
                                 test_eval_result = NULL,
                                 n_test           = 0L,
                                 fit_result       = NULL) {

    # ---- Header -------------------------------------------------------------
    cat("class         : MaxEnt\n")
    cat(sprintf("species       : %s\n", species))

    # ---- Sample counts ------------------------------------------------------
    cat(sprintf("n presence    : %d\n", as.integer(n_presence)))
    cat(sprintf("n background  : %d\n", as.integer(n_background)))
    if (as.integer(n_test) > 0L) {
        cat(sprintf("n test        : %d\n", as.integer(n_test)))
    }

    # ---- Training statistics ------------------------------------------------
    cat("\nTraining statistics\n")
    if (!is.null(eval_result$auc)) {
        cat(sprintf("  AUC             : %.4f\n", eval_result$auc))
    }
    if (!is.null(fit_result$loss)) {
        # fit_result$loss is the regularized training gain (negative log-loss)
        cat(sprintf("  Gain            : %.4f\n", fit_result$loss))
    }
    entropy_val <- if (!is.null(fit_result$entropy))
        fit_result$entropy else NULL
    if (!is.null(entropy_val)) {
        cat(sprintf("  Entropy         : %.4f\n", entropy_val))
    }

    # ---- Test statistics ----------------------------------------------------
    if (!is.null(test_eval_result) && !is.null(test_eval_result$auc)) {
        cat("\nTest statistics\n")
        cat(sprintf("  AUC             : %.4f\n", test_eval_result$auc))
    }

    # ---- Variable contributions ---------------------------------------------
    if (!is.null(contributions_df) && nrow(contributions_df) > 0 &&
        !is.null(perm_imp_df)      && nrow(perm_imp_df) > 0) {

        merged <- merge(contributions_df, perm_imp_df, by = "name", all = TRUE)
        merged[is.na(merged)] <- 0
        merged <- merged[order(merged$contribution, decreasing = TRUE), ]

        cat("\nVariable contributions\n")
        cat(sprintf("  %-20s  %s  %s\n",
                    "Variable",
                    "Contribution (%)",
                    "Permutation importance (%)"))
        for (i in seq_len(nrow(merged))) {
            cat(sprintf("  %-20s  %16.1f  %25.1f\n",
                        merged$name[i],
                        merged$contribution[i],
                        merged$permutation_importance[i]))
        }
    }
    cat("\n")

    # ---- Return metrics invisibly -------------------------------------------
    invisible(list(
        species          = species,
        n_presence       = as.integer(n_presence),
        n_background     = as.integer(n_background),
        n_test           = as.integer(n_test),
        training_auc     = eval_result$auc,
        test_auc         = if (!is.null(test_eval_result))
                               test_eval_result$auc else NULL,
        gain             = fit_result$loss,  # regularized training gain
        entropy          = fit_result$entropy,
        contributions    = contributions_df,
        perm_importance  = perm_imp_df
    ))
}

#' @importFrom methods slotNames slot
NULL

# Internal helper: plan row-wise blocks for streaming a SpatRaster.
#
# Returns a list with elements:
#   $n      integer, number of blocks
#   $row    integer vector of 1-based starting rows per block
#   $nrows  integer vector of row counts per block
#
# Tries `terra::blocks()` first (terra >= 1.6), falls back to
# `terra::blockSize()` (terra 1.5), and finally to a single-block plan
# that reads the entire raster in one shot.
.maxent_plan_blocks <- function(rast) {
    nr <- terra::nrow(rast)
    if (exists("blocks", where = asNamespace("terra"), inherits = FALSE)) {
        blocks_fn <- get("blocks", envir = asNamespace("terra"),
                         inherits = FALSE)
        bk <- tryCatch(blocks_fn(rast), error = function(e) NULL)
        if (!is.null(bk) && !is.null(bk$n) && bk$n > 0L) {
            return(list(n     = as.integer(bk$n),
                        row   = as.integer(bk$row),
                        nrows = as.integer(bk$nrows)))
        }
    }
    if (exists("blockSize", where = asNamespace("terra"), inherits = FALSE)) {
        block_size_fn <- get("blockSize", envir = asNamespace("terra"),
                             inherits = FALSE)
        bk <- tryCatch(block_size_fn(rast), error = function(e) NULL)
        if (!is.null(bk) && !is.null(bk$n) && bk$n > 0L) {
            return(list(n     = as.integer(bk$n),
                        row   = as.integer(bk$row),
                        nrows = as.integer(bk$nrows)))
        }
    }
    list(n = 1L, row = 1L, nrows = as.integer(nr))
}

.maxent_get_spatraster_xptr <- function(rast) {
    slots <- methods::slotNames(rast)
    for (nm in c("ptr", "pntr", "pnt")) {
        if (nm %in% slots) {
            xp <- methods::slot(rast, nm)
            if (typeof(xp) == "externalptr") return(xp)
            # terra >= 1.8: slot holds an Rcpp module object (S4) whose
            # .xData environment contains the real external pointer as
            # ".pointer".
            if (isS4(xp)) {
                env <- tryCatch(methods::slot(xp, ".xData"), error = function(e) NULL)
                if (is.environment(env) && exists(".pointer", envir = env, inherits = FALSE)) {
                    inner <- get(".pointer", envir = env, inherits = FALSE)
                    if (typeof(inner) == "externalptr") return(inner)
                }
            }
        }
    }
    stop("Could not locate SpatRaster external pointer slot ('ptr'/'pntr'/'pnt').")
}

.maxent_make_spatraster_stream <- function(rast) {
    blocks <- .maxent_plan_blocks(rast)
    n_layers <- terra::nlyr(rast)

    terra::readStart(rast)
    on.exit(try(terra::readStop(rast), silent = TRUE), add = TRUE)
    num_points <- 0L
    for (i in seq_len(blocks$n)) {
        m <- terra::readValues(rast,
                               row   = blocks$row[i],
                               nrows = blocks$nrows[i],
                               mat   = TRUE)
        num_points <- num_points + sum(stats::complete.cases(m))
    }
    terra::readStop(rast)

    state <- new.env(parent = emptyenv())
    state$block_idx <- 0L
    state$open      <- FALSE
    state$blocks    <- blocks

    open_stream <- function() {
        if (!state$open) {
            terra::readStart(rast)
            state$open <- TRUE
        }
    }
    close_stream <- function() {
        if (state$open) {
            try(terra::readStop(rast), silent = TRUE)
            state$open <- FALSE
        }
    }
    reset_fn <- function() {
        close_stream()
        state$block_idx <- 0L
        open_stream()
        invisible(NULL)
    }
    next_tile_fn <- function() {
        while (state$open && state$block_idx < state$blocks$n) {
            state$block_idx <- state$block_idx + 1L
            i <- state$block_idx
            m <- terra::readValues(rast,
                                   row   = state$blocks$row[i],
                                   nrows = state$blocks$nrows[i],
                                   mat   = TRUE)
            keep <- stats::complete.cases(m)
            if (any(keep)) return(m[keep, , drop = FALSE])
        }
        close_stream()
        matrix(numeric(0), nrow = 0L, ncol = n_layers)
    }

    list(
        num_points = as.integer(num_points),
        num_layers = as.integer(n_layers),
        layer_names = names(rast),
        next_tile_fn = next_tile_fn,
        reset_fn = reset_fn,
        close_stream = close_stream
    )
}

#' Create a FeaturedSpace directly from a terra SpatRaster
#'
#' Streams a \code{terra::SpatRaster} block-by-block into the C++
#' \code{FeaturedSpace} streaming constructor, filtering NA cells on the fly
#' and handing the concatenated (\code{num_points} x \code{nlyr(rast)})
#' matrix to \code{\link{maxent_generate_features}} so feature objects are
#' built inside C++ without the R caller having to materialise the full
#' raster stack first.
#'
#' This is the Phase E.2 entry point described in
#' \code{docs/ARCHITECTURE_terra_raster.md} (sections 3.1 and 3.2). The
#' result is bit-for-bit identical to the dense path
#' (\code{\link{maxent_generate_features}} -> \code{\link{maxent_featured_space}})
#' when the same data, sample indices, and feature configuration are used.
#'
#' @param rast A \code{terra::SpatRaster} with one layer per environmental
#'   variable. Layers must be named; names are passed through to
#'   \code{\link{maxent_generate_features}} so generated features carry the
#'   same identifiers as the layer names (e.g. \code{bio1^2}, \code{bio1*bio2}).
#' @param sample_indices Integer vector: 0-based indices of occurrence
#'   samples in the concatenated stream of finite background cells. See
#'   \code{\link{maxent_raster_sample_indices}} for a helper that converts
#'   occurrence \code{data.frame}s into these indices.
#' @param feature_types Character vector of feature types to generate.
#'   Any subset of \code{c("linear", "quadratic", "product", "threshold",
#'   "hinge")}. Defaults to all types.
#' @param n_thresholds Integer; number of threshold knots per variable.
#' @param n_hinges Integer; number of hinge knots per variable.
#' @param enable_streaming_eval Logical; when \code{TRUE}, enables
#'   streaming-evaluation mode after object construction.
#' @return External pointer to a \code{FeaturedSpace} object.
#' @seealso \code{\link{maxent_featured_space}},
#'   \code{\link{maxent_featured_space_from_callback}},
#'   \code{\link{maxent_generate_features}}.
#' @export
#' @examples
#' \donttest{
#' stack_path <- system.file("extdata", "stack_1_12_crop.rds",
#'                          package = "maxentcpp")
#' r <- terra::unwrap(readRDS(stack_path))
#' fs <- maxent_featured_space_from_rast(
#'     r,
#'     sample_indices = c(0L, 1L, 2L),
#'     feature_types  = c("linear", "quadratic")
#' )
#' }
maxent_featured_space_from_rast <- function(
    rast,
    sample_indices,
    feature_types = c("linear", "quadratic", "product", "threshold", "hinge"),
    n_thresholds  = 10L,
    n_hinges      = 10L,
    enable_streaming_eval = TRUE)
{
    if (!requireNamespace("terra", quietly = TRUE)) {
        stop("maxent_featured_space_from_rast requires the 'terra' package.")
    }
    if (!inherits(rast, "SpatRaster")) {
        stop("`rast` must be a terra::SpatRaster.")
    }

    layer_names <- names(rast)
    n_layers    <- length(layer_names)
    if (n_layers == 0L) {
        stop("`rast` has 0 layers.")
    }

    stream <- .maxent_make_spatraster_stream(rast)
    num_points <- stream$num_points

    if (num_points == 0L) {
        stop("`rast` contains no finite background cells.")
    }

    sample_indices <- as.integer(sample_indices)
    if (length(sample_indices) == 0L) {
        stop("`sample_indices` must contain at least one occurrence index.")
    }
    if (any(sample_indices < 0L | sample_indices >= num_points)) {
        stop(sprintf(
            "All sample_indices must lie in [0, %d); got range [%d, %d].",
            num_points, min(sample_indices), max(sample_indices)))
    }

    on.exit(stream$close_stream(), add = TRUE)
    maxent_featured_space_from_spatraster_callback(
        preserved_rast  = rast,
        rast_xptr       = .maxent_get_spatraster_xptr(rast),
        num_points      = stream$num_points,
        num_layers      = stream$num_layers,
        layer_names     = stream$layer_names,
        sample_indices  = sample_indices,
        feature_types   = as.character(feature_types),
        n_thresholds    = as.integer(n_thresholds),
        n_hinges        = as.integer(n_hinges),
        next_tile_fn    = stream$next_tile_fn,
        reset_fn        = stream$reset_fn,
        enable_streaming_eval = isTRUE(enable_streaming_eval),
        use_cache = TRUE)
}

#' Convert occurrence coordinates to 0-based stream indices
#'
#' Maps occurrence cell indices (as produced by \code{terra::cellFromXY})
#' into 0-based indices within the concatenated stream of finite
#' (non-NA) background cells emitted by
#' \code{\link{maxent_featured_space_from_rast}}.
#'
#' @param rast A \code{terra::SpatRaster}.
#' @param cells Integer vector of 1-based cell indices
#'   (as returned by \code{terra::cellFromXY}).
#' @return Integer vector of 0-based indices within the finite-cell stream.
#'   Occurrence cells that fall on NA raster values (and therefore never
#'   appear in the stream) are silently dropped and a warning is emitted.
#' @export
maxent_raster_sample_indices <- function(rast, cells) {
    if (!requireNamespace("terra", quietly = TRUE)) {
        stop("maxent_raster_sample_indices requires the 'terra' package.")
    }
    if (!inherits(rast, "SpatRaster")) {
        stop("`rast` must be a terra::SpatRaster.")
    }
    cells <- as.integer(cells)
    # Build the finite-cell mask by streaming blocks once.
    blocks <- .maxent_plan_blocks(rast)
    terra::readStart(rast)
    on.exit(try(terra::readStop(rast), silent = TRUE), add = TRUE)
    n_cells <- as.integer(terra::ncell(rast))
    is_finite <- logical(n_cells)
    row_offset <- 0L
    ncol_r <- terra::ncol(rast)
    for (i in seq_len(blocks$n)) {
        m <- terra::readValues(rast,
                               row   = blocks$row[i],
                               nrows = blocks$nrows[i],
                               mat   = TRUE)
        block_finite <- stats::complete.cases(m)
        start <- (blocks$row[i] - 1L) * ncol_r + 1L
        end   <- start + length(block_finite) - 1L
        is_finite[start:end] <- block_finite
    }
    terra::readStop(rast)

    # Convert 1-based full-grid cell index -> 0-based finite-cell-stream index.
    finite_prefix <- cumsum(is_finite) - 1L   # 0-based position within finite stream
    valid <- !is.na(cells) & cells >= 1L & cells <= n_cells
    valid[valid] <- is_finite[cells[valid]]
    if (!all(valid)) {
        warning(sprintf(
            "%d occurrence cell(s) fall on NA raster values; they are dropped.",
            sum(!valid)))
    }
    finite_prefix[cells[valid]]
}

#' Extract environmental values for occurrences from a terra SpatRaster
#'
#' Streams finite raster rows through the same callback-backed provider path
#' used by \code{\link{maxent_featured_space_from_rast}}, then returns values
#' at occurrence locations.
#'
#' @param rast A \code{terra::SpatRaster}.
#' @param occurrences Either a two-column matrix/data.frame (x,y) or a vector
#'   of 1-based raster cell indices.
#' @param lon_col Longitude column name when \code{occurrences} is a data.frame.
#' @param lat_col Latitude column name when \code{occurrences} is a data.frame.
#' @return Numeric matrix with one row per retained occurrence and one column
#'   per raster layer.
#' @export
maxent_extract_occurrence_env_terra <- function(
    rast,
    occurrences,
    lon_col = "longitude",
    lat_col = "latitude") {
    if (!requireNamespace("terra", quietly = TRUE)) {
        stop("maxent_extract_occurrence_env_terra requires the 'terra' package.")
    }
    if (!inherits(rast, "SpatRaster")) {
        stop("`rast` must be a terra::SpatRaster.")
    }
    cells <- NULL
    if (is.numeric(occurrences) || is.integer(occurrences)) {
        cells <- as.integer(occurrences)
    } else if (is.matrix(occurrences) && ncol(occurrences) >= 2L) {
        cells <- terra::cellFromXY(rast, occurrences[, 1:2, drop = FALSE])
    } else if (is.data.frame(occurrences)) {
        if (!(lon_col %in% names(occurrences) && lat_col %in% names(occurrences))) {
            stop("occurrences data.frame must contain lon_col and lat_col.")
        }
        xy <- cbind(occurrences[[lon_col]], occurrences[[lat_col]])
        cells <- terra::cellFromXY(rast, xy)
    } else {
        stop("`occurrences` must be cells, a 2-column matrix, or a data.frame.")
    }

    idx <- maxent_raster_sample_indices(rast, cells)
    if (length(idx) == 0L) {
        return(matrix(numeric(0), nrow = 0L, ncol = terra::nlyr(rast)))
    }
    stream <- .maxent_make_spatraster_stream(rast)
    on.exit(stream$close_stream(), add = TRUE)
    out <- maxent_extract_occurrence_from_callback(
        num_points = stream$num_points,
        num_layers = stream$num_layers,
        occurrence_indices = idx,
        next_tile_fn = stream$next_tile_fn,
        reset_fn = stream$reset_fn,
        preserved_rast = rast)
    colnames(out) <- names(rast)
    out
}

#' Train MaxEnt directly from a terra SpatRaster
#'
#' High-level convenience wrapper that builds a streaming FeaturedSpace from
#' \code{rast}, maps occurrence locations to finite-stream sample indices, and
#' trains with \code{\link{maxent_fit}}.
#'
#' @param rast A \code{terra::SpatRaster}.
#' @param occurrences Either a two-column matrix/data.frame (x,y) or a vector
#'   of 1-based raster cell indices.
#' @param lon_col Longitude column name when \code{occurrences} is a data.frame.
#' @param lat_col Latitude column name when \code{occurrences} is a data.frame.
#' @param feature_types Feature type set passed to feature generation.
#' @param n_thresholds Number of threshold knots.
#' @param n_hinges Number of hinge knots.
#' @param max_iter Maximum training iterations.
#' @param convergence Convergence threshold.
#' @param beta_multiplier Regularization multiplier.
#' @param min_deviation Minimum deviation floor.
#' @return A named list with training results plus \code{model} and
#'   \code{sample_indices}.
#' @export
maxent_train_terra <- function(
    rast,
    occurrences,
    lon_col = "longitude",
    lat_col = "latitude",
    feature_types = c("linear", "quadratic", "product", "threshold", "hinge"),
    n_thresholds = 10L,
    n_hinges = 10L,
    max_iter = 500L,
    convergence = 1e-5,
    beta_multiplier = 1.0,
    min_deviation = 0.001) {
    if (!requireNamespace("terra", quietly = TRUE)) {
        stop("Package 'terra' is required. Install it with install.packages('terra').")
    }
    cells <- NULL
    if (is.numeric(occurrences) || is.integer(occurrences)) {
        cells <- as.integer(occurrences)
    } else if (is.matrix(occurrences) && ncol(occurrences) >= 2L) {
        cells <- terra::cellFromXY(rast, occurrences[, 1:2, drop = FALSE])
    } else if (is.data.frame(occurrences)) {
        if (!(lon_col %in% names(occurrences) && lat_col %in% names(occurrences))) {
            stop("occurrences data.frame must contain lon_col and lat_col.")
        }
        xy <- cbind(occurrences[[lon_col]], occurrences[[lat_col]])
        cells <- terra::cellFromXY(rast, xy)
    } else {
        stop("`occurrences` must be cells, a 2-column matrix, or a data.frame.")
    }
    sample_indices <- maxent_raster_sample_indices(rast, cells)
    if (length(sample_indices) == 0L) {
        stop("No valid occurrence points remain after finite-cell filtering.")
    }
    fs <- maxent_featured_space_from_rast(
        rast = rast,
        sample_indices = sample_indices,
        feature_types = feature_types,
        n_thresholds = n_thresholds,
        n_hinges = n_hinges,
        enable_streaming_eval = TRUE)
    fit <- maxent_fit(
        featured_space = fs,
        max_iter = max_iter,
        convergence = convergence,
        beta_multiplier = beta_multiplier,
        min_deviation = min_deviation)
    fit$model <- fs
    fit$sample_indices <- sample_indices
    fit
}

# Phase E.2: SpatRaster streaming provider bit-for-bit equivalence test.
#
# Compares the streaming `maxent_featured_space_from_rast()` path against the
# dense `maxent_featured_space()` path on the same SpatRaster data.  Both
# paths must agree at 1e-14 on weights / loss / entropy / lambdas.

# skip_if_no_terra() is provided by helper-terra.R

# -- Helpers ------------------------------------------------------------------

# Build a small deterministic 2-layer raster.  20x3 = 60 cells.
build_test_raster <- function(n_rows = 20L, n_cols = 3L) {
    skip_if_no_terra()
    set.seed(42L)
    n_cells <- n_rows * n_cols
    bio1  <- seq(10, 30, length.out = n_cells) + rnorm(n_cells, 0, 0.5)
    bio12 <- seq(500, 2000, length.out = n_cells) + rnorm(n_cells, 0, 20)
    r1 <- terra::rast(nrows = n_rows, ncols = n_cols,
                      xmin = 0, xmax = n_cols, ymin = 0, ymax = n_rows,
                      vals = bio1)
    names(r1) <- "bio1"
    r2 <- terra::rast(nrows = n_rows, ncols = n_cols,
                      xmin = 0, xmax = n_cols, ymin = 0, ymax = n_rows,
                      vals = bio12)
    names(r2) <- "bio12"
    rast <- c(r1, r2)
    rast
}

# Drain a SpatRaster into a dense matrix (num_points x num_layers), dropping
# rows that contain NAs, in the same order the streaming path would emit.
raster_to_dense <- function(rast) {
    vals <- terra::values(rast, mat = TRUE)
    vals[stats::complete.cases(vals), , drop = FALSE]
}

train_args <- list(
    max_iter        = 50L,
    convergence     = 1e-5,
    beta_multiplier = 1.0,
    min_deviation   = 0.001)

fit_dense <- function(rast, sample_indices, feature_types) {
    dense <- raster_to_dense(rast)
    features <- maxent_generate_features(
        data  = stats::setNames(as.list(as.data.frame(dense)), names(rast)),
        types = feature_types,
        n_thresholds = 10L,
        n_hinges     = 10L)
    fs <- maxent_featured_space(
        num_points     = nrow(dense),
        sample_indices = sample_indices,
        features       = features)
    res <- do.call(maxent_fit, c(list(featured_space = fs), train_args))
    list(fs = fs, res = res)
}

fit_streaming <- function(rast, sample_indices, feature_types) {
    fs <- maxent_featured_space_from_rast(
        rast           = rast,
        sample_indices = sample_indices,
        feature_types  = feature_types,
        n_thresholds   = 10L,
        n_hinges       = 10L)
    res <- do.call(maxent_fit, c(list(featured_space = fs), train_args))
    list(fs = fs, res = res)
}

# -- Tests --------------------------------------------------------------------

test_that("streaming path matches dense path bit-for-bit (linear + quadratic)", {
    skip_if_no_terra()
    rast <- build_test_raster()
    sample_indices <- c(3L, 11L, 29L, 41L, 55L)
    feature_types  <- c("linear", "quadratic")

    dense    <- fit_dense(rast,   sample_indices, feature_types)
    streamed <- fit_streaming(rast, sample_indices, feature_types)

    expect_equal(streamed$res$lambdas, dense$res$lambdas, tolerance = 1e-14)
    expect_equal(streamed$res$loss,    dense$res$loss,    tolerance = 1e-14)
    expect_equal(streamed$res$entropy, dense$res$entropy, tolerance = 1e-14)
    expect_equal(streamed$res$iterations, dense$res$iterations)
})

test_that("streaming path matches dense path (all feature types)", {
    skip_if_no_terra()
    rast <- build_test_raster()
    sample_indices <- c(3L, 11L, 29L, 41L, 55L)
    feature_types  <- c("linear", "quadratic", "product", "threshold", "hinge")

    dense    <- fit_dense(rast,   sample_indices, feature_types)
    streamed <- fit_streaming(rast, sample_indices, feature_types)

    expect_equal(streamed$res$lambdas, dense$res$lambdas, tolerance = 1e-14)
    expect_equal(streamed$res$loss,    dense$res$loss,    tolerance = 1e-14)
    expect_equal(streamed$res$entropy, dense$res$entropy, tolerance = 1e-14)
})

test_that("streaming path survives NA cells identical to dense path", {
    skip_if_no_terra()
    rast <- build_test_raster()
    # Poke some NAs into bio12.
    na_cells <- c(4L, 17L, 33L, 51L)
    terra::values(rast[[2]])[na_cells] <- NA_real_

    # Any sample index must land on a finite cell.  Finite-cell count is
    # (60 - length(na_cells)) = 56.
    sample_indices <- c(3L, 10L, 25L, 40L, 50L)
    feature_types  <- c("linear", "quadratic")

    dense    <- fit_dense(rast,   sample_indices, feature_types)
    streamed <- fit_streaming(rast, sample_indices, feature_types)

    expect_equal(streamed$res$lambdas, dense$res$lambdas, tolerance = 1e-14)
    expect_equal(streamed$res$loss,    dense$res$loss,    tolerance = 1e-14)
})

test_that("maxent_raster_sample_indices maps cells into the finite stream", {
    skip_if_no_terra()
    rast <- build_test_raster()
    # Poke NAs into bio1 cells 5,10,15 (1-based); they should be dropped.
    terra::values(rast[[1]])[c(5L, 10L, 15L)] <- NA_real_

    # Pick three occurrence cells, two finite (1,20) and one on an NA (10).
    cells <- c(1L, 10L, 20L)
    expect_warning(
        idx <- maxent_raster_sample_indices(rast, cells),
        "drop"
    )
    # cell 1  -> cumsum(is_finite)[1]  - 1 = 1  - 1 = 0
    # cell 10 -> NA, dropped
    # cell 20 -> cumsum(is_finite)[20] - 1 = 17 - 1 = 16
    expect_equal(idx, c(0L, 16L))
})

test_that("streaming path handles mid-stream all-NA blocks without truncation", {
    # Regression test for Devin Review BUG_pr-review-job-a516d323_0001:
    # if a full block contains only NA cells, next_tile_fn() used to emit a
    # 0-row tile which the C++ streaming constructor interpreted as
    # end-of-stream, truncating the background matrix and throwing
    # "BackgroundProvider emitted fewer rows than declared".  The fix is to
    # skip over all-NA blocks inside next_tile_fn() instead of emitting them.
    skip_if_no_terra()
    rast <- build_test_raster(n_rows = 20L, n_cols = 3L)
    # Force several small blocks so one of them can be fully NA.  1 row per
    # block = 20 blocks, each with 3 cells.
    terra::terraOptions(memfrac = 1e-6)
    on.exit(terra::terraOptions(memfrac = 0.6), add = TRUE)
    # Wipe an entire interior row (cells 25-27 = row 9, 1-based indices) to
    # create an all-NA block between finite ones.
    terra::values(rast[[1]])[25:27] <- NA_real_
    terra::values(rast[[2]])[25:27] <- NA_real_

    sample_indices <- c(3L, 10L, 25L, 40L, 50L)
    feature_types  <- c("linear", "quadratic")

    dense    <- fit_dense(rast,   sample_indices, feature_types)
    # Must not throw "BackgroundProvider emitted fewer rows..."
    streamed <- fit_streaming(rast, sample_indices, feature_types)

    expect_equal(streamed$res$lambdas, dense$res$lambdas, tolerance = 1e-14)
    expect_equal(streamed$res$loss,    dense$res$loss,    tolerance = 1e-14)
})

test_that("maxent_raster_sample_indices handles NA cells without error", {
    # Regression test for Devin Review BUG_pr-review-job-a516d323_0002:
    # NA entries in `cells` used to propagate into `valid` via
    # `NA >= 1L` -> NA and `is_finite[NA]` -> NA, which then made
    # `if (!all(valid))` raise "missing value where TRUE/FALSE needed".
    skip_if_no_terra()
    rast <- build_test_raster()
    # Cells returned by terra::cellFromXY() for out-of-extent points are NA;
    # make sure the helper treats them the same as on-NA cells: drop + warn.
    cells <- c(1L, NA_integer_, 20L)
    expect_warning(
        idx <- maxent_raster_sample_indices(rast, cells),
        "drop"
    )
    expect_equal(idx, c(0L, 19L))
})

test_that("maxent_extract_occurrence_env_terra matches terra::extract", {
    skip_if_no_terra()
    rast <- build_test_raster()
    pts <- data.frame(longitude = c(0.5, 1.5, 2.5),
                      latitude  = c(1.5, 5.5, 10.5))

    got <- maxent_extract_occurrence_env_terra(
        rast,
        occurrences = pts,
        lon_col = "longitude",
        lat_col = "latitude")
    expect_equal(colnames(got), names(rast))

    cells <- terra::cellFromXY(rast, as.matrix(pts[, c("longitude", "latitude")]))
    ref <- terra::values(rast, mat = TRUE)[cells, , drop = FALSE]
    expect_equal(unname(got), unname(ref), tolerance = 1e-14)
})

test_that("maxent_train_terra matches manual streaming fit", {
    skip_if_no_terra()
    rast <- build_test_raster()
    pts <- data.frame(longitude = c(0.5, 1.5, 2.5, 0.5, 2.5),
                      latitude  = c(1.5, 5.5, 10.5, 12.5, 18.5))
    feature_types <- c("linear", "quadratic")

    cells <- terra::cellFromXY(rast, as.matrix(pts[, c("longitude", "latitude")]))
    idx <- maxent_raster_sample_indices(rast, cells)

    manual_fs <- maxent_featured_space_from_rast(
        rast = rast,
        sample_indices = idx,
        feature_types = feature_types,
        n_thresholds = 10L,
        n_hinges = 10L,
        enable_streaming_eval = TRUE)
    manual <- maxent_fit(manual_fs, max_iter = 50L, convergence = 1e-5,
                         beta_multiplier = 1.0, min_deviation = 0.001)

    auto <- maxent_train_terra(
        rast = rast,
        occurrences = pts,
        lon_col = "longitude",
        lat_col = "latitude",
        feature_types = feature_types,
        n_thresholds = 10L,
        n_hinges = 10L,
        max_iter = 50L,
        convergence = 1e-5,
        beta_multiplier = 1.0,
        min_deviation = 0.001)

    expect_equal(auto$sample_indices, idx)
    expect_equal(auto$lambdas, manual$lambdas, tolerance = 1e-14)
    expect_equal(auto$loss, manual$loss, tolerance = 1e-14)
    expect_equal(auto$entropy, manual$entropy, tolerance = 1e-14)
    expect_true(inherits(auto$model, "externalptr"))
})

#' @useDynLib maxentcpp, .registration = TRUE
#' @importFrom Rcpp sourceCpp
NULL

#' Create a Linear Feature
#'
#' Creates a linear (normalized) feature that evaluates to
#' \code{(values[i] - min) / (max - min)}. Returns 0 when \code{min == max}.
#'
#' @param values Numeric vector of environmental variable values.
#' @param name Character string: feature name/identifier.
#' @param min_val Minimum value for normalization. If \code{NULL} (default),
#'   computed as \code{min(values)}.
#' @param max_val Maximum value for normalization. If \code{NULL} (default),
#'   computed as \code{max(values)}.
#' @return External pointer to a LinearFeature C++ object.
#' @export
#' @examples
#' vals <- c(0, 5, 10, 3)
#' f <- maxent_linear_feature(vals, "temperature")
#' maxent_feature_eval(f, 1)  # index 1 (R) -> 0-based index 0
maxent_linear_feature <- function(values, name, min_val = NULL, max_val = NULL) {
    values <- as.numeric(values)
    if (is.null(min_val)) min_val <- min(values)
    if (is.null(max_val)) max_val <- max(values)
    create_linear_feature(values, name, min_val, max_val)
}

#' Create a Quadratic Feature
#'
#' Creates a quadratic feature that evaluates to \code{linear_val^2} where
#' \code{linear_val = (values[i] - min) / (max - min)}.
#'
#' @param values Numeric vector of environmental variable values.
#' @param name Character string: feature name/identifier.
#' @param min_val Minimum value for normalization. If \code{NULL} (default),
#'   computed as \code{min(values)}.
#' @param max_val Maximum value for normalization. If \code{NULL} (default),
#'   computed as \code{max(values)}.
#' @return External pointer to a QuadraticFeature C++ object.
#' @export
#' @examples
#' vals <- c(0, 5, 10)
#' f <- maxent_quadratic_feature(vals, "temperature_sq")
#' maxent_feature_eval(f, 2)  # evaluates at index 2 (1-based -> 0-based: 1)
maxent_quadratic_feature <- function(values, name, min_val = NULL, max_val = NULL) {
    values <- as.numeric(values)
    if (is.null(min_val)) min_val <- min(values)
    if (is.null(max_val)) max_val <- max(values)
    create_quadratic_feature(values, name, min_val, max_val)
}

#' Create a Product Feature
#'
#' Creates an interaction feature between two environmental variables:
#' \code{eval(i) = norm(values1[i]) * norm(values2[i])}.
#'
#' @param values1 Numeric vector for the first environmental variable.
#' @param values2 Numeric vector for the second environmental variable
#'   (must have the same length as \code{values1}).
#' @param name Character string: feature name/identifier.
#' @param min1 Minimum of the first variable. If \code{NULL} (default),
#'   computed as \code{min(values1)}.
#' @param max1 Maximum of the first variable. If \code{NULL} (default),
#'   computed as \code{max(values1)}.
#' @param min2 Minimum of the second variable. If \code{NULL} (default),
#'   computed as \code{min(values2)}.
#' @param max2 Maximum of the second variable. If \code{NULL} (default),
#'   computed as \code{max(values2)}.
#' @return External pointer to a ProductFeature C++ object.
#' @export
#' @examples
#' temp <- c(0, 5, 10)
#' prec <- c(100, 200, 150)
#' f <- maxent_product_feature(temp, prec, "temp_x_prec")
#' maxent_feature_eval(f, 2)
maxent_product_feature <- function(values1, values2, name,
                                   min1 = NULL, max1 = NULL,
                                   min2 = NULL, max2 = NULL) {
    values1 <- as.numeric(values1)
    values2 <- as.numeric(values2)
    if (length(values1) != length(values2)) {
        stop("values1 and values2 must have the same length")
    }
    if (is.null(min1)) min1 <- min(values1)
    if (is.null(max1)) max1 <- max(values1)
    if (is.null(min2)) min2 <- min(values2)
    if (is.null(max2)) max2 <- max(values2)
    create_product_feature(values1, values2, name, min1, max1, min2, max2)
}

#' Create a Threshold Feature
#'
#' Creates a binary step feature: \code{eval(i) = 1} if
#' \code{values[i] > threshold}, else \code{0}.
#'
#' @param values Numeric vector of environmental variable values.
#' @param name Character string: feature name/identifier.
#' @param threshold The threshold value.
#' @return External pointer to a ThresholdFeature C++ object.
#' @export
#' @examples
#' vals <- c(1, 5, 10, 3)
#' f <- maxent_threshold_feature(vals, "temperature_threshold", threshold = 5)
#' maxent_feature_eval(f, 3)  # values[3] = 10 > 5 -> 1
maxent_threshold_feature <- function(values, name, threshold) {
    values <- as.numeric(values)
    create_threshold_feature(values, name, threshold)
}

#' Create a Hinge Feature
#'
#' Creates a piecewise-linear hinge feature.
#'
#' \strong{Forward hinge} (\code{reverse = FALSE}):
#' \code{eval(i) = max(0, (values[i] - min_knot) / (max_knot - min_knot))}
#'
#' \strong{Reverse hinge} (\code{reverse = TRUE}):
#' \code{eval(i) = max(0, (max_knot - values[i]) / (max_knot - min_knot))}
#'
#' @param values Numeric vector of environmental variable values.
#' @param name Character string: feature name/identifier.
#' @param min_knot Lower knot of the hinge.
#' @param max_knot Upper knot of the hinge (must be strictly greater than
#'   \code{min_knot}).
#' @param reverse Logical; if \code{TRUE}, use reverse hinge. Default is
#'   \code{FALSE} (forward hinge).
#' @return External pointer to a HingeFeature C++ object.
#' @export
#' @examples
#' vals <- c(0, 5, 10, 3)
#' f <- maxent_hinge_feature(vals, "temperature_hinge", min_knot = 2, max_knot = 8)
#' maxent_feature_eval(f, 2)  # values[2] = 5 -> (5-2)/(8-2) = 0.5
maxent_hinge_feature <- function(values, name, min_knot, max_knot, reverse = FALSE) {
    values <- as.numeric(values)
    if (min_knot >= max_knot) {
        stop("min_knot must be strictly less than max_knot")
    }
    create_hinge_feature(values, name, min_knot, max_knot, reverse)
}

#' Evaluate a Feature at a Given Index
#'
#' Evaluates the feature function at the specified index (1-based, as is
#' standard in R). Internally converts to 0-based indexing for C++.
#'
#' @param feature External pointer to a Feature C++ object.
#' @param index 1-based integer index into the data vector.
#' @return Numeric scalar: the feature value at that index.
#' @export
#' @examples
#' vals <- c(0, 5, 10)
#' f <- maxent_linear_feature(vals, "temp")
#' maxent_feature_eval(f, 2)  # 5/10 = 0.5
maxent_feature_eval <- function(feature, index) {
    if (length(index) != 1L || is.na(index) || !is.finite(index)) {
        stop("index must be a single finite value")
    }

    index_int <- as.integer(index)
    if (!identical(as.numeric(index_int), as.numeric(index))) {
        stop("index must be a whole number")
    }

    size <- maxent_feature_info(feature)$size
    if (index_int < 1L || index_int > size) {
        stop(sprintf("index must be between 1 and %d", size))
    }

    feature_eval(feature, index_int - 1L)
}

#' Get Feature Properties
#'
#' Returns a list of metadata for a feature object.
#'
#' @param feature External pointer to a Feature C++ object.
#' @return Named list with elements: \code{name}, \code{type}, \code{lambda},
#'   \code{min}, \code{max}, \code{size}.
#' @export
#' @examples
#' vals <- c(0, 5, 10)
#' f <- maxent_linear_feature(vals, "temp")
#' maxent_feature_info(f)
maxent_feature_info <- function(feature) {
    feature_get_info(feature)
}

#' Generate Features from Environmental Variable Data
#'
#' Automatically generates all configured feature types from one or more
#' environmental variable vectors.
#'
#' @param data Named list of numeric vectors, one per environmental variable.
#' @param types Character vector of feature types to generate. Valid values:
#'   \code{"linear"}, \code{"quadratic"}, \code{"product"},
#'   \code{"threshold"}, \code{"hinge"}. Defaults to all types.
#' @param n_thresholds Integer; number of threshold knots per variable
#'   (default: 10).
#' @param n_hinges Integer; number of hinge knots per variable (default: 10).
#' @return Named list of external pointers to Feature C++ objects.
#' @export
#' @examples
#' env_data <- list(
#'   temperature   = c(15, 20, 25, 18, 22),
#'   precipitation = c(100, 200, 150, 80, 300)
#' )
#' features <- maxent_generate_features(env_data, types = c("linear", "hinge"))
#' length(features)
maxent_generate_features <- function(
    data,
    types = c("linear", "quadratic", "product", "threshold", "hinge"),
    n_thresholds = 10L,
    n_hinges = 10L)
{
    if (!is.list(data) || is.null(names(data))) {
        stop("data must be a named list of numeric vectors")
    }
    data <- lapply(data, as.numeric)
    generate_features(data, types, as.integer(n_thresholds), as.integer(n_hinges))
}

test_that("LinearFeature creation and evaluation work", {
    skip_if_not_installed("maxentcpp")

    vals <- c(0.0, 5.0, 10.0, 3.0)
    f <- maxent_linear_feature(vals, "temperature", min_val = 0.0, max_val = 10.0)
    expect_true(!is.null(f))

    # 1-based indexing in R
    expect_equal(maxent_feature_eval(f, 1), 0.0)   # (0 - 0) / 10 = 0
    expect_equal(maxent_feature_eval(f, 2), 0.5)   # (5 - 0) / 10 = 0.5
    expect_equal(maxent_feature_eval(f, 3), 1.0)   # (10 - 0) / 10 = 1
    expect_equal(maxent_feature_eval(f, 4), 0.3)   # (3 - 0) / 10 = 0.3
})

test_that("LinearFeature auto-computes min/max", {
    skip_if_not_installed("maxentcpp")

    vals <- c(2.0, 4.0, 6.0, 8.0)
    f <- maxent_linear_feature(vals, "temp")  # no min_val/max_val
    info <- maxent_feature_info(f)
    expect_equal(info$min, 2.0)
    expect_equal(info$max, 8.0)
    expect_equal(maxent_feature_eval(f, 1), 0.0)  # (2-2)/(8-2)
    expect_equal(maxent_feature_eval(f, 4), 1.0)  # (8-2)/(8-2)
})

test_that("LinearFeature min == max returns 0", {
    skip_if_not_installed("maxentcpp")

    vals <- c(5.0, 5.0, 5.0)
    f <- maxent_linear_feature(vals, "constant", min_val = 5.0, max_val = 5.0)
    expect_equal(maxent_feature_eval(f, 1), 0.0)
    expect_equal(maxent_feature_eval(f, 2), 0.0)
})

test_that("QuadraticFeature creation and evaluation work", {
    skip_if_not_installed("maxentcpp")

    vals <- c(0.0, 5.0, 10.0)
    f <- maxent_quadratic_feature(vals, "temperature_sq", min_val = 0.0, max_val = 10.0)
    expect_true(!is.null(f))

    expect_equal(maxent_feature_eval(f, 1), 0.0)   # (0/10)^2
    expect_equal(maxent_feature_eval(f, 2), 0.25)  # (5/10)^2
    expect_equal(maxent_feature_eval(f, 3), 1.0)   # (10/10)^2
})

test_that("QuadraticFeature min == max returns 0", {
    skip_if_not_installed("maxentcpp")

    vals <- c(3.0, 3.0)
    f <- maxent_quadratic_feature(vals, "flat_sq", min_val = 3.0, max_val = 3.0)
    expect_equal(maxent_feature_eval(f, 1), 0.0)
})

test_that("ProductFeature creation and evaluation work", {
    skip_if_not_installed("maxentcpp")

    v1 <- c(0.0, 5.0, 10.0)
    v2 <- c(0.0, 2.0, 4.0)
    f <- maxent_product_feature(v1, v2, "temp_x_prec",
                                min1 = 0, max1 = 10,
                                min2 = 0, max2 = 4)
    expect_true(!is.null(f))

    expect_equal(maxent_feature_eval(f, 1), 0.0)   # 0 * 0
    expect_equal(maxent_feature_eval(f, 2), 0.25)  # 0.5 * 0.5
    expect_equal(maxent_feature_eval(f, 3), 1.0)   # 1 * 1
})

test_that("ProductFeature auto-computes min/max", {
    skip_if_not_installed("maxentcpp")

    v1 <- c(0.0, 5.0, 10.0)
    v2 <- c(100.0, 200.0, 150.0)
    f <- maxent_product_feature(v1, v2, "temp_x_prec")
    expect_true(!is.null(f))
    # Just check it runs without error
    val <- maxent_feature_eval(f, 1)
    expect_true(is.numeric(val))
})

test_that("ProductFeature mismatched lengths raise error", {
    skip_if_not_installed("maxentcpp")

    expect_error(
        maxent_product_feature(c(1, 2, 3), c(1, 2), "bad"),
        "same length"
    )
})

test_that("ThresholdFeature creation and evaluation work", {
    skip_if_not_installed("maxentcpp")

    vals <- c(1.0, 5.0, 10.0, 5.0)
    f <- maxent_threshold_feature(vals, "temp_threshold", threshold = 5.0)
    expect_true(!is.null(f))

    expect_equal(maxent_feature_eval(f, 1), 0.0)  # 1 < 5 → No
    expect_equal(maxent_feature_eval(f, 2), 1.0)  # 5 >= 5 → Yes (non-strict, matches Java)
    expect_equal(maxent_feature_eval(f, 3), 1.0)  # 10 > 5 → Yes
    expect_equal(maxent_feature_eval(f, 4), 1.0)  # 5 >= 5 → Yes (non-strict, matches Java)
})

test_that("HingeFeature (forward) creation and evaluation work", {
    skip_if_not_installed("maxentcpp")

    vals <- c(0.0, 5.0, 10.0, 3.0)
    f <- maxent_hinge_feature(vals, "temp_hinge", min_knot = 2.0, max_knot = 8.0)
    expect_true(!is.null(f))

    expect_equal(maxent_feature_eval(f, 1), 0.0)             # 0 <= 2 → 0
    expect_equal(maxent_feature_eval(f, 2), 0.5)             # (5-2)/(8-2) = 0.5
    expect_equal(maxent_feature_eval(f, 3), 8.0 / 6.0)      # (10-2)/(8-2)
    expect_equal(maxent_feature_eval(f, 4), 1.0 / 6.0)      # (3-2)/(8-2)
})

test_that("HingeFeature (reverse) creation and evaluation work", {
    skip_if_not_installed("maxentcpp")

    vals <- c(0.0, 5.0, 10.0, 8.0)
    f <- maxent_hinge_feature(vals, "temp_rev_hinge",
                              min_knot = 2.0, max_knot = 8.0,
                              reverse = TRUE)
    expect_true(!is.null(f))

    expect_equal(maxent_feature_eval(f, 1), 8.0 / 6.0)  # (8-0)/(8-2)
    expect_equal(maxent_feature_eval(f, 2), 0.5)         # (8-5)/(8-2)
    expect_equal(maxent_feature_eval(f, 3), 0.0)         # 10 >= 8 → 0
    expect_equal(maxent_feature_eval(f, 4), 0.0)         # 8 < 8? No → 0
})

test_that("HingeFeature invalid knots raise error", {
    skip_if_not_installed("maxentcpp")

    expect_error(
        maxent_hinge_feature(c(1, 2, 3), "bad", min_knot = 5, max_knot = 3),
        "min_knot"
    )
})

test_that("feature_get_info returns correct metadata", {
    skip_if_not_installed("maxentcpp")

    vals <- c(0.0, 5.0, 10.0)
    f <- maxent_linear_feature(vals, "my_feature", min_val = 0.0, max_val = 10.0)
    info <- maxent_feature_info(f)

    expect_equal(info$name,   "my_feature")
    expect_equal(info$type,   "linear")
    expect_equal(info$lambda, 0.0)
    expect_equal(info$min,    0.0)
    expect_equal(info$max,    10.0)
    expect_equal(info$size,   3L)
})

test_that("maxent_feature_eval validates index bounds", {
    skip_if_not_installed("maxentcpp")

    f <- maxent_linear_feature(c(0, 5, 10), "temp")

    expect_error(maxent_feature_eval(f, 0), "between 1 and 3")
    expect_error(maxent_feature_eval(f, 4), "between 1 and 3")
    expect_error(maxent_feature_eval(f, NA_real_), "single finite value")
    expect_error(maxent_feature_eval(f, 1.5), "whole number")
})

test_that("maxent_generate_features generates expected number of features", {
    skip_if_not_installed("maxentcpp")

    data <- list(
        temperature   = c(15.0, 20.0, 25.0, 18.0, 22.0),
        precipitation = c(100.0, 200.0, 150.0, 80.0, 300.0)
    )

    features <- maxent_generate_features(
        data,
        types = c("linear", "quadratic"),
        n_thresholds = 0,
        n_hinges = 0
    )

    # 2 variables × (1 linear + 1 quadratic) = 4
    expect_equal(length(features), 4L)

    types <- sapply(features, function(f) maxent_feature_info(f)$type)
    expect_true("linear" %in% types)
    expect_true("quadratic" %in% types)
})

test_that("maxent_generate_features: all types", {
    skip_if_not_installed("maxentcpp")

    data <- list(
        temp = c(10.0, 20.0, 30.0),
        prec = c(50.0, 100.0, 200.0)
    )

    features <- maxent_generate_features(
        data,
        types = c("linear", "quadratic", "product", "threshold", "hinge"),
        n_thresholds = 3,
        n_hinges = 2
    )

    # Per variable: linear(1) + quadratic(1) + threshold(3) + hinge_fwd(2) + hinge_rev(2) = 9
    # Product pairs: 1
    # Total: 9*2 + 1 = 19
    expect_equal(length(features), 19L)

    types <- sapply(features, function(f) maxent_feature_info(f)$type)
    expect_true("linear"        %in% types)
    expect_true("quadratic"     %in% types)
    expect_true("product"       %in% types)
    expect_true("threshold"     %in% types)
    expect_true("hinge"         %in% types)
    expect_true("reverse_hinge" %in% types)
})

test_that("maxent_generate_features: single variable, linear only", {
    skip_if_not_installed("maxentcpp")

    data <- list(elevation = c(100.0, 500.0, 1000.0))
    features <- maxent_generate_features(data, types = "linear",
                                         n_thresholds = 0, n_hinges = 0)
    expect_equal(length(features), 1L)
    expect_equal(maxent_feature_info(features[[1]])$type, "linear")
})

test_that("maxent_generate_features: unnamed list raises error", {
    skip_if_not_installed("maxentcpp")

    expect_error(
        maxent_generate_features(list(c(1, 2, 3))),
        "named list"
    )
})

#' Generate the Maxent Canonical Color Ramp
#'
#' Produces a vector of hex color strings matching the color ramp used by the
#' Java Maxent software (\code{density/Display.java::showColor()}).  The
#' default palette cycles Red → Yellow → Green → Cyan → Blue across 1020
#' steps, with higher values rendered as red and lower values as blue.
#'
#' @param n Integer: number of colors to generate (default 1020, matching the
#'   Java implementation).
#' @param mode Character: one of \code{"plain"} (default),
#'   \code{"log"} (logarithmic spacing), \code{"blackandwhite"}
#'   (greyscale, white = high, black = low), or \code{"redandyellow"}
#'   (red–yellow ramp).
#' @return Character vector of \code{n} hex color strings of the form
#'   \code{"#RRGGBB"}.
#' @export
#' @examples
#' pal <- maxent_color_ramp(1020)
#' pal[1]     # "#FF0000"  (max value → red)
#' pal[510]   # near "#00FF00" (mid-point → green)
#' pal[1020]  # "#0000FF"  (min value → blue)
maxent_color_ramp <- function(n = 1020L, mode = "plain") {
    n <- as.integer(n)
    if (n < 1L) stop("'n' must be >= 1")

    mode <- match.arg(mode, c("plain", "log", "blackandwhite", "redandyellow"))

    if (mode == "blackandwhite") {
        # White (high, index 1) → Black (low, index n)
        vals <- seq(255, 0, length.out = n)
        v    <- as.integer(round(vals))
        return(sprintf("#%02X%02X%02X", v, v, v))
    }

    if (mode == "redandyellow") {
        # Red (low) → Yellow (high)
        vals <- seq(0, 1, length.out = n)
        r    <- as.integer(255L)
        g    <- as.integer(round(vals * 255))
        b    <- as.integer(0L)
        return(sprintf("#%02X%02X%02X", r, g, b))
    }

    # Positions in [0, 1]: index 1 = max (red), index n = min (blue).
    # Replicate Java logic:  show_color divides [0, 255*4) into 4 bands.
    #   Band 0 (255..192): R=255,  G=0..255,  B=0
    #   Band 1 (192..128): R=255..0, G=255, B=0
    #   Band 2 (128..64):  R=0,   G=255, B=0..255
    #   Band 3 (64..0):    R=0,   G=255..0, B=255
    # frac=0 (index 1) → band 0 → red (high); frac=1 (index n) → band 3 → blue (low).

    # Fractional position from 0 (index 1, red/max) to 1 (index n, blue/min)
    frac <- if (mode == "log") {
        # log spacing: more resolution near low values
        log_pos <- seq(0, 1, length.out = n)
        (exp(log_pos) - 1) / (exp(1) - 1)
    } else {
        seq(0, 1, length.out = n)
    }

    # v in [0, n-1]: 0 = max (red), n-1 = min (blue)
    v    <- frac * (n - 1L)
    band <- floor(v / (n / 4))
    pos  <- (v %% (n / 4)) / (n / 4 - 1L)    # 0..1 within band
    pos  <- pmin(pmax(pos, 0), 1)

    r <- g <- b <- integer(n)

    # Band 0: R=255 G=0→255 B=0
    sel <- band == 0
    r[sel] <- 255L; g[sel] <- as.integer(round(pos[sel] * 255)); b[sel] <- 0L

    # Band 1: R=255→0 G=255 B=0
    sel <- band == 1
    r[sel] <- as.integer(round((1 - pos[sel]) * 255)); g[sel] <- 255L; b[sel] <- 0L

    # Band 2: R=0 G=255 B=0→255
    sel <- band == 2
    r[sel] <- 0L; g[sel] <- 255L; b[sel] <- as.integer(round(pos[sel] * 255))

    # Band 3 (and overflow): R=0 G=255→0 B=255
    sel <- band >= 3
    r[sel] <- 0L; g[sel] <- as.integer(round((1 - pos[sel]) * 255)); b[sel] <- 255L

    # Clamp
    r <- pmin(pmax(r, 0L), 255L)
    g <- pmin(pmax(g, 0L), 255L)
    b <- pmin(pmax(b, 0L), 255L)

    # Index 1 = max (red), index n = min (blue)
    sprintf("#%02X%02X%02X", r, g, b)
}

test_that("maxent_color_ramp returns correct length and type", {
    pal <- maxent_color_ramp(1020)
    expect_length(pal, 1020)
    expect_true(is.character(pal))
    expect_true(all(grepl("^#[0-9A-F]{6}$", pal)))
})

test_that("maxent_color_ramp first element is red (max value)", {
    pal <- maxent_color_ramp(1020)
    expect_equal(pal[1], "#FF0000")
})

test_that("maxent_color_ramp last element is blue (min value)", {
    pal <- maxent_color_ramp(1020)
    expect_equal(pal[1020], "#0000FF")
})

test_that("maxent_color_ramp mid-point is near green", {
    pal <- maxent_color_ramp(1020)
    # Around index 510 the ramp passes through green (#00FF00)
    rgb_mid <- grDevices::col2rgb(pal[510])
    expect_true(rgb_mid["green", 1] > 200)
})

test_that("maxent_color_ramp custom n works", {
    pal <- maxent_color_ramp(100)
    expect_length(pal, 100)
})

test_that("maxent_color_ramp blackandwhite mode", {
    pal <- maxent_color_ramp(100, mode = "blackandwhite")
    expect_length(pal, 100)
    # First should be whitest, last should be darkest (or vice versa)
    rgb_first <- grDevices::col2rgb(pal[1])
    rgb_last  <- grDevices::col2rgb(pal[100])
    # In blackandwhite: first = white (high), last = black (low)
    # The ramp builds low → high then reverses, so [1] should be whiter
    expect_true(mean(rgb_first) > mean(rgb_last))
})

test_that("maxent_color_ramp redandyellow mode", {
    pal <- maxent_color_ramp(10, mode = "redandyellow")
    expect_length(pal, 10)
    expect_true(all(grepl("^#[0-9A-F]{6}$", pal)))
})

test_that("maxent_color_ramp log mode produces valid colors", {
    pal <- maxent_color_ramp(50, mode = "log")
    expect_length(pal, 50)
    expect_true(all(grepl("^#[0-9A-F]{6}$", pal)))
})

test_that("maxent_color_ramp n=1 does not error", {
    pal <- maxent_color_ramp(1)
    expect_length(pal, 1)
})

test_that("maxent_color_ramp errors on n < 1", {
    expect_error(maxent_color_ramp(0))
})

# Shared helpers for tests that depend on terra / PROJ.

skip_if_no_terra <- function() {
    testthat::skip_if_not_installed("terra")
}

# Skip when the PROJ database is not available (e.g. macOS CI runners
# where proj.db is not installed).  Without proj.db, terra cannot resolve
# CRS strings like "EPSG:4326" and throws "[rast] empty srs".
skip_if_no_proj <- function() {
    skip_if_no_terra()
    has_proj <- tryCatch({
        r <- terra::rast(nrows = 1L, ncols = 1L, crs = "EPSG:4326")
        nchar(terra::crs(r)) > 0L
    }, warning = function(w) {
        if (grepl("proj", conditionMessage(w), ignore.case = TRUE))
            return(FALSE)
        TRUE
    }, error = function(e) FALSE)
    if (!has_proj) testthat::skip("PROJ database (proj.db) not available")
}

## Runs the quickstart example script as part of the package test suite.
## This ensures the complete end-to-end workflow stays in sync with the
## bundled data and exported API.

test_that("quickstart example runs without errors", {
    skip_if_not_installed("maxentcpp")
    skip_if_no_proj()

    script <- system.file("examples", "quickstart.R",
                          package = "maxentcpp")
    expect_true(nchar(script) > 0, info = "quickstart.R not found in package")

    # Run the script in a clean environment so it does not pollute the test
    # namespace, and capture any messages/warnings via tryCatch.
    env <- new.env(parent = globalenv())
    expect_no_error(source(script, local = env))
})

#' @useDynLib maxentcpp, .registration = TRUE
#' @importFrom Rcpp sourceCpp
NULL

#' Create a MaxEnt Sample Object
#'
#' Creates a sample point representing a species occurrence location.
#'
#' @param lon Longitude coordinate
#' @param lat Latitude coordinate
#' @param name Sample identifier/name
#' @param dim Optional GridDimension object to calculate row/col indices
#' @return Sample object (external pointer)
#' @export
#' @examples
#' sample <- maxent_sample(lon = -118.5, lat = 36.5, name = "site1")
maxent_sample <- function(lon, lat, name = "", dim = NULL) {
    if (is.null(dim)) {
        # Create sample with dummy row/col if no dimension provided
        create_sample(0, 0, 0, lat, lon, name)
    } else {
        rc <- coords_to_rowcol(dim, lon, lat)
        create_sample(0, rc[1], rc[2], lat, lon, name)
    }
}

#' Create a Grid Dimension Object
#'
#' Defines the spatial extent and resolution of a raster grid.
#'
#' @param nrows Number of rows
#' @param ncols Number of columns
#' @param xll X coordinate of lower-left corner
#' @param yll Y coordinate of lower-left corner
#' @param cellsize Cell size (square cells)
#' @return GridDimension object (external pointer)
#' @export
#' @examples
#' dim <- maxent_dimension(
#'   nrows = 100, ncols = 100,
#'   xll = -120.0, yll = 35.0,
#'   cellsize = 0.1
#' )
maxent_dimension <- function(nrows, ncols, xll, yll, cellsize) {
    create_grid_dimension(nrows, ncols, xll, yll, cellsize)
}

#' Create a Grid Object
#'
#' Creates a raster grid for storing environmental variable data.
#'
#' @param dim GridDimension object defining spatial extent
#' @param name Grid layer name
#' @param nodata_value Value representing missing data (default: -9999)
#' @return Grid object (external pointer)
#' @export
#' @examples
#' dim <- maxent_dimension(100, 100, -120, 35, 0.1)
#' grid <- maxent_grid(dim, "temperature")
maxent_grid <- function(dim, name = "", nodata_value = -9999) {
    create_grid_float(dim, name, nodata_value)
}

#' Get Grid Information
#'
#' Retrieves properties of a grid object.
#'
#' @param grid Grid object
#' @return List with grid properties
#' @keywords internal
grid_info <- function(grid) {
    # Get dimension info if available
    tryCatch({
        list(
            name = "grid",  # Would need to add getter for this
            rows = attr(grid, "rows"),
            cols = attr(grid, "cols")
        )
    }, error = function(e) {
        list(error = "Could not retrieve grid info")
    })
}

#' Convert Grid to Raster Matrix
#'
#' Extracts grid data as an R matrix.
#'
#' @param grid Grid object
#' @return Numeric matrix
#' @keywords internal
as_matrix <- function(grid) {
    grid_to_matrix(grid)
}

#' Set Grid from Matrix
#'
#' Populates grid with values from an R matrix.
#'
#' @param grid Grid object
#' @param mat Numeric matrix of values
#' @return Invisibly returns the grid object.
#' @keywords internal
set_grid_matrix <- function(grid, mat) {
    grid_from_matrix(grid, mat)
    invisible(grid)
}

#' Print Sample Information
#'
#' @param x Sample object
#' @param ... Additional arguments (ignored)
#' @return Invisibly returns \code{x}.
#' @export
print.maxent_sample <- function(x, ...) {
    info <- get_sample_info(x)
    cat("MaxEnt Sample\n")
    cat(sprintf("  Name: %s\n", info$name))
    cat(sprintf("  Location: (%.4f, %.4f)\n", info$lon, info$lat))
    cat(sprintf("  Grid indices: [%d, %d]\n", info$row, info$col))
    invisible(x)
}

#' Read Species Occurrence Data
#'
#' Reads species presence records from a CSV file or an R \code{data.frame}
#' and converts them to row/column indices suitable for
#' \code{\link{maxent_featured_space}}.
#'
#' @param file_or_df Either a character path to a CSV file, or a
#'   \code{data.frame} already loaded in R (e.g. from
#'   \code{rgbif::occ_data()}, GBIF download, or \code{read.csv()}).
#' @param dim  A GridDimension object (from \code{\link{maxent_dimension}})
#'   defining the study area grid.
#' @param lon_col Character: name of the longitude column
#'   (default \code{"longitude"}).
#' @param lat_col Character: name of the latitude column
#'   (default \code{"latitude"}).
#' @param name_col Character or \code{NULL}: column for sample names.
#'   If \code{NULL} (default), sequential names are generated.
#' @return A named list with:
#'   \describe{
#'     \item{samples}{List of Sample external pointers.}
#'     \item{rows}{Integer vector of row indices (0-based).}
#'     \item{cols}{Integer vector of column indices (0-based).}
#'     \item{indices}{Integer vector of 0-based flat indices
#'       (\code{row * ncols + col}), suitable for
#'       \code{\link{maxent_featured_space}}.}
#'   }
#' @export
#' @examples
#' dim <- maxent_dimension(100, 100, -120, 35, 0.1)
#' occ <- data.frame(longitude = c(-118.5, -119.0),
#'                   latitude  = c(36.5, 37.0))
#' result <- maxent_read_occurrences(occ, dim)
#' result$indices  # 0-based flat indices for FeaturedSpace
maxent_read_occurrences <- function(file_or_df, dim, lon_col = "longitude",
                                    lat_col = "latitude", name_col = NULL) {
    # Load data
    if (is.character(file_or_df) && length(file_or_df) == 1L) {
        df <- utils::read.csv(file_or_df, stringsAsFactors = FALSE)
    } else if (is.data.frame(file_or_df)) {
        df <- file_or_df
    } else {
        stop("'file_or_df' must be a file path (character) or a data.frame")
    }

    if (!lon_col %in% names(df)) {
        stop("Column '", lon_col, "' not found in data")
    }
    if (!lat_col %in% names(df)) {
        stop("Column '", lat_col, "' not found in data")
    }

    lons <- as.numeric(df[[lon_col]])
    lats <- as.numeric(df[[lat_col]])

    # Drop rows with missing coordinates
    valid <- !is.na(lons) & !is.na(lats)
    lons <- lons[valid]
    lats <- lats[valid]

    if (length(lons) == 0L) {
        stop("No valid occurrence records after removing NA coordinates")
    }

    # Generate names
    if (!is.null(name_col) && name_col %in% names(df)) {
        sample_names <- as.character(df[[name_col]][valid])
    } else {
        sample_names <- paste0("occ_", seq_along(lons))
    }

    # Get dimension info to compute flat indices
    dim_info <- get_grid_dimension_info(dim)
    ncols <- dim_info$ncols

    samples <- vector("list", length(lons))
    rows_vec <- integer(length(lons))
    cols_vec <- integer(length(lons))
    indices  <- integer(length(lons))

    for (i in seq_along(lons)) {
        rc <- coords_to_rowcol(dim, lons[i], lats[i])
        rows_vec[i] <- rc[1]
        cols_vec[i] <- rc[2]
        indices[i]  <- rc[1] * ncols + rc[2]
        samples[[i]] <- create_sample(0L, rc[1], rc[2], lats[i], lons[i],
                                      sample_names[i])
    }

    list(
        samples = samples,
        rows    = rows_vec,
        cols    = cols_vec,
        indices = indices
    )
}

#' Generate Background Sample Indices
#'
#' Randomly selects valid (non-NODATA) cells from a reference grid for use
#' as background points in MaxEnt modeling.
#'
#' @param grid External pointer to a GridFloat object used as a reference
#'   (e.g. an environmental layer). Only cells with valid data are eligible.
#' @param n  Integer: number of background points to sample
#'   (default \code{10000L}).
#' @param seed  Integer or \code{NULL}: random seed for reproducibility.
#'   If \code{NULL} (default), no seed is set.
#' @return A named list with:
#'   \describe{
#'     \item{rows}{Integer vector of row indices (0-based).}
#'     \item{cols}{Integer vector of column indices (0-based).}
#'     \item{indices}{Integer vector of 0-based flat indices
#'       (\code{row * ncols + col}), suitable for
#'       \code{\link{maxent_featured_space}}.}
#'   }
#' @export
#' @examples
#' \donttest{
#' dim <- maxent_dimension(100, 100, -120, 35, 0.1)
#' grid <- maxent_grid(dim, "env1")
#' bg <- maxent_background_indices(grid, n = 5000, seed = 42)
#' bg$indices  # 0-based flat indices for FeaturedSpace
#' }
maxent_background_indices <- function(grid, n = 10000L, seed = NULL) {
    info <- maxent_grid_info(grid)
    nrows <- info$nrows
    ncols <- info$ncols

    mat <- maxent_grid_to_matrix(grid)

    # Identify valid (non-NA) cells
    valid_cells <- which(!is.na(mat), arr.ind = TRUE)
    if (nrow(valid_cells) == 0L) {
        stop("No valid (non-NODATA) cells in the grid")
    }

    n <- min(as.integer(n), nrow(valid_cells))
    if (!is.null(seed)) set.seed(seed)

    chosen <- sample.int(nrow(valid_cells), size = n)

    # valid_cells uses 1-based row/col; convert to 0-based
    rows_0 <- as.integer(valid_cells[chosen, 1] - 1L)
    cols_0 <- as.integer(valid_cells[chosen, 2] - 1L)
    flat   <- rows_0 * ncols + cols_0

    list(
        rows    = rows_0,
        cols    = cols_0,
        indices = flat
    )
}

test_that("maxent_read_occurrences works with data.frame", {
    skip_if_not_installed("maxentcpp")

    dim <- maxent_dimension(100, 100, -120, 35, 0.1)

    occ_df <- data.frame(
        longitude = c(-119.5, -119.0, -118.5),
        latitude  = c(36.0, 37.0, 38.0)
    )

    result <- maxent_read_occurrences(occ_df, dim)

    expect_true(is.list(result))
    expect_equal(length(result$samples), 3)
    expect_equal(length(result$rows), 3)
    expect_equal(length(result$cols), 3)
    expect_equal(length(result$indices), 3)

    # All indices should be non-negative
    expect_true(all(result$rows >= 0))
    expect_true(all(result$cols >= 0))
    expect_true(all(result$indices >= 0))

    # indices should be row * ncols + col
    for (i in seq_along(result$indices)) {
        expect_equal(result$indices[i], result$rows[i] * 100 + result$cols[i])
    }
})

test_that("maxent_read_occurrences works with CSV file", {
    skip_if_not_installed("maxentcpp")

    dim <- maxent_dimension(100, 100, -120, 35, 0.1)

    tmp <- tempfile(fileext = ".csv")
    on.exit(unlink(tmp), add = TRUE)

    writeLines(c("species,lon,lat",
                 "oak,-119.5,36.0",
                 "pine,-119.0,37.0"), tmp)

    result <- maxent_read_occurrences(tmp, dim,
                                      lon_col = "lon", lat_col = "lat")

    expect_equal(length(result$samples), 2)
    expect_equal(length(result$indices), 2)
})

test_that("maxent_read_occurrences handles custom name column", {
    skip_if_not_installed("maxentcpp")

    dim <- maxent_dimension(100, 100, -120, 35, 0.1)

    occ_df <- data.frame(
        longitude = c(-119.5, -119.0),
        latitude  = c(36.0, 37.0),
        site_name = c("site_A", "site_B")
    )

    result <- maxent_read_occurrences(occ_df, dim, name_col = "site_name")

    info1 <- get_sample_info(result$samples[[1]])
    info2 <- get_sample_info(result$samples[[2]])

    expect_equal(info1$name, "site_A")
    expect_equal(info2$name, "site_B")
})

test_that("maxent_read_occurrences drops NA coordinates", {
    skip_if_not_installed("maxentcpp")

    dim <- maxent_dimension(100, 100, -120, 35, 0.1)

    occ_df <- data.frame(
        longitude = c(-119.5, NA, -118.5),
        latitude  = c(36.0, 37.0, NA)
    )

    result <- maxent_read_occurrences(occ_df, dim)

    # Only the first row has valid lon+lat
    expect_equal(length(result$samples), 1)
    expect_equal(length(result$indices), 1)
})

test_that("maxent_read_occurrences errors on missing columns", {
    skip_if_not_installed("maxentcpp")

    dim <- maxent_dimension(100, 100, -120, 35, 0.1)

    occ_df <- data.frame(x = 1, y = 2)

    expect_error(maxent_read_occurrences(occ_df, dim),
                 "longitude")
})

test_that("maxent_background_indices returns valid indices", {
    skip_if_not_installed("maxentcpp")

    # Create a grid with some valid and some NA cells
    mat <- matrix(runif(100), nrow = 10, ncol = 10)
    mat[1, 1] <- NA
    mat[5, 5] <- NA

    g <- maxent_grid_from_matrix(mat, xll = 0, yll = 0, cellsize = 1)

    bg <- maxent_background_indices(g, n = 20, seed = 42)

    expect_true(is.list(bg))
    expect_equal(length(bg$rows), 20)
    expect_equal(length(bg$cols), 20)
    expect_equal(length(bg$indices), 20)

    # All indices should be non-negative
    expect_true(all(bg$rows >= 0))
    expect_true(all(bg$cols >= 0))
    expect_true(all(bg$indices >= 0))

    # None of the selected cells should be NA
    for (i in seq_along(bg$rows)) {
        val <- grid_get_value(g, bg$rows[i], bg$cols[i])
        expect_false(is.na(val))
    }
})

test_that("maxent_background_indices respects seed", {
    skip_if_not_installed("maxentcpp")

    mat <- matrix(runif(100), nrow = 10, ncol = 10)
    g <- maxent_grid_from_matrix(mat, xll = 0, yll = 0, cellsize = 1)

    bg1 <- maxent_background_indices(g, n = 20, seed = 123)
    bg2 <- maxent_background_indices(g, n = 20, seed = 123)

    expect_equal(bg1$indices, bg2$indices)
})

test_that("maxent_background_indices caps n at valid cell count", {
    skip_if_not_installed("maxentcpp")

    mat <- matrix(c(1, 2, NA, 4), nrow = 2, ncol = 2)
    g <- maxent_grid_from_matrix(mat, xll = 0, yll = 0, cellsize = 1)

    bg <- maxent_background_indices(g, n = 100, seed = 1)

    # Only 3 valid cells, so should get at most 3
    expect_equal(length(bg$indices), 3)
})

# ---- Stage 3: Response Curve PNG Writer -------------------------------------

#' Write Response Curve PNGs for All Variables
#'
#' Generates response curve plots for each environmental variable and saves
#' them as PNG files, replicating \code{density/ResponsePlot.java::makeplot()}.
#' A full-size PNG and an optional thumbnail are written for every variable.
#'
#' @param model         External pointer to a trained FeaturedSpace object.
#' @param env_grids     List of external pointers to Grid<float> objects.
#' @param feature_names Character vector of environment variable names
#'   (one entry per element of \code{env_grids}).
#' @param output_dir    Character: directory under which a \code{plots/}
#'   sub-directory will be created.
#' @param species       Character: species name (used in file names).
#' @param var_indices   Integer vector of 0-based variable indices to plot.
#'   Defaults to all variables.
#' @param n_steps       Integer: number of steps in each curve (default 100).
#' @param thumbnail     Logical: also write a 210 × 140 pixel thumbnail PNG
#'   (default \code{TRUE}).
#' @param write_dat     Logical: also write a tab-delimited \code{.dat} file
#'   of the curve data (default \code{FALSE}).
#' @return Invisibly returns a named list of file paths written.
#' @export
#' @examples
#' \donttest{
#' maxent_plot_response_curves(
#'   model, list(g1, g2), c("bio1", "bio12"),
#'   output_dir = tempdir(), species = "Sp1")
#' }
maxent_plot_response_curves <- function(model, env_grids, feature_names,
                                        output_dir, species,
                                        var_indices = NULL,
                                        n_steps     = 100L,
                                        thumbnail   = TRUE,
                                        write_dat   = FALSE) {
    feature_names <- as.character(feature_names)
    n_vars <- length(feature_names)

    if (is.null(var_indices)) {
        var_indices <- seq_len(n_vars) - 1L   # 0-based
    }
    var_indices <- as.integer(var_indices)

    plots_dir <- file.path(output_dir, "plots")
    if (!dir.exists(plots_dir)) dir.create(plots_dir, recursive = TRUE)

    file_list <- list()

    for (vi in var_indices) {
        var_name <- feature_names[vi + 1L]
        curve    <- maxent_response_curve(model, env_grids, feature_names,
                                          var_index = vi,
                                          n_steps   = as.integer(n_steps))

        # Full-size plot
        full_path <- file.path(plots_dir,
                               paste0(species, "_", var_name, ".png"))
        grDevices::png(full_path, width = 600L, height = 400L)
        oldpar <- graphics::par(mar = c(4, 4, 2, 1))
        graphics::plot(curve$value, curve$prediction,
                       type = "l", lwd = 2L, col = "#1B7837",
                       xlab = var_name,
                       ylab = "Cloglog prediction",
                       main = paste("Response curve:", var_name),
                       ylim = c(0, 1))
        graphics::par(oldpar)
        grDevices::dev.off()
        file_list[[paste0(var_name, "_full")]] <- full_path

        # Thumbnail
        if (thumbnail) {
            thumb_path <- file.path(plots_dir,
                                    paste0(species, "_", var_name, "_thumb.png"))
            grDevices::png(thumb_path, width = 210L, height = 140L)
            oldpar_th <- graphics::par(mar = c(2, 2, 1, 0.5))
            graphics::plot(curve$value, curve$prediction,
                           type = "l", lwd = 1L, col = "#1B7837",
                           xlab = "", ylab = "",
                           main = var_name, cex.main = 0.75,
                           ylim = c(0, 1))
            graphics::par(oldpar_th)
            grDevices::dev.off()
            file_list[[paste0(var_name, "_thumb")]] <- thumb_path
        }

        # Optional .dat file
        if (write_dat) {
            dat_path <- file.path(plots_dir,
                                  paste0(species, "_", var_name, ".dat"))
            utils::write.table(curve, dat_path, sep = "\t",
                               row.names = FALSE, quote = FALSE)
            file_list[[paste0(var_name, "_dat")]] <- dat_path
        }
    }

    invisible(file_list)
}


# ---- Stage 4: Variable Importance Bar Chart ---------------------------------

#' Write a Variable Importance Bar Chart PNG
#'
#' Creates a horizontal bar chart comparing percent contribution and
#' permutation importance for each environmental variable, replicating
#' \code{density/Runner.java::makeJackknifePlots()}.
#'
#' @param contributions_df A data.frame with columns \code{name} and
#'   \code{contribution} (from \code{\link{maxent_percent_contribution}}).
#' @param perm_imp_df      A data.frame with columns \code{name} and
#'   \code{permutation_importance} (from
#'   \code{\link{maxent_permutation_importance}}).
#' @param species          Character: species name (used in the filename).
#' @param output_dir       Character: directory under which a \code{plots/}
#'   sub-directory will be created.
#' @return Invisibly returns the path to the PNG file.
#' @export
#' @examples
#' \donttest{
#' maxent_plot_variable_importance(contrib, perm_imp,
#'   species = "Sp1", output_dir = tempdir())
#' }
maxent_plot_variable_importance <- function(contributions_df, perm_imp_df,
                                            species, output_dir) {
    plots_dir <- file.path(output_dir, "plots")
    if (!dir.exists(plots_dir)) dir.create(plots_dir, recursive = TRUE)

    # Merge by name and sort by contribution (descending)
    merged <- merge(contributions_df, perm_imp_df, by = "name", all = TRUE)
    merged[is.na(merged)] <- 0
    merged <- merged[order(merged$contribution, decreasing = FALSE), ]

    n_vars <- nrow(merged)
    png_path <- file.path(plots_dir, paste0(species, "_varimp.png"))

    fig_height <- max(300L, 60L * n_vars + 100L)
    grDevices::png(png_path, width = 600L, height = fig_height)
    oldpar <- graphics::par(mar = c(4, max(nchar(merged$name)) * 0.55, 3, 1))
    on.exit({graphics::par(oldpar); grDevices::dev.off()}, add = TRUE)

    bar_mat <- rbind(merged$contribution,
                     merged$permutation_importance)
    colnames(bar_mat) <- merged$name

    graphics::barplot(bar_mat,
                      beside    = TRUE,
                      horiz     = TRUE,
                      las       = 1,
                      col       = c("#2166AC", "#D6604D"),
                      border    = NA,
                      xlab      = "Percent (%)",
                      main      = paste("Variable importance:", species),
                      xlim      = c(0, 100),
                      names.arg = merged$name,
                      legend.text = c("Percent contribution",
                                      "Permutation importance"),
                      args.legend = list(x = "bottomright",
                                         bty = "n",
                                         cex = 0.8))
    invisible(png_path)
}

#' Species occurrence points from Abeillia abeillei presence points after
#' download and clean from GBIF. This is a hummingbird example.
#' A dataset with three variables. Contains scientific name, longitude and
#' latitude.
#' @format A data frame with 73 rows and 3 variables:
#' \describe{
#'   \item{species}{Species name}
#'   \item{long}{Decimal longitude geographical coordinate, in degrees}
#'   \item{lat}{Decimal latitude geographical coordinate, in degrees}
#' }
#' @source <doi:10.1016/j.ecolmodel.2021.109823>
"example_occ_df"

#' Train a MaxEnt Model with the Java-Equivalent Sequential Optimizer
#'
#' Runs the full \code{density.Sequential} optimizer ported from the
#' original Java Maxent 3.4.4 on a FeaturedSpace.  Unlike
#' \code{maxent_fit()}, which uses the historical \code{goodAlpha}-only
#' loop, this trainer reproduces the real Java optimizer: feature
#' selection via \code{deltaLossBound}, Newton step with 1-D line search
#' (\code{searchAlpha}), every-10-iteration \code{doParallelUpdate} with
#' undo on loss-violating batch steps, and per-feature state tracking.
#'
#' The lambda trajectory produced by this trainer matches the Java
#' oracle's to \eqn{10^{-6}} on \eqn{\left\|\Delta\lambda\right\|_\infty}
#' at every checkpoint, on both the symmetric and asymmetric fixtures in
#' \code{maxentcppCompTest} — see
#' \code{docs/ARCHITECTURE_xtensor_openmp.md} and the Phase C baseline
#' report for details.
#'
#' @param featured_space External pointer to a FeaturedSpace object (from
#'   \code{maxent_featured_space()}).
#' @param max_iter Maximum number of iterations (default \code{500L}).
#' @param convergence Convergence threshold on the per-20-iteration loss
#'   improvement (default \code{1e-5}).  Ignored when
#'   \code{disable_convergence_test = TRUE}.
#' @param beta_multiplier Regularization multiplier (default \code{1.0}).
#' @param min_deviation Minimum sample-deviation floor used in
#'   regularization (default \code{0.001}).
#' @param parallel_update_frequency Iteration frequency at which
#'   \code{doParallelUpdate} runs (default \code{10L}, matching Java).
#' @param disable_convergence_test Logical: when \code{TRUE}, the loop
#'   runs a fixed \code{max_iter} iterations with no early stop.  This
#'   is required for deterministic per-iteration trajectory comparisons
#'   against the Java oracle.  Default \code{FALSE}.
#' @param trajectory_iterations Integer vector of 1-based iteration
#'   indices at which to capture \code{(loss, entropy, lambdas)}
#'   snapshots.  May be empty.  Snapshots outside
#'   \code{[1, max_iter]} are silently dropped.
#' @return Named list with:
#'   \describe{
#'     \item{loss}{Final regularized loss (scalar).}
#'     \item{entropy}{Shannon entropy of the trained distribution.}
#'     \item{iterations}{Number of training iterations completed.}
#'     \item{converged}{Logical: whether the convergence threshold was
#'       reached (always \code{FALSE} when
#'       \code{disable_convergence_test = TRUE}).}
#'     \item{lambdas}{Numeric vector of final lambda values.}
#'     \item{trajectory}{A \code{data.frame} with one row per captured
#'       checkpoint and columns \code{iteration, loss, entropy,
#'       lambda_0, lambda_1, ..., lambda_{J-1}}.}
#'   }
#' @seealso \code{\link{maxent_fit}}, \code{\link{maxent_featured_space}}.
#' @export
#' @examples
#' \donttest{
#' fs  <- maxent_featured_space(100L, 90:99, list(f))
#' res <- maxent_sequential_fit(
#'     fs,
#'     max_iter                 = 500L,
#'     disable_convergence_test = TRUE,
#'     trajectory_iterations    = c(1L, 2L, 5L, 10L, 50L, 100L, 500L))
#' print(res$trajectory)
#' }
maxent_sequential_fit <- function(featured_space,
                                  max_iter                  = 500L,
                                  convergence               = 1e-5,
                                  beta_multiplier           = 1.0,
                                  min_deviation             = 0.001,
                                  parallel_update_frequency = 10L,
                                  disable_convergence_test  = FALSE,
                                  trajectory_iterations     = integer(0)) {
    maxent_sequential_train(
        featured_space,
        as.integer(max_iter),
        as.double(convergence),
        as.double(beta_multiplier),
        as.double(min_deviation),
        as.integer(parallel_update_frequency),
        isTRUE(disable_convergence_test),
        as.integer(trajectory_iterations)
    )
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

#' Create a FeaturedSpace object
#'
#' Constructs a MaxEnt FeaturedSpace from background point count,
#' occurrence sample indices, and a list of Feature objects.
#'
#' @param num_points     Integer: number of background points.
#' @param sample_indices Integer vector: 0-based indices of occurrence samples
#'   in the background array.
#' @param feature_ptrs   List of external pointers to Feature objects
#'   (from \code{create_linear_feature()} etc.).
#' @param bias_weights   Optional numeric vector of per-point bias weights
#'   (length \code{num_points}).  When supplied, background density is
#'   computed as \code{bias[i] * exp(lp[i] - lpn)} instead of the standard
#'   \code{exp(lp[i] - lpn)}.  Pass an empty vector (default) for uniform
#'   (unbiased) background.
#' @return External pointer to a FeaturedSpace object.
#' @export
maxent_featured_space_create <- function(num_points, sample_indices, feature_ptrs, bias_weights = numeric(0)) {
    .Call(`_maxentcpp_maxent_featured_space_create`, num_points, sample_indices, feature_ptrs, bias_weights)
}

#' Train a FeaturedSpace model
#'
#' Runs the sequential coordinate-ascent MaxEnt optimization.
#'
#' @param fs_ptr            External pointer to a FeaturedSpace object.
#' @param max_iter          Maximum number of training iterations (default 500).
#' @param convergence       Convergence threshold (default 1e-5).
#' @param beta_multiplier   Regularization multiplier (default 1.0).
#' @param min_deviation     Minimum sample deviation floor (default 0.001).
#' @return Named list with elements: \code{loss}, \code{entropy},
#'   \code{iterations}, \code{converged}, \code{lambdas}.
#' @export
maxent_train <- function(fs_ptr, max_iter = 500L, convergence = 1e-5, beta_multiplier = 1.0, min_deviation = 0.001) {
    .Call(`_maxentcpp_maxent_train`, fs_ptr, max_iter, convergence, beta_multiplier, min_deviation)
}

#' Predict with a trained FeaturedSpace model
#'
#' Computes raw Gibbs distribution scores for new environmental data.
#'
#' @param fs_ptr    External pointer to a trained FeaturedSpace object.
#' @param new_data  Numeric matrix with one row per new point and one column
#'   per feature (values must be pre-evaluated, i.e., the feature
#'   transformation already applied).
#' @return Numeric vector of raw scores (unnormalized).
#' @export
maxent_predict <- function(fs_ptr, new_data) {
    .Call(`_maxentcpp_maxent_predict`, fs_ptr, new_data)
}

#' Get current distribution weights from a FeaturedSpace
#'
#' @param fs_ptr External pointer to a FeaturedSpace object.
#' @return Numeric vector of normalized weights (sums to 1).
#' @export
maxent_get_weights <- function(fs_ptr) {
    .Call(`_maxentcpp_maxent_get_weights`, fs_ptr)
}

#' Get entropy of a FeaturedSpace distribution
#'
#' @param fs_ptr External pointer to a FeaturedSpace object.
#' @return Shannon entropy (non-negative scalar).
#' @export
maxent_get_entropy <- function(fs_ptr) {
    .Call(`_maxentcpp_maxent_get_entropy`, fs_ptr)
}

#' Get current loss of a FeaturedSpace
#'
#' @param fs_ptr External pointer to a FeaturedSpace object.
#' @return Scalar loss value (negative log-likelihood).
#' @export
maxent_get_loss <- function(fs_ptr) {
    .Call(`_maxentcpp_maxent_get_loss`, fs_ptr)
}

#' Write feature lambdas to a file
#'
#' Saves the trained model coefficients in CSV format compatible with
#' the original Java Maxent .lambdas file format.
#'
#' @param fs_ptr   External pointer to a trained FeaturedSpace object.
#' @param filename Character: path to the output file.
#' @return Called for side effects; returns invisibly.
#' @export
maxent_write_lambdas <- function(fs_ptr, filename) {
    invisible(.Call(`_maxentcpp_maxent_write_lambdas`, fs_ptr, filename))
}

#' Read feature lambdas from a file
#'
#' Restores model coefficients from a .lambdas file.  The FeaturedSpace
#' must have been created with the same features (same names and order).
#'
#' @param fs_ptr   External pointer to a FeaturedSpace object.
#' @param filename Character: path to the lambdas file.
#' @return Called for side effects; returns invisibly.
#' @export
maxent_read_lambdas <- function(fs_ptr, filename) {
    invisible(.Call(`_maxentcpp_maxent_read_lambdas`, fs_ptr, filename))
}

#' Set sample expectations for a FeaturedSpace
#'
#' Computes sample expectations and regularization deviations for each feature.
#' Called automatically by \code{maxent_train()}, but exposed for advanced use.
#'
#' @param fs_ptr          External pointer to a FeaturedSpace object.
#' @param beta_multiplier Regularization multiplier (default 1.0).
#' @param min_deviation   Minimum sample deviation (default 0.001).
#' @return Called for side effects; returns invisibly.
#' @export
maxent_set_sample_expectations <- function(fs_ptr, beta_multiplier = 1.0, min_deviation = 0.001) {
    invisible(.Call(`_maxentcpp_maxent_set_sample_expectations`, fs_ptr, beta_multiplier, min_deviation))
}

#' Train a FeaturedSpace with the Sequential optimizer (trajectory-capable)
#'
#' Runs the full \code{density.Sequential} optimizer ported from the original
#' Java Maxent, with optional per-iteration trajectory snapshots.  Unlike
#' \code{maxent_train()} which uses a \code{goodAlpha}-only loop, this trainer
#' reproduces the real Java optimizer's feature-selection (\code{deltaLossBound}),
#' Newton step, 1-D line search, and every-10-iter \code{doParallelUpdate}
#' with undo on loss-violating batch steps.
#'
#' @param fs_ptr                   External pointer to a FeaturedSpace object.
#' @param max_iter                 Maximum number of iterations (default 500).
#' @param convergence              Convergence threshold on \code{newLoss - oldLoss}
#'   (default \code{1e-5}).  Ignored when \code{disable_convergence_test=TRUE}.
#' @param beta_multiplier          Regularization multiplier (default 1.0).
#' @param min_deviation            Minimum sample deviation floor (default 0.001).
#' @param parallel_update_frequency Iteration frequency at which
#'   \code{doParallelUpdate} runs (default 10, matching Java).
#' @param disable_convergence_test When \code{TRUE}, the loop runs a fixed
#'   \code{max_iter} iterations with no early stop.  Needed for deterministic
#'   per-iteration trajectory comparisons against the Java oracle.
#' @param trajectory_iterations    Integer vector of 1-based iteration indices
#'   at which to capture \code{(loss, entropy, lambdas)} snapshots.  May be
#'   empty.  Snapshots outside \code{[1, max_iter]} are silently dropped.
#' @return Named list with elements:
#'   \code{loss}, \code{entropy}, \code{iterations}, \code{converged},
#'   \code{lambdas}, and \code{trajectory} — a data.frame with columns
#'   \code{iteration, loss, entropy, lambda_0, ..., lambda_{J-1}} holding
#'   one row per requested checkpoint that was actually reached.
#' @keywords internal
maxent_sequential_train <- function(fs_ptr, max_iter = 500L, convergence = 1e-5, beta_multiplier = 1.0, min_deviation = 0.001, parallel_update_frequency = 10L, disable_convergence_test = FALSE, trajectory_iterations = as.integer( c())) {
    .Call(`_maxentcpp_maxent_sequential_train`, fs_ptr, max_iter, convergence, beta_multiplier, min_deviation, parallel_update_frequency, disable_convergence_test, trajectory_iterations)
}

#' Get FeaturedSpace metadata
#'
#' @param fs_ptr External pointer to a FeaturedSpace object.
#' @return Named list with num_points, num_samples, num_features,
#'   density_normalizer, linear_predictor_normalizer, has_bias.
#' @export
maxent_featured_space_info <- function(fs_ptr) {
    .Call(`_maxentcpp_maxent_featured_space_info`, fs_ptr)
}

#' Enable or disable streaming-eval mode on a FeaturedSpace
#'
#' @param fs_ptr External pointer to a FeaturedSpace object.
#' @param enabled Logical flag.
#' @return Called for side effects; returns invisibly.
#' @keywords internal
maxent_set_streaming_eval <- function(fs_ptr, enabled = TRUE) {
    invisible(.Call(`_maxentcpp_maxent_set_streaming_eval`, fs_ptr, enabled))
}

#' Create a FeaturedSpace from a streaming background provider
#'
#' Builds a \code{FeaturedSpace} by draining a callback-style background
#' provider (typically backed by a \code{terra::SpatRaster} block loop)
#' one tile at a time, then constructing features from the concatenated
#' \code{(num_points x num_layers)} matrix via \code{\link{generate_features}}.
#'
#' This is the C++ entry point used by
#' \code{\link{maxent_featured_space_from_rast}}; end users should prefer
#' that R-level wrapper.
#'
#' @param num_points      Integer: total number of finite background points
#'   the provider will emit (sum of \code{nrow()} across tiles).
#' @param num_layers      Integer: number of environmental variables / raster
#'   layers per tile row.
#' @param layer_names     Character vector of length \code{num_layers}
#'   giving the name for each layer / column (used when generating feature
#'   names such as \code{bio1^2}, \code{bio1*bio2}).
#' @param sample_indices  Integer vector: 0-based indices of occurrence
#'   samples in the concatenated background stream.
#' @param feature_types   Character vector of feature types to generate.
#'   See \code{\link{maxent_generate_features}}.
#' @param n_thresholds    Number of threshold knots per variable.
#' @param n_hinges        Number of hinge knots per variable.
#' @param next_tile_fn    R function returning the next tile as a numeric
#'   matrix (nrow x num_layers), or a 0-row matrix / \code{NULL} when the
#'   stream is exhausted. The function is responsible for filtering out
#'   rows containing NAs.
#' @param reset_fn        R function (no arguments) that rewinds the
#'   underlying stream to the beginning.
#' @param use_cache Logical; wrap the callback stream in
#'   \code{CachingBackgroundProvider}.
#' @return External pointer to a \code{FeaturedSpace} object.
#' @export
maxent_featured_space_from_callback <- function(num_points, num_layers, layer_names, sample_indices, feature_types, n_thresholds, n_hinges, next_tile_fn, reset_fn, use_cache = TRUE) {
    .Call(`_maxentcpp_maxent_featured_space_from_callback`, num_points, num_layers, layer_names, sample_indices, feature_types, n_thresholds, n_hinges, next_tile_fn, reset_fn, use_cache)
}

#' Create a FeaturedSpace from a terra SpatRaster callback stream
#'
#' Same as \code{maxent_featured_space_from_callback()}, but also preserves
#' the source \code{SpatRaster} S4 object for provider lifetime safety and
#' can enable streaming-eval mode immediately.
#'
#' @param preserved_rast A terra \code{SpatRaster} S4 object to preserve.
#' @param rast_xptr External pointer extracted from the raster object
#'   (used for contract validation).
#' @param num_points Integer; total number of finite background points.
#' @param num_layers Integer; number of environmental variables per row.
#' @param layer_names Character vector of layer names (length num_layers).
#' @param sample_indices Integer vector; 0-based indices of presence samples.
#' @param feature_types Character vector of feature types to generate.
#' @param n_thresholds Integer; number of threshold knots per variable.
#' @param n_hinges Integer; number of hinge knots per variable.
#' @param next_tile_fn R function returning the next tile or NULL/0-row matrix.
#' @param reset_fn R function (no args) that rewinds the underlying stream.
#' @param enable_streaming_eval Logical; whether to enable streaming eval mode.
#' @param use_cache Logical; wrap the stream in \code{CachingBackgroundProvider}.
#' @return External pointer to a FeaturedSpace object.
#' @keywords internal
maxent_featured_space_from_spatraster_callback <- function(preserved_rast, rast_xptr, num_points, num_layers, layer_names, sample_indices, feature_types, n_thresholds, n_hinges, next_tile_fn, reset_fn, enable_streaming_eval = TRUE, use_cache = TRUE) {
    .Call(`_maxentcpp_maxent_featured_space_from_spatraster_callback`, preserved_rast, rast_xptr, num_points, num_layers, layer_names, sample_indices, feature_types, n_thresholds, n_hinges, next_tile_fn, reset_fn, enable_streaming_eval, use_cache)
}

#' Extract occurrence environmental values from a callback stream
#'
#' Drains a callback background stream once and returns the rows at the
#' requested 0-based finite-stream indices.
#'
#' @param num_points Integer; total finite background rows in stream.
#' @param num_layers Integer; number of environmental variables per row.
#' @param occurrence_indices Integer vector of 0-based finite-stream indices.
#' @param next_tile_fn R function returning the next tile or NULL/0-row matrix.
#' @param reset_fn R function (no args) that rewinds the underlying stream.
#' @param preserved_rast Optional SpatRaster object to preserve during read.
#' @return Numeric matrix with one row per requested occurrence index.
#' @keywords internal
maxent_extract_occurrence_from_callback <- function(num_points, num_layers, occurrence_indices, next_tile_fn, reset_fn, preserved_rast = NULL) {
    .Call(`_maxentcpp_maxent_extract_occurrence_from_callback`, num_points, num_layers, occurrence_indices, next_tile_fn, reset_fn, preserved_rast)
}

#' Create a LinearFeature object
#'
#' Creates a linear (normalized) feature: eval(i) = (values[i] - min) / (max - min).
#' Returns 0 when min == max.
#'
#' @param values Numeric vector of environmental variable values
#' @param name Feature name/identifier
#' @param min_val Minimum value for normalization
#' @param max_val Maximum value for normalization
#' @return External pointer to LinearFeature object
#' @export
create_linear_feature <- function(values, name, min_val, max_val) {
    .Call(`_maxentcpp_create_linear_feature`, values, name, min_val, max_val)
}

#' Create a QuadraticFeature object
#'
#' Creates a quadratic feature: eval(i) = linear_val^2 where
#' linear_val = (values[i] - min) / (max - min).
#'
#' @param values Numeric vector of environmental variable values
#' @param name Feature name/identifier
#' @param min_val Minimum value for normalization
#' @param max_val Maximum value for normalization
#' @return External pointer to QuadraticFeature object
#' @export
create_quadratic_feature <- function(values, name, min_val, max_val) {
    .Call(`_maxentcpp_create_quadratic_feature`, values, name, min_val, max_val)
}

#' Create a ProductFeature object
#'
#' Creates a product (interaction) feature between two environmental variables:
#' eval(i) = norm(values1[i]) * norm(values2[i]).
#'
#' @param values1 Numeric vector for the first environmental variable
#' @param values2 Numeric vector for the second environmental variable
#' @param name Feature name/identifier
#' @param min1 Minimum of the first variable
#' @param max1 Maximum of the first variable
#' @param min2 Minimum of the second variable
#' @param max2 Maximum of the second variable
#' @return External pointer to ProductFeature object
#' @export
create_product_feature <- function(values1, values2, name, min1, max1, min2, max2) {
    .Call(`_maxentcpp_create_product_feature`, values1, values2, name, min1, max1, min2, max2)
}

#' Create a ThresholdFeature object
#'
#' Creates a binary step feature: eval(i) = 1.0 if values[i] > threshold, else 0.0.
#'
#' @param values Numeric vector of environmental variable values
#' @param name Feature name/identifier
#' @param threshold The threshold value
#' @return External pointer to ThresholdFeature object
#' @export
create_threshold_feature <- function(values, name, threshold) {
    .Call(`_maxentcpp_create_threshold_feature`, values, name, threshold)
}

#' Create a HingeFeature object
#'
#' Creates a piecewise-linear hinge feature.
#' Forward hinge: eval(i) = (values[i] > min_knot) ? (values[i] - min_knot) / (max_knot - min_knot) : 0.
#' Reverse hinge: eval(i) = (values[i] < max_knot) ? (max_knot - values[i]) / (max_knot - min_knot) : 0.
#'
#' @param values Numeric vector of environmental variable values
#' @param name Feature name/identifier
#' @param min_knot Lower knot of the hinge
#' @param max_knot Upper knot of the hinge (must be > min_knot)
#' @param is_reverse If TRUE, use reverse hinge; if FALSE (default), use forward hinge
#' @return External pointer to HingeFeature object
#' @export
create_hinge_feature <- function(values, name, min_knot, max_knot, is_reverse = FALSE) {
    .Call(`_maxentcpp_create_hinge_feature`, values, name, min_knot, max_knot, is_reverse)
}

#' Evaluate a feature at a given index
#'
#' @param feature_ptr External pointer to a Feature object
#' @param index 0-based index into the data vector
#' @return Feature value at that index (double)
#' @export
feature_eval <- function(feature_ptr, index) {
    .Call(`_maxentcpp_feature_eval`, feature_ptr, index)
}

#' Get feature metadata
#'
#' Returns a list with name, type, lambda, min, and max values.
#'
#' @param feature_ptr External pointer to a Feature object
#' @return Named list with feature properties
#' @export
feature_get_info <- function(feature_ptr) {
    .Call(`_maxentcpp_feature_get_info`, feature_ptr)
}

#' Generate features from a list of environmental variable vectors
#'
#' Generates all configured feature types (linear, quadratic, product,
#' threshold, hinge) from the supplied data vectors.
#'
#' @param data_list Named list of numeric vectors, one per environmental variable
#' @param feature_types Character vector of feature types to generate.
#'   Valid values: "linear", "quadratic", "product", "threshold", "hinge".
#'   Defaults to all types.
#' @param n_thresholds Number of threshold knots per variable (default: 10)
#' @param n_hinges Number of hinge knots per variable (default: 10)
#' @return List of external pointers to Feature objects
#' @export
generate_features <- function(data_list, feature_types = as.character( c("linear", "quadratic",                                                    "product", "threshold", "hinge")), n_thresholds = 10L, n_hinges = 10L) {
    .Call(`_maxentcpp_generate_features`, data_list, feature_types, n_thresholds, n_hinges)
}

#' Read an ESRI ASCII grid (.asc) file
#'
#' Reads an ASC raster file and returns a float Grid external pointer.
#'
#' @param filename Path to the .asc file.
#' @return External pointer to a GridFloat object.
#' @export
grid_read_asc <- function(filename) {
    .Call(`_maxentcpp_grid_read_asc`, filename)
}

#' Write a float Grid to an ESRI ASCII (.asc) file
#'
#' @param grid_ptr   External pointer to a GridFloat object.
#' @param filename   Output file path.
#' @param scientific Logical: use scientific notation (default TRUE).
#' @return Called for side effects; returns invisibly.
#' @export
grid_write_asc <- function(grid_ptr, filename, scientific = TRUE) {
    invisible(.Call(`_maxentcpp_grid_write_asc`, grid_ptr, filename, scientific))
}

#' Read a grid from file (auto-detect format)
#'
#' Currently supports .asc format. Returns a float Grid.
#'
#' @param filename Path to the grid file.
#' @return External pointer to a GridFloat object.
#' @export
grid_read_file <- function(filename) {
    .Call(`_maxentcpp_grid_read_file`, filename)
}

#' Get grid information for an ASC-loaded grid
#'
#' Returns metadata and basic statistics for a GridFloat.
#'
#' @param grid_ptr External pointer to a GridFloat object.
#' @return Named list with nrows, ncols, xll, yll, cellsize, nodata,
#'   name, count_data.
#' @export
grid_float_info <- function(grid_ptr) {
    .Call(`_maxentcpp_grid_float_info`, grid_ptr)
}

#' Extract the data from a GridFloat as an R matrix
#'
#' @param grid_ptr External pointer to a GridFloat object.
#' @return Numeric matrix (nrows × ncols). NODATA cells are set to NA.
#' @export
grid_float_to_matrix <- function(grid_ptr) {
    .Call(`_maxentcpp_grid_float_to_matrix`, grid_ptr)
}

#' Create a GridFloat from an R matrix
#'
#' @param mat       Numeric matrix with grid data (NA becomes NODATA).
#' @param xll       X coordinate of lower-left corner.
#' @param yll       Y coordinate of lower-left corner.
#' @param cellsize  Cell size.
#' @param nodata    NODATA value (default -9999).
#' @param name      Grid name (default "").
#' @return External pointer to a GridFloat object.
#' @export
grid_float_from_matrix <- function(mat, xll, yll, cellsize, nodata = -9999.0, name = "") {
    .Call(`_maxentcpp_grid_float_from_matrix`, mat, xll, yll, cellsize, nodata, name)
}

#' Open a CSV file for reading
#'
#' @param filename   Path to the CSV file.
#' @param has_header Logical: first line contains column names (default TRUE).
#' @return External pointer to a CsvReader object.
#' @export
csv_open <- function(filename, has_header = TRUE) {
    .Call(`_maxentcpp_csv_open`, filename, has_header)
}

#' Get CSV column headers
#'
#' @param reader_ptr External pointer to a CsvReader object.
#' @return Character vector of header names.
#' @export
csv_headers <- function(reader_ptr) {
    .Call(`_maxentcpp_csv_headers`, reader_ptr)
}

#' Read the next CSV record
#'
#' @param reader_ptr External pointer to a CsvReader object.
#' @return Character vector of field values, or NULL on EOF.
#' @export
csv_next_record <- function(reader_ptr) {
    .Call(`_maxentcpp_csv_next_record`, reader_ptr)
}

#' Read an entire named column as doubles
#'
#' Reads from the current position to EOF.
#'
#' @param reader_ptr External pointer to a CsvReader object.
#' @param field      Column name.
#' @return Numeric vector.
#' @export
csv_read_double_column <- function(reader_ptr, field) {
    .Call(`_maxentcpp_csv_read_double_column`, reader_ptr, field)
}

#' Close a CsvReader
#'
#' @param reader_ptr External pointer to a CsvReader object.
#' @return Called for side effects; returns invisibly.
#' @export
csv_close <- function(reader_ptr) {
    invisible(.Call(`_maxentcpp_csv_close`, reader_ptr))
}

#' Open a CSV file for writing
#'
#' @param filename  Path to the output CSV file.
#' @param append    Logical: append to existing file (default FALSE).
#' @param precision Number of decimal places for doubles (default 4).
#' @return External pointer to a CsvWriter object.
#' @export
csv_writer_open <- function(filename, append = FALSE, precision = 4L) {
    .Call(`_maxentcpp_csv_writer_open`, filename, append, precision)
}

#' Write a value into the current row of a CSV
#'
#' @param writer_ptr External pointer to a CsvWriter object.
#' @param column     Column name.
#' @param value      Value to write (character).
#' @return Called for side effects; returns invisibly.
#' @export
csv_writer_print <- function(writer_ptr, column, value) {
    invisible(.Call(`_maxentcpp_csv_writer_print`, writer_ptr, column, value))
}

#' Write a numeric value into the current row of a CSV
#'
#' @param writer_ptr External pointer to a CsvWriter object.
#' @param column     Column name.
#' @param value      Numeric value.
#' @return Called for side effects; returns invisibly.
#' @export
csv_writer_print_double <- function(writer_ptr, column, value) {
    invisible(.Call(`_maxentcpp_csv_writer_print_double`, writer_ptr, column, value))
}

#' Flush the current CSV row to disk
#'
#' @param writer_ptr External pointer to a CsvWriter object.
#' @return Called for side effects; returns invisibly.
#' @export
csv_writer_println <- function(writer_ptr) {
    invisible(.Call(`_maxentcpp_csv_writer_println`, writer_ptr))
}

#' Close a CsvWriter
#'
#' @param writer_ptr External pointer to a CsvWriter object.
#' @return Called for side effects; returns invisibly.
#' @export
csv_writer_close <- function(writer_ptr) {
    invisible(.Call(`_maxentcpp_csv_writer_close`, writer_ptr))
}

#' Create a Layer metadata object
#'
#' @param name      Layer name.
#' @param type_str  Type string: "Continuous", "Categorical", "Bias",
#'   "Mask", "Probability", "Cumulative", "DebiasAvg", or "Unknown".
#' @return External pointer to a Layer object.
#' @export
create_layer <- function(name, type_str) {
    .Call(`_maxentcpp_create_layer`, name, type_str)
}

#' Get layer metadata
#'
#' @param layer_ptr External pointer to a Layer object.
#' @return Named list with name and type (string).
#' @export
get_layer_info <- function(layer_ptr) {
    .Call(`_maxentcpp_get_layer_info`, layer_ptr)
}

#' Extract a layer name from a file path
#'
#' Strips directory and extension, e.g. "/data/bio1.asc" → "bio1".
#'
#' @param path File path.
#' @return Character: layer name.
#' @export
layer_name_from_path <- function(path) {
    .Call(`_maxentcpp_layer_name_from_path`, path)
}

#' Create a GridDimension object
#'
#' Creates a grid dimension specifying the spatial extent and resolution
#' 
#' @param nrows Number of rows
#' @param ncols Number of columns
#' @param xll X coordinate of lower-left corner
#' @param yll Y coordinate of lower-left corner
#' @param cellsize Cell size (assumed square cells)
#' @return External pointer to GridDimension object
#' @export
create_grid_dimension <- function(nrows, ncols, xll, yll, cellsize) {
    .Call(`_maxentcpp_create_grid_dimension`, nrows, ncols, xll, yll, cellsize)
}

#' Get grid dimension properties
#'
#' @param dim_ptr External pointer to GridDimension
#' @return List with dimension properties
#' @export
get_grid_dimension_info <- function(dim_ptr) {
    .Call(`_maxentcpp_get_grid_dimension_info`, dim_ptr)
}

#' Convert geographic coordinates to row/col
#'
#' @param dim_ptr External pointer to GridDimension
#' @param lon Longitude
#' @param lat Latitude
#' @return Integer vector with row and column indices
#' @export
coords_to_rowcol <- function(dim_ptr, lon, lat) {
    .Call(`_maxentcpp_coords_to_rowcol`, dim_ptr, lon, lat)
}

#' Create a Sample object
#'
#' Creates a species occurrence sample point
#'
#' @param point Point identifier
#' @param row Row index in grid
#' @param col Column index in grid
#' @param lat Latitude coordinate
#' @param lon Longitude coordinate
#' @param name Sample name/identifier
#' @return External pointer to Sample object
#' @export
create_sample <- function(point, row, col, lat, lon, name) {
    .Call(`_maxentcpp_create_sample`, point, row, col, lat, lon, name)
}

#' Get sample properties
#'
#' @param sample_ptr External pointer to Sample
#' @return List with sample properties
#' @export
get_sample_info <- function(sample_ptr) {
    .Call(`_maxentcpp_get_sample_info`, sample_ptr)
}

#' Set feature value on a sample
#'
#' @param sample_ptr External pointer to Sample
#' @param feature_name Name of the feature
#' @param value Feature value
#' @return Called for side effects; returns invisibly.
#' @export
sample_set_feature <- function(sample_ptr, feature_name, value) {
    invisible(.Call(`_maxentcpp_sample_set_feature`, sample_ptr, feature_name, value))
}

#' Get feature value from a sample
#'
#' @param sample_ptr External pointer to Sample
#' @param feature_name Name of the feature
#' @param default_val Default value if feature not found
#' @return Feature value
#' @export
sample_get_feature <- function(sample_ptr, feature_name, default_val = 0.0) {
    .Call(`_maxentcpp_sample_get_feature`, sample_ptr, feature_name, default_val)
}

#' Create a float Grid object
#'
#' Creates a raster grid for environmental variables
#'
#' @param dim_ptr External pointer to GridDimension
#' @param name Grid layer name
#' @param nodata_value Value representing missing data
#' @return External pointer to Grid<float> object
#' @export
create_grid_float <- function(dim_ptr, name, nodata_value = -9999.0) {
    .Call(`_maxentcpp_create_grid_float`, dim_ptr, name, nodata_value)
}

#' Get grid value
#'
#' @param grid_ptr External pointer to Grid
#' @param row Row index
#' @param col Column index
#' @return Grid value
#' @export
grid_get_value <- function(grid_ptr, row, col) {
    .Call(`_maxentcpp_grid_get_value`, grid_ptr, row, col)
}

#' Set grid value
#'
#' @param grid_ptr External pointer to Grid
#' @param row Row index
#' @param col Column index
#' @param value Value to set
#' @return Called for side effects; returns invisibly.
#' @export
grid_set_value <- function(grid_ptr, row, col, value) {
    invisible(.Call(`_maxentcpp_grid_set_value`, grid_ptr, row, col, value))
}

#' Check if grid cell has data
#'
#' @param grid_ptr External pointer to Grid
#' @param row Row index
#' @param col Column index
#' @return TRUE if cell has valid data
#' @export
grid_has_data <- function(grid_ptr, row, col) {
    .Call(`_maxentcpp_grid_has_data`, grid_ptr, row, col)
}

#' Get grid as matrix
#'
#' @param grid_ptr External pointer to Grid
#' @return Numeric matrix of grid values
#' @export
grid_to_matrix <- function(grid_ptr) {
    .Call(`_maxentcpp_grid_to_matrix`, grid_ptr)
}

#' Extract multiple values from a Grid at given row/col indices (batch)
#'
#' @param grid_ptr External pointer to a Grid<float>.
#' @param rows     Integer vector of 0-based row indices.
#' @param cols     Integer vector of 0-based column indices.
#' @return Numeric vector of extracted values (NA for NODATA cells).
#' @export
grid_get_values_batch <- function(grid_ptr, rows, cols) {
    .Call(`_maxentcpp_grid_get_values_batch`, grid_ptr, rows, cols)
}

#' Set grid from matrix
#'
#' @param grid_ptr External pointer to Grid
#' @param mat Numeric matrix of values
#' @return Called for side effects; returns invisibly.
#' @export
grid_from_matrix <- function(grid_ptr, mat) {
    invisible(.Call(`_maxentcpp_grid_from_matrix`, grid_ptr, mat))
}

#' Compute Permutation Importance
#'
#' Measures how much each environmental variable contributes to model
#' prediction quality by permuting each variable and measuring AUC drop.
#'
#' @param fs_ptr         External pointer to a FeaturedSpace object.
#' @param grid_ptrs      List of external pointers to Grid<float> objects.
#' @param feature_names  Character vector of environment variable names.
#' @param presence_rows  Integer vector of presence site row indices.
#' @param presence_cols  Integer vector of presence site column indices.
#' @param absence_rows   Integer vector of absence site row indices.
#' @param absence_cols   Integer vector of absence site column indices.
#' @param seed           Random seed for reproducibility.
#' @return A data.frame with columns: name, permutation_importance.
#' @export
compute_permutation_importance <- function(fs_ptr, grid_ptrs, feature_names, presence_rows, presence_cols, absence_rows, absence_cols, seed = 42L) {
    .Call(`_maxentcpp_compute_permutation_importance`, fs_ptr, grid_ptrs, feature_names, presence_rows, presence_cols, absence_rows, absence_cols, seed)
}

#' Compute Percent Contribution
#'
#' Computes variable contribution based on sum of absolute lambda values
#' for features derived from each variable.
#'
#' @param fs_ptr         External pointer to a FeaturedSpace object.
#' @param feature_names  Character vector of base variable names.
#' @return A data.frame with columns: name, contribution.
#' @export
compute_percent_contribution <- function(fs_ptr, feature_names) {
    .Call(`_maxentcpp_compute_percent_contribution`, fs_ptr, feature_names)
}

#' Compute Java-compatible Marginal Response Curve
#'
#' Like compute_response_curve() but applies the Java Maxent cloglog:
#'   cloglog_java = 1 - exp(-exp(H) * raw_java)
#'
#' This matches the response curves produced by Java Maxent and dismo.
#'
#' @param fs_ptr         External pointer to a FeaturedSpace object.
#' @param grid_ptrs      List of external pointers to Grid<float> objects.
#' @param feature_names  Character vector of environment variable names.
#' @param var_index      0-based index of the variable to vary.
#' @param n_steps        Number of steps across the variable range.
#' @return A data.frame with columns: value, prediction.
#' @export
compute_response_curve <- function(fs_ptr, grid_ptrs, feature_names, var_index, n_steps = 100L) {
    .Call(`_maxentcpp_compute_response_curve`, fs_ptr, grid_ptrs, feature_names, var_index, n_steps)
}

#' Compute Java-compatible Response Curve with Fixed Values
#'
#' Like compute_response_curve_fixed() but applies the Java Maxent cloglog:
#'   cloglog_java = 1 - exp(-exp(H) * raw_java)
#'
#' @param fs_ptr         External pointer to a FeaturedSpace object.
#' @param fixed_values   Numeric vector of fixed values for each variable.
#' @param feature_names  Character vector of environment variable names.
#' @param var_index      0-based index of the variable to vary.
#' @param var_min        Minimum value of the target variable.
#' @param var_max        Maximum value of the target variable.
#' @param n_steps        Number of steps.
#' @return A data.frame with columns: value, prediction.
#' @export
compute_response_curve_fixed <- function(fs_ptr, fixed_values, feature_names, var_index, var_min, var_max, n_steps = 100L) {
    .Call(`_maxentcpp_compute_response_curve_fixed`, fs_ptr, fixed_values, feature_names, var_index, var_min, var_max, n_steps)
}

#' Clamp Environmental Grids
#'
#' Restricts environmental variable values to the training range.
#' Returns clamped grids and a clamping indicator grid.
#'
#' @param grid_ptrs List of external pointers to Grid<float> objects.
#' @param var_mins  Numeric vector of minimum training values per variable.
#' @param var_maxs  Numeric vector of maximum training values per variable.
#' @return A named list with:
#'   \describe{
#'     \item{clamped_grids}{List of external pointers to clamped Grid<float>}
#'     \item{clamp_grid}{External pointer to Grid<float> with clamping magnitudes}
#'   }
#' @export
clamp_grids <- function(grid_ptrs, var_mins, var_maxs) {
    .Call(`_maxentcpp_clamp_grids`, grid_ptrs, var_mins, var_maxs)
}

#' Compute Variable Ranges from Grids
#'
#' Scans all valid cells to determine min/max of each variable.
#'
#' @param grid_ptrs List of external pointers to Grid<float> objects.
#' @return A data.frame with columns: min, max.
#' @export
compute_variable_ranges <- function(grid_ptrs) {
    .Call(`_maxentcpp_compute_variable_ranges`, grid_ptrs)
}

#' Compute MESS (Multivariate Environmental Similarity Surface)
#'
#' Measures how similar each cell is to the training environment.
#' Negative values indicate novel (non-analog) conditions.
#'
#' @param grid_ptrs         List of external pointers to Grid<float> objects.
#' @param reference_values  List of numeric vectors with reference values
#'   for each variable (e.g. values at training sites).
#' @param feature_names     Character vector of variable names.
#' @return A named list with:
#'   \describe{
#'     \item{mess_grid}{External pointer to Grid<float> with MESS values}
#'     \item{mod_grid}{External pointer to Grid<float> with MoD variable index (1-based)}
#'   }
#' @export
compute_mess <- function(grid_ptrs, reference_values, feature_names) {
    .Call(`_maxentcpp_compute_mess`, grid_ptrs, reference_values, feature_names)
}

#' Compute MESS from Min/Max Ranges
#'
#' Simplified MESS using only the min/max of reference data.
#'
#' @param grid_ptrs List of external pointers to Grid<float> objects.
#' @param var_mins  Numeric vector of minimum reference values.
#' @param var_maxs  Numeric vector of maximum reference values.
#' @return A named list with mess_grid and mod_grid (external pointers).
#' @export
compute_mess_range <- function(grid_ptrs, var_mins, var_maxs) {
    .Call(`_maxentcpp_compute_mess_range`, grid_ptrs, var_mins, var_maxs)
}

#' Compute AUC (Area Under the ROC Curve)
#'
#' Computes the Wilcoxon-Mann-Whitney AUC statistic from prediction scores
#' at presence and absence sites.
#'
#' @param presence Numeric vector of prediction scores at presence sites.
#' @param absence  Numeric vector of prediction scores at absence sites.
#' @return A named list with elements: auc, max_kappa, max_kappa_thresh.
#' @export
eval_auc <- function(presence, absence) {
    .Call(`_maxentcpp_eval_auc`, presence, absence)
}

#' Compute Pearson Correlation
#'
#' Computes Pearson correlation coefficient between two numeric vectors.
#'
#' @param x Numeric vector.
#' @param y Numeric vector (same length as x).
#' @return Correlation coefficient in [-1, 1].
#' @export
eval_correlation <- function(x, y) {
    .Call(`_maxentcpp_eval_correlation`, x, y)
}

#' Compute Log-Loss
#'
#' Computes average cross-entropy log-loss from predictions at
#' presence and absence sites.
#'
#' @param presence Numeric vector of prediction scores at presence sites.
#' @param absence  Numeric vector of prediction scores at absence sites.
#' @return Average log-loss value.
#' @export
eval_logloss <- function(presence, absence) {
    .Call(`_maxentcpp_eval_logloss`, presence, absence)
}

#' Compute Mean Squared Error
#'
#' Presence sites contribute (1-pred)^2, absence sites contribute pred^2.
#'
#' @param presence Numeric vector of prediction scores at presence sites.
#' @param absence  Numeric vector of prediction scores at absence sites.
#' @return Mean squared error.
#' @export
eval_square_error <- function(presence, absence) {
    .Call(`_maxentcpp_eval_square_error`, presence, absence)
}

#' Compute Misclassification Rate
#'
#' Fraction of misclassified samples at threshold 0.5.
#'
#' @param presence Numeric vector of prediction scores at presence sites.
#' @param absence  Numeric vector of prediction scores at absence sites.
#' @return Misclassification rate in [0, 1].
#' @export
eval_misclassification <- function(presence, absence) {
    .Call(`_maxentcpp_eval_misclassification`, presence, absence)
}

#' Full Model Evaluation
#'
#' Computes all evaluation metrics at once: AUC, correlation, log-loss,
#' squared error, misclassification, max kappa, and prevalence.
#'
#' @param presence Numeric vector of prediction scores at presence sites.
#' @param absence  Numeric vector of prediction scores at absence sites.
#' @return A named list with: auc, max_kappa, max_kappa_thresh,
#'   correlation, square_error, logloss, misclassification, prevalence.
#' @export
eval_model <- function(presence, absence) {
    .Call(`_maxentcpp_eval_model`, presence, absence)
}

#' Project Model onto Grids (Java-compatible raw output)
#'
#' Applies a trained FeaturedSpace model to produce Java Maxent "raw" scores:
#'   raw_java = exp(lp - lpNorm) / densityNorm
#'
#' This matches the raw output of the Java Maxent software and the dismo
#' R package.
#'
#' @param fs_ptr         External pointer to a FeaturedSpace object.
#' @param grid_ptrs      List of external pointers to Grid<float> objects.
#' @param feature_names  Character vector of environment variable names.
#' @return External pointer to a Grid<float> with Java raw scores.
#' @export
project_raw <- function(fs_ptr, grid_ptrs, feature_names) {
    .Call(`_maxentcpp_project_raw`, fs_ptr, grid_ptrs, feature_names)
}

#' Project Model onto Grids (Java-compatible cloglog output)
#'
#' Applies the Java Maxent cloglog formula:
#'   cloglog_java = 1 - exp(-exp(H) * raw_java)
#'
#' where H is the model entropy and raw_java is the normalised probability.
#' This matches the cloglog output of Java Maxent and dismo.
#'
#' @param fs_ptr         External pointer to a FeaturedSpace object.
#' @param grid_ptrs      List of external pointers to Grid<float> objects.
#' @param feature_names  Character vector of environment variable names.
#' @return External pointer to a Grid<float> with Java cloglog scores in [0, 1].
#' @export
project_cloglog <- function(fs_ptr, grid_ptrs, feature_names) {
    .Call(`_maxentcpp_project_cloglog`, fs_ptr, grid_ptrs, feature_names)
}

#' Project Model onto Grids (Java-compatible logistic output)
#'
#' Applies the Java Maxent logistic formula:
#'   logistic_java = (exp(H) * raw_java) / (1 + exp(H) * raw_java)
#'
#' where H is the model entropy and raw_java is the normalised probability.
#' This matches the logistic output of Java Maxent and dismo.
#'
#' @param fs_ptr         External pointer to a FeaturedSpace object.
#' @param grid_ptrs      List of external pointers to Grid<float> objects.
#' @param feature_names  Character vector of environment variable names.
#' @return External pointer to a Grid<float> with Java logistic scores in [0, 1].
#' @export
project_logistic <- function(fs_ptr, grid_ptrs, feature_names) {
    .Call(`_maxentcpp_project_logistic`, fs_ptr, grid_ptrs, feature_names)
}

#' Extract Java-compatible Raw Predictions at Sample Locations
#'
#' Gets Java Maxent raw scores (raw_java = exp(lp - lpNorm) / densityNorm)
#' at specific grid cell locations.
#'
#' @param fs_ptr         External pointer to a FeaturedSpace object.
#' @param grid_ptrs      List of external pointers to Grid<float> objects.
#' @param feature_names  Character vector of environment variable names.
#' @param rows           Integer vector of row indices.
#' @param cols           Integer vector of column indices.
#' @return Numeric vector of Java raw scores. NaN for NODATA cells.
#' @export
extract_predictions_raw <- function(fs_ptr, grid_ptrs, feature_names, rows, cols) {
    .Call(`_maxentcpp_extract_predictions_raw`, fs_ptr, grid_ptrs, feature_names, rows, cols)
}

# Tests for grid_get_values_batch() — batch grid value extraction

make_test_grid <- function() {
    dim <- maxent_dimension(nrows = 3, ncols = 3, xll = 0, yll = 0, cellsize = 1)
    g <- maxent_grid(dim, "test")
    idx <- 0L
    for (r in 0:2) {
        for (cc in 0:2) {
            grid_set_value(g, r, cc, as.double(idx))
            idx <- idx + 1L
        }
    }
    g
}

test_that("grid_get_values_batch matches grid_get_value for valid cells", {
    skip_if_not_installed("maxentcpp")

    g <- make_test_grid()
    rows <- as.integer(c(0, 0, 1, 2))
    cols <- as.integer(c(0, 2, 1, 2))

    batch <- grid_get_values_batch(g, rows, cols)

    for (i in seq_along(rows)) {
        expect_equal(batch[i], grid_get_value(g, rows[i], cols[i]),
                     tolerance = 1e-6,
                     info = paste("mismatch at position", i))
    }
})

test_that("grid_get_values_batch returns NA for NODATA cells", {
    skip_if_not_installed("maxentcpp")

    dim <- maxent_dimension(nrows = 2, ncols = 2, xll = 0, yll = 0, cellsize = 1)
    g <- maxent_grid(dim, "nodata_test")
    grid_set_value(g, 0, 0, 1.0)
    # (0,1), (1,0), (1,1) are left as NODATA (never set)

    batch <- grid_get_values_batch(g, as.integer(c(0, 0, 1, 1)),
                                       as.integer(c(0, 1, 0, 1)))

    expect_equal(batch[1], 1.0, tolerance = 1e-6)
    expect_true(is.na(batch[2]))
    expect_true(is.na(batch[3]))
    expect_true(is.na(batch[4]))
})

test_that("grid_get_values_batch errors on mismatched rows/cols lengths", {
    skip_if_not_installed("maxentcpp")

    g <- make_test_grid()
    expect_error(
        grid_get_values_batch(g, as.integer(c(0, 1)), as.integer(c(0))),
        "same length"
    )
})

test_that("grid_get_values_batch errors on out-of-bounds indices", {
    skip_if_not_installed("maxentcpp")

    g <- make_test_grid()   # 3x3 grid, valid rows/cols: 0..2
    expect_error(
        grid_get_values_batch(g, as.integer(c(0, 5)), as.integer(c(0, 0))),
        "out of bounds"
    )
    expect_error(
        grid_get_values_batch(g, as.integer(c(0, -1)), as.integer(c(0, 0))),
        "out of bounds"
    )
})

test_that("grid_get_values_batch handles an empty index vector", {
    skip_if_not_installed("maxentcpp")

    g <- make_test_grid()
    result <- grid_get_values_batch(g, integer(0), integer(0))
    expect_length(result, 0)
})

test_that("GridDimension creation works", {
    skip_if_not_installed("maxentcpp")
    
    dim <- create_grid_dimension(100, 100, -120.0, 35.0, 0.1)
    expect_true(!is.null(dim))
    
    info <- get_grid_dimension_info(dim)
    expect_equal(info$nrows, 100)
    expect_equal(info$ncols, 100)
    expect_equal(info$xll, -120.0)
    expect_equal(info$yll, 35.0)
    expect_equal(info$cellsize, 0.1)
    expect_equal(info$size, 10000)
})

test_that("Coordinate conversion works", {
    skip_if_not_installed("maxentcpp")
    
    dim <- create_grid_dimension(100, 100, -120.0, 35.0, 0.1)
    rc <- coords_to_rowcol(dim, -119.5, 36.0)
    
    expect_length(rc, 2)
    expect_true(rc[1] >= 0 && rc[1] < 100)
    expect_true(rc[2] >= 0 && rc[2] < 100)
})

test_that("Sample creation works", {
    skip_if_not_installed("maxentcpp")
    
    sample <- create_sample(0, 50, 50, 36.5, -118.5, "test")
    expect_true(!is.null(sample))
    
    info <- get_sample_info(sample)
    expect_equal(info$point, 0)
    expect_equal(info$row, 50)
    expect_equal(info$col, 50)
    expect_equal(info$lat, 36.5)
    expect_equal(info$lon, -118.5)
    expect_equal(info$name, "test")
})

test_that("Sample features work", {
    skip_if_not_installed("maxentcpp")
    
    sample <- create_sample(0, 50, 50, 36.5, -118.5, "test")
    
    # Set features
    sample_set_feature(sample, "temperature", 25.5)
    sample_set_feature(sample, "precipitation", 100.0)
    
    # Get features
    temp <- sample_get_feature(sample, "temperature")
    precip <- sample_get_feature(sample, "precipitation")
    missing <- sample_get_feature(sample, "elevation", -999.0)
    
    expect_equal(temp, 25.5)
    expect_equal(precip, 100.0)
    expect_equal(missing, -999.0)
})

test_that("Grid creation and manipulation work", {
    skip_if_not_installed("maxentcpp")
    
    dim <- create_grid_dimension(10, 10, -120.0, 35.0, 0.1)
    grid <- create_grid_float(dim, "test_grid", -9999.0)
    
    expect_true(!is.null(grid))
    
    # Set and get values
    grid_set_value(grid, 5, 5, 42.5)
    val <- grid_get_value(grid, 5, 5)
    expect_equal(val, 42.5)
    
    # Check has_data
    expect_true(grid_has_data(grid, 5, 5))
})

test_that("Grid matrix conversion works", {
    skip_if_not_installed("maxentcpp")
    
    dim <- create_grid_dimension(5, 5, -120.0, 35.0, 0.1)
    grid <- create_grid_float(dim, "test_grid")
    
    # Create a test matrix
    test_mat <- matrix(1:25, nrow = 5, ncol = 5, byrow = TRUE)
    
    # Set grid from matrix
    grid_from_matrix(grid, test_mat)
    
    # Convert back to matrix
    result_mat <- grid_to_matrix(grid)
    
    expect_equal(dim(result_mat), c(5, 5))
    expect_equal(result_mat, test_mat)
})

test_that("High-level R functions work", {
    skip_if_not_installed("maxentcpp")
    
    # Create dimension
    dim <- maxent_dimension(20, 20, -120, 35, 0.1)
    expect_true(!is.null(dim))
    
    # Create sample
    sample <- maxent_sample(lon = -119.5, lat = 36.0, name = "site1", dim = dim)
    expect_true(!is.null(sample))
    
    # Create grid
    grid <- maxent_grid(dim, "temperature")
    expect_true(!is.null(grid))
    
    # Test matrix operations
    mat <- as_matrix(grid)
    expect_equal(dim(mat), c(20, 20))
})

# Tests for Java Maxent numerical equivalency fixes:
#  Phase 1: raw_java capped at 1.0 for novel environments
#  Phase 2: Feature clamping for out-of-range env values
#  Phase 3: ThresholdFeature uses non-strict >= comparison
#  Phase 4: read_lambdas() preserves entropy and normalizers

# ---------------------------------------------------------------------------
# Helpers
# ---------------------------------------------------------------------------

# A tiny model trained on 3 points (bio values 0.2, 0.5, 0.8) with 9 bg pts.
make_tiny_model <- function() {
    vals <- seq(0, 1, length.out = 9)
    env_data <- list(bio1 = vals)
    features <- maxent_generate_features(env_data, types = "linear",
                                         n_thresholds = 0L, n_hinges = 0L)
    model <- maxent_featured_space(9L, c(2L, 5L, 8L), features)
    maxent_fit(model, max_iter = 300L, convergence = 1e-6, beta_multiplier = 1.0)
    model
}

# ---------------------------------------------------------------------------
# Phase 1: raw_java cap at 1.0
# ---------------------------------------------------------------------------

test_that("raw_java is capped at 1.0 for in-range values", {
    skip_if_not_installed("maxentcpp")

    model <- make_tiny_model()
    info  <- maxent_featured_space_info(model)
    dn    <- info$density_normalizer

    # In-range env value should produce raw_java <= 1 without capping
    dim  <- maxent_dimension(nrows = 1L, ncols = 3L, xll = 0, yll = 0, cellsize = 1)
    g    <- maxent_grid(dim, "bio1")
    grid_set_value(g, 0L, 0L, 0.0)
    grid_set_value(g, 0L, 1L, 0.5)
    grid_set_value(g, 0L, 2L, 1.0)

    raw_java_grid <- maxent_project_raw(model, list(g), c("bio1"))
    raw_java_mat  <- maxent_grid_to_matrix(raw_java_grid)

    expect_true(all(raw_java_mat <= 1.0))
    expect_true(all(raw_java_mat >= 0.0))
})

test_that("extreme out-of-range env values produce the same raw_java as the training max boundary", {
    skip_if_not_installed("maxentcpp")

    model <- make_tiny_model()

    # With Phase 2 feature clamping, env value 1e6 is clamped to the training
    # max (1.0) and should produce the same prediction as env = 1.0.
    # The Phase 1 cap at 1.0 still applies as a safety bound.
    dim     <- maxent_dimension(nrows = 1L, ncols = 1L, xll = 0, yll = 0, cellsize = 1)
    g       <- maxent_grid(dim, "bio1")
    g_max   <- maxent_grid(dim, "bio1")
    grid_set_value(g,     0L, 0L, 1e6)   # extreme out-of-range value
    grid_set_value(g_max, 0L, 0L, 1.0)   # max training value

    raw_java_grid <- maxent_project_raw(model, list(g),     c("bio1"))
    raw_max_grid  <- maxent_project_raw(model, list(g_max), c("bio1"))
    raw_java_mat  <- maxent_grid_to_matrix(raw_java_grid)
    raw_max       <- maxent_grid_to_matrix(raw_max_grid)[1, 1]

    # Extreme values are clamped to the boundary, producing the same prediction
    expect_equal(raw_java_mat[1, 1], raw_max, tolerance = 1e-9)
    # The cap invariant: raw_java is always bounded by 1.0
    expect_true(raw_java_mat[1, 1] <= 1.0)
})

# ---------------------------------------------------------------------------
# Phase 2: Feature clamping
# ---------------------------------------------------------------------------

test_that("LinearFeature eval_from_env clamps out-of-range values to [0,1]", {
    skip_if_not_installed("maxentcpp")

    model <- make_tiny_model()

    # A grid where cell 0 has value below training range, cell 1 is in range,
    # cell 2 is above training range.
    dim <- maxent_dimension(nrows = 1L, ncols = 3L, xll = 0, yll = 0, cellsize = 1)
    g   <- maxent_grid(dim, "bio1")
    grid_set_value(g, 0L, 0L, -10.0)   # below min (0)
    grid_set_value(g, 0L, 1L,  0.5)    # in range
    grid_set_value(g, 0L, 2L,  10.0)   # above max (1)

    # With clamping, below-range cells produce the same prediction as the min,
    # and above-range cells produce the same prediction as the max.
    dim_min <- maxent_dimension(nrows = 1L, ncols = 1L, xll = 0, yll = 0, cellsize = 1)
    g_min   <- maxent_grid(dim_min, "bio1")
    g_max   <- maxent_grid(dim_min, "bio1")
    grid_set_value(g_min, 0L, 0L, 0.0)
    grid_set_value(g_max, 0L, 0L, 1.0)

    raw_java_grid <- maxent_project_raw(model, list(g),     c("bio1"))
    raw_min_grid  <- maxent_project_raw(model, list(g_min), c("bio1"))
    raw_max_grid  <- maxent_project_raw(model, list(g_max), c("bio1"))

    raw_java_mat <- maxent_grid_to_matrix(raw_java_grid)
    raw_min      <- maxent_grid_to_matrix(raw_min_grid)[1, 1]
    raw_max      <- maxent_grid_to_matrix(raw_max_grid)[1, 1]

    expect_equal(raw_java_mat[1, 1], raw_min, tolerance = 1e-9)  # below → same as min
    expect_equal(raw_java_mat[1, 3], raw_max, tolerance = 1e-9)  # above → same as max (capped at 1)
})

test_that("cloglog_java stays in [0,1] for extreme env values", {
    skip_if_not_installed("maxentcpp")

    model <- make_tiny_model()

    dim <- maxent_dimension(nrows = 1L, ncols = 2L, xll = 0, yll = 0, cellsize = 1)
    g   <- maxent_grid(dim, "bio1")
    grid_set_value(g, 0L, 0L, -1e6)
    grid_set_value(g, 0L, 1L,  1e6)

    cloglog_grid <- maxent_project_cloglog(model, list(g), c("bio1"))
    cloglog_mat  <- maxent_grid_to_matrix(cloglog_grid)

    expect_true(all(cloglog_mat >= 0.0 & cloglog_mat <= 1.0))
})

# ---------------------------------------------------------------------------
# Phase 3: ThresholdFeature >= semantics
# ---------------------------------------------------------------------------

test_that("ThresholdFeature returns 1.0 when value equals threshold", {
    skip_if_not_installed("maxentcpp")

    vals <- c(3.0, 5.0, 7.0)
    f    <- maxent_threshold_feature(vals, "thresh", threshold = 5.0)

    expect_equal(maxent_feature_eval(f, 1L), 0.0)  # 3.0 < 5.0 → 0
    expect_equal(maxent_feature_eval(f, 2L), 1.0)  # 5.0 == 5.0 → 1 (non-strict >=)
    expect_equal(maxent_feature_eval(f, 3L), 1.0)  # 7.0 > 5.0 → 1
})

test_that("ThresholdFeature returns 0.0 for value strictly below threshold", {
    skip_if_not_installed("maxentcpp")

    vals <- c(4.9, 5.0, 5.1)
    f    <- maxent_threshold_feature(vals, "thresh2", threshold = 5.0)

    expect_equal(maxent_feature_eval(f, 1L), 0.0)  # 4.9 < 5.0
    expect_equal(maxent_feature_eval(f, 2L), 1.0)  # 5.0 == 5.0
    expect_equal(maxent_feature_eval(f, 3L), 1.0)  # 5.1 > 5.0
})

# ---------------------------------------------------------------------------
# Phase 4: read_lambdas() preserves entropy and normalizers
# ---------------------------------------------------------------------------

test_that("read_lambdas restores entropy matching the trained model", {
    skip_if_not_installed("maxentcpp")

    model <- make_tiny_model()

    H_trained <- maxent_model_entropy(model)
    info_trained <- maxent_featured_space_info(model)
    lpn_trained  <- info_trained$linear_predictor_normalizer
    dn_trained   <- info_trained$density_normalizer

    # Write and read back into a fresh model
    lambda_file <- tempfile(fileext = ".csv")
    on.exit(unlink(lambda_file), add = TRUE)
    maxent_write_lambdas(model, lambda_file)

    # Build an identical empty featured space and read lambdas into it
    vals_empty <- seq(0, 1, length.out = 9)
    env_data   <- list(bio1 = vals_empty)
    features2  <- maxent_generate_features(env_data, types = "linear",
                                           n_thresholds = 0L, n_hinges = 0L)
    model2 <- maxent_featured_space(9L, c(2L, 5L, 8L), features2)
    maxent_read_lambdas(model2, lambda_file)

    H2    <- maxent_model_entropy(model2)
    info2 <- maxent_featured_space_info(model2)

    expect_equal(H2,                      H_trained,   tolerance = 1e-9)
    expect_equal(info2$density_normalizer,          dn_trained,  tolerance = 1e-9)
    expect_equal(info2$linear_predictor_normalizer, lpn_trained, tolerance = 1e-9)
})

test_that("predictions from read_lambdas model match original", {
    skip_if_not_installed("maxentcpp")

    model <- make_tiny_model()

    lambda_file <- tempfile(fileext = ".csv")
    on.exit(unlink(lambda_file), add = TRUE)
    maxent_write_lambdas(model, lambda_file)

    vals_empty <- seq(0, 1, length.out = 9)
    env_data   <- list(bio1 = vals_empty)
    features2  <- maxent_generate_features(env_data, types = "linear",
                                           n_thresholds = 0L, n_hinges = 0L)
    model2 <- maxent_featured_space(9L, c(2L, 5L, 8L), features2)
    maxent_read_lambdas(model2, lambda_file)

    # Build a small grid and compare cloglog_java predictions
    dim <- maxent_dimension(nrows = 1L, ncols = 5L, xll = 0, yll = 0, cellsize = 0.25)
    g   <- maxent_grid(dim, "bio1")
    for (cc in 0:4) grid_set_value(g, 0L, as.integer(cc), cc * 0.25)

    pred1 <- maxent_grid_to_matrix(
        maxent_project_cloglog(model,  list(g), c("bio1")))
    pred2 <- maxent_grid_to_matrix(
        maxent_project_cloglog(model2, list(g), c("bio1")))

    expect_equal(pred1, pred2, tolerance = 1e-6)
})

#' Compute Permutation Importance
#'
#' Measures how much each environmental variable contributes to model
#' prediction quality by permuting each variable and measuring AUC drop.
#' Results are normalised so that all importances sum to 100%.
#'
#' @param model         External pointer to a FeaturedSpace object.
#' @param env_grids     List of external pointers to Grid<float> objects.
#' @param feature_names Character vector of environment variable names,
#'   matching the order of env_grids.
#' @param presence_rows Integer vector of presence site row indices (0-based).
#' @param presence_cols Integer vector of presence site column indices (0-based).
#' @param absence_rows  Integer vector of absence/background site row indices.
#' @param absence_cols  Integer vector of absence/background site column indices.
#' @param seed          Random seed for reproducibility (default 42).
#' @return A data.frame with columns:
#'   \describe{
#'     \item{name}{Variable name}
#'     \item{permutation_importance}{Normalised importance (\%)}
#'   }
#' @export
#' @examples
#' \donttest{
#' imp <- maxent_permutation_importance(model, list(g1, g2),
#'          c("temp", "precip"),
#'          pres_rows, pres_cols, abs_rows, abs_cols)
#' imp  # data.frame with name and permutation_importance
#' }
maxent_permutation_importance <- function(model, env_grids, feature_names,
                                          presence_rows, presence_cols,
                                          absence_rows, absence_cols,
                                          seed = 42L) {
    compute_permutation_importance(
        model, env_grids, as.character(feature_names),
        as.integer(presence_rows), as.integer(presence_cols),
        as.integer(absence_rows), as.integer(absence_cols),
        as.integer(seed))
}

#' Compute Percent Contribution
#'
#' Computes variable contribution based on sum of absolute lambda values
#' for features derived from each variable. Results sum to 100%.
#'
#' @param model         External pointer to a FeaturedSpace object.
#' @param feature_names Character vector of base variable names.
#' @return A data.frame with columns:
#'   \describe{
#'     \item{name}{Variable name}
#'     \item{contribution}{Percent contribution}
#'   }
#' @export
#' @examples
#' n   <- 50L
#' idx <- c(5L, 15L, 25L, 35L, 45L)
#' env <- list(temp = runif(n), precip = runif(n))
#' feats <- maxent_generate_features(env, types = "linear")
#' model <- maxent_featured_space(n, idx, feats)
#' maxent_fit(model, max_iter = 100, convergence = 1e-3)
#' contrib <- maxent_percent_contribution(model, c("temp", "precip"))
#' contrib  # data.frame with name and contribution
maxent_percent_contribution <- function(model, feature_names) {
    compute_percent_contribution(model, as.character(feature_names))
}

#' Compute Marginal Response Curve
#'
#' Generates a response curve by varying one environmental variable from its
#' minimum to maximum value while holding all other variables at their mean.
#' Applies the Java Maxent cloglog transform, matching the output of Java
#' Maxent and \pkg{dismo}:
#' \deqn{cloglog = 1 - exp(-exp(H) \cdot raw)}
#'
#' @param model         External pointer to a FeaturedSpace object.
#' @param env_grids     List of external pointers to Grid<float> objects.
#' @param feature_names Character vector of environment variable names.
#' @param var_index     0-based index of the variable to vary.
#' @param n_steps       Number of steps across the variable range (default 100).
#' @return A data.frame with columns:
#'   \describe{
#'     \item{value}{Environmental variable value}
#'     \item{prediction}{Cloglog-transformed prediction}
#'   }
#' @export
#' @examples
#' \donttest{
#' curve <- maxent_response_curve(model, list(g1, g2),
#'            c("temp", "precip"), var_index = 0)
#' plot(curve$value, curve$prediction, type = "l")
#' }
#' @seealso \code{\link{maxent_response_curve_fixed}}
maxent_response_curve <- function(model, env_grids, feature_names,
                                  var_index, n_steps = 100L) {
    compute_response_curve(
        model, env_grids, as.character(feature_names),
        as.integer(var_index), as.integer(n_steps))
}

#' Compute Response Curve with Fixed Values
#'
#' Like \code{maxent_response_curve} but the user supplies explicit fixed
#' values for the non-target variables. Applies the Java Maxent cloglog
#' transform, matching the output of Java Maxent and \pkg{dismo}:
#' \deqn{cloglog = 1 - exp(-exp(H) \cdot raw)}
#'
#' @param model         External pointer to a FeaturedSpace object.
#' @param fixed_values  Numeric vector of fixed values for each variable.
#' @param feature_names Character vector of environment variable names.
#' @param var_index     0-based index of the variable to vary.
#' @param var_min       Minimum value of the target variable.
#' @param var_max       Maximum value of the target variable.
#' @param n_steps       Number of steps (default 100).
#' @return A data.frame with columns: value, prediction.
#' @export
#' @seealso \code{\link{maxent_response_curve}}
maxent_response_curve_fixed <- function(model, fixed_values, feature_names,
                                        var_index, var_min, var_max,
                                        n_steps = 100L) {
    compute_response_curve_fixed(
        model, as.numeric(fixed_values), as.character(feature_names),
        as.integer(var_index), as.numeric(var_min), as.numeric(var_max),
        as.integer(n_steps))
}

#' Clamp Environmental Grids
#'
#' Restricts environmental variable values to the range observed during
#' training. Values below min are set to min, values above max are set to max.
#' A clamping grid records the total absolute clamping at each cell.
#'
#' @param env_grids List of external pointers to Grid<float> objects.
#' @param var_mins  Numeric vector of minimum training values per variable.
#' @param var_maxs  Numeric vector of maximum training values per variable.
#' @return A named list with:
#'   \describe{
#'     \item{clamped_grids}{List of external pointers to clamped grids}
#'     \item{clamp_grid}{External pointer to clamping magnitude grid}
#'   }
#' @export
#' @examples
#' \donttest{
#' result <- maxent_clamp(list(g1, g2), c(0, 50), c(30, 200))
#' clamped <- result$clamped_grids
#' clamp_mat <- maxent_grid_to_matrix(result$clamp_grid)
#' }
maxent_clamp <- function(env_grids, var_mins, var_maxs) {
    clamp_grids(env_grids, as.numeric(var_mins), as.numeric(var_maxs))
}

#' Compute Variable Ranges from Grids
#'
#' Scans all valid cells in the grids to determine the [min, max] range
#' of each environmental variable.
#'
#' @param env_grids List of external pointers to Grid<float> objects.
#' @return A data.frame with columns: min, max (one row per variable).
#' @export
maxent_variable_ranges <- function(env_grids) {
    compute_variable_ranges(env_grids)
}

#' Compute MESS (Multivariate Environmental Similarity Surface)
#'
#' Measures how similar each cell is to the training environment using the
#' full distribution of reference values. Negative MESS values indicate
#' novel (non-analog) environments.
#'
#' @param env_grids        List of external pointers to Grid<float> objects.
#' @param reference_values List of numeric vectors with reference values
#'   for each variable (e.g. values at training sites).
#' @param feature_names    Character vector of variable names.
#' @return A named list with:
#'   \describe{
#'     \item{mess_grid}{External pointer to Grid<float> with MESS values}
#'     \item{mod_grid}{External pointer to Grid<float> with Most Dissimilar
#'       Variable index (1-based)}
#'   }
#' @export
#' @examples
#' \donttest{
#' result <- maxent_mess(list(g1, g2),
#'             list(temp_train_vals, precip_train_vals),
#'             c("temp", "precip"))
#' mess_mat <- maxent_grid_to_matrix(result$mess_grid)
#' mod_mat  <- maxent_grid_to_matrix(result$mod_grid)
#' }
maxent_mess <- function(env_grids, reference_values, feature_names) {
    compute_mess(env_grids, reference_values, as.character(feature_names))
}

#' Compute MESS from Min/Max Ranges
#'
#' Simplified MESS analysis using only the min/max of the reference data
#' rather than the full distribution.
#'
#' @param env_grids List of external pointers to Grid<float> objects.
#' @param var_mins  Numeric vector of minimum reference values per variable.
#' @param var_maxs  Numeric vector of maximum reference values per variable.
#' @return A named list with mess_grid and mod_grid (external pointers).
#' @export
maxent_mess_range <- function(env_grids, var_mins, var_maxs) {
    compute_mess_range(env_grids, as.numeric(var_mins), as.numeric(var_maxs))
}

test_that("Percent contribution returns correct structure", {
    skip_if_not_installed("maxentcpp")

    # Create a simple model
    dim <- maxent_dimension(nrows = 3, ncols = 3, xll = 0, yll = 0, cellsize = 1)
    g0 <- maxent_grid(dim, "env0")
    g1 <- maxent_grid(dim, "env1")
    idx <- 0
    for (r in 0:2) {
        for (cc in 0:2) {
            grid_set_value(g0, r, cc, idx / 8.0)
            grid_set_value(g1, r, cc, 0.5)
            idx <- idx + 1
        }
    }

    vals0 <- sapply(0:8, function(i) i / 8.0)
    vals1 <- rep(0.5, 9)

    # Use named list with correct API: types vector, n_thresholds, n_hinges
    env_data <- list(env0 = vals0, env1 = vals1)
    features <- maxent_generate_features(env_data, types = "linear",
                                         n_thresholds = 0, n_hinges = 0)

    model <- maxent_featured_space(9, c(6L, 7L, 8L), features)
    maxent_fit(model, max_iter = 100, convergence = 0.001, beta_multiplier = 1.0)

    contrib <- maxent_percent_contribution(model, c("env0", "env1"))

    expect_true(is.data.frame(contrib))
    expect_true("name" %in% names(contrib))
    expect_true("contribution" %in% names(contrib))
    expect_equal(nrow(contrib), 2)
    expect_equal(sum(contrib$contribution), 100.0, tolerance = 1e-6)
})

test_that("Percent contribution handles prefix-colliding names (bio1/bio12)", {
    skip_if_not_installed("maxentcpp")

    # bio12 starts with "bio1" — old prefix-matching assigned all bio12
    # features to bio1, resulting in 0% for bio12.
    vals1  <- seq(0, 1, length.out = 9)
    vals12 <- seq(1, 0, length.out = 9)

    env_data <- list(bio1 = vals1, bio12 = vals12)
    features <- maxent_generate_features(env_data, types = "linear",
                                         n_thresholds = 0, n_hinges = 0)

    model <- maxent_featured_space(9, c(6L, 7L, 8L), features)
    maxent_fit(model, max_iter = 100, convergence = 0.001, beta_multiplier = 1.0)

    contrib <- maxent_percent_contribution(model, c("bio1", "bio12"))

    expect_equal(nrow(contrib), 2)
    expect_equal(sum(contrib$contribution), 100.0, tolerance = 1e-6)
    # bio12 must NOT be zero when its features have non-zero lambdas
    expect_true(contrib$contribution[contrib$name == "bio12"] > 0)
})

test_that("Clamping works correctly via R wrapper", {
    skip_if_not_installed("maxentcpp")

    dim <- maxent_dimension(nrows = 1, ncols = 3, xll = 0, yll = 0, cellsize = 1)
    g0 <- maxent_grid(dim, "temp")
    grid_set_value(g0, 0, 0, -5.0)   # below min
    grid_set_value(g0, 0, 1, 10.0)   # within range
    grid_set_value(g0, 0, 2, 30.0)   # above max

    result <- maxent_clamp(list(g0), c(0.0), c(20.0))

    expect_true("clamped_grids" %in% names(result))
    expect_true("clamp_grid" %in% names(result))
    expect_length(result$clamped_grids, 1)

    # Check clamped values via grid_to_matrix
    clamped_mat <- grid_float_to_matrix(result$clamped_grids[[1]])
    expect_equal(clamped_mat[1, 1], 0.0, tolerance = 1e-6)
    expect_equal(clamped_mat[1, 2], 10.0, tolerance = 1e-6)
    expect_equal(clamped_mat[1, 3], 20.0, tolerance = 1e-6)

    # Check clamping magnitude grid
    clamp_mat <- grid_float_to_matrix(result$clamp_grid)
    expect_equal(clamp_mat[1, 1], 5.0, tolerance = 1e-6)
    expect_equal(clamp_mat[1, 2], 0.0, tolerance = 1e-6)
    expect_equal(clamp_mat[1, 3], 10.0, tolerance = 1e-6)
})

test_that("Variable ranges are computed correctly", {
    skip_if_not_installed("maxentcpp")

    dim <- maxent_dimension(nrows = 2, ncols = 2, xll = 0, yll = 0, cellsize = 1)
    g0 <- maxent_grid(dim, "temp")
    grid_set_value(g0, 0, 0, 5.0)
    grid_set_value(g0, 0, 1, 15.0)
    grid_set_value(g0, 1, 0, 10.0)
    grid_set_value(g0, 1, 1, 25.0)

    ranges <- maxent_variable_ranges(list(g0))

    expect_true(is.data.frame(ranges))
    expect_equal(ranges$min[1], 5.0, tolerance = 1e-6)
    expect_equal(ranges$max[1], 25.0, tolerance = 1e-6)
})

test_that("MESS range detects novel environments", {
    skip_if_not_installed("maxentcpp")

    dim <- maxent_dimension(nrows = 1, ncols = 3, xll = 0, yll = 0, cellsize = 1)
    g0 <- maxent_grid(dim, "temp")
    grid_set_value(g0, 0, 0, -5.0)   # below min → novel
    grid_set_value(g0, 0, 1, 10.0)   # within range
    grid_set_value(g0, 0, 2, 25.0)   # above max → novel

    result <- maxent_mess_range(list(g0), c(0.0), c(20.0))

    expect_true("mess_grid" %in% names(result))
    expect_true("mod_grid" %in% names(result))

    mess_mat <- grid_float_to_matrix(result$mess_grid)
    expect_true(mess_mat[1, 1] < 0)   # novel
    expect_true(mess_mat[1, 2] >= 0)  # within range
    expect_true(mess_mat[1, 3] < 0)   # novel
})

test_that("MESS with full reference values works", {
    skip_if_not_installed("maxentcpp")

    dim <- maxent_dimension(nrows = 1, ncols = 2, xll = 0, yll = 0, cellsize = 1)
    g0 <- maxent_grid(dim, "temp")
    grid_set_value(g0, 0, 0, 10.0)   # within range
    grid_set_value(g0, 0, 1, -20.0)  # novel

    ref <- list(c(0, 5, 10, 15, 20))

    result <- maxent_mess(list(g0), ref, c("temp"))

    mess_mat <- grid_float_to_matrix(result$mess_grid)
    expect_true(mess_mat[1, 1] >= 0)   # within reference
    expect_true(mess_mat[1, 2] < 0)    # novel
})

test_that("Response curve returns correct structure", {
    skip_if_not_installed("maxentcpp")

    dim <- maxent_dimension(nrows = 3, ncols = 3, xll = 0, yll = 0, cellsize = 1)
    g0 <- maxent_grid(dim, "env0")
    g1 <- maxent_grid(dim, "env1")
    idx <- 0
    for (r in 0:2) {
        for (cc in 0:2) {
            grid_set_value(g0, r, cc, idx / 8.0)
            grid_set_value(g1, r, cc, 0.5)
            idx <- idx + 1
        }
    }

    vals0 <- sapply(0:8, function(i) i / 8.0)
    vals1 <- rep(0.5, 9)

    # Use named list with correct API: types vector, n_thresholds, n_hinges
    env_data <- list(env0 = vals0, env1 = vals1)
    features <- maxent_generate_features(env_data, types = "linear",
                                         n_thresholds = 0, n_hinges = 0)

    model <- maxent_featured_space(9, c(6L, 7L, 8L), features)
    maxent_fit(model, max_iter = 100, convergence = 0.001, beta_multiplier = 1.0)

    curve <- maxent_response_curve(model, list(g0, g1), c("env0", "env1"),
                                   var_index = 0, n_steps = 20)

    expect_true(is.data.frame(curve))
    expect_true("value" %in% names(curve))
    expect_true("prediction" %in% names(curve))
    expect_equal(nrow(curve), 20)
    expect_true(all(curve$prediction >= 0))
    expect_true(all(curve$prediction <= 1))
})

test_that("Projection works with multi-type features (quadratic + hinge)", {
    skip_if_not_installed("maxentcpp")

    dim <- maxent_dimension(nrows = 3, ncols = 3, xll = 0, yll = 0, cellsize = 1)
    g0 <- maxent_grid(dim, "env0")
    g1 <- maxent_grid(dim, "env1")
    idx <- 0
    for (r in 0:2) {
        for (cc in 0:2) {
            grid_set_value(g0, r, cc, idx / 8.0)
            grid_set_value(g1, r, cc, 1.0 - idx / 8.0)
            idx <- idx + 1
        }
    }

    vals0 <- sapply(0:8, function(i) i / 8.0)
    vals1 <- 1.0 - vals0

    env_data <- list(env0 = vals0, env1 = vals1)
    # Generate features with linear, quadratic, and hinge types
    features <- maxent_generate_features(env_data,
                                         types = c("linear", "quadratic", "hinge"),
                                         n_thresholds = 0, n_hinges = 2)

    # Should have more than 2 features: 2 linear + 2 quadratic + 2*2*2 hinges = 12
    expect_true(length(features) > 2)

    model <- maxent_featured_space(9, c(6L, 7L, 8L), features)
    maxent_fit(model, max_iter = 100, convergence = 0.001, beta_multiplier = 1.0)

    # Projection with 2 grids should work despite having 12+ features
    pred <- maxent_project_cloglog(model, list(g0, g1), c("env0", "env1"))
    expect_true(!is.null(pred))

    mat <- maxent_grid_to_matrix(pred)
    expect_equal(dim(mat), c(3, 3))
    expect_true(all(!is.na(mat)))
    expect_true(all(mat >= 0))
    expect_true(all(mat <= 1))
})

test_that("extract_predictions_raw works with multi-type features", {
    skip_if_not_installed("maxentcpp")

    dim <- maxent_dimension(nrows = 3, ncols = 3, xll = 0, yll = 0, cellsize = 1)
    g0 <- maxent_grid(dim, "env0")
    g1 <- maxent_grid(dim, "env1")
    idx <- 0
    for (r in 0:2) {
        for (cc in 0:2) {
            grid_set_value(g0, r, cc, idx / 8.0)
            grid_set_value(g1, r, cc, 1.0 - idx / 8.0)
            idx <- idx + 1
        }
    }

    vals0 <- sapply(0:8, function(i) i / 8.0)
    vals1 <- 1.0 - vals0

    env_data <- list(env0 = vals0, env1 = vals1)
    features <- maxent_generate_features(env_data,
                                         types = c("linear", "quadratic", "hinge"),
                                         n_thresholds = 0, n_hinges = 2)

    model <- maxent_featured_space(9, c(6L, 7L, 8L), features)
    maxent_fit(model, max_iter = 100, convergence = 0.001, beta_multiplier = 1.0)

    preds <- maxent_extract_predictions_raw(model, list(g0, g1), c("env0", "env1"),
                                        c(0L, 1L, 2L), c(0L, 1L, 2L))
    expect_equal(length(preds), 3)
    expect_true(all(is.finite(preds)))
    expect_true(all(preds >= 0))
})

test_that("Response curve works with multi-type features", {
    skip_if_not_installed("maxentcpp")

    dim <- maxent_dimension(nrows = 3, ncols = 3, xll = 0, yll = 0, cellsize = 1)
    g0 <- maxent_grid(dim, "env0")
    g1 <- maxent_grid(dim, "env1")
    idx <- 0
    for (r in 0:2) {
        for (cc in 0:2) {
            grid_set_value(g0, r, cc, idx / 8.0)
            grid_set_value(g1, r, cc, 1.0 - idx / 8.0)
            idx <- idx + 1
        }
    }

    vals0 <- sapply(0:8, function(i) i / 8.0)
    vals1 <- 1.0 - vals0

    env_data <- list(env0 = vals0, env1 = vals1)
    features <- maxent_generate_features(env_data,
                                         types = c("linear", "quadratic", "hinge"),
                                         n_thresholds = 0, n_hinges = 2)

    model <- maxent_featured_space(9, c(6L, 7L, 8L), features)
    maxent_fit(model, max_iter = 100, convergence = 0.001, beta_multiplier = 1.0)

    curve <- maxent_response_curve(model, list(g0, g1), c("env0", "env1"),
                                   var_index = 0, n_steps = 20)

    expect_true(is.data.frame(curve))
    expect_equal(nrow(curve), 20)
    expect_true(all(curve$prediction >= 0))
    expect_true(all(curve$prediction <= 1))
})

#' Read an ESRI ASCII Grid File
#'
#' Reads a \code{.asc} raster file into a GridFloat object.
#'
#' @param filename  Character: path to the \code{.asc} file.
#' @return External pointer to a GridFloat C++ object.
#' @export
#' @examples
#' \donttest{
#' g <- maxent_read_asc("bio1.asc")
#' info <- maxent_grid_info(g)
#' print(info)
#' }
maxent_read_asc <- function(filename) {
    grid_read_asc(as.character(filename))
}

#' Read a Grid File (Auto-Detect Format)
#'
#' Reads a raster grid file. The format is detected from the file extension.
#' Currently supports \code{.asc} (ESRI ASCII Grid).
#'
#' @param filename  Character: path to the grid file.
#' @return External pointer to a GridFloat C++ object.
#' @export
maxent_read_grid <- function(filename) {
    grid_read_file(as.character(filename))
}

#' Write a Grid to ESRI ASCII Format
#'
#' Writes a GridFloat to a \code{.asc} file.
#'
#' @param grid       External pointer to a GridFloat object.
#' @param filename   Character: output file path.
#' @param scientific Logical: use scientific notation for floating-point
#'   values (default \code{TRUE}).
#' @return Invisibly returns the output file path.
#' @export
#' @examples
#' \donttest{
#' maxent_write_asc(g, tempfile(fileext = ".asc"))
#' }
maxent_write_asc <- function(grid, filename, scientific = TRUE) {
    grid_write_asc(grid, as.character(filename), as.logical(scientific))
    invisible(filename)
}

#' Get Grid Information
#'
#' Returns metadata about a GridFloat object read from a file.
#'
#' @param grid  External pointer to a GridFloat object.
#' @return Named list with: \code{nrows}, \code{ncols}, \code{xll},
#'   \code{yll}, \code{cellsize}, \code{nodata}, \code{name},
#'   \code{count_data}.
#' @export
maxent_grid_info <- function(grid) {
    grid_float_info(grid)
}

#' Convert Grid to R Matrix
#'
#' Extracts the data from a GridFloat as an R numeric matrix.
#' NODATA cells are converted to \code{NA}.
#'
#' @param grid  External pointer to a GridFloat object.
#' @return Numeric matrix (\code{nrows} \eqn{\times} \code{ncols}).
#' @export
maxent_grid_to_matrix <- function(grid) {
    grid_float_to_matrix(grid)
}

#' Create a Grid from an R Matrix
#'
#' Builds a GridFloat from a numeric matrix and spatial parameters.
#' \code{NA} values in the matrix are stored as the \code{nodata} value.
#'
#' @param mat       Numeric matrix.
#' @param xll       X coordinate of lower-left corner.
#' @param yll       Y coordinate of lower-left corner.
#' @param cellsize  Cell size.
#' @param nodata    NODATA sentinel value (default \code{-9999}).
#' @param name      Grid name (default \code{""}).
#' @return External pointer to a GridFloat C++ object.
#' @export
maxent_grid_from_matrix <- function(mat, xll, yll, cellsize,
                                    nodata = -9999, name = "") {
    grid_float_from_matrix(as.matrix(mat),
                           as.double(xll), as.double(yll),
                           as.double(cellsize), as.double(nodata),
                           as.character(name))
}

#' Convert a terra SpatRaster to a maxentcpp GridFloat
#'
#' Converts a single-layer \code{terra::SpatRaster} into a maxentcpp
#' GridFloat external pointer. The raster must have exactly one layer and
#' square cells (equal x and y resolution).
#'
#' @param r A single-layer \code{terra::SpatRaster} object.
#' @param name Character: grid name (default: layer name from the raster).
#' @return External pointer to a GridFloat C++ object.
#' @details
#' \code{NA} values in the raster are stored as NODATA (sentinel \code{-9999}).
#'
#' Requires the \pkg{terra} package (listed in Suggests).
#'
#' For the \pkg{raster} package, an equivalent workflow is:
#' \preformatted{
#'   library(raster)
#'   r <- raster("bio1.tif")
#'   mat <- as.matrix(r)
#'   e <- extent(r)
#'   g <- maxent_grid_from_matrix(mat, xll = e@@xmin, yll = e@@ymin,
#'                                cellsize = res(r)[1],
#'                                name = names(r))
#' }
#' @export
#' @examples
#' \donttest{
#' stack_path <- system.file("extdata", "stack_1_12_crop.rds",
#'                          package = "maxentcpp")
#' r <- terra::unwrap(readRDS(stack_path))
#' g <- maxent_grid_from_terra(r[[1]])
#' maxent_grid_info(g)
#' }
maxent_grid_from_terra <- function(r, name = NULL) {
    if (!requireNamespace("terra", quietly = TRUE)) {
        stop("Package 'terra' is required. Install it with install.packages('terra').")
    }
    if (!inherits(r, "SpatRaster")) {
        stop("'r' must be a terra::SpatRaster object")
    }
    if (terra::nlyr(r) != 1L) {
        stop("'r' must have exactly one layer; got ", terra::nlyr(r))
    }

    res_xy <- terra::res(r)
    if (abs(res_xy[1] - res_xy[2]) > 1e-10 * max(res_xy)) {
        stop("Grid cells must be square (equal x and y resolution)")
    }
    cellsize <- res_xy[1]

    e <- terra::ext(r)
    xll <- e[1]  # xmin
    yll <- e[3]  # ymin

    if (is.null(name)) {
        name <- names(r)[1]
    }

    # terra stores values top-to-bottom (row 1 = northernmost) which matches
    # the ESRI ASCII / maxentcpp convention.
    mat <- terra::as.matrix(r, wide = TRUE)

    maxent_grid_from_matrix(mat, xll = xll, yll = yll, cellsize = cellsize,
                            nodata = -9999, name = name)
}

#' Convert a maxentcpp GridFloat to a terra SpatRaster
#'
#' Creates a \code{terra::SpatRaster} from a maxentcpp GridFloat external
#' pointer.
#'
#' @param grid  External pointer to a GridFloat C++ object.
#' @param crs   Character: coordinate reference system string
#'   (default \code{"EPSG:4326"}, i.e. WGS 84 longitude/latitude).
#' @return A single-layer \code{terra::SpatRaster}.
#' @details
#' Requires the \pkg{terra} package (listed in Suggests).
#'
#' For the \pkg{raster} package, an equivalent workflow is:
#' \preformatted{
#'   library(raster)
#'   info <- maxent_grid_info(grid)
#'   mat  <- maxent_grid_to_matrix(grid)
#'   r <- raster(mat,
#'               xmn = info$xll,
#'               xmx = info$xll + info$ncols * info$cellsize,
#'               ymn = info$yll,
#'               ymx = info$yll + info$nrows * info$cellsize,
#'               crs = "+proj=longlat +datum=WGS84")
#' }
#' @export
#' @examples
#' \donttest{
#' stack_path <- system.file("extdata", "stack_1_12_crop.rds",
#'                          package = "maxentcpp")
#' r <- terra::unwrap(readRDS(stack_path))
#' g <- maxent_grid_from_terra(r[[1]])
#' r2 <- maxent_grid_to_terra(g)
#' }
maxent_grid_to_terra <- function(grid, crs = "EPSG:4326") {
    if (!requireNamespace("terra", quietly = TRUE)) {
        stop("Package 'terra' is required. Install it with install.packages('terra').")
    }

    info <- maxent_grid_info(grid)
    mat  <- maxent_grid_to_matrix(grid)

    xmin <- info$xll
    xmax <- info$xll + info$ncols * info$cellsize
    ymin <- info$yll
    ymax <- info$yll + info$nrows * info$cellsize

    r <- terra::rast(
        nrows = info$nrows,
        ncols = info$ncols,
        xmin  = xmin,
        xmax  = xmax,
        ymin  = ymin,
        ymax  = ymax,
        crs   = crs
    )
    names(r) <- info$name

    # terra::values() expects a vector in row-major order (top-to-bottom),
    # which is what as.vector(t(mat)) produces.
    terra::values(r) <- as.vector(t(mat))

    r
}

#' Open a CSV File for Reading
#'
#' Opens a CSV file and reads column headers.
#'
#' @param filename    Character: path to the CSV file.
#' @param has_header  Logical: first line is header (default \code{TRUE}).
#' @return External pointer to a CsvReader C++ object.
#' @export
maxent_csv_open <- function(filename, has_header = TRUE) {
    csv_open(as.character(filename), as.logical(has_header))
}

#' Get CSV Column Headers
#'
#' @param reader  External pointer to a CsvReader object.
#' @return Character vector of column names.
#' @export
maxent_csv_headers <- function(reader) {
    csv_headers(reader)
}

#' Read the Next CSV Record
#'
#' @param reader  External pointer to a CsvReader object.
#' @return Character vector of field values, or \code{NULL} on EOF.
#' @export
maxent_csv_next <- function(reader) {
    csv_next_record(reader)
}

#' Read an Entire Column as Doubles
#'
#' Reads from the current file position to EOF.
#'
#' @param reader  External pointer to a CsvReader object.
#' @param field   Character: column name.
#' @return Numeric vector.
#' @export
maxent_csv_read_column <- function(reader, field) {
    csv_read_double_column(reader, as.character(field))
}

#' Close a CSV Reader
#'
#' @param reader  External pointer to a CsvReader object.
#' @return Invisibly returns \code{NULL}.
#' @export
maxent_csv_close <- function(reader) {
    csv_close(reader)
    invisible(NULL)
}

#' Create a Layer Metadata Object
#'
#' @param name      Character: layer name.
#' @param type      Character: layer type. One of \code{"Continuous"},
#'   \code{"Categorical"}, \code{"Bias"}, \code{"Mask"},
#'   \code{"Probability"}, \code{"Cumulative"}, \code{"DebiasAvg"},
#'   or \code{"Unknown"}.
#' @return External pointer to a Layer C++ object.
#' @export
maxent_layer <- function(name, type = "Continuous") {
    create_layer(as.character(name), as.character(type))
}

#' Get Layer Metadata
#'
#' @param layer  External pointer to a Layer object.
#' @return Named list with \code{name} and \code{type}.
#' @export
maxent_layer_info <- function(layer) {
    get_layer_info(layer)
}

#' Extract Layer Name from a File Path
#'
#' Strips directory prefix and extension.
#'
#' @param path  Character: file path.
#' @return Character: layer name (e.g., \code{"/data/bio1.asc"} → \code{"bio1"}).
#' @export
maxent_layer_name <- function(path) {
    layer_name_from_path(as.character(path))
}

#' Open a CSV File for Writing
#'
#' Opens a CSV file and returns a writer object. Use \code{\link{maxent_csv_write}},
#' \code{\link{maxent_csv_write_num}}, \code{\link{maxent_csv_write_row}}, and
#' \code{\link{maxent_csv_write_close}} to write data and close the file.
#'
#' @param filename   Character: output file path.
#' @param append     Logical: append to an existing file (default \code{FALSE}).
#' @param precision  Integer: number of decimal places for numeric values
#'   (default \code{4}).
#' @return External pointer to a CsvWriter C++ object.
#' @export
maxent_csv_write_open <- function(filename, append = FALSE, precision = 4L) {
    csv_writer_open(as.character(filename), as.logical(append),
                    as.integer(precision))
}

#' Write a String Value to the Current CSV Row
#'
#' Adds a \code{column = value} pair (as a character string) to the current
#' row buffer. Call \code{\link{maxent_csv_write_row}} to flush the row.
#'
#' @param writer  External pointer to a CsvWriter object.
#' @param column  Character: column name.
#' @param value   Character: value to write.
#' @return Invisibly returns the writer object.
#' @export
maxent_csv_write <- function(writer, column, value) {
    csv_writer_print(writer, as.character(column), as.character(value))
    invisible(writer)
}

#' Write a Numeric Value to the Current CSV Row
#'
#' Adds a \code{column = value} pair (as a double) to the current row buffer.
#' Call \code{\link{maxent_csv_write_row}} to flush the row.
#'
#' @param writer  External pointer to a CsvWriter object.
#' @param column  Character: column name.
#' @param value   Numeric: value to write.
#' @return Invisibly returns the writer object.
#' @export
maxent_csv_write_num <- function(writer, column, value) {
    csv_writer_print_double(writer, as.character(column), as.double(value))
    invisible(writer)
}

#' Flush the Current Row to the CSV File
#'
#' Writes all buffered column values as a single CSV row and starts a new
#' row buffer.
#'
#' @param writer  External pointer to a CsvWriter object.
#' @return Invisibly returns the writer object.
#' @export
maxent_csv_write_row <- function(writer) {
    csv_writer_println(writer)
    invisible(writer)
}

#' Close a CSV Writer
#'
#' Flushes any pending data and closes the CSV file.
#'
#' @param writer  External pointer to a CsvWriter object.
#' @return Invisibly returns \code{NULL}.
#' @export
maxent_csv_write_close <- function(writer) {
    csv_writer_close(writer)
    invisible(NULL)
}

test_that("Layer can be created and queried", {
    skip_if_not_installed("maxentcpp")

    l <- maxent_layer("bio1", "Continuous")
    info <- maxent_layer_info(l)

    expect_equal(info$name, "bio1")
    expect_equal(info$type, "Continuous")
})

test_that("Layer type is case-insensitive", {
    skip_if_not_installed("maxentcpp")

    l <- maxent_layer("mask", "MASK")
    expect_equal(maxent_layer_info(l)$type, "Mask")

    l2 <- maxent_layer("cat", "categorical")
    expect_equal(maxent_layer_info(l2)$type, "Categorical")
})

test_that("layer_name_from_path strips path and extension", {
    skip_if_not_installed("maxentcpp")

    expect_equal(maxent_layer_name("/data/bio1.asc"), "bio1")
    expect_equal(maxent_layer_name("temperature.tif"), "temperature")
    expect_equal(maxent_layer_name("noext"), "noext")
})

test_that("ASC round-trip: write then read", {
    skip_if_not_installed("maxentcpp")

    # Create a small grid from a matrix
    mat <- matrix(c(1.0, 2.0, 3.0,
                     4.0, NA,  6.0), nrow = 2, byrow = TRUE)
    g <- maxent_grid_from_matrix(mat, xll = 0, yll = 0, cellsize = 1,
                                 nodata = -9999, name = "test")

    tmp <- tempfile(fileext = ".asc")
    on.exit(unlink(tmp), add = TRUE)

    maxent_write_asc(g, tmp, scientific = FALSE)

    g2   <- maxent_read_asc(tmp)
    info <- maxent_grid_info(g2)

    expect_equal(info$nrows, 2)
    expect_equal(info$ncols, 3)
    expect_equal(info$cellsize, 1.0)
    expect_equal(info$count_data, 5)  # one NA cell

    mat2 <- maxent_grid_to_matrix(g2)
    expect_equal(mat2[1, 1], 1.0)
    expect_equal(mat2[1, 3], 3.0)
    expect_equal(mat2[2, 3], 6.0)
    expect_true(is.na(mat2[2, 2]))
})

test_that("read_grid auto-detects .asc format", {
    skip_if_not_installed("maxentcpp")

    mat <- matrix(c(10, 20, 30, 40), nrow = 2)
    g <- maxent_grid_from_matrix(mat, xll = -118, yll = 34,
                                 cellsize = 0.5, name = "elev")

    tmp <- tempfile(fileext = ".asc")
    on.exit(unlink(tmp), add = TRUE)
    maxent_write_asc(g, tmp)

    g2 <- maxent_read_grid(tmp)
    info <- maxent_grid_info(g2)
    expect_equal(info$nrows, 2)
    expect_equal(info$ncols, 2)
})

test_that("CSV reader reads headers and records", {
    skip_if_not_installed("maxentcpp")

    tmp <- tempfile(fileext = ".csv")
    on.exit(unlink(tmp), add = TRUE)
    writeLines(c("species,lon,lat",
                 "oak,-118.5,36.5",
                 "pine,-119.0,37.0"), tmp)

    reader <- maxent_csv_open(tmp)
    hdrs   <- maxent_csv_headers(reader)
    expect_equal(length(hdrs), 3)
    expect_equal(hdrs[1], "species")

    rec <- maxent_csv_next(reader)
    expect_equal(rec[1], "oak")

    rec2 <- maxent_csv_next(reader)
    expect_equal(rec2[1], "pine")

    rec3 <- maxent_csv_next(reader)
    expect_null(rec3)

    maxent_csv_close(reader)
})

test_that("CSV read_double_column returns correct values", {
    skip_if_not_installed("maxentcpp")

    tmp <- tempfile(fileext = ".csv")
    on.exit(unlink(tmp), add = TRUE)
    writeLines(c("x,y,temp",
                 "1,2,25.5",
                 "3,4,26.0",
                 "5,6,27.5"), tmp)

    reader <- maxent_csv_open(tmp)
    temps  <- maxent_csv_read_column(reader, "temp")
    expect_equal(length(temps), 3)
    expect_equal(temps[1], 25.5)
    expect_equal(temps[3], 27.5)
    maxent_csv_close(reader)
})

test_that("Grid from matrix and back preserves values", {
    skip_if_not_installed("maxentcpp")

    m <- matrix(runif(20), nrow = 4, ncol = 5)
    m[2, 3] <- NA  # one nodata cell

    g   <- maxent_grid_from_matrix(m, xll = -120, yll = 35,
                                    cellsize = 0.1)
    m2  <- maxent_grid_to_matrix(g)

    # Non-NA cells should match
    for (i in 1:4) {
        for (j in 1:5) {
            if (i == 2 && j == 3) {
                expect_true(is.na(m2[i, j]))
            } else {
                expect_equal(m2[i, j], m[i, j], tolerance = 1e-5)
            }
        }
    }
})

test_that("grid_float_info returns correct metadata", {
    skip_if_not_installed("maxentcpp")

    m    <- matrix(1:12, nrow = 3, ncol = 4)
    g    <- maxent_grid_from_matrix(m, xll = 0, yll = 0,
                                    cellsize = 0.5, name = "testgrid")
    info <- maxent_grid_info(g)

    expect_equal(info$nrows, 3)
    expect_equal(info$ncols, 4)
    expect_equal(info$xll, 0)
    expect_equal(info$cellsize, 0.5)
    expect_equal(info$name, "testgrid")
    expect_equal(info$count_data, 12)
})

test_that("CSV writer writes string and numeric values", {
    skip_if_not_installed("maxentcpp")

    tmp <- tempfile(fileext = ".csv")
    on.exit(unlink(tmp), add = TRUE)

    w <- maxent_csv_write_open(tmp)
    maxent_csv_write(w, "species", "oak")
    maxent_csv_write_num(w, "score", 0.85)
    maxent_csv_write_row(w)
    maxent_csv_write(w, "species", "pine")
    maxent_csv_write_num(w, "score", 0.72)
    maxent_csv_write_row(w)
    maxent_csv_write_close(w)

    # Verify the written file
    reader <- maxent_csv_open(tmp)
    hdrs   <- maxent_csv_headers(reader)
    expect_equal(length(hdrs), 2)
    expect_true("species" %in% hdrs)
    expect_true("score" %in% hdrs)

    rec1 <- maxent_csv_next(reader)
    expect_equal(rec1[which(hdrs == "species")], "oak")

    rec2 <- maxent_csv_next(reader)
    expect_equal(rec2[which(hdrs == "species")], "pine")

    maxent_csv_close(reader)
})

test_that("CSV writer round-trip preserves numeric precision", {
    skip_if_not_installed("maxentcpp")

    tmp <- tempfile(fileext = ".csv")
    on.exit(unlink(tmp), add = TRUE)

    values <- c(1.2345, 98.7654, -0.0001)

    w <- maxent_csv_write_open(tmp, precision = 4L)
    for (v in values) {
        maxent_csv_write_num(w, "value", v)
        maxent_csv_write_row(w)
    }
    maxent_csv_write_close(w)

    reader <- maxent_csv_open(tmp)
    result <- maxent_csv_read_column(reader, "value")
    maxent_csv_close(reader)

    expect_equal(length(result), 3)
    expect_equal(result[1], 1.2345, tolerance = 1e-4)
    expect_equal(result[2], 98.7654, tolerance = 1e-4)
    expect_equal(result[3], -0.0001, tolerance = 1e-4)
})

test_that("CSV writer append mode adds rows to existing file", {
    skip_if_not_installed("maxentcpp")

    tmp <- tempfile(fileext = ".csv")
    on.exit(unlink(tmp), add = TRUE)

    # Write first batch
    w1 <- maxent_csv_write_open(tmp, append = FALSE)
    maxent_csv_write(w1, "name", "alpha")
    maxent_csv_write_row(w1)
    maxent_csv_write_close(w1)

    # Append second batch
    w2 <- maxent_csv_write_open(tmp, append = TRUE)
    maxent_csv_write(w2, "name", "beta")
    maxent_csv_write_row(w2)
    maxent_csv_write_close(w2)

    reader <- maxent_csv_open(tmp)
    rec1 <- maxent_csv_next(reader)
    rec2 <- maxent_csv_next(reader)
    maxent_csv_close(reader)

    expect_equal(rec1[1], "alpha")
    expect_equal(rec2[1], "beta")
})

library(testthat)
test_that("maxent_grid_from_terra creates grid from SpatRaster", {
    skip_if_not_installed("maxentcpp")
    skip_if_no_proj()

    library(terra)

    # Create a simple SpatRaster
    r <- rast(xmin = -120, xmax = -116,
              ymin = 35, ymax = 40, res = 1,
              crs = "EPSG:4326")
    values(r) <- 1:(nrow(r)*ncol(r))
    names(r) <- "test_layer"

    g <- maxent_grid_from_terra(r)
    info <- maxent_grid_info(g)

    expect_equal(info$nrows, 5)
    expect_equal(info$ncols, 4)
    expect_equal(info$xll, -120)
    expect_equal(info$yll, 35)
    expect_equal(info$cellsize, 1.0)
    expect_equal(info$name, "test_layer")
    expect_equal(info$count_data, 20)
})

test_that("maxent_grid_to_terra creates SpatRaster from grid", {
    skip_if_not_installed("maxentcpp")
    skip_if_no_proj()

    library(terra)

    mat <- matrix(c(1, 2, 3, 4, NA, 6), nrow = 2, byrow = TRUE)
    g <- maxent_grid_from_matrix(mat, xll = -120, yll = 35,
                                 cellsize = 0.5, name = "bio1")

    r <- maxent_grid_to_terra(g)

    expect_true(inherits(r, "SpatRaster"))
    expect_equal(nrow(r), 2)
    expect_equal(ncol(r), 3)
    expect_equal(names(r), "bio1")

    e <- ext(r)
    expect_equal(unname(e[1]), -120)   # xmin
    expect_equal(unname(e[2]), -118.5) # xmax
    expect_equal(unname(e[3]), 35)     # ymin
    expect_equal(unname(e[4]), 36)     # ymax
    expect_equal(as.numeric(e[1]), -120)   # xmin
    expect_equal(as.numeric(e[2]), -118.5) # xmax
    expect_equal(as.numeric(e[3]), 35)     # ymin
    expect_equal(as.numeric(e[4]), 36)     # ymax

    vals <- as.matrix(r, wide = TRUE)
    expect_equal(vals[1, 1], 1)
    expect_equal(vals[1, 3], 3)
    expect_equal(vals[2, 3], 6)
    expect_true(is.na(vals[2, 2]))
})

test_that("terra round-trip preserves values", {
    skip_if_not_installed("maxentcpp")
    skip_if_no_proj()

    library(terra)

    # Create a raster with known values including NA
    r_orig <- rast(nrows = 3, ncols = 4,
                   xmin = 0, xmax = 4,
                   ymin = 0, ymax = 3,
                   crs = "EPSG:4326")
    vals <- c(1.5, 2.5, NA, 4.5, 5.5, 6.5, 7.5, 8.5, 9.5, 10.5, 11.5, 12.5)
    values(r_orig) <- vals
    names(r_orig) <- "roundtrip"

    # terra → maxentcpp → terra
    g <- maxent_grid_from_terra(r_orig)
    r_back <- maxent_grid_to_terra(g)

    mat_orig <- as.matrix(r_orig, wide = TRUE)
    mat_back <- as.matrix(r_back, wide = TRUE)

    # Compare non-NA cells
    for (i in 1:3) {
        for (j in 1:4) {
            if (is.na(mat_orig[i, j])) {
                expect_true(is.na(mat_back[i, j]))
            } else {
                expect_equal(mat_back[i, j], mat_orig[i, j], tolerance = 1e-5)
            }
        }
    }
})

test_that("maxent_grid_from_terra rejects multi-layer raster", {
    skip_if_not_installed("maxentcpp")
    skip_if_no_terra()

    library(terra)

    r <- rast(nrows = 2, ncols = 2, nlyr = 2,
              xmin = 0, xmax = 2, ymin = 0, ymax = 2)
    values(r) <- 1:8

    expect_error(maxent_grid_from_terra(r), "exactly one layer")
})

test_that("maxent_grid_from_terra rejects non-SpatRaster", {
    skip_if_not_installed("maxentcpp")

    expect_error(maxent_grid_from_terra(matrix(1:4, 2, 2)),
                 "SpatRaster")
})

#' Compute AUC (Area Under the ROC Curve)
#'
#' Computes the Wilcoxon-Mann-Whitney AUC statistic from prediction scores
#' at presence and absence sites. Also returns max-Kappa and its threshold.
#'
#' @param presence Numeric vector of prediction scores at presence sites.
#' @param absence  Numeric vector of prediction scores at absence sites.
#' @return A named list with elements:
#'   \describe{
#'     \item{auc}{AUC value in [0, 1]}
#'     \item{max_kappa}{Best Cohen's Kappa found}
#'     \item{max_kappa_thresh}{Threshold at best Kappa}
#'   }
#' @export
#' @examples
#' result <- maxent_auc(c(0.8, 0.9, 1.0), c(0.1, 0.2, 0.3))
#' result$auc  # 1.0
maxent_auc <- function(presence, absence) {
    eval_auc(as.numeric(presence), as.numeric(absence))
}

#' Compute Pearson Correlation
#'
#' Computes the Pearson correlation coefficient between two numeric vectors.
#'
#' @param x Numeric vector.
#' @param y Numeric vector (same length as x).
#' @return Correlation coefficient in [-1, 1].
#' @export
#' @examples
#' maxent_correlation(c(1, 2, 3), c(2, 4, 6))  # 1.0
maxent_correlation <- function(x, y) {
    eval_correlation(as.numeric(x), as.numeric(y))
}

#' Compute Log-Loss
#'
#' Computes the average cross-entropy log-loss from predictions at
#' presence and absence sites.
#'
#' @param presence Numeric vector of prediction scores at presence sites.
#' @param absence  Numeric vector of prediction scores at absence sites.
#' @return Average log-loss value (lower is better).
#' @export
#' @examples
#' maxent_logloss(c(0.9, 0.8), c(0.1, 0.2))
maxent_logloss <- function(presence, absence) {
    eval_logloss(as.numeric(presence), as.numeric(absence))
}

#' Compute Mean Squared Error
#'
#' Presence sites contribute (1 - pred)^2, absence sites contribute pred^2.
#'
#' @param presence Numeric vector of prediction scores at presence sites.
#' @param absence  Numeric vector of prediction scores at absence sites.
#' @return Mean squared error.
#' @export
#' @examples
#' maxent_square_error(c(1.0, 1.0), c(0.0, 0.0))  # 0.0
maxent_square_error <- function(presence, absence) {
    eval_square_error(as.numeric(presence), as.numeric(absence))
}

#' Compute Misclassification Rate
#'
#' Fraction of misclassified samples at threshold 0.5.
#'
#' @param presence Numeric vector of prediction scores at presence sites.
#' @param absence  Numeric vector of prediction scores at absence sites.
#' @return Misclassification rate in [0, 1].
#' @export
#' @examples
#' maxent_misclassification(c(0.8, 0.9), c(0.1, 0.2))  # 0.0
maxent_misclassification <- function(presence, absence) {
    eval_misclassification(as.numeric(presence), as.numeric(absence))
}

#' Full Model Evaluation
#'
#' Computes all evaluation metrics at once: AUC, correlation, log-loss,
#' squared error, misclassification, max Kappa, and prevalence.
#'
#' @param presence Numeric vector of prediction scores at presence sites.
#' @param absence  Numeric vector of prediction scores at absence sites.
#' @return A named list with:
#'   \describe{
#'     \item{auc}{AUC value}
#'     \item{max_kappa}{Best Cohen's Kappa}
#'     \item{max_kappa_thresh}{Threshold at best Kappa}
#'     \item{correlation}{Pearson correlation with labels}
#'     \item{square_error}{Mean squared error}
#'     \item{logloss}{Cross-entropy log-loss}
#'     \item{misclassification}{Misclassification rate}
#'     \item{prevalence}{Fraction of presence sites}
#'   }
#' @export
#' @examples
#' res <- maxent_evaluate(c(0.9, 0.85, 0.95), c(0.1, 0.15, 0.2))
#' res$auc          # 1.0
#' res$correlation  # > 0
maxent_evaluate <- function(presence, absence) {
    eval_model(as.numeric(presence), as.numeric(absence))
}

# ---------------------------------------------------------------------------
# Java-compatible prediction APIs (match Java Maxent / dismo output)
# These are now the default implementations.
# ---------------------------------------------------------------------------

#' Project Model onto Grids (Raw Output)
#'
#' Produces Java Maxent "raw" scores for every grid cell:
#' \deqn{raw = \frac{exp(lp - lpNorm)}{densityNorm}}
#'
#' This matches the raw output of the Java Maxent software and the
#' \pkg{dismo} R package.  See \code{vignette("PREDICTION_SCALES")} or
#' \code{docs/PREDICTION_SCALES.md} for a full description of prediction
#' scales.
#'
#' @param model         External pointer to a FeaturedSpace object.
#' @param env_grids     List of external pointers to Grid<float> objects.
#' @param feature_names Character vector of environment variable names.
#' @return External pointer to a Grid<float> with Java raw scores.
#' @export
#' @seealso \code{\link{maxent_project_cloglog}},
#'   \code{\link{maxent_project_logistic}}
maxent_project_raw <- function(model, env_grids, feature_names) {
    project_raw(model, env_grids, as.character(feature_names))
}

#' Project Model onto Grids (Cloglog Output)
#'
#' Applies the Java Maxent cloglog transform to grid predictions:
#' \deqn{cloglog = 1 - exp(-exp(H) \cdot raw)}
#'
#' where \eqn{H} is the model entropy and \eqn{raw} is the
#' normalised probability.  This matches the cloglog output of Java
#' Maxent and \pkg{dismo}, and is the recommended output for reporting.
#'
#' @param model         External pointer to a FeaturedSpace object.
#' @param env_grids     List of external pointers to Grid<float> objects.
#' @param feature_names Character vector of environment variable names.
#' @return External pointer to a Grid<float> with cloglog scores in [0, 1].
#' @export
#' @seealso \code{\link{maxent_project_raw}},
#'   \code{\link{maxent_project_logistic}}
maxent_project_cloglog <- function(model, env_grids, feature_names) {
    project_cloglog(model, env_grids, as.character(feature_names))
}

#' Project Model onto Grids (Logistic Output)
#'
#' Applies the Java Maxent logistic transform to grid predictions:
#' \deqn{logistic = \frac{exp(H) \cdot raw}{1 + exp(H) \cdot raw}}
#'
#' where \eqn{H} is the model entropy and \eqn{raw} is the
#' normalised probability.  This matches the logistic output of Java
#' Maxent and \pkg{dismo}.
#'
#' @param model         External pointer to a FeaturedSpace object.
#' @param env_grids     List of external pointers to Grid<float> objects.
#' @param feature_names Character vector of environment variable names.
#' @return External pointer to a Grid<float> with logistic scores in [0, 1].
#' @export
#' @seealso \code{\link{maxent_project_raw}},
#'   \code{\link{maxent_project_cloglog}}
maxent_project_logistic <- function(model, env_grids, feature_names) {
    project_logistic(model, env_grids, as.character(feature_names))
}

#' Extract Raw Predictions at Sample Locations
#'
#' Gets Java Maxent raw scores at specific grid cell locations:
#' \deqn{raw = \frac{exp(lp - lpNorm)}{densityNorm}}
#'
#' This matches the raw output of Java Maxent and \pkg{dismo}.
#'
#' @param model         External pointer to a FeaturedSpace object.
#' @param env_grids     List of external pointers to Grid<float> objects.
#' @param feature_names Character vector of environment variable names.
#' @param rows          Integer vector of row indices.
#' @param cols          Integer vector of column indices.
#' @return Numeric vector of Java raw scores. NaN for NODATA cells.
#' @export
#' @seealso \code{\link{maxent_project_raw}}
maxent_extract_predictions_raw <- function(model, env_grids, feature_names,
                                           rows, cols) {
    extract_predictions_raw(model, env_grids, as.character(feature_names),
                            as.integer(rows), as.integer(cols))
}

#' Extract Cloglog Predictions at Sample Locations
#'
#' Gets Java Maxent cloglog scores at specific grid cell locations:
#' \deqn{cloglog = 1 - exp(-exp(H) \cdot raw)}
#'
#' @param model         External pointer to a FeaturedSpace object.
#' @param env_grids     List of external pointers to Grid<float> objects.
#' @param feature_names Character vector of environment variable names.
#' @param rows          Integer vector of row indices.
#' @param cols          Integer vector of column indices.
#' @return Numeric vector of cloglog scores in [0, 1]. NaN for NODATA cells.
#' @export
#' @seealso \code{\link{maxent_extract_predictions_raw}},
#'   \code{\link{maxent_project_cloglog}}
maxent_extract_predictions_cloglog <- function(model, env_grids,
                                               feature_names, rows, cols) {
    raw <- maxent_extract_predictions_raw(
        model, env_grids, feature_names, rows, cols)
    H    <- maxent_model_entropy(model)
    expH <- exp(H)
    1.0 - exp(-expH * raw)
}

#' Extract Logistic Predictions at Sample Locations
#'
#' Gets Java Maxent logistic scores at specific grid cell locations:
#' \deqn{logistic = \frac{exp(H) \cdot raw}{1 + exp(H) \cdot raw}}
#'
#' @param model         External pointer to a FeaturedSpace object.
#' @param env_grids     List of external pointers to Grid<float> objects.
#' @param feature_names Character vector of environment variable names.
#' @param rows          Integer vector of row indices.
#' @param cols          Integer vector of column indices.
#' @return Numeric vector of logistic scores in [0, 1]. NaN for NODATA cells.
#' @export
#' @seealso \code{\link{maxent_extract_predictions_raw}},
#'   \code{\link{maxent_project_logistic}}
maxent_extract_predictions_logistic <- function(model, env_grids,
                                                feature_names, rows, cols) {
    raw   <- maxent_extract_predictions_raw(
        model, env_grids, feature_names, rows, cols)
    H     <- maxent_model_entropy(model)
    expH  <- exp(H)
    scaled <- expH * raw
    scaled / (1.0 + scaled)
}

test_that("AUC gives 1.0 for perfect separation", {
    skip_if_not_installed("maxentcpp")

    res <- maxent_auc(c(0.8, 0.9, 1.0), c(0.1, 0.2, 0.3))
    expect_equal(res$auc, 1.0, tolerance = 1e-9)
    expect_true(res$max_kappa >= 0)
})

test_that("AUC gives 0.5 for no discrimination", {
    skip_if_not_installed("maxentcpp")

    res <- maxent_auc(c(0.5, 0.5, 0.5), c(0.5, 0.5, 0.5))
    expect_equal(res$auc, 0.5, tolerance = 1e-9)
})

test_that("Pearson correlation works correctly", {
    skip_if_not_installed("maxentcpp")

    # Perfect positive
    r <- maxent_correlation(c(1, 2, 3, 4, 5), c(2, 4, 6, 8, 10))
    expect_equal(r, 1.0, tolerance = 1e-9)

    # Perfect negative
    r2 <- maxent_correlation(c(1, 2, 3, 4, 5), c(10, 8, 6, 4, 2))
    expect_equal(r2, -1.0, tolerance = 1e-9)
})

test_that("Log-loss is 0 for perfect predictions", {
    skip_if_not_installed("maxentcpp")

    ll <- maxent_logloss(c(1.0, 1.0), c(0.0, 0.0))
    expect_equal(ll, 0.0, tolerance = 1e-9)
})

test_that("Log-loss clamps finite predictions outside [0, 1]", {
    skip_if_not_installed("maxentcpp")

    ll <- maxent_logloss(c(-0.2, 1.2), c(1.4, -0.3))
    expect_true(is.finite(ll))
    expect_true(ll >= 0)
})

test_that("Log-loss rejects non-finite predictions", {
    skip_if_not_installed("maxentcpp")

    expect_error(maxent_logloss(c(0.5, NA_real_), c(0.2)), "must be finite")
    expect_error(maxent_logloss(c(0.5), c(Inf)), "must be finite")
})

test_that("Squared error is 0 for perfect predictions", {
    skip_if_not_installed("maxentcpp")

    se <- maxent_square_error(c(1.0, 1.0), c(0.0, 0.0))
    expect_equal(se, 0.0, tolerance = 1e-9)
})

test_that("Misclassification is 0 for correct predictions", {
    skip_if_not_installed("maxentcpp")

    mc <- maxent_misclassification(c(0.8, 0.9), c(0.1, 0.2))
    expect_equal(mc, 0.0, tolerance = 1e-9)
})

test_that("Misclassification is 1 for all wrong", {
    skip_if_not_installed("maxentcpp")

    mc <- maxent_misclassification(c(0.1, 0.2), c(0.8, 0.9))
    expect_equal(mc, 1.0, tolerance = 1e-9)
})

test_that("Full evaluate returns all metrics", {
    skip_if_not_installed("maxentcpp")

    res <- maxent_evaluate(c(0.9, 0.85, 0.95), c(0.1, 0.15, 0.2))
    expect_true("auc" %in% names(res))
    expect_true("max_kappa" %in% names(res))
    expect_true("correlation" %in% names(res))
    expect_true("square_error" %in% names(res))
    expect_true("logloss" %in% names(res))
    expect_true("misclassification" %in% names(res))
    expect_true("prevalence" %in% names(res))

    expect_equal(res$auc, 1.0, tolerance = 1e-9)
    expect_equal(res$prevalence, 0.5, tolerance = 1e-9)
    expect_true(res$correlation > 0)
})

test_that("AUC handles partial overlap", {
    skip_if_not_installed("maxentcpp")

    res <- maxent_auc(c(0.6, 0.7, 0.8), c(0.3, 0.5, 0.65))
    expect_true(res$auc > 0.5)
    expect_true(res$auc < 1.0)
})

# Tests for cloglog and logistic projection transforms (un-scaled)

# Helper: build a tiny trained model and matching grids.
make_projection_model <- function() {
    dim <- maxent_dimension(nrows = 3, ncols = 3, xll = 0, yll = 0, cellsize = 1)
    g0 <- maxent_grid(dim, "env0")
    g1 <- maxent_grid(dim, "env1")
    idx <- 0L
    for (r in 0:2) {
        for (cc in 0:2) {
            grid_set_value(g0, r, cc, idx / 8.0)
            grid_set_value(g1, r, cc, 0.5)
            idx <- idx + 1L
        }
    }

    vals0 <- sapply(0:8, function(i) i / 8.0)
    vals1 <- rep(0.5, 9)
    env_data <- list(env0 = vals0, env1 = vals1)
    features <- maxent_generate_features(env_data, types = "linear",
                                         n_thresholds = 0L, n_hinges = 0L)
    model <- maxent_featured_space(9L, c(6L, 7L, 8L), features)
    maxent_fit(model, max_iter = 200L, convergence = 1e-5, beta_multiplier = 1.0)
    list(model = model, grids = list(g0, g1), names = c("env0", "env1"))
}

test_that("cloglog projection values are in [0, 1]", {
    skip_if_not_installed("maxentcpp")

    m <- make_projection_model()

    cloglog_grid <- maxent_project_cloglog(m$model, m$grids, m$names)
    cloglog_mat  <- maxent_grid_to_matrix(cloglog_grid)

    expect_true(all(cloglog_mat >= 0 & cloglog_mat <= 1))
})

test_that("logistic projection values are in [0, 1]", {
    skip_if_not_installed("maxentcpp")

    m <- make_projection_model()

    logistic_grid <- maxent_project_logistic(m$model, m$grids, m$names)
    logistic_mat  <- maxent_grid_to_matrix(logistic_grid)

    expect_true(all(logistic_mat >= 0 & logistic_mat <= 1))
})

test_that("cloglog and logistic values are in [0, 1] and increase with raw", {
    skip_if_not_installed("maxentcpp")

    m <- make_projection_model()

    raw_grid      <- maxent_project_raw(m$model, m$grids, m$names)
    cloglog_grid  <- maxent_project_cloglog(m$model, m$grids, m$names)
    logistic_grid <- maxent_project_logistic(m$model, m$grids, m$names)

    raw_mat      <- maxent_grid_to_matrix(raw_grid)
    cloglog_mat  <- maxent_grid_to_matrix(cloglog_grid)
    logistic_mat <- maxent_grid_to_matrix(logistic_grid)

    expect_true(all(cloglog_mat  >= 0 & cloglog_mat  <= 1))
    expect_true(all(logistic_mat >= 0 & logistic_mat <= 1))

    # Transforms are monotonically increasing: higher raw → higher output
    # Use Spearman rank correlation (robust to nonlinearity)
    expect_gt(cor(as.vector(raw_mat), as.vector(cloglog_mat),  method = "spearman"), 0.99)
    expect_gt(cor(as.vector(raw_mat), as.vector(logistic_mat), method = "spearman"), 0.99)
})

# ============================================================================
# Tests for Java-compatible prediction APIs
# ============================================================================

test_that("cloglog_java matches 1 - exp(-exp(H) * raw_java)", {
    skip_if_not_installed("maxentcpp")

    m <- make_projection_model()

    raw_java_grid     <- maxent_project_raw(m$model, m$grids, m$names)
    cloglog_java_grid <- maxent_project_cloglog(m$model, m$grids, m$names)

    raw_java_mat     <- maxent_grid_to_matrix(raw_java_grid)
    cloglog_java_mat <- maxent_grid_to_matrix(cloglog_java_grid)

    H    <- maxent_model_entropy(m$model)
    expH <- exp(H)

    expected <- 1.0 - exp(-expH * raw_java_mat)
    expect_equal(cloglog_java_mat, expected, tolerance = 1e-5)
    expect_true(all(cloglog_java_mat >= 0 & cloglog_java_mat <= 1))
})

test_that("logistic_java matches (exp(H)*raw_java) / (1 + exp(H)*raw_java)", {
    skip_if_not_installed("maxentcpp")

    m <- make_projection_model()

    raw_java_grid      <- maxent_project_raw(m$model, m$grids, m$names)
    logistic_java_grid <- maxent_project_logistic(m$model, m$grids, m$names)

    raw_java_mat      <- maxent_grid_to_matrix(raw_java_grid)
    logistic_java_mat <- maxent_grid_to_matrix(logistic_java_grid)

    H    <- maxent_model_entropy(m$model)
    expH <- exp(H)
    scaled   <- expH * raw_java_mat
    expected <- scaled / (1.0 + scaled)

    expect_equal(logistic_java_mat, expected, tolerance = 1e-5)
    expect_true(all(logistic_java_mat >= 0 & logistic_java_mat <= 1))
})

test_that("extract_predictions_raw_java == project_raw_java sampled at points", {
    skip_if_not_installed("maxentcpp")

    m <- make_projection_model()

    # Use a few grid points: rows and cols (0-based)
    rows <- c(0L, 1L, 2L)
    cols <- c(0L, 1L, 2L)

    # Extract via point extraction
    pt_raw_java <- maxent_extract_predictions_raw(
        m$model, m$grids, m$names, rows, cols)

    # Extract from grid
    raw_java_grid <- maxent_project_raw(m$model, m$grids, m$names)
    raw_java_mat  <- maxent_grid_to_matrix(raw_java_grid)

    # maxent_grid_to_matrix gives matrix with row 0 at top
    grid_vals <- mapply(function(r, c) raw_java_mat[r + 1L, c + 1L], rows, cols)

    expect_equal(pt_raw_java, grid_vals, tolerance = 1e-5)
})

test_that("extract_predictions_cloglog_java matches formula", {
    skip_if_not_installed("maxentcpp")

    m <- make_projection_model()

    rows <- c(0L, 1L, 2L)
    cols <- c(0L, 1L, 2L)

    pt_raw_java    <- maxent_extract_predictions_raw(
        m$model, m$grids, m$names, rows, cols)
    pt_cloglog_java <- maxent_extract_predictions_cloglog(
        m$model, m$grids, m$names, rows, cols)

    H    <- maxent_model_entropy(m$model)
    expH <- exp(H)
    expected <- 1.0 - exp(-expH * pt_raw_java)

    expect_equal(pt_cloglog_java, expected, tolerance = 1e-5)
})

test_that("extract_predictions_logistic_java matches formula", {
    skip_if_not_installed("maxentcpp")

    m <- make_projection_model()

    rows <- c(0L, 1L, 2L)
    cols <- c(0L, 1L, 2L)

    pt_raw_java      <- maxent_extract_predictions_raw(
        m$model, m$grids, m$names, rows, cols)
    pt_logistic_java <- maxent_extract_predictions_logistic(
        m$model, m$grids, m$names, rows, cols)

    H    <- maxent_model_entropy(m$model)
    expH <- exp(H)
    scaled   <- expH * pt_raw_java
    expected <- scaled / (1.0 + scaled)

    expect_equal(pt_logistic_java, expected, tolerance = 1e-5)
})

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grid_io.R
\name{maxent_layer_name}
\alias{maxent_layer_name}
\title{Extract Layer Name from a File Path}
\usage{
maxent_layer_name(path)
}
\arguments{
\item{path}{Character: file path.}
}
\value{
Character: layer name (e.g., \code{"/data/bio1.asc"} → \code{"bio1"}).
}
\description{
Strips directory prefix and extension.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{maxent_get_weights}
\alias{maxent_get_weights}
\title{Get current distribution weights from a FeaturedSpace}
\usage{
maxent_get_weights(fs_ptr)
}
\arguments{
\item{fs_ptr}{External pointer to a FeaturedSpace object.}
}
\value{
Numeric vector of normalized weights (sums to 1).
}
\description{
Get current distribution weights from a FeaturedSpace
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/featured_space.R
\name{maxent_predict_model}
\alias{maxent_predict_model}
\title{Predict with a Trained MaxEnt Model}
\usage{
maxent_predict_model(featured_space, newdata)
}
\arguments{
\item{featured_space}{External pointer to a trained FeaturedSpace object.}

\item{newdata}{Numeric matrix: one row per new point, one column per
feature.  Column values must be the \emph{already-evaluated} feature
values (e.g., from running \code{maxent_feature_eval()} for each feature
and each point).}
}
\value{
Numeric vector of raw (unnormalized) prediction scores.
}
\description{
Computes raw Gibbs distribution scores for new environmental data, using
the feature lambdas stored in the trained FeaturedSpace.
}
\examples{
\donttest{
# After training, predict on 5 new points with 2 features each
newdata <- matrix(runif(10), nrow = 5, ncol = 2)
preds <- maxent_predict_model(fs, newdata)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/example_occ_df.R
\docType{data}
\name{example_occ_df}
\alias{example_occ_df}
\title{Species occurrence points from Abeillia abeillei presence points after
download and clean from GBIF. This is a hummingbird example.
A dataset with three variables. Contains scientific name, longitude and
latitude.}
\format{
A data frame with 73 rows and 3 variables:
\describe{
  \item{species}{Species name}
  \item{long}{Decimal longitude geographical coordinate, in degrees}
  \item{lat}{Decimal latitude geographical coordinate, in degrees}
}
}
\source{
<doi:10.1016/j.ecolmodel.2021.109823>
}
\usage{
example_occ_df
}
\description{
Species occurrence points from Abeillia abeillei presence points after
download and clean from GBIF. This is a hummingbird example.
A dataset with three variables. Contains scientific name, longitude and
latitude.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grid_io.R
\name{maxent_csv_read_column}
\alias{maxent_csv_read_column}
\title{Read an Entire Column as Doubles}
\usage{
maxent_csv_read_column(reader, field)
}
\arguments{
\item{reader}{External pointer to a CsvReader object.}

\item{field}{Character: column name.}
}
\value{
Numeric vector.
}
\description{
Reads from the current file position to EOF.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grid_io.R
\name{maxent_grid_info}
\alias{maxent_grid_info}
\title{Get Grid Information}
\usage{
maxent_grid_info(grid)
}
\arguments{
\item{grid}{External pointer to a GridFloat object.}
}
\value{
Named list with: \code{nrows}, \code{ncols}, \code{xll},
  \code{yll}, \code{cellsize}, \code{nodata}, \code{name},
  \code{count_data}.
}
\description{
Returns metadata about a GridFloat object read from a file.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{create_sample}
\alias{create_sample}
\title{Create a Sample object}
\usage{
create_sample(point, row, col, lat, lon, name)
}
\arguments{
\item{point}{Point identifier}

\item{row}{Row index in grid}

\item{col}{Column index in grid}

\item{lat}{Latitude coordinate}

\item{lon}{Longitude coordinate}

\item{name}{Sample name/identifier}
}
\value{
External pointer to Sample object
}
\description{
Creates a species occurrence sample point
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{compute_variable_ranges}
\alias{compute_variable_ranges}
\title{Compute Variable Ranges from Grids}
\usage{
compute_variable_ranges(grid_ptrs)
}
\arguments{
\item{grid_ptrs}{List of external pointers to Grid<float> objects.}
}
\value{
A data.frame with columns: min, max.
}
\description{
Scans all valid cells to determine min/max of each variable.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grid_io.R
\name{maxent_read_grid}
\alias{maxent_read_grid}
\title{Read a Grid File (Auto-Detect Format)}
\usage{
maxent_read_grid(filename)
}
\arguments{
\item{filename}{Character: path to the grid file.}
}
\value{
External pointer to a GridFloat C++ object.
}
\description{
Reads a raster grid file. The format is detected from the file extension.
Currently supports \code{.asc} (ESRI ASCII Grid).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{sample_get_feature}
\alias{sample_get_feature}
\title{Get feature value from a sample}
\usage{
sample_get_feature(sample_ptr, feature_name, default_val = 0)
}
\arguments{
\item{sample_ptr}{External pointer to Sample}

\item{feature_name}{Name of the feature}

\item{default_val}{Default value if feature not found}
}
\value{
Feature value
}
\description{
Get feature value from a sample
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/maxent.R
\name{print.maxent_sample}
\alias{print.maxent_sample}
\title{Print Sample Information}
\usage{
\method{print}{maxent_sample}(x, ...)
}
\arguments{
\item{x}{Sample object}

\item{...}{Additional arguments (ignored)}
}
\value{
Invisibly returns \code{x}.
}
\description{
Print Sample Information
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{grid_read_asc}
\alias{grid_read_asc}
\title{Read an ESRI ASCII grid (.asc) file}
\usage{
grid_read_asc(filename)
}
\arguments{
\item{filename}{Path to the .asc file.}
}
\value{
External pointer to a GridFloat object.
}
\description{
Reads an ASC raster file and returns a float Grid external pointer.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{csv_writer_open}
\alias{csv_writer_open}
\title{Open a CSV file for writing}
\usage{
csv_writer_open(filename, append = FALSE, precision = 4L)
}
\arguments{
\item{filename}{Path to the output CSV file.}

\item{append}{Logical: append to existing file (default FALSE).}

\item{precision}{Number of decimal places for doubles (default 4).}
}
\value{
External pointer to a CsvWriter object.
}
\description{
Open a CSV file for writing
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/featured_space.R
\name{maxent_model_loss}
\alias{maxent_model_loss}
\title{Get Model Loss}
\usage{
maxent_model_loss(featured_space)
}
\arguments{
\item{featured_space}{External pointer to a FeaturedSpace object.}
}
\value{
Scalar: current loss value.
}
\description{
Returns the current (negative log-likelihood) loss of the model.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grid_io.R
\name{maxent_read_asc}
\alias{maxent_read_asc}
\title{Read an ESRI ASCII Grid File}
\usage{
maxent_read_asc(filename)
}
\arguments{
\item{filename}{Character: path to the \code{.asc} file.}
}
\value{
External pointer to a GridFloat C++ object.
}
\description{
Reads a \code{.asc} raster file into a GridFloat object.
}
\examples{
\donttest{
g <- maxent_read_asc("bio1.asc")
info <- maxent_grid_info(g)
print(info)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{eval_auc}
\alias{eval_auc}
\title{Compute AUC (Area Under the ROC Curve)}
\usage{
eval_auc(presence, absence)
}
\arguments{
\item{presence}{Numeric vector of prediction scores at presence sites.}

\item{absence}{Numeric vector of prediction scores at absence sites.}
}
\value{
A named list with elements: auc, max_kappa, max_kappa_thresh.
}
\description{
Computes the Wilcoxon-Mann-Whitney AUC statistic from prediction scores
at presence and absence sites.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{compute_percent_contribution}
\alias{compute_percent_contribution}
\title{Compute Percent Contribution}
\usage{
compute_percent_contribution(fs_ptr, feature_names)
}
\arguments{
\item{fs_ptr}{External pointer to a FeaturedSpace object.}

\item{feature_names}{Character vector of base variable names.}
}
\value{
A data.frame with columns: name, contribution.
}
\description{
Computes variable contribution based on sum of absolute lambda values
for features derived from each variable.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{maxent_featured_space_info}
\alias{maxent_featured_space_info}
\title{Get FeaturedSpace metadata}
\usage{
maxent_featured_space_info(fs_ptr)
}
\arguments{
\item{fs_ptr}{External pointer to a FeaturedSpace object.}
}
\value{
Named list with num_points, num_samples, num_features,
  density_normalizer, linear_predictor_normalizer, has_bias.
}
\description{
Get FeaturedSpace metadata
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{compute_response_curve}
\alias{compute_response_curve}
\title{Compute Java-compatible Marginal Response Curve}
\usage{
compute_response_curve(
  fs_ptr,
  grid_ptrs,
  feature_names,
  var_index,
  n_steps = 100L
)
}
\arguments{
\item{fs_ptr}{External pointer to a FeaturedSpace object.}

\item{grid_ptrs}{List of external pointers to Grid<float> objects.}

\item{feature_names}{Character vector of environment variable names.}

\item{var_index}{0-based index of the variable to vary.}

\item{n_steps}{Number of steps across the variable range.}
}
\value{
A data.frame with columns: value, prediction.
}
\description{
Like compute_response_curve() but applies the Java Maxent cloglog:
  cloglog_java = 1 - exp(-exp(H) * raw_java)
}
\details{
This matches the response curves produced by Java Maxent and dismo.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/featured_space.R
\name{maxent_space_info}
\alias{maxent_space_info}
\title{Get FeaturedSpace Information}
\usage{
maxent_space_info(featured_space)
}
\arguments{
\item{featured_space}{External pointer to a FeaturedSpace object.}
}
\value{
Named list with: \code{num_points}, \code{num_samples},
  \code{num_features}, \code{density_normalizer},
  \code{linear_predictor_normalizer}.
}
\description{
Returns basic metadata about a FeaturedSpace object.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/features.R
\name{maxent_feature_info}
\alias{maxent_feature_info}
\title{Get Feature Properties}
\usage{
maxent_feature_info(feature)
}
\arguments{
\item{feature}{External pointer to a Feature C++ object.}
}
\value{
Named list with elements: \code{name}, \code{type}, \code{lambda},
  \code{min}, \code{max}, \code{size}.
}
\description{
Returns a list of metadata for a feature object.
}
\examples{
vals <- c(0, 5, 10)
f <- maxent_linear_feature(vals, "temp")
maxent_feature_info(f)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_diagnostics.R
\name{maxent_clamp}
\alias{maxent_clamp}
\title{Clamp Environmental Grids}
\usage{
maxent_clamp(env_grids, var_mins, var_maxs)
}
\arguments{
\item{env_grids}{List of external pointers to Grid<float> objects.}

\item{var_mins}{Numeric vector of minimum training values per variable.}

\item{var_maxs}{Numeric vector of maximum training values per variable.}
}
\value{
A named list with:
  \describe{
    \item{clamped_grids}{List of external pointers to clamped grids}
    \item{clamp_grid}{External pointer to clamping magnitude grid}
  }
}
\description{
Restricts environmental variable values to the range observed during
training. Values below min are set to min, values above max are set to max.
A clamping grid records the total absolute clamping at each cell.
}
\examples{
\donttest{
result <- maxent_clamp(list(g1, g2), c(0, 50), c(30, 200))
clamped <- result$clamped_grids
clamp_mat <- maxent_grid_to_matrix(result$clamp_grid)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/maxent_outputs.R
\name{maxent_write_omission_csv}
\alias{maxent_write_omission_csv}
\title{Write an Omission / Threshold CSV}
\usage{
maxent_write_omission_csv(
  model,
  env_grids,
  feature_names,
  presence_rows,
  presence_cols,
  output_dir,
  species,
  test_rows = NULL,
  test_cols = NULL
)
}
\arguments{
\item{model}{External pointer to a trained FeaturedSpace object.}

\item{env_grids}{List of external pointers to Grid<float> objects.}

\item{feature_names}{Character vector of environment variable names.}

\item{presence_rows}{Integer vector of presence row indices (0-based).}

\item{presence_cols}{Integer vector of presence column indices (0-based).}

\item{output_dir}{Character: directory for output files.}

\item{species}{Character: species name (used in the filename).}

\item{test_rows}{Integer vector of test row indices (0-based) or
\code{NULL} (default).}

\item{test_cols}{Integer vector of test column indices (0-based) or
\code{NULL} (default).}
}
\value{
Invisibly returns the path to the written CSV file.
}
\description{
Computes cumulative predictions at all background + presence locations and
derives nine standard threshold statistics, writing the result to
\code{<output_dir>/<species>_omission.csv}.  The output format mirrors
\code{density/Runner.java::writeCumulativeIndex()}.
}
\examples{
\donttest{
maxent_write_omission_csv(model, list(g1, g2), c("bio1", "bio12"),
  pres_rows, pres_cols, output_dir = tempdir(), species = "Sp1")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grid_io.R
\name{maxent_csv_headers}
\alias{maxent_csv_headers}
\title{Get CSV Column Headers}
\usage{
maxent_csv_headers(reader)
}
\arguments{
\item{reader}{External pointer to a CsvReader object.}
}
\value{
Character vector of column names.
}
\description{
Get CSV Column Headers
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{grid_has_data}
\alias{grid_has_data}
\title{Check if grid cell has data}
\usage{
grid_has_data(grid_ptr, row, col)
}
\arguments{
\item{grid_ptr}{External pointer to Grid}

\item{row}{Row index}

\item{col}{Column index}
}
\value{
TRUE if cell has valid data
}
\description{
Check if grid cell has data
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{compute_mess}
\alias{compute_mess}
\title{Compute MESS (Multivariate Environmental Similarity Surface)}
\usage{
compute_mess(grid_ptrs, reference_values, feature_names)
}
\arguments{
\item{grid_ptrs}{List of external pointers to Grid<float> objects.}

\item{reference_values}{List of numeric vectors with reference values
for each variable (e.g. values at training sites).}

\item{feature_names}{Character vector of variable names.}
}
\value{
A named list with:
  \describe{
    \item{mess_grid}{External pointer to Grid<float> with MESS values}
    \item{mod_grid}{External pointer to Grid<float> with MoD variable index (1-based)}
  }
}
\description{
Measures how similar each cell is to the training environment.
Negative values indicate novel (non-analog) conditions.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/featured_space_raster.R
\name{maxent_train_terra}
\alias{maxent_train_terra}
\title{Train MaxEnt directly from a terra SpatRaster}
\usage{
maxent_train_terra(
  rast,
  occurrences,
  lon_col = "longitude",
  lat_col = "latitude",
  feature_types = c("linear", "quadratic", "product", "threshold", "hinge"),
  n_thresholds = 10L,
  n_hinges = 10L,
  max_iter = 500L,
  convergence = 1e-05,
  beta_multiplier = 1,
  min_deviation = 0.001
)
}
\arguments{
\item{rast}{A \code{terra::SpatRaster}.}

\item{occurrences}{Either a two-column matrix/data.frame (x,y) or a vector
of 1-based raster cell indices.}

\item{lon_col}{Longitude column name when \code{occurrences} is a data.frame.}

\item{lat_col}{Latitude column name when \code{occurrences} is a data.frame.}

\item{feature_types}{Feature type set passed to feature generation.}

\item{n_thresholds}{Number of threshold knots.}

\item{n_hinges}{Number of hinge knots.}

\item{max_iter}{Maximum training iterations.}

\item{convergence}{Convergence threshold.}

\item{beta_multiplier}{Regularization multiplier.}

\item{min_deviation}{Minimum deviation floor.}
}
\value{
A named list with training results plus \code{model} and
  \code{sample_indices}.
}
\description{
High-level convenience wrapper that builds a streaming FeaturedSpace from
\code{rast}, maps occurrence locations to finite-stream sample indices, and
trains with \code{\link{maxent_fit}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_evaluation.R
\name{maxent_misclassification}
\alias{maxent_misclassification}
\title{Compute Misclassification Rate}
\usage{
maxent_misclassification(presence, absence)
}
\arguments{
\item{presence}{Numeric vector of prediction scores at presence sites.}

\item{absence}{Numeric vector of prediction scores at absence sites.}
}
\value{
Misclassification rate in [0, 1].
}
\description{
Fraction of misclassified samples at threshold 0.5.
}
\examples{
maxent_misclassification(c(0.8, 0.9), c(0.1, 0.2))  # 0.0
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{csv_writer_print}
\alias{csv_writer_print}
\title{Write a value into the current row of a CSV}
\usage{
csv_writer_print(writer_ptr, column, value)
}
\arguments{
\item{writer_ptr}{External pointer to a CsvWriter object.}

\item{column}{Column name.}

\item{value}{Value to write (character).}
}
\value{
Called for side effects; returns invisibly.
}
\description{
Write a value into the current row of a CSV
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grid_io.R
\name{maxent_grid_to_matrix}
\alias{maxent_grid_to_matrix}
\title{Convert Grid to R Matrix}
\usage{
maxent_grid_to_matrix(grid)
}
\arguments{
\item{grid}{External pointer to a GridFloat object.}
}
\value{
Numeric matrix (\code{nrows} \eqn{\times} \code{ncols}).
}
\description{
Extracts the data from a GridFloat as an R numeric matrix.
NODATA cells are converted to \code{NA}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{compute_permutation_importance}
\alias{compute_permutation_importance}
\title{Compute Permutation Importance}
\usage{
compute_permutation_importance(
  fs_ptr,
  grid_ptrs,
  feature_names,
  presence_rows,
  presence_cols,
  absence_rows,
  absence_cols,
  seed = 42L
)
}
\arguments{
\item{fs_ptr}{External pointer to a FeaturedSpace object.}

\item{grid_ptrs}{List of external pointers to Grid<float> objects.}

\item{feature_names}{Character vector of environment variable names.}

\item{presence_rows}{Integer vector of presence site row indices.}

\item{presence_cols}{Integer vector of presence site column indices.}

\item{absence_rows}{Integer vector of absence site row indices.}

\item{absence_cols}{Integer vector of absence site column indices.}

\item{seed}{Random seed for reproducibility.}
}
\value{
A data.frame with columns: name, permutation_importance.
}
\description{
Measures how much each environmental variable contributes to model
prediction quality by permuting each variable and measuring AUC drop.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/featured_space_raster.R
\name{maxent_raster_sample_indices}
\alias{maxent_raster_sample_indices}
\title{Convert occurrence coordinates to 0-based stream indices}
\usage{
maxent_raster_sample_indices(rast, cells)
}
\arguments{
\item{rast}{A \code{terra::SpatRaster}.}

\item{cells}{Integer vector of 1-based cell indices
(as returned by \code{terra::cellFromXY}).}
}
\value{
Integer vector of 0-based indices within the finite-cell stream.
  Occurrence cells that fall on NA raster values (and therefore never
  appear in the stream) are silently dropped and a warning is emitted.
}
\description{
Maps occurrence cell indices (as produced by \code{terra::cellFromXY})
into 0-based indices within the concatenated stream of finite
(non-NA) background cells emitted by
\code{\link{maxent_featured_space_from_rast}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{eval_model}
\alias{eval_model}
\title{Full Model Evaluation}
\usage{
eval_model(presence, absence)
}
\arguments{
\item{presence}{Numeric vector of prediction scores at presence sites.}

\item{absence}{Numeric vector of prediction scores at absence sites.}
}
\value{
A named list with: auc, max_kappa, max_kappa_thresh,
  correlation, square_error, logloss, misclassification, prevalence.
}
\description{
Computes all evaluation metrics at once: AUC, correlation, log-loss,
squared error, misclassification, max kappa, and prevalence.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{csv_read_double_column}
\alias{csv_read_double_column}
\title{Read an entire named column as doubles}
\usage{
csv_read_double_column(reader_ptr, field)
}
\arguments{
\item{reader_ptr}{External pointer to a CsvReader object.}

\item{field}{Column name.}
}
\value{
Numeric vector.
}
\description{
Reads from the current position to EOF.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{grid_get_value}
\alias{grid_get_value}
\title{Get grid value}
\usage{
grid_get_value(grid_ptr, row, col)
}
\arguments{
\item{grid_ptr}{External pointer to Grid}

\item{row}{Row index}

\item{col}{Column index}
}
\value{
Grid value
}
\description{
Get grid value
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequential.R
\name{maxent_sequential_fit}
\alias{maxent_sequential_fit}
\title{Train a MaxEnt Model with the Java-Equivalent Sequential Optimizer}
\usage{
maxent_sequential_fit(
  featured_space,
  max_iter = 500L,
  convergence = 1e-05,
  beta_multiplier = 1,
  min_deviation = 0.001,
  parallel_update_frequency = 10L,
  disable_convergence_test = FALSE,
  trajectory_iterations = integer(0)
)
}
\arguments{
\item{featured_space}{External pointer to a FeaturedSpace object (from
\code{maxent_featured_space()}).}

\item{max_iter}{Maximum number of iterations (default \code{500L}).}

\item{convergence}{Convergence threshold on the per-20-iteration loss
improvement (default \code{1e-5}).  Ignored when
\code{disable_convergence_test = TRUE}.}

\item{beta_multiplier}{Regularization multiplier (default \code{1.0}).}

\item{min_deviation}{Minimum sample-deviation floor used in
regularization (default \code{0.001}).}

\item{parallel_update_frequency}{Iteration frequency at which
\code{doParallelUpdate} runs (default \code{10L}, matching Java).}

\item{disable_convergence_test}{Logical: when \code{TRUE}, the loop
runs a fixed \code{max_iter} iterations with no early stop.  This
is required for deterministic per-iteration trajectory comparisons
against the Java oracle.  Default \code{FALSE}.}

\item{trajectory_iterations}{Integer vector of 1-based iteration
indices at which to capture \code{(loss, entropy, lambdas)}
snapshots.  May be empty.  Snapshots outside
\code{[1, max_iter]} are silently dropped.}
}
\value{
Named list with:
  \describe{
    \item{loss}{Final regularized loss (scalar).}
    \item{entropy}{Shannon entropy of the trained distribution.}
    \item{iterations}{Number of training iterations completed.}
    \item{converged}{Logical: whether the convergence threshold was
      reached (always \code{FALSE} when
      \code{disable_convergence_test = TRUE}).}
    \item{lambdas}{Numeric vector of final lambda values.}
    \item{trajectory}{A \code{data.frame} with one row per captured
      checkpoint and columns \code{iteration, loss, entropy,
      lambda_0, lambda_1, ..., lambda_{J-1}}.}
  }
}
\description{
Runs the full \code{density.Sequential} optimizer ported from the
original Java Maxent 3.4.4 on a FeaturedSpace.  Unlike
\code{maxent_fit()}, which uses the historical \code{goodAlpha}-only
loop, this trainer reproduces the real Java optimizer: feature
selection via \code{deltaLossBound}, Newton step with 1-D line search
(\code{searchAlpha}), every-10-iteration \code{doParallelUpdate} with
undo on loss-violating batch steps, and per-feature state tracking.
}
\details{
The lambda trajectory produced by this trainer matches the Java
oracle's to \eqn{10^{-6}} on \eqn{\left\|\Delta\lambda\right\|_\infty}
at every checkpoint, on both the symmetric and asymmetric fixtures in
\code{maxentcppCompTest} — see
\code{docs/ARCHITECTURE_xtensor_openmp.md} and the Phase C baseline
report for details.
}
\examples{
\donttest{
fs  <- maxent_featured_space(100L, 90:99, list(f))
res <- maxent_sequential_fit(
    fs,
    max_iter                 = 500L,
    disable_convergence_test = TRUE,
    trajectory_iterations    = c(1L, 2L, 5L, 10L, 50L, 100L, 500L))
print(res$trajectory)
}
}
\seealso{
\code{\link{maxent_fit}}, \code{\link{maxent_featured_space}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_diagnostics.R
\name{maxent_mess}
\alias{maxent_mess}
\title{Compute MESS (Multivariate Environmental Similarity Surface)}
\usage{
maxent_mess(env_grids, reference_values, feature_names)
}
\arguments{
\item{env_grids}{List of external pointers to Grid<float> objects.}

\item{reference_values}{List of numeric vectors with reference values
for each variable (e.g. values at training sites).}

\item{feature_names}{Character vector of variable names.}
}
\value{
A named list with:
  \describe{
    \item{mess_grid}{External pointer to Grid<float> with MESS values}
    \item{mod_grid}{External pointer to Grid<float> with Most Dissimilar
      Variable index (1-based)}
  }
}
\description{
Measures how similar each cell is to the training environment using the
full distribution of reference values. Negative MESS values indicate
novel (non-analog) environments.
}
\examples{
\donttest{
result <- maxent_mess(list(g1, g2),
            list(temp_train_vals, precip_train_vals),
            c("temp", "precip"))
mess_mat <- maxent_grid_to_matrix(result$mess_grid)
mod_mat  <- maxent_grid_to_matrix(result$mod_grid)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{csv_open}
\alias{csv_open}
\title{Open a CSV file for reading}
\usage{
csv_open(filename, has_header = TRUE)
}
\arguments{
\item{filename}{Path to the CSV file.}

\item{has_header}{Logical: first line contains column names (default TRUE).}
}
\value{
External pointer to a CsvReader object.
}
\description{
Open a CSV file for reading
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{maxent_predict}
\alias{maxent_predict}
\title{Predict with a trained FeaturedSpace model}
\usage{
maxent_predict(fs_ptr, new_data)
}
\arguments{
\item{fs_ptr}{External pointer to a trained FeaturedSpace object.}

\item{new_data}{Numeric matrix with one row per new point and one column
per feature (values must be pre-evaluated, i.e., the feature
transformation already applied).}
}
\value{
Numeric vector of raw scores (unnormalized).
}
\description{
Computes raw Gibbs distribution scores for new environmental data.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grid_io.R
\name{maxent_csv_write_row}
\alias{maxent_csv_write_row}
\title{Flush the Current Row to the CSV File}
\usage{
maxent_csv_write_row(writer)
}
\arguments{
\item{writer}{External pointer to a CsvWriter object.}
}
\value{
Invisibly returns the writer object.
}
\description{
Writes all buffered column values as a single CSV row and starts a new
row buffer.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{get_grid_dimension_info}
\alias{get_grid_dimension_info}
\title{Get grid dimension properties}
\usage{
get_grid_dimension_info(dim_ptr)
}
\arguments{
\item{dim_ptr}{External pointer to GridDimension}
}
\value{
List with dimension properties
}
\description{
Get grid dimension properties
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/featured_space.R
\name{maxent_featured_space}
\alias{maxent_featured_space}
\title{Create a FeaturedSpace Object}
\usage{
maxent_featured_space(
  num_points,
  sample_indices,
  features,
  bias_weights = NULL
)
}
\arguments{
\item{num_points}{Integer: total number of background points.}

\item{sample_indices}{Integer vector: 0-based indices (into the background
array) of the occurrence sample locations.}

\item{features}{List of external pointers to Feature objects, as
returned by \code{maxent_linear_feature()}, \code{maxent_hinge_feature()},
\code{maxent_generate_features()}, etc.}

\item{bias_weights}{Optional numeric vector of per-point bias weights
(length \code{num_points}).  When supplied, background density is computed
as \code{bias[i] * exp(lp[i] - lpn)} instead of the standard
\code{exp(lp[i] - lpn)}.  This mirrors Java Maxent's \code{biasFile}
parameter.  Pass \code{NULL} (default) for uniform (unbiased) background.}
}
\value{
External pointer to a FeaturedSpace C++ object.
}
\description{
Constructs a MaxEnt FeaturedSpace from background point data, occurrence
sample indices, and a list of pre-built Feature objects.
}
\examples{
n   <- 100L
idx <- 90:99  # 0-based sample indices
env <- seq(0, 1, length.out = n)
f   <- maxent_linear_feature(env, "env1")
fs  <- maxent_featured_space(n, idx, list(f))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{grid_write_asc}
\alias{grid_write_asc}
\title{Write a float Grid to an ESRI ASCII (.asc) file}
\usage{
grid_write_asc(grid_ptr, filename, scientific = TRUE)
}
\arguments{
\item{grid_ptr}{External pointer to a GridFloat object.}

\item{filename}{Output file path.}

\item{scientific}{Logical: use scientific notation (default TRUE).}
}
\value{
Called for side effects; returns invisibly.
}
\description{
Write a float Grid to an ESRI ASCII (.asc) file
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/maxent_palette.R
\name{maxent_color_ramp}
\alias{maxent_color_ramp}
\title{Generate the Maxent Canonical Color Ramp}
\usage{
maxent_color_ramp(n = 1020L, mode = "plain")
}
\arguments{
\item{n}{Integer: number of colors to generate (default 1020, matching the
Java implementation).}

\item{mode}{Character: one of \code{"plain"} (default),
\code{"log"} (logarithmic spacing), \code{"blackandwhite"}
(greyscale, white = high, black = low), or \code{"redandyellow"}
(red–yellow ramp).}
}
\value{
Character vector of \code{n} hex color strings of the form
  \code{"#RRGGBB"}.
}
\description{
Produces a vector of hex color strings matching the color ramp used by the
Java Maxent software (\code{density/Display.java::showColor()}).  The
default palette cycles Red → Yellow → Green → Cyan → Blue across 1020
steps, with higher values rendered as red and lower values as blue.
}
\examples{
pal <- maxent_color_ramp(1020)
pal[1]     # "#FF0000"  (max value → red)
pal[510]   # near "#00FF00" (mid-point → green)
pal[1020]  # "#0000FF"  (min value → blue)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/maxent_plots.R
\name{maxent_plot_variable_importance}
\alias{maxent_plot_variable_importance}
\title{Write a Variable Importance Bar Chart PNG}
\usage{
maxent_plot_variable_importance(
  contributions_df,
  perm_imp_df,
  species,
  output_dir
)
}
\arguments{
\item{contributions_df}{A data.frame with columns \code{name} and
\code{contribution} (from \code{\link{maxent_percent_contribution}}).}

\item{perm_imp_df}{A data.frame with columns \code{name} and
\code{permutation_importance} (from
\code{\link{maxent_permutation_importance}}).}

\item{species}{Character: species name (used in the filename).}

\item{output_dir}{Character: directory under which a \code{plots/}
sub-directory will be created.}
}
\value{
Invisibly returns the path to the PNG file.
}
\description{
Creates a horizontal bar chart comparing percent contribution and
permutation importance for each environmental variable, replicating
\code{density/Runner.java::makeJackknifePlots()}.
}
\examples{
\donttest{
maxent_plot_variable_importance(contrib, perm_imp,
  species = "Sp1", output_dir = tempdir())
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_evaluation.R
\name{maxent_project_raw}
\alias{maxent_project_raw}
\title{Project Model onto Grids (Raw Output)}
\usage{
maxent_project_raw(model, env_grids, feature_names)
}
\arguments{
\item{model}{External pointer to a FeaturedSpace object.}

\item{env_grids}{List of external pointers to Grid<float> objects.}

\item{feature_names}{Character vector of environment variable names.}
}
\value{
External pointer to a Grid<float> with Java raw scores.
}
\description{
Produces Java Maxent "raw" scores for every grid cell:
\deqn{raw = \frac{exp(lp - lpNorm)}{densityNorm}}
}
\details{
This matches the raw output of the Java Maxent software and the
\pkg{dismo} R package.  See \code{vignette("PREDICTION_SCALES")} or
\code{docs/PREDICTION_SCALES.md} for a full description of prediction
scales.
}
\seealso{
\code{\link{maxent_project_cloglog}},
  \code{\link{maxent_project_logistic}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/maxent.R
\name{maxent_sample}
\alias{maxent_sample}
\title{Create a MaxEnt Sample Object}
\usage{
maxent_sample(lon, lat, name = "", dim = NULL)
}
\arguments{
\item{lon}{Longitude coordinate}

\item{lat}{Latitude coordinate}

\item{name}{Sample identifier/name}

\item{dim}{Optional GridDimension object to calculate row/col indices}
}
\value{
Sample object (external pointer)
}
\description{
Creates a sample point representing a species occurrence location.
}
\examples{
sample <- maxent_sample(lon = -118.5, lat = 36.5, name = "site1")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{coords_to_rowcol}
\alias{coords_to_rowcol}
\title{Convert geographic coordinates to row/col}
\usage{
coords_to_rowcol(dim_ptr, lon, lat)
}
\arguments{
\item{dim_ptr}{External pointer to GridDimension}

\item{lon}{Longitude}

\item{lat}{Latitude}
}
\value{
Integer vector with row and column indices
}
\description{
Convert geographic coordinates to row/col
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/featured_space.R
\name{maxent_extract_lambdas}
\alias{maxent_extract_lambdas}
\title{Extract Lambda Values from a Maxent Run Result}
\usage{
maxent_extract_lambdas(run_result)
}
\arguments{
\item{run_result}{A named list as returned by \code{\link{maxent_run}}.
Must contain an element \code{model} holding the trained FeaturedSpace
external pointer.}
}
\value{
A named numeric vector of lambda values, one per feature.  Names
  are the feature names as stored in the model.
}
\description{
Retrieves the trained lambda (weight) coefficients from the output of
\code{\link{maxent_run}}, returning them as a named numeric vector.
Each element is the lambda value for the corresponding feature, and the
name of each element is the feature name (e.g. \code{"bio1"},
\code{"bio1^2"}, \code{"bio1'1"}).
}
\details{
The function reads lambdas directly from the trained model object, so
it also works correctly after reloading coefficients with
\code{\link{maxent_load_lambdas}}.
}
\examples{
\donttest{
stack_path      <- system.file("extdata", "stack_1_12_crop.rds",
                               package = "maxentcpp")
example_rasters <- terra::unwrap(readRDS(stack_path))
grids <- list(
  bio1  = maxent_grid_from_terra(example_rasters[[1]]),
  bio12 = maxent_grid_from_terra(example_rasters[[2]])
)
data(example_occ_df)

result <- maxent_run(
  species    = "Abeillia_abeillei",
  env_grids  = grids,
  occ_df     = example_occ_df,
  output_dir = tempdir(),
  lon_col    = "long",
  lat_col    = "lat")

lambdas <- maxent_extract_lambdas(result)
print(lambdas)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/features.R
\name{maxent_product_feature}
\alias{maxent_product_feature}
\title{Create a Product Feature}
\usage{
maxent_product_feature(
  values1,
  values2,
  name,
  min1 = NULL,
  max1 = NULL,
  min2 = NULL,
  max2 = NULL
)
}
\arguments{
\item{values1}{Numeric vector for the first environmental variable.}

\item{values2}{Numeric vector for the second environmental variable
(must have the same length as \code{values1}).}

\item{name}{Character string: feature name/identifier.}

\item{min1}{Minimum of the first variable. If \code{NULL} (default),
computed as \code{min(values1)}.}

\item{max1}{Maximum of the first variable. If \code{NULL} (default),
computed as \code{max(values1)}.}

\item{min2}{Minimum of the second variable. If \code{NULL} (default),
computed as \code{min(values2)}.}

\item{max2}{Maximum of the second variable. If \code{NULL} (default),
computed as \code{max(values2)}.}
}
\value{
External pointer to a ProductFeature C++ object.
}
\description{
Creates an interaction feature between two environmental variables:
\code{eval(i) = norm(values1[i]) * norm(values2[i])}.
}
\examples{
temp <- c(0, 5, 10)
prec <- c(100, 200, 150)
f <- maxent_product_feature(temp, prec, "temp_x_prec")
maxent_feature_eval(f, 2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/maxent.R
\name{set_grid_matrix}
\alias{set_grid_matrix}
\title{Set Grid from Matrix}
\usage{
set_grid_matrix(grid, mat)
}
\arguments{
\item{grid}{Grid object}

\item{mat}{Numeric matrix of values}
}
\value{
Invisibly returns the grid object.
}
\description{
Populates grid with values from an R matrix.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{grid_read_file}
\alias{grid_read_file}
\title{Read a grid from file (auto-detect format)}
\usage{
grid_read_file(filename)
}
\arguments{
\item{filename}{Path to the grid file.}
}
\value{
External pointer to a GridFloat object.
}
\description{
Currently supports .asc format. Returns a float Grid.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_diagnostics.R
\name{maxent_response_curve}
\alias{maxent_response_curve}
\title{Compute Marginal Response Curve}
\usage{
maxent_response_curve(
  model,
  env_grids,
  feature_names,
  var_index,
  n_steps = 100L
)
}
\arguments{
\item{model}{External pointer to a FeaturedSpace object.}

\item{env_grids}{List of external pointers to Grid<float> objects.}

\item{feature_names}{Character vector of environment variable names.}

\item{var_index}{0-based index of the variable to vary.}

\item{n_steps}{Number of steps across the variable range (default 100).}
}
\value{
A data.frame with columns:
  \describe{
    \item{value}{Environmental variable value}
    \item{prediction}{Cloglog-transformed prediction}
  }
}
\description{
Generates a response curve by varying one environmental variable from its
minimum to maximum value while holding all other variables at their mean.
Applies the Java Maxent cloglog transform, matching the output of Java
Maxent and \pkg{dismo}:
\deqn{cloglog = 1 - exp(-exp(H) \cdot raw)}
}
\examples{
\donttest{
curve <- maxent_response_curve(model, list(g1, g2),
           c("temp", "precip"), var_index = 0)
plot(curve$value, curve$prediction, type = "l")
}
}
\seealso{
\code{\link{maxent_response_curve_fixed}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{grid_set_value}
\alias{grid_set_value}
\title{Set grid value}
\usage{
grid_set_value(grid_ptr, row, col, value)
}
\arguments{
\item{grid_ptr}{External pointer to Grid}

\item{row}{Row index}

\item{col}{Column index}

\item{value}{Value to set}
}
\value{
Called for side effects; returns invisibly.
}
\description{
Set grid value
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{csv_close}
\alias{csv_close}
\title{Close a CsvReader}
\usage{
csv_close(reader_ptr)
}
\arguments{
\item{reader_ptr}{External pointer to a CsvReader object.}
}
\value{
Called for side effects; returns invisibly.
}
\description{
Close a CsvReader
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grid_io.R
\name{maxent_grid_from_terra}
\alias{maxent_grid_from_terra}
\title{Convert a terra SpatRaster to a maxentcpp GridFloat}
\usage{
maxent_grid_from_terra(r, name = NULL)
}
\arguments{
\item{r}{A single-layer \code{terra::SpatRaster} object.}

\item{name}{Character: grid name (default: layer name from the raster).}
}
\value{
External pointer to a GridFloat C++ object.
}
\description{
Converts a single-layer \code{terra::SpatRaster} into a maxentcpp
GridFloat external pointer. The raster must have exactly one layer and
square cells (equal x and y resolution).
}
\details{
\code{NA} values in the raster are stored as NODATA (sentinel \code{-9999}).

Requires the \pkg{terra} package (listed in Suggests).

For the \pkg{raster} package, an equivalent workflow is:
\preformatted{
  library(raster)
  r <- raster("bio1.tif")
  mat <- as.matrix(r)
  e <- extent(r)
  g <- maxent_grid_from_matrix(mat, xll = e@xmin, yll = e@ymin,
                               cellsize = res(r)[1],
                               name = names(r))
}
}
\examples{
\donttest{
stack_path <- system.file("extdata", "stack_1_12_crop.rds",
                         package = "maxentcpp")
r <- terra::unwrap(readRDS(stack_path))
g <- maxent_grid_from_terra(r[[1]])
maxent_grid_info(g)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_diagnostics.R
\name{maxent_response_curve_fixed}
\alias{maxent_response_curve_fixed}
\title{Compute Response Curve with Fixed Values}
\usage{
maxent_response_curve_fixed(
  model,
  fixed_values,
  feature_names,
  var_index,
  var_min,
  var_max,
  n_steps = 100L
)
}
\arguments{
\item{model}{External pointer to a FeaturedSpace object.}

\item{fixed_values}{Numeric vector of fixed values for each variable.}

\item{feature_names}{Character vector of environment variable names.}

\item{var_index}{0-based index of the variable to vary.}

\item{var_min}{Minimum value of the target variable.}

\item{var_max}{Maximum value of the target variable.}

\item{n_steps}{Number of steps (default 100).}
}
\value{
A data.frame with columns: value, prediction.
}
\description{
Like \code{maxent_response_curve} but the user supplies explicit fixed
values for the non-target variables. Applies the Java Maxent cloglog
transform, matching the output of Java Maxent and \pkg{dismo}:
\deqn{cloglog = 1 - exp(-exp(H) \cdot raw)}
}
\seealso{
\code{\link{maxent_response_curve}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_evaluation.R
\name{maxent_auc}
\alias{maxent_auc}
\title{Compute AUC (Area Under the ROC Curve)}
\usage{
maxent_auc(presence, absence)
}
\arguments{
\item{presence}{Numeric vector of prediction scores at presence sites.}

\item{absence}{Numeric vector of prediction scores at absence sites.}
}
\value{
A named list with elements:
  \describe{
    \item{auc}{AUC value in [0, 1]}
    \item{max_kappa}{Best Cohen's Kappa found}
    \item{max_kappa_thresh}{Threshold at best Kappa}
  }
}
\description{
Computes the Wilcoxon-Mann-Whitney AUC statistic from prediction scores
at presence and absence sites. Also returns max-Kappa and its threshold.
}
\examples{
result <- maxent_auc(c(0.8, 0.9, 1.0), c(0.1, 0.2, 0.3))
result$auc  # 1.0
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/maxent.R
\name{maxent_grid}
\alias{maxent_grid}
\title{Create a Grid Object}
\usage{
maxent_grid(dim, name = "", nodata_value = -9999)
}
\arguments{
\item{dim}{GridDimension object defining spatial extent}

\item{name}{Grid layer name}

\item{nodata_value}{Value representing missing data (default: -9999)}
}
\value{
Grid object (external pointer)
}
\description{
Creates a raster grid for storing environmental variable data.
}
\examples{
dim <- maxent_dimension(100, 100, -120, 35, 0.1)
grid <- maxent_grid(dim, "temperature")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{extract_predictions_raw}
\alias{extract_predictions_raw}
\title{Extract Java-compatible Raw Predictions at Sample Locations}
\usage{
extract_predictions_raw(fs_ptr, grid_ptrs, feature_names, rows, cols)
}
\arguments{
\item{fs_ptr}{External pointer to a FeaturedSpace object.}

\item{grid_ptrs}{List of external pointers to Grid<float> objects.}

\item{feature_names}{Character vector of environment variable names.}

\item{rows}{Integer vector of row indices.}

\item{cols}{Integer vector of column indices.}
}
\value{
Numeric vector of Java raw scores. NaN for NODATA cells.
}
\description{
Gets Java Maxent raw scores (raw_java = exp(lp - lpNorm) / densityNorm)
at specific grid cell locations.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grid_io.R
\name{maxent_grid_from_matrix}
\alias{maxent_grid_from_matrix}
\title{Create a Grid from an R Matrix}
\usage{
maxent_grid_from_matrix(mat, xll, yll, cellsize, nodata = -9999, name = "")
}
\arguments{
\item{mat}{Numeric matrix.}

\item{xll}{X coordinate of lower-left corner.}

\item{yll}{Y coordinate of lower-left corner.}

\item{cellsize}{Cell size.}

\item{nodata}{NODATA sentinel value (default \code{-9999}).}

\item{name}{Grid name (default \code{""}).}
}
\value{
External pointer to a GridFloat C++ object.
}
\description{
Builds a GridFloat from a numeric matrix and spatial parameters.
\code{NA} values in the matrix are stored as the \code{nodata} value.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{create_grid_dimension}
\alias{create_grid_dimension}
\title{Create a GridDimension object}
\usage{
create_grid_dimension(nrows, ncols, xll, yll, cellsize)
}
\arguments{
\item{nrows}{Number of rows}

\item{ncols}{Number of columns}

\item{xll}{X coordinate of lower-left corner}

\item{yll}{Y coordinate of lower-left corner}

\item{cellsize}{Cell size (assumed square cells)}
}
\value{
External pointer to GridDimension object
}
\description{
Creates a grid dimension specifying the spatial extent and resolution
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/features.R
\name{maxent_generate_features}
\alias{maxent_generate_features}
\title{Generate Features from Environmental Variable Data}
\usage{
maxent_generate_features(
  data,
  types = c("linear", "quadratic", "product", "threshold", "hinge"),
  n_thresholds = 10L,
  n_hinges = 10L
)
}
\arguments{
\item{data}{Named list of numeric vectors, one per environmental variable.}

\item{types}{Character vector of feature types to generate. Valid values:
\code{"linear"}, \code{"quadratic"}, \code{"product"},
\code{"threshold"}, \code{"hinge"}. Defaults to all types.}

\item{n_thresholds}{Integer; number of threshold knots per variable
(default: 10).}

\item{n_hinges}{Integer; number of hinge knots per variable (default: 10).}
}
\value{
Named list of external pointers to Feature C++ objects.
}
\description{
Automatically generates all configured feature types from one or more
environmental variable vectors.
}
\examples{
env_data <- list(
  temperature   = c(15, 20, 25, 18, 22),
  precipitation = c(100, 200, 150, 80, 300)
)
features <- maxent_generate_features(env_data, types = c("linear", "hinge"))
length(features)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/features.R
\name{maxent_linear_feature}
\alias{maxent_linear_feature}
\title{Create a Linear Feature}
\usage{
maxent_linear_feature(values, name, min_val = NULL, max_val = NULL)
}
\arguments{
\item{values}{Numeric vector of environmental variable values.}

\item{name}{Character string: feature name/identifier.}

\item{min_val}{Minimum value for normalization. If \code{NULL} (default),
computed as \code{min(values)}.}

\item{max_val}{Maximum value for normalization. If \code{NULL} (default),
computed as \code{max(values)}.}
}
\value{
External pointer to a LinearFeature C++ object.
}
\description{
Creates a linear (normalized) feature that evaluates to
\code{(values[i] - min) / (max - min)}. Returns 0 when \code{min == max}.
}
\examples{
vals <- c(0, 5, 10, 3)
f <- maxent_linear_feature(vals, "temperature")
maxent_feature_eval(f, 1)  # index 1 (R) -> 0-based index 0
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{maxent_featured_space_create}
\alias{maxent_featured_space_create}
\title{Create a FeaturedSpace object}
\usage{
maxent_featured_space_create(
  num_points,
  sample_indices,
  feature_ptrs,
  bias_weights = numeric(0)
)
}
\arguments{
\item{num_points}{Integer: number of background points.}

\item{sample_indices}{Integer vector: 0-based indices of occurrence samples
in the background array.}

\item{feature_ptrs}{List of external pointers to Feature objects
(from \code{create_linear_feature()} etc.).}

\item{bias_weights}{Optional numeric vector of per-point bias weights
(length \code{num_points}).  When supplied, background density is
computed as \code{bias[i] * exp(lp[i] - lpn)} instead of the standard
\code{exp(lp[i] - lpn)}.  Pass an empty vector (default) for uniform
(unbiased) background.}
}
\value{
External pointer to a FeaturedSpace object.
}
\description{
Constructs a MaxEnt FeaturedSpace from background point count,
occurrence sample indices, and a list of Feature objects.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{project_raw}
\alias{project_raw}
\title{Project Model onto Grids (Java-compatible raw output)}
\usage{
project_raw(fs_ptr, grid_ptrs, feature_names)
}
\arguments{
\item{fs_ptr}{External pointer to a FeaturedSpace object.}

\item{grid_ptrs}{List of external pointers to Grid<float> objects.}

\item{feature_names}{Character vector of environment variable names.}
}
\value{
External pointer to a Grid<float> with Java raw scores.
}
\description{
Applies a trained FeaturedSpace model to produce Java Maxent "raw" scores:
  raw_java = exp(lp - lpNorm) / densityNorm
}
\details{
This matches the raw output of the Java Maxent software and the dismo
R package.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_diagnostics.R
\name{maxent_variable_ranges}
\alias{maxent_variable_ranges}
\title{Compute Variable Ranges from Grids}
\usage{
maxent_variable_ranges(env_grids)
}
\arguments{
\item{env_grids}{List of external pointers to Grid<float> objects.}
}
\value{
A data.frame with columns: min, max (one row per variable).
}
\description{
Scans all valid cells in the grids to determine the [min, max] range
of each environmental variable.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{layer_name_from_path}
\alias{layer_name_from_path}
\title{Extract a layer name from a file path}
\usage{
layer_name_from_path(path)
}
\arguments{
\item{path}{File path.}
}
\value{
Character: layer name.
}
\description{
Strips directory and extension, e.g. "/data/bio1.asc" → "bio1".
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{project_cloglog}
\alias{project_cloglog}
\title{Project Model onto Grids (Java-compatible cloglog output)}
\usage{
project_cloglog(fs_ptr, grid_ptrs, feature_names)
}
\arguments{
\item{fs_ptr}{External pointer to a FeaturedSpace object.}

\item{grid_ptrs}{List of external pointers to Grid<float> objects.}

\item{feature_names}{Character vector of environment variable names.}
}
\value{
External pointer to a Grid<float> with Java cloglog scores in [0, 1].
}
\description{
Applies the Java Maxent cloglog formula:
  cloglog_java = 1 - exp(-exp(H) * raw_java)
}
\details{
where H is the model entropy and raw_java is the normalised probability.
This matches the cloglog output of Java Maxent and dismo.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{feature_get_info}
\alias{feature_get_info}
\title{Get feature metadata}
\usage{
feature_get_info(feature_ptr)
}
\arguments{
\item{feature_ptr}{External pointer to a Feature object}
}
\value{
Named list with feature properties
}
\description{
Returns a list with name, type, lambda, min, and max values.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grid_io.R
\name{maxent_csv_close}
\alias{maxent_csv_close}
\title{Close a CSV Reader}
\usage{
maxent_csv_close(reader)
}
\arguments{
\item{reader}{External pointer to a CsvReader object.}
}
\value{
Invisibly returns \code{NULL}.
}
\description{
Close a CSV Reader
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/maxent_outputs.R
\name{maxent_write_prediction_png}
\alias{maxent_write_prediction_png}
\title{Write a Maxent Prediction Grid as a PNG Image}
\usage{
maxent_write_prediction_png(
  grid,
  filename,
  presence_rows = NULL,
  presence_cols = NULL,
  test_rows = NULL,
  test_cols = NULL,
  mode = "plain",
  legend = TRUE,
  width = 800L,
  height = 600L
)
}
\arguments{
\item{grid}{External pointer to a GridFloat prediction grid (e.g.
from \code{\link{maxent_project_cloglog}}).}

\item{filename}{Character: path for the output PNG file.}

\item{presence_rows}{Integer vector of presence row indices (0-based) or
\code{NULL} (default).}

\item{presence_cols}{Integer vector of presence column indices (0-based) or
\code{NULL} (default).}

\item{test_rows}{Integer vector of test-set row indices (0-based) or
\code{NULL} (default).}

\item{test_cols}{Integer vector of test-set column indices (0-based) or
\code{NULL} (default).}

\item{mode}{Colour mode passed to \code{\link{maxent_color_ramp}}:
one of \code{"plain"} (default), \code{"log"}, \code{"blackandwhite"},
or \code{"redandyellow"}.}

\item{legend}{Logical: draw a colour-bar legend (default \code{TRUE}).}

\item{width}{Integer: PNG width in pixels (default 800).}

\item{height}{Integer: PNG height in pixels (default 600).}
}
\value{
Invisibly returns \code{filename}.
}
\description{
Converts a \code{GridFloat} prediction grid to a colour PNG image using
the canonical Maxent colour ramp (red = high, blue = low).  Optionally
overlays presence and test-point locations, and renders a small legend.
}
\examples{
\donttest{
pred <- maxent_project_cloglog(model, list(g1, g2), c("bio1", "bio12"))
maxent_write_prediction_png(pred, tempfile(fileext = ".png"))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_diagnostics.R
\name{maxent_percent_contribution}
\alias{maxent_percent_contribution}
\title{Compute Percent Contribution}
\usage{
maxent_percent_contribution(model, feature_names)
}
\arguments{
\item{model}{External pointer to a FeaturedSpace object.}

\item{feature_names}{Character vector of base variable names.}
}
\value{
A data.frame with columns:
  \describe{
    \item{name}{Variable name}
    \item{contribution}{Percent contribution}
  }
}
\description{
Computes variable contribution based on sum of absolute lambda values
for features derived from each variable. Results sum to 100%.
}
\examples{
n   <- 50L
idx <- c(5L, 15L, 25L, 35L, 45L)
env <- list(temp = runif(n), precip = runif(n))
feats <- maxent_generate_features(env, types = "linear")
model <- maxent_featured_space(n, idx, feats)
maxent_fit(model, max_iter = 100, convergence = 1e-3)
contrib <- maxent_percent_contribution(model, c("temp", "precip"))
contrib  # data.frame with name and contribution
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{project_logistic}
\alias{project_logistic}
\title{Project Model onto Grids (Java-compatible logistic output)}
\usage{
project_logistic(fs_ptr, grid_ptrs, feature_names)
}
\arguments{
\item{fs_ptr}{External pointer to a FeaturedSpace object.}

\item{grid_ptrs}{List of external pointers to Grid<float> objects.}

\item{feature_names}{Character vector of environment variable names.}
}
\value{
External pointer to a Grid<float> with Java logistic scores in [0, 1].
}
\description{
Applies the Java Maxent logistic formula:
  logistic_java = (exp(H) * raw_java) / (1 + exp(H) * raw_java)
}
\details{
where H is the model entropy and raw_java is the normalised probability.
This matches the logistic output of Java Maxent and dismo.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/features.R
\name{maxent_threshold_feature}
\alias{maxent_threshold_feature}
\title{Create a Threshold Feature}
\usage{
maxent_threshold_feature(values, name, threshold)
}
\arguments{
\item{values}{Numeric vector of environmental variable values.}

\item{name}{Character string: feature name/identifier.}

\item{threshold}{The threshold value.}
}
\value{
External pointer to a ThresholdFeature C++ object.
}
\description{
Creates a binary step feature: \code{eval(i) = 1} if
\code{values[i] > threshold}, else \code{0}.
}
\examples{
vals <- c(1, 5, 10, 3)
f <- maxent_threshold_feature(vals, "temperature_threshold", threshold = 5)
maxent_feature_eval(f, 3)  # values[3] = 10 > 5 -> 1
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/maxent.R
\name{maxent_dimension}
\alias{maxent_dimension}
\title{Create a Grid Dimension Object}
\usage{
maxent_dimension(nrows, ncols, xll, yll, cellsize)
}
\arguments{
\item{nrows}{Number of rows}

\item{ncols}{Number of columns}

\item{xll}{X coordinate of lower-left corner}

\item{yll}{Y coordinate of lower-left corner}

\item{cellsize}{Cell size (square cells)}
}
\value{
GridDimension object (external pointer)
}
\description{
Defines the spatial extent and resolution of a raster grid.
}
\examples{
dim <- maxent_dimension(
  nrows = 100, ncols = 100,
  xll = -120.0, yll = 35.0,
  cellsize = 0.1
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{create_grid_float}
\alias{create_grid_float}
\title{Create a float Grid object}
\usage{
create_grid_float(dim_ptr, name, nodata_value = -9999)
}
\arguments{
\item{dim_ptr}{External pointer to GridDimension}

\item{name}{Grid layer name}

\item{nodata_value}{Value representing missing data}
}
\value{
External pointer to Grid<float> object
}
\description{
Creates a raster grid for environmental variables
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{clamp_grids}
\alias{clamp_grids}
\title{Clamp Environmental Grids}
\usage{
clamp_grids(grid_ptrs, var_mins, var_maxs)
}
\arguments{
\item{grid_ptrs}{List of external pointers to Grid<float> objects.}

\item{var_mins}{Numeric vector of minimum training values per variable.}

\item{var_maxs}{Numeric vector of maximum training values per variable.}
}
\value{
A named list with:
  \describe{
    \item{clamped_grids}{List of external pointers to clamped Grid<float>}
    \item{clamp_grid}{External pointer to Grid<float> with clamping magnitudes}
  }
}
\description{
Restricts environmental variable values to the training range.
Returns clamped grids and a clamping indicator grid.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{maxent_featured_space_from_spatraster_callback}
\alias{maxent_featured_space_from_spatraster_callback}
\title{Create a FeaturedSpace from a terra SpatRaster callback stream}
\usage{
maxent_featured_space_from_spatraster_callback(
  preserved_rast,
  rast_xptr,
  num_points,
  num_layers,
  layer_names,
  sample_indices,
  feature_types,
  n_thresholds,
  n_hinges,
  next_tile_fn,
  reset_fn,
  enable_streaming_eval = TRUE,
  use_cache = TRUE
)
}
\arguments{
\item{preserved_rast}{A terra \code{SpatRaster} S4 object to preserve.}

\item{rast_xptr}{External pointer extracted from the raster object
(used for contract validation).}

\item{num_points}{Integer; total number of finite background points.}

\item{num_layers}{Integer; number of environmental variables per row.}

\item{layer_names}{Character vector of layer names (length num_layers).}

\item{sample_indices}{Integer vector; 0-based indices of presence samples.}

\item{feature_types}{Character vector of feature types to generate.}

\item{n_thresholds}{Integer; number of threshold knots per variable.}

\item{n_hinges}{Integer; number of hinge knots per variable.}

\item{next_tile_fn}{R function returning the next tile or NULL/0-row matrix.}

\item{reset_fn}{R function (no args) that rewinds the underlying stream.}

\item{enable_streaming_eval}{Logical; whether to enable streaming eval mode.}

\item{use_cache}{Logical; wrap the stream in \code{CachingBackgroundProvider}.}
}
\value{
External pointer to a FeaturedSpace object.
}
\description{
Same as \code{maxent_featured_space_from_callback()}, but also preserves
the source \code{SpatRaster} S4 object for provider lifetime safety and
can enable streaming-eval mode immediately.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grid_io.R
\name{maxent_csv_write_close}
\alias{maxent_csv_write_close}
\title{Close a CSV Writer}
\usage{
maxent_csv_write_close(writer)
}
\arguments{
\item{writer}{External pointer to a CsvWriter object.}
}
\value{
Invisibly returns \code{NULL}.
}
\description{
Flushes any pending data and closes the CSV file.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/maxent_outputs.R
\name{maxent_append_results_csv}
\alias{maxent_append_results_csv}
\title{Append a Row to maxentResults.csv}
\usage{
maxent_append_results_csv(
  results_file,
  species,
  n_training,
  n_test = 0L,
  training_gain,
  training_auc,
  test_gain = NA_real_,
  test_auc = NA_real_,
  entropy,
  contributions_df,
  perm_imp_df
)
}
\arguments{
\item{results_file}{Character: path to the results CSV file.}

\item{species}{Character: species name.}

\item{n_training}{Integer: number of training presence points.}

\item{n_test}{Integer: number of test presence points (default 0).}

\item{training_gain}{Numeric: regularized training gain.}

\item{training_auc}{Numeric: training AUC.}

\item{test_gain}{Numeric or \code{NA}: regularized test gain.}

\item{test_auc}{Numeric or \code{NA}: test AUC.}

\item{entropy}{Numeric: model entropy.}

\item{contributions_df}{A data.frame with columns \code{name} and
\code{contribution} (from \code{\link{maxent_percent_contribution}}).}

\item{perm_imp_df}{A data.frame with columns \code{name} and
\code{permutation_importance} (from
\code{\link{maxent_permutation_importance}}).}
}
\value{
Invisibly returns \code{results_file}.
}
\description{
Appends one row of species-level summary statistics to a
\code{maxentResults.csv} file, creating the file (with a header) if it does
not yet exist.  Column names match the Java Maxent output for
cross-tool compatibility.
}
\examples{
\donttest{
maxent_append_results_csv(
  file.path(tempdir(), "maxentResults.csv"),
  species = "Sp1", n_training = 50L, training_gain = 1.23,
  training_auc = 0.95, entropy = 6.7,
  contributions_df = contrib, perm_imp_df = perm_imp)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{create_layer}
\alias{create_layer}
\title{Create a Layer metadata object}
\usage{
create_layer(name, type_str)
}
\arguments{
\item{name}{Layer name.}

\item{type_str}{Type string: "Continuous", "Categorical", "Bias",
"Mask", "Probability", "Cumulative", "DebiasAvg", or "Unknown".}
}
\value{
External pointer to a Layer object.
}
\description{
Create a Layer metadata object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grid_io.R
\name{maxent_layer_info}
\alias{maxent_layer_info}
\title{Get Layer Metadata}
\usage{
maxent_layer_info(layer)
}
\arguments{
\item{layer}{External pointer to a Layer object.}
}
\value{
Named list with \code{name} and \code{type}.
}
\description{
Get Layer Metadata
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{csv_writer_print_double}
\alias{csv_writer_print_double}
\title{Write a numeric value into the current row of a CSV}
\usage{
csv_writer_print_double(writer_ptr, column, value)
}
\arguments{
\item{writer_ptr}{External pointer to a CsvWriter object.}

\item{column}{Column name.}

\item{value}{Numeric value.}
}
\value{
Called for side effects; returns invisibly.
}
\description{
Write a numeric value into the current row of a CSV
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/maxent_plots.R
\name{maxent_plot_response_curves}
\alias{maxent_plot_response_curves}
\title{Write Response Curve PNGs for All Variables}
\usage{
maxent_plot_response_curves(
  model,
  env_grids,
  feature_names,
  output_dir,
  species,
  var_indices = NULL,
  n_steps = 100L,
  thumbnail = TRUE,
  write_dat = FALSE
)
}
\arguments{
\item{model}{External pointer to a trained FeaturedSpace object.}

\item{env_grids}{List of external pointers to Grid<float> objects.}

\item{feature_names}{Character vector of environment variable names
(one entry per element of \code{env_grids}).}

\item{output_dir}{Character: directory under which a \code{plots/}
sub-directory will be created.}

\item{species}{Character: species name (used in file names).}

\item{var_indices}{Integer vector of 0-based variable indices to plot.
Defaults to all variables.}

\item{n_steps}{Integer: number of steps in each curve (default 100).}

\item{thumbnail}{Logical: also write a 210 × 140 pixel thumbnail PNG
(default \code{TRUE}).}

\item{write_dat}{Logical: also write a tab-delimited \code{.dat} file
of the curve data (default \code{FALSE}).}
}
\value{
Invisibly returns a named list of file paths written.
}
\description{
Generates response curve plots for each environmental variable and saves
them as PNG files, replicating \code{density/ResponsePlot.java::makeplot()}.
A full-size PNG and an optional thumbnail are written for every variable.
}
\examples{
\donttest{
maxent_plot_response_curves(
  model, list(g1, g2), c("bio1", "bio12"),
  output_dir = tempdir(), species = "Sp1")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{grid_to_matrix}
\alias{grid_to_matrix}
\title{Get grid as matrix}
\usage{
grid_to_matrix(grid_ptr)
}
\arguments{
\item{grid_ptr}{External pointer to Grid}
}
\value{
Numeric matrix of grid values
}
\description{
Get grid as matrix
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{maxent_train}
\alias{maxent_train}
\title{Train a FeaturedSpace model}
\usage{
maxent_train(
  fs_ptr,
  max_iter = 500L,
  convergence = 1e-05,
  beta_multiplier = 1,
  min_deviation = 0.001
)
}
\arguments{
\item{fs_ptr}{External pointer to a FeaturedSpace object.}

\item{max_iter}{Maximum number of training iterations (default 500).}

\item{convergence}{Convergence threshold (default 1e-5).}

\item{beta_multiplier}{Regularization multiplier (default 1.0).}

\item{min_deviation}{Minimum sample deviation floor (default 0.001).}
}
\value{
Named list with elements: \code{loss}, \code{entropy},
  \code{iterations}, \code{converged}, \code{lambdas}.
}
\description{
Runs the sequential coordinate-ascent MaxEnt optimization.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grid_io.R
\name{maxent_layer}
\alias{maxent_layer}
\title{Create a Layer Metadata Object}
\usage{
maxent_layer(name, type = "Continuous")
}
\arguments{
\item{name}{Character: layer name.}

\item{type}{Character: layer type. One of \code{"Continuous"},
\code{"Categorical"}, \code{"Bias"}, \code{"Mask"},
\code{"Probability"}, \code{"Cumulative"}, \code{"DebiasAvg"},
or \code{"Unknown"}.}
}
\value{
External pointer to a Layer C++ object.
}
\description{
Create a Layer Metadata Object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{csv_writer_println}
\alias{csv_writer_println}
\title{Flush the current CSV row to disk}
\usage{
csv_writer_println(writer_ptr)
}
\arguments{
\item{writer_ptr}{External pointer to a CsvWriter object.}
}
\value{
Called for side effects; returns invisibly.
}
\description{
Flush the current CSV row to disk
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/featured_space.R
\name{maxent_fit}
\alias{maxent_fit}
\title{Train a MaxEnt Model}
\usage{
maxent_fit(
  featured_space,
  max_iter = 500L,
  convergence = 1e-05,
  beta_multiplier = 1,
  min_deviation = 0.001
)
}
\arguments{
\item{featured_space}{External pointer to a FeaturedSpace object (from
\code{maxent_featured_space()}).}

\item{max_iter}{Maximum number of training iterations (default 500).}

\item{convergence}{Convergence threshold: stop when the per-20-iteration
loss improvement is below this value (default 1e-5).}

\item{beta_multiplier}{Regularization multiplier (default 1.0). Higher
values increase regularization strength.}

\item{min_deviation}{Minimum sample deviation floor used in regularization
(default 0.001).}
}
\value{
Named list with:
  \describe{
    \item{loss}{Final regularized loss (scalar).}
    \item{entropy}{Shannon entropy of the trained distribution.}
    \item{iterations}{Number of training iterations completed.}
    \item{converged}{Logical: whether the convergence threshold was reached.}
    \item{lambdas}{Numeric vector of final lambda (weight) values.}
  }
}
\description{
Runs the sequential coordinate-ascent MaxEnt optimization on a FeaturedSpace.
}
\examples{
n   <- 50L
idx <- c(5L, 15L, 25L, 35L, 45L)
env <- list(bio1 = runif(n), bio12 = runif(n))
feats <- maxent_generate_features(env, types = "linear")
fs  <- maxent_featured_space(n, idx, feats)
result <- maxent_fit(fs, max_iter = 100, convergence = 1e-5)
result$loss
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{maxent_write_lambdas}
\alias{maxent_write_lambdas}
\title{Write feature lambdas to a file}
\usage{
maxent_write_lambdas(fs_ptr, filename)
}
\arguments{
\item{fs_ptr}{External pointer to a trained FeaturedSpace object.}

\item{filename}{Character: path to the output file.}
}
\value{
Called for side effects; returns invisibly.
}
\description{
Saves the trained model coefficients in CSV format compatible with
the original Java Maxent .lambdas file format.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{get_layer_info}
\alias{get_layer_info}
\title{Get layer metadata}
\usage{
get_layer_info(layer_ptr)
}
\arguments{
\item{layer_ptr}{External pointer to a Layer object.}
}
\value{
Named list with name and type (string).
}
\description{
Get layer metadata
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{generate_features}
\alias{generate_features}
\title{Generate features from a list of environmental variable vectors}
\usage{
generate_features(
  data_list,
  feature_types = as.character(c("linear", "quadratic", "product", "threshold", "hinge")),
  n_thresholds = 10L,
  n_hinges = 10L
)
}
\arguments{
\item{data_list}{Named list of numeric vectors, one per environmental variable}

\item{feature_types}{Character vector of feature types to generate.
Valid values: "linear", "quadratic", "product", "threshold", "hinge".
Defaults to all types.}

\item{n_thresholds}{Number of threshold knots per variable (default: 10)}

\item{n_hinges}{Number of hinge knots per variable (default: 10)}
}
\value{
List of external pointers to Feature objects
}
\description{
Generates all configured feature types (linear, quadratic, product,
threshold, hinge) from the supplied data vectors.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grid_io.R
\name{maxent_csv_write_open}
\alias{maxent_csv_write_open}
\title{Open a CSV File for Writing}
\usage{
maxent_csv_write_open(filename, append = FALSE, precision = 4L)
}
\arguments{
\item{filename}{Character: output file path.}

\item{append}{Logical: append to an existing file (default \code{FALSE}).}

\item{precision}{Integer: number of decimal places for numeric values
(default \code{4}).}
}
\value{
External pointer to a CsvWriter C++ object.
}
\description{
Opens a CSV file and returns a writer object. Use \code{\link{maxent_csv_write}},
\code{\link{maxent_csv_write_num}}, \code{\link{maxent_csv_write_row}}, and
\code{\link{maxent_csv_write_close}} to write data and close the file.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{maxent_set_sample_expectations}
\alias{maxent_set_sample_expectations}
\title{Set sample expectations for a FeaturedSpace}
\usage{
maxent_set_sample_expectations(
  fs_ptr,
  beta_multiplier = 1,
  min_deviation = 0.001
)
}
\arguments{
\item{fs_ptr}{External pointer to a FeaturedSpace object.}

\item{beta_multiplier}{Regularization multiplier (default 1.0).}

\item{min_deviation}{Minimum sample deviation (default 0.001).}
}
\value{
Called for side effects; returns invisibly.
}
\description{
Computes sample expectations and regularization deviations for each feature.
Called automatically by \code{maxent_train()}, but exposed for advanced use.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{grid_float_from_matrix}
\alias{grid_float_from_matrix}
\title{Create a GridFloat from an R matrix}
\usage{
grid_float_from_matrix(mat, xll, yll, cellsize, nodata = -9999, name = "")
}
\arguments{
\item{mat}{Numeric matrix with grid data (NA becomes NODATA).}

\item{xll}{X coordinate of lower-left corner.}

\item{yll}{Y coordinate of lower-left corner.}

\item{cellsize}{Cell size.}

\item{nodata}{NODATA value (default -9999).}

\item{name}{Grid name (default "").}
}
\value{
External pointer to a GridFloat object.
}
\description{
Create a GridFloat from an R matrix
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{maxent_set_streaming_eval}
\alias{maxent_set_streaming_eval}
\title{Enable or disable streaming-eval mode on a FeaturedSpace}
\usage{
maxent_set_streaming_eval(fs_ptr, enabled = TRUE)
}
\arguments{
\item{fs_ptr}{External pointer to a FeaturedSpace object.}

\item{enabled}{Logical flag.}
}
\value{
Called for side effects; returns invisibly.
}
\description{
Enable or disable streaming-eval mode on a FeaturedSpace
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{create_hinge_feature}
\alias{create_hinge_feature}
\title{Create a HingeFeature object}
\usage{
create_hinge_feature(values, name, min_knot, max_knot, is_reverse = FALSE)
}
\arguments{
\item{values}{Numeric vector of environmental variable values}

\item{name}{Feature name/identifier}

\item{min_knot}{Lower knot of the hinge}

\item{max_knot}{Upper knot of the hinge (must be > min_knot)}

\item{is_reverse}{If TRUE, use reverse hinge; if FALSE (default), use forward hinge}
}
\value{
External pointer to HingeFeature object
}
\description{
Creates a piecewise-linear hinge feature.
Forward hinge: eval(i) = (values[i] > min_knot) ? (values[i] - min_knot) / (max_knot - min_knot) : 0.
Reverse hinge: eval(i) = (values[i] < max_knot) ? (max_knot - values[i]) / (max_knot - min_knot) : 0.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grid_io.R
\name{maxent_write_asc}
\alias{maxent_write_asc}
\title{Write a Grid to ESRI ASCII Format}
\usage{
maxent_write_asc(grid, filename, scientific = TRUE)
}
\arguments{
\item{grid}{External pointer to a GridFloat object.}

\item{filename}{Character: output file path.}

\item{scientific}{Logical: use scientific notation for floating-point
values (default \code{TRUE}).}
}
\value{
Invisibly returns the output file path.
}
\description{
Writes a GridFloat to a \code{.asc} file.
}
\examples{
\donttest{
maxent_write_asc(g, tempfile(fileext = ".asc"))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grid_io.R
\name{maxent_csv_open}
\alias{maxent_csv_open}
\title{Open a CSV File for Reading}
\usage{
maxent_csv_open(filename, has_header = TRUE)
}
\arguments{
\item{filename}{Character: path to the CSV file.}

\item{has_header}{Logical: first line is header (default \code{TRUE}).}
}
\value{
External pointer to a CsvReader C++ object.
}
\description{
Opens a CSV file and reads column headers.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{grid_get_values_batch}
\alias{grid_get_values_batch}
\title{Extract multiple values from a Grid at given row/col indices (batch)}
\usage{
grid_get_values_batch(grid_ptr, rows, cols)
}
\arguments{
\item{grid_ptr}{External pointer to a Grid<float>.}

\item{rows}{Integer vector of 0-based row indices.}

\item{cols}{Integer vector of 0-based column indices.}
}
\value{
Numeric vector of extracted values (NA for NODATA cells).
}
\description{
Extract multiple values from a Grid at given row/col indices (batch)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/featured_space.R
\name{maxent_save_lambdas}
\alias{maxent_save_lambdas}
\title{Save Model Lambdas to File}
\usage{
maxent_save_lambdas(featured_space, file)
}
\arguments{
\item{featured_space}{External pointer to a trained FeaturedSpace object.}

\item{file}{Character: path to the output file.}
}
\value{
Invisibly returns the output file path.
}
\description{
Writes the trained model coefficients (lambdas) to a CSV file in
the standard Maxent .lambdas format.
}
\examples{
\donttest{
maxent_save_lambdas(fs, tempfile(fileext = ".lambdas"))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_evaluation.R
\name{maxent_square_error}
\alias{maxent_square_error}
\title{Compute Mean Squared Error}
\usage{
maxent_square_error(presence, absence)
}
\arguments{
\item{presence}{Numeric vector of prediction scores at presence sites.}

\item{absence}{Numeric vector of prediction scores at absence sites.}
}
\value{
Mean squared error.
}
\description{
Presence sites contribute (1 - pred)^2, absence sites contribute pred^2.
}
\examples{
maxent_square_error(c(1.0, 1.0), c(0.0, 0.0))  # 0.0
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{eval_square_error}
\alias{eval_square_error}
\title{Compute Mean Squared Error}
\usage{
eval_square_error(presence, absence)
}
\arguments{
\item{presence}{Numeric vector of prediction scores at presence sites.}

\item{absence}{Numeric vector of prediction scores at absence sites.}
}
\value{
Mean squared error.
}
\description{
Presence sites contribute (1-pred)^2, absence sites contribute pred^2.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{grid_float_info}
\alias{grid_float_info}
\title{Get grid information for an ASC-loaded grid}
\usage{
grid_float_info(grid_ptr)
}
\arguments{
\item{grid_ptr}{External pointer to a GridFloat object.}
}
\value{
Named list with nrows, ncols, xll, yll, cellsize, nodata,
  name, count_data.
}
\description{
Returns metadata and basic statistics for a GridFloat.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{maxent_sequential_train}
\alias{maxent_sequential_train}
\title{Train a FeaturedSpace with the Sequential optimizer (trajectory-capable)}
\usage{
maxent_sequential_train(
  fs_ptr,
  max_iter = 500L,
  convergence = 1e-05,
  beta_multiplier = 1,
  min_deviation = 0.001,
  parallel_update_frequency = 10L,
  disable_convergence_test = FALSE,
  trajectory_iterations = as.integer(c())
)
}
\arguments{
\item{fs_ptr}{External pointer to a FeaturedSpace object.}

\item{max_iter}{Maximum number of iterations (default 500).}

\item{convergence}{Convergence threshold on \code{newLoss - oldLoss}
(default \code{1e-5}).  Ignored when \code{disable_convergence_test=TRUE}.}

\item{beta_multiplier}{Regularization multiplier (default 1.0).}

\item{min_deviation}{Minimum sample deviation floor (default 0.001).}

\item{parallel_update_frequency}{Iteration frequency at which
\code{doParallelUpdate} runs (default 10, matching Java).}

\item{disable_convergence_test}{When \code{TRUE}, the loop runs a fixed
\code{max_iter} iterations with no early stop.  Needed for deterministic
per-iteration trajectory comparisons against the Java oracle.}

\item{trajectory_iterations}{Integer vector of 1-based iteration indices
at which to capture \code{(loss, entropy, lambdas)} snapshots.  May be
empty.  Snapshots outside \code{[1, max_iter]} are silently dropped.}
}
\value{
Named list with elements:
  \code{loss}, \code{entropy}, \code{iterations}, \code{converged},
  \code{lambdas}, and \code{trajectory} — a data.frame with columns
  \code{iteration, loss, entropy, lambda_0, ..., lambda_{J-1}} holding
  one row per requested checkpoint that was actually reached.
}
\description{
Runs the full \code{density.Sequential} optimizer ported from the original
Java Maxent, with optional per-iteration trajectory snapshots.  Unlike
\code{maxent_train()} which uses a \code{goodAlpha}-only loop, this trainer
reproduces the real Java optimizer's feature-selection (\code{deltaLossBound}),
Newton step, 1-D line search, and every-10-iter \code{doParallelUpdate}
with undo on loss-violating batch steps.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{get_sample_info}
\alias{get_sample_info}
\title{Get sample properties}
\usage{
get_sample_info(sample_ptr)
}
\arguments{
\item{sample_ptr}{External pointer to Sample}
}
\value{
List with sample properties
}
\description{
Get sample properties
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{grid_from_matrix}
\alias{grid_from_matrix}
\title{Set grid from matrix}
\usage{
grid_from_matrix(grid_ptr, mat)
}
\arguments{
\item{grid_ptr}{External pointer to Grid}

\item{mat}{Numeric matrix of values}
}
\value{
Called for side effects; returns invisibly.
}
\description{
Set grid from matrix
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/featured_space.R
\name{maxent_load_lambdas}
\alias{maxent_load_lambdas}
\title{Load Model Lambdas from File}
\usage{
maxent_load_lambdas(featured_space, file)
}
\arguments{
\item{featured_space}{External pointer to a FeaturedSpace object (must have
features with the same names as those in the file).}

\item{file}{Character: path to the lambdas file.}
}
\value{
Invisibly returns the FeaturedSpace external pointer (updated
  in-place).
}
\description{
Reads model coefficients from a .lambdas file and applies them to
a FeaturedSpace that was created with the same features.
}
\examples{
\donttest{
maxent_load_lambdas(fs, tempfile(fileext = ".lambdas"))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/maxent_report.R
\name{maxent_print_results}
\alias{maxent_print_results}
\title{Print Maxent Model Results}
\usage{
maxent_print_results(
  species,
  eval_result,
  contributions_df,
  perm_imp_df,
  n_presence,
  n_background,
  test_eval_result = NULL,
  n_test = 0L,
  fit_result = NULL
)
}
\arguments{
\item{species}{Character: species name.}

\item{eval_result}{Named list returned by \code{\link{maxent_evaluate}}
(training predictions vs background).}

\item{contributions_df}{Data.frame with columns \code{name} and
\code{contribution} (from \code{\link{maxent_percent_contribution}}).}

\item{perm_imp_df}{Data.frame with columns \code{name} and
\code{permutation_importance} (from
\code{\link{maxent_permutation_importance}}).}

\item{n_presence}{Integer: number of training presence records.}

\item{n_background}{Integer: number of background records.}

\item{test_eval_result}{Named list returned by \code{\link{maxent_evaluate}}
for test data, or \code{NULL} (default).}

\item{n_test}{Integer: number of test presence records
(default \code{0L}).}

\item{fit_result}{Named list returned by \code{\link{maxent_fit}} or
\code{NULL}.  Used to report regularized training gain and entropy.}
}
\value{
Invisibly returns a named list with all reported metrics.
}
\description{
Prints a summary of Maxent model performance metrics to the console,
replicating the style of the \pkg{dismo} package's MaxEnt output.  The
report includes sample counts, training (and optionally test) evaluation
statistics, and a ranked variable-contributions table.
}
\examples{
\donttest{
maxent_print_results(
  species          = "Sp1",
  eval_result      = maxent_evaluate(pres_preds, bg_preds),
  contributions_df = contrib,
  perm_imp_df      = perm_imp,
  n_presence       = length(pres_rows),
  n_background     = length(bg_rows))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_diagnostics.R
\name{maxent_permutation_importance}
\alias{maxent_permutation_importance}
\title{Compute Permutation Importance}
\usage{
maxent_permutation_importance(
  model,
  env_grids,
  feature_names,
  presence_rows,
  presence_cols,
  absence_rows,
  absence_cols,
  seed = 42L
)
}
\arguments{
\item{model}{External pointer to a FeaturedSpace object.}

\item{env_grids}{List of external pointers to Grid<float> objects.}

\item{feature_names}{Character vector of environment variable names,
matching the order of env_grids.}

\item{presence_rows}{Integer vector of presence site row indices (0-based).}

\item{presence_cols}{Integer vector of presence site column indices (0-based).}

\item{absence_rows}{Integer vector of absence/background site row indices.}

\item{absence_cols}{Integer vector of absence/background site column indices.}

\item{seed}{Random seed for reproducibility (default 42).}
}
\value{
A data.frame with columns:
  \describe{
    \item{name}{Variable name}
    \item{permutation_importance}{Normalised importance (\%)}
  }
}
\description{
Measures how much each environmental variable contributes to model
prediction quality by permuting each variable and measuring AUC drop.
Results are normalised so that all importances sum to 100%.
}
\examples{
\donttest{
imp <- maxent_permutation_importance(model, list(g1, g2),
         c("temp", "precip"),
         pres_rows, pres_cols, abs_rows, abs_cols)
imp  # data.frame with name and permutation_importance
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{maxent_featured_space_from_callback}
\alias{maxent_featured_space_from_callback}
\title{Create a FeaturedSpace from a streaming background provider}
\usage{
maxent_featured_space_from_callback(
  num_points,
  num_layers,
  layer_names,
  sample_indices,
  feature_types,
  n_thresholds,
  n_hinges,
  next_tile_fn,
  reset_fn,
  use_cache = TRUE
)
}
\arguments{
\item{num_points}{Integer: total number of finite background points
the provider will emit (sum of \code{nrow()} across tiles).}

\item{num_layers}{Integer: number of environmental variables / raster
layers per tile row.}

\item{layer_names}{Character vector of length \code{num_layers}
giving the name for each layer / column (used when generating feature
names such as \code{bio1^2}, \code{bio1*bio2}).}

\item{sample_indices}{Integer vector: 0-based indices of occurrence
samples in the concatenated background stream.}

\item{feature_types}{Character vector of feature types to generate.
See \code{\link{maxent_generate_features}}.}

\item{n_thresholds}{Number of threshold knots per variable.}

\item{n_hinges}{Number of hinge knots per variable.}

\item{next_tile_fn}{R function returning the next tile as a numeric
matrix (nrow x num_layers), or a 0-row matrix / \code{NULL} when the
stream is exhausted. The function is responsible for filtering out
rows containing NAs.}

\item{reset_fn}{R function (no arguments) that rewinds the
underlying stream to the beginning.}

\item{use_cache}{Logical; wrap the callback stream in
\code{CachingBackgroundProvider}.}
}
\value{
External pointer to a \code{FeaturedSpace} object.
}
\description{
Builds a \code{FeaturedSpace} by draining a callback-style background
provider (typically backed by a \code{terra::SpatRaster} block loop)
one tile at a time, then constructing features from the concatenated
\code{(num_points x num_layers)} matrix via \code{\link{generate_features}}.
}
\details{
This is the C++ entry point used by
\code{\link{maxent_featured_space_from_rast}}; end users should prefer
that R-level wrapper.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{compute_mess_range}
\alias{compute_mess_range}
\title{Compute MESS from Min/Max Ranges}
\usage{
compute_mess_range(grid_ptrs, var_mins, var_maxs)
}
\arguments{
\item{grid_ptrs}{List of external pointers to Grid<float> objects.}

\item{var_mins}{Numeric vector of minimum reference values.}

\item{var_maxs}{Numeric vector of maximum reference values.}
}
\value{
A named list with mess_grid and mod_grid (external pointers).
}
\description{
Simplified MESS using only the min/max of reference data.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/features.R
\name{maxent_hinge_feature}
\alias{maxent_hinge_feature}
\title{Create a Hinge Feature}
\usage{
maxent_hinge_feature(values, name, min_knot, max_knot, reverse = FALSE)
}
\arguments{
\item{values}{Numeric vector of environmental variable values.}

\item{name}{Character string: feature name/identifier.}

\item{min_knot}{Lower knot of the hinge.}

\item{max_knot}{Upper knot of the hinge (must be strictly greater than
\code{min_knot}).}

\item{reverse}{Logical; if \code{TRUE}, use reverse hinge. Default is
\code{FALSE} (forward hinge).}
}
\value{
External pointer to a HingeFeature C++ object.
}
\description{
Creates a piecewise-linear hinge feature.
}
\details{
\strong{Forward hinge} (\code{reverse = FALSE}):
\code{eval(i) = max(0, (values[i] - min_knot) / (max_knot - min_knot))}

\strong{Reverse hinge} (\code{reverse = TRUE}):
\code{eval(i) = max(0, (max_knot - values[i]) / (max_knot - min_knot))}
}
\examples{
vals <- c(0, 5, 10, 3)
f <- maxent_hinge_feature(vals, "temperature_hinge", min_knot = 2, max_knot = 8)
maxent_feature_eval(f, 2)  # values[2] = 5 -> (5-2)/(8-2) = 0.5
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{csv_next_record}
\alias{csv_next_record}
\title{Read the next CSV record}
\usage{
csv_next_record(reader_ptr)
}
\arguments{
\item{reader_ptr}{External pointer to a CsvReader object.}
}
\value{
Character vector of field values, or NULL on EOF.
}
\description{
Read the next CSV record
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{feature_eval}
\alias{feature_eval}
\title{Evaluate a feature at a given index}
\usage{
feature_eval(feature_ptr, index)
}
\arguments{
\item{feature_ptr}{External pointer to a Feature object}

\item{index}{0-based index into the data vector}
}
\value{
Feature value at that index (double)
}
\description{
Evaluate a feature at a given index
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grid_io.R
\name{maxent_csv_next}
\alias{maxent_csv_next}
\title{Read the Next CSV Record}
\usage{
maxent_csv_next(reader)
}
\arguments{
\item{reader}{External pointer to a CsvReader object.}
}
\value{
Character vector of field values, or \code{NULL} on EOF.
}
\description{
Read the Next CSV Record
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{maxent_get_entropy}
\alias{maxent_get_entropy}
\title{Get entropy of a FeaturedSpace distribution}
\usage{
maxent_get_entropy(fs_ptr)
}
\arguments{
\item{fs_ptr}{External pointer to a FeaturedSpace object.}
}
\value{
Shannon entropy (non-negative scalar).
}
\description{
Get entropy of a FeaturedSpace distribution
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grid_io.R
\name{maxent_csv_write}
\alias{maxent_csv_write}
\title{Write a String Value to the Current CSV Row}
\usage{
maxent_csv_write(writer, column, value)
}
\arguments{
\item{writer}{External pointer to a CsvWriter object.}

\item{column}{Character: column name.}

\item{value}{Character: value to write.}
}
\value{
Invisibly returns the writer object.
}
\description{
Adds a \code{column = value} pair (as a character string) to the current
row buffer. Call \code{\link{maxent_csv_write_row}} to flush the row.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_evaluation.R
\name{maxent_evaluate}
\alias{maxent_evaluate}
\title{Full Model Evaluation}
\usage{
maxent_evaluate(presence, absence)
}
\arguments{
\item{presence}{Numeric vector of prediction scores at presence sites.}

\item{absence}{Numeric vector of prediction scores at absence sites.}
}
\value{
A named list with:
  \describe{
    \item{auc}{AUC value}
    \item{max_kappa}{Best Cohen's Kappa}
    \item{max_kappa_thresh}{Threshold at best Kappa}
    \item{correlation}{Pearson correlation with labels}
    \item{square_error}{Mean squared error}
    \item{logloss}{Cross-entropy log-loss}
    \item{misclassification}{Misclassification rate}
    \item{prevalence}{Fraction of presence sites}
  }
}
\description{
Computes all evaluation metrics at once: AUC, correlation, log-loss,
squared error, misclassification, max Kappa, and prevalence.
}
\examples{
res <- maxent_evaluate(c(0.9, 0.85, 0.95), c(0.1, 0.15, 0.2))
res$auc          # 1.0
res$correlation  # > 0
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{create_threshold_feature}
\alias{create_threshold_feature}
\title{Create a ThresholdFeature object}
\usage{
create_threshold_feature(values, name, threshold)
}
\arguments{
\item{values}{Numeric vector of environmental variable values}

\item{name}{Feature name/identifier}

\item{threshold}{The threshold value}
}
\value{
External pointer to ThresholdFeature object
}
\description{
Creates a binary step feature: eval(i) = 1.0 if values[i] > threshold, else 0.0.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/maxent.R
\name{as_matrix}
\alias{as_matrix}
\title{Convert Grid to Raster Matrix}
\usage{
as_matrix(grid)
}
\arguments{
\item{grid}{Grid object}
}
\value{
Numeric matrix
}
\description{
Extracts grid data as an R matrix.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/featured_space_raster.R
\name{maxent_extract_occurrence_env_terra}
\alias{maxent_extract_occurrence_env_terra}
\title{Extract environmental values for occurrences from a terra SpatRaster}
\usage{
maxent_extract_occurrence_env_terra(
  rast,
  occurrences,
  lon_col = "longitude",
  lat_col = "latitude"
)
}
\arguments{
\item{rast}{A \code{terra::SpatRaster}.}

\item{occurrences}{Either a two-column matrix/data.frame (x,y) or a vector
of 1-based raster cell indices.}

\item{lon_col}{Longitude column name when \code{occurrences} is a data.frame.}

\item{lat_col}{Latitude column name when \code{occurrences} is a data.frame.}
}
\value{
Numeric matrix with one row per retained occurrence and one column
  per raster layer.
}
\description{
Streams finite raster rows through the same callback-backed provider path
used by \code{\link{maxent_featured_space_from_rast}}, then returns values
at occurrence locations.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_evaluation.R
\name{maxent_project_cloglog}
\alias{maxent_project_cloglog}
\title{Project Model onto Grids (Cloglog Output)}
\usage{
maxent_project_cloglog(model, env_grids, feature_names)
}
\arguments{
\item{model}{External pointer to a FeaturedSpace object.}

\item{env_grids}{List of external pointers to Grid<float> objects.}

\item{feature_names}{Character vector of environment variable names.}
}
\value{
External pointer to a Grid<float> with cloglog scores in [0, 1].
}
\description{
Applies the Java Maxent cloglog transform to grid predictions:
\deqn{cloglog = 1 - exp(-exp(H) \cdot raw)}
}
\details{
where \eqn{H} is the model entropy and \eqn{raw} is the
normalised probability.  This matches the cloglog output of Java
Maxent and \pkg{dismo}, and is the recommended output for reporting.
}
\seealso{
\code{\link{maxent_project_raw}},
  \code{\link{maxent_project_logistic}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{sample_set_feature}
\alias{sample_set_feature}
\title{Set feature value on a sample}
\usage{
sample_set_feature(sample_ptr, feature_name, value)
}
\arguments{
\item{sample_ptr}{External pointer to Sample}

\item{feature_name}{Name of the feature}

\item{value}{Feature value}
}
\value{
Called for side effects; returns invisibly.
}
\description{
Set feature value on a sample
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grid_io.R
\name{maxent_grid_to_terra}
\alias{maxent_grid_to_terra}
\title{Convert a maxentcpp GridFloat to a terra SpatRaster}
\usage{
maxent_grid_to_terra(grid, crs = "EPSG:4326")
}
\arguments{
\item{grid}{External pointer to a GridFloat C++ object.}

\item{crs}{Character: coordinate reference system string
(default \code{"EPSG:4326"}, i.e. WGS 84 longitude/latitude).}
}
\value{
A single-layer \code{terra::SpatRaster}.
}
\description{
Creates a \code{terra::SpatRaster} from a maxentcpp GridFloat external
pointer.
}
\details{
Requires the \pkg{terra} package (listed in Suggests).

For the \pkg{raster} package, an equivalent workflow is:
\preformatted{
  library(raster)
  info <- maxent_grid_info(grid)
  mat  <- maxent_grid_to_matrix(grid)
  r <- raster(mat,
              xmn = info$xll,
              xmx = info$xll + info$ncols * info$cellsize,
              ymn = info$yll,
              ymx = info$yll + info$nrows * info$cellsize,
              crs = "+proj=longlat +datum=WGS84")
}
}
\examples{
\donttest{
stack_path <- system.file("extdata", "stack_1_12_crop.rds",
                         package = "maxentcpp")
r <- terra::unwrap(readRDS(stack_path))
g <- maxent_grid_from_terra(r[[1]])
r2 <- maxent_grid_to_terra(g)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{csv_headers}
\alias{csv_headers}
\title{Get CSV column headers}
\usage{
csv_headers(reader_ptr)
}
\arguments{
\item{reader_ptr}{External pointer to a CsvReader object.}
}
\value{
Character vector of header names.
}
\description{
Get CSV column headers
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/maxent.R
\name{maxent_background_indices}
\alias{maxent_background_indices}
\title{Generate Background Sample Indices}
\usage{
maxent_background_indices(grid, n = 10000L, seed = NULL)
}
\arguments{
\item{grid}{External pointer to a GridFloat object used as a reference
(e.g. an environmental layer). Only cells with valid data are eligible.}

\item{n}{Integer: number of background points to sample
(default \code{10000L}).}

\item{seed}{Integer or \code{NULL}: random seed for reproducibility.
If \code{NULL} (default), no seed is set.}
}
\value{
A named list with:
  \describe{
    \item{rows}{Integer vector of row indices (0-based).}
    \item{cols}{Integer vector of column indices (0-based).}
    \item{indices}{Integer vector of 0-based flat indices
      (\code{row * ncols + col}), suitable for
      \code{\link{maxent_featured_space}}.}
  }
}
\description{
Randomly selects valid (non-NODATA) cells from a reference grid for use
as background points in MaxEnt modeling.
}
\examples{
\donttest{
dim <- maxent_dimension(100, 100, -120, 35, 0.1)
grid <- maxent_grid(dim, "env1")
bg <- maxent_background_indices(grid, n = 5000, seed = 42)
bg$indices  # 0-based flat indices for FeaturedSpace
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{eval_correlation}
\alias{eval_correlation}
\title{Compute Pearson Correlation}
\usage{
eval_correlation(x, y)
}
\arguments{
\item{x}{Numeric vector.}

\item{y}{Numeric vector (same length as x).}
}
\value{
Correlation coefficient in [-1, 1].
}
\description{
Computes Pearson correlation coefficient between two numeric vectors.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_diagnostics.R
\name{maxent_mess_range}
\alias{maxent_mess_range}
\title{Compute MESS from Min/Max Ranges}
\usage{
maxent_mess_range(env_grids, var_mins, var_maxs)
}
\arguments{
\item{env_grids}{List of external pointers to Grid<float> objects.}

\item{var_mins}{Numeric vector of minimum reference values per variable.}

\item{var_maxs}{Numeric vector of maximum reference values per variable.}
}
\value{
A named list with mess_grid and mod_grid (external pointers).
}
\description{
Simplified MESS analysis using only the min/max of the reference data
rather than the full distribution.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{maxent_read_lambdas}
\alias{maxent_read_lambdas}
\title{Read feature lambdas from a file}
\usage{
maxent_read_lambdas(fs_ptr, filename)
}
\arguments{
\item{fs_ptr}{External pointer to a FeaturedSpace object.}

\item{filename}{Character: path to the lambdas file.}
}
\value{
Called for side effects; returns invisibly.
}
\description{
Restores model coefficients from a .lambdas file.  The FeaturedSpace
must have been created with the same features (same names and order).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{grid_float_to_matrix}
\alias{grid_float_to_matrix}
\title{Extract the data from a GridFloat as an R matrix}
\usage{
grid_float_to_matrix(grid_ptr)
}
\arguments{
\item{grid_ptr}{External pointer to a GridFloat object.}
}
\value{
Numeric matrix (nrows × ncols). NODATA cells are set to NA.
}
\description{
Extract the data from a GridFloat as an R matrix
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{create_linear_feature}
\alias{create_linear_feature}
\title{Create a LinearFeature object}
\usage{
create_linear_feature(values, name, min_val, max_val)
}
\arguments{
\item{values}{Numeric vector of environmental variable values}

\item{name}{Feature name/identifier}

\item{min_val}{Minimum value for normalization}

\item{max_val}{Maximum value for normalization}
}
\value{
External pointer to LinearFeature object
}
\description{
Creates a linear (normalized) feature: eval(i) = (values[i] - min) / (max - min).
Returns 0 when min == max.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/featured_space.R
\name{maxent_model_entropy}
\alias{maxent_model_entropy}
\title{Get Model Entropy}
\usage{
maxent_model_entropy(featured_space)
}
\arguments{
\item{featured_space}{External pointer to a FeaturedSpace object.}
}
\value{
Scalar: Shannon entropy (non-negative).
}
\description{
Returns the Shannon entropy of the current MaxEnt distribution.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/features.R
\name{maxent_quadratic_feature}
\alias{maxent_quadratic_feature}
\title{Create a Quadratic Feature}
\usage{
maxent_quadratic_feature(values, name, min_val = NULL, max_val = NULL)
}
\arguments{
\item{values}{Numeric vector of environmental variable values.}

\item{name}{Character string: feature name/identifier.}

\item{min_val}{Minimum value for normalization. If \code{NULL} (default),
computed as \code{min(values)}.}

\item{max_val}{Maximum value for normalization. If \code{NULL} (default),
computed as \code{max(values)}.}
}
\value{
External pointer to a QuadraticFeature C++ object.
}
\description{
Creates a quadratic feature that evaluates to \code{linear_val^2} where
\code{linear_val = (values[i] - min) / (max - min)}.
}
\examples{
vals <- c(0, 5, 10)
f <- maxent_quadratic_feature(vals, "temperature_sq")
maxent_feature_eval(f, 2)  # evaluates at index 2 (1-based -> 0-based: 1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_evaluation.R
\name{maxent_extract_predictions_raw}
\alias{maxent_extract_predictions_raw}
\title{Extract Raw Predictions at Sample Locations}
\usage{
maxent_extract_predictions_raw(model, env_grids, feature_names, rows, cols)
}
\arguments{
\item{model}{External pointer to a FeaturedSpace object.}

\item{env_grids}{List of external pointers to Grid<float> objects.}

\item{feature_names}{Character vector of environment variable names.}

\item{rows}{Integer vector of row indices.}

\item{cols}{Integer vector of column indices.}
}
\value{
Numeric vector of Java raw scores. NaN for NODATA cells.
}
\description{
Gets Java Maxent raw scores at specific grid cell locations:
\deqn{raw = \frac{exp(lp - lpNorm)}{densityNorm}}
}
\details{
This matches the raw output of Java Maxent and \pkg{dismo}.
}
\seealso{
\code{\link{maxent_project_raw}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{create_product_feature}
\alias{create_product_feature}
\title{Create a ProductFeature object}
\usage{
create_product_feature(values1, values2, name, min1, max1, min2, max2)
}
\arguments{
\item{values1}{Numeric vector for the first environmental variable}

\item{values2}{Numeric vector for the second environmental variable}

\item{name}{Feature name/identifier}

\item{min1}{Minimum of the first variable}

\item{max1}{Maximum of the first variable}

\item{min2}{Minimum of the second variable}

\item{max2}{Maximum of the second variable}
}
\value{
External pointer to ProductFeature object
}
\description{
Creates a product (interaction) feature between two environmental variables:
eval(i) = norm(values1[i]) * norm(values2[i]).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_evaluation.R
\name{maxent_extract_predictions_cloglog}
\alias{maxent_extract_predictions_cloglog}
\title{Extract Cloglog Predictions at Sample Locations}
\usage{
maxent_extract_predictions_cloglog(model, env_grids, feature_names, rows, cols)
}
\arguments{
\item{model}{External pointer to a FeaturedSpace object.}

\item{env_grids}{List of external pointers to Grid<float> objects.}

\item{feature_names}{Character vector of environment variable names.}

\item{rows}{Integer vector of row indices.}

\item{cols}{Integer vector of column indices.}
}
\value{
Numeric vector of cloglog scores in [0, 1]. NaN for NODATA cells.
}
\description{
Gets Java Maxent cloglog scores at specific grid cell locations:
\deqn{cloglog = 1 - exp(-exp(H) \cdot raw)}
}
\seealso{
\code{\link{maxent_extract_predictions_raw}},
  \code{\link{maxent_project_cloglog}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/maxent_run.R
\name{maxent_run}
\alias{maxent_run}
\title{Run a Complete Maxent Species Distribution Modelling Workflow}
\usage{
maxent_run(
  species,
  env_grids,
  occ_df,
  output_dir,
  lon_col = "long",
  lat_col = "lat",
  n_background = 10000L,
  types = c("linear", "quadratic", "hinge"),
  n_hinges = 15L,
  max_iter = 500L,
  seed = 42L,
  bias_weights = NULL,
  response_curves = TRUE,
  pictures = TRUE
)
}
\arguments{
\item{species}{Character: species name (used in file names and the
console report).}

\item{env_grids}{Named list of external pointers to Grid<float> objects.
The \emph{names} of the list are used as the environmental variable names.}

\item{occ_df}{A \code{data.frame} (or CSV path) with occurrence
records.}

\item{output_dir}{Character: directory where all output files are written
(created if it does not exist).}

\item{lon_col}{Character: longitude column name in \code{occ_df}
(default \code{"long"}).}

\item{lat_col}{Character: latitude column name in \code{occ_df}
(default \code{"lat"}).}

\item{n_background}{Integer: number of background points (default 10000).}

\item{types}{Character vector of feature types passed to
\code{\link{maxent_generate_features}} (default
\code{c("linear", "quadratic", "hinge")}).}

\item{n_hinges}{Integer: number of hinge knots (default 15).}

\item{max_iter}{Integer: maximum training iterations (default 500).}

\item{seed}{Integer: random seed for background sampling
(default 42).}

\item{bias_weights}{Optional numeric vector of per-background-point bias
weights.  Must have length equal to \code{n_background} (the background
points only, not including occurrence points).  When supplied, the
background density is weighted by \code{bias[i] * exp(lp[i] - lpn)},
mirroring Java Maxent's \code{biasFile}.  Pass \code{NULL} (default)
for uniform (unbiased) background.}

\item{response_curves}{Logical: write response curve PNGs
(default \code{TRUE}).}

\item{pictures}{Logical: write prediction map PNG (default \code{TRUE}).}
}
\value{
A named list with:
  \describe{
    \item{model}{Trained FeaturedSpace external pointer.}
    \item{fit_result}{List returned by \code{\link{maxent_fit}}.}
    \item{evaluation}{List returned by \code{\link{maxent_evaluate}}.}
    \item{contributions}{Data.frame of percent contributions.}
    \item{permutation_importance}{Data.frame of permutation importances.}
    \item{output_dir}{The output directory used.}
  }
}
\description{
Provides a single high-level entry point that mirrors running the Java
Maxent GUI with one click.  Starting from raw environmental grids and
occurrence records, the function:
\enumerate{
  \item Reads and validates occurrence records.
  \item Samples background points.
  \item Extracts environmental values and builds features.
  \item Trains the Maxent model.
  \item Projects the model onto the full landscape (cloglog output).
  \item Evaluates model performance (AUC).
  \item Computes percent contribution and permutation importance.
  \item Writes all standard output files (lambdas, prediction PNG, response
    curve PNGs, omission CSV, sample-predictions CSV, and
    \code{maxentResults.csv}).
  \item Prints a performance summary to the console.
}
}
\examples{
\donttest{
stack_path      <- system.file("extdata", "stack_1_12_crop.rds",
                               package = "maxentcpp")
example_rasters <- terra::unwrap(readRDS(stack_path))
grids <- list(
  bio1  = maxent_grid_from_terra(example_rasters[[1]]),
  bio12 = maxent_grid_from_terra(example_rasters[[2]])
)
data(example_occ_df)

result <- maxent_run(
  species    = "Abeillia_abeillei",
  env_grids  = grids,
  occ_df     = example_occ_df,
  output_dir = tempdir(),
  lon_col    = "long",
  lat_col    = "lat")

result$evaluation$auc
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/maxent.R
\name{maxent_read_occurrences}
\alias{maxent_read_occurrences}
\title{Read Species Occurrence Data}
\usage{
maxent_read_occurrences(
  file_or_df,
  dim,
  lon_col = "longitude",
  lat_col = "latitude",
  name_col = NULL
)
}
\arguments{
\item{file_or_df}{Either a character path to a CSV file, or a
\code{data.frame} already loaded in R (e.g. from
\code{rgbif::occ_data()}, GBIF download, or \code{read.csv()}).}

\item{dim}{A GridDimension object (from \code{\link{maxent_dimension}})
defining the study area grid.}

\item{lon_col}{Character: name of the longitude column
(default \code{"longitude"}).}

\item{lat_col}{Character: name of the latitude column
(default \code{"latitude"}).}

\item{name_col}{Character or \code{NULL}: column for sample names.
If \code{NULL} (default), sequential names are generated.}
}
\value{
A named list with:
  \describe{
    \item{samples}{List of Sample external pointers.}
    \item{rows}{Integer vector of row indices (0-based).}
    \item{cols}{Integer vector of column indices (0-based).}
    \item{indices}{Integer vector of 0-based flat indices
      (\code{row * ncols + col}), suitable for
      \code{\link{maxent_featured_space}}.}
  }
}
\description{
Reads species presence records from a CSV file or an R \code{data.frame}
and converts them to row/column indices suitable for
\code{\link{maxent_featured_space}}.
}
\examples{
dim <- maxent_dimension(100, 100, -120, 35, 0.1)
occ <- data.frame(longitude = c(-118.5, -119.0),
                  latitude  = c(36.5, 37.0))
result <- maxent_read_occurrences(occ, dim)
result$indices  # 0-based flat indices for FeaturedSpace
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/featured_space_raster.R
\name{maxent_featured_space_from_rast}
\alias{maxent_featured_space_from_rast}
\title{Create a FeaturedSpace directly from a terra SpatRaster}
\usage{
maxent_featured_space_from_rast(
  rast,
  sample_indices,
  feature_types = c("linear", "quadratic", "product", "threshold", "hinge"),
  n_thresholds = 10L,
  n_hinges = 10L,
  enable_streaming_eval = TRUE
)
}
\arguments{
\item{rast}{A \code{terra::SpatRaster} with one layer per environmental
variable. Layers must be named; names are passed through to
\code{\link{maxent_generate_features}} so generated features carry the
same identifiers as the layer names (e.g. \code{bio1^2}, \code{bio1*bio2}).}

\item{sample_indices}{Integer vector: 0-based indices of occurrence
samples in the concatenated stream of finite background cells. See
\code{\link{maxent_raster_sample_indices}} for a helper that converts
occurrence \code{data.frame}s into these indices.}

\item{feature_types}{Character vector of feature types to generate.
Any subset of \code{c("linear", "quadratic", "product", "threshold",
"hinge")}. Defaults to all types.}

\item{n_thresholds}{Integer; number of threshold knots per variable.}

\item{n_hinges}{Integer; number of hinge knots per variable.}

\item{enable_streaming_eval}{Logical; when \code{TRUE}, enables
streaming-evaluation mode after object construction.}
}
\value{
External pointer to a \code{FeaturedSpace} object.
}
\description{
Streams a \code{terra::SpatRaster} block-by-block into the C++
\code{FeaturedSpace} streaming constructor, filtering NA cells on the fly
and handing the concatenated (\code{num_points} x \code{nlyr(rast)})
matrix to \code{\link{maxent_generate_features}} so feature objects are
built inside C++ without the R caller having to materialise the full
raster stack first.
}
\details{
This is the Phase E.2 entry point described in
\code{docs/ARCHITECTURE_terra_raster.md} (sections 3.1 and 3.2). The
result is bit-for-bit identical to the dense path
(\code{\link{maxent_generate_features}} -> \code{\link{maxent_featured_space}})
when the same data, sample indices, and feature configuration are used.
}
\examples{
\donttest{
stack_path <- system.file("extdata", "stack_1_12_crop.rds",
                         package = "maxentcpp")
r <- terra::unwrap(readRDS(stack_path))
fs <- maxent_featured_space_from_rast(
    r,
    sample_indices = c(0L, 1L, 2L),
    feature_types  = c("linear", "quadratic")
)
}
}
\seealso{
\code{\link{maxent_featured_space}},
  \code{\link{maxent_featured_space_from_callback}},
  \code{\link{maxent_generate_features}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{compute_response_curve_fixed}
\alias{compute_response_curve_fixed}
\title{Compute Java-compatible Response Curve with Fixed Values}
\usage{
compute_response_curve_fixed(
  fs_ptr,
  fixed_values,
  feature_names,
  var_index,
  var_min,
  var_max,
  n_steps = 100L
)
}
\arguments{
\item{fs_ptr}{External pointer to a FeaturedSpace object.}

\item{fixed_values}{Numeric vector of fixed values for each variable.}

\item{feature_names}{Character vector of environment variable names.}

\item{var_index}{0-based index of the variable to vary.}

\item{var_min}{Minimum value of the target variable.}

\item{var_max}{Maximum value of the target variable.}

\item{n_steps}{Number of steps.}
}
\value{
A data.frame with columns: value, prediction.
}
\description{
Like compute_response_curve_fixed() but applies the Java Maxent cloglog:
  cloglog_java = 1 - exp(-exp(H) * raw_java)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_evaluation.R
\name{maxent_extract_predictions_logistic}
\alias{maxent_extract_predictions_logistic}
\title{Extract Logistic Predictions at Sample Locations}
\usage{
maxent_extract_predictions_logistic(
  model,
  env_grids,
  feature_names,
  rows,
  cols
)
}
\arguments{
\item{model}{External pointer to a FeaturedSpace object.}

\item{env_grids}{List of external pointers to Grid<float> objects.}

\item{feature_names}{Character vector of environment variable names.}

\item{rows}{Integer vector of row indices.}

\item{cols}{Integer vector of column indices.}
}
\value{
Numeric vector of logistic scores in [0, 1]. NaN for NODATA cells.
}
\description{
Gets Java Maxent logistic scores at specific grid cell locations:
\deqn{logistic = \frac{exp(H) \cdot raw}{1 + exp(H) \cdot raw}}
}
\seealso{
\code{\link{maxent_extract_predictions_raw}},
  \code{\link{maxent_project_logistic}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/features.R
\name{maxent_feature_eval}
\alias{maxent_feature_eval}
\title{Evaluate a Feature at a Given Index}
\usage{
maxent_feature_eval(feature, index)
}
\arguments{
\item{feature}{External pointer to a Feature C++ object.}

\item{index}{1-based integer index into the data vector.}
}
\value{
Numeric scalar: the feature value at that index.
}
\description{
Evaluates the feature function at the specified index (1-based, as is
standard in R). Internally converts to 0-based indexing for C++.
}
\examples{
vals <- c(0, 5, 10)
f <- maxent_linear_feature(vals, "temp")
maxent_feature_eval(f, 2)  # 5/10 = 0.5
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{eval_logloss}
\alias{eval_logloss}
\title{Compute Log-Loss}
\usage{
eval_logloss(presence, absence)
}
\arguments{
\item{presence}{Numeric vector of prediction scores at presence sites.}

\item{absence}{Numeric vector of prediction scores at absence sites.}
}
\value{
Average log-loss value.
}
\description{
Computes average cross-entropy log-loss from predictions at
presence and absence sites.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/featured_space.R
\name{maxent_model_weights}
\alias{maxent_model_weights}
\title{Get Model Distribution Weights}
\usage{
maxent_model_weights(featured_space)
}
\arguments{
\item{featured_space}{External pointer to a FeaturedSpace object.}
}
\value{
Numeric vector of weights that sum to 1.
}
\description{
Returns the normalized probability weights of the current MaxEnt
distribution over the background points.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{maxent_extract_occurrence_from_callback}
\alias{maxent_extract_occurrence_from_callback}
\title{Extract occurrence environmental values from a callback stream}
\usage{
maxent_extract_occurrence_from_callback(
  num_points,
  num_layers,
  occurrence_indices,
  next_tile_fn,
  reset_fn,
  preserved_rast = NULL
)
}
\arguments{
\item{num_points}{Integer; total finite background rows in stream.}

\item{num_layers}{Integer; number of environmental variables per row.}

\item{occurrence_indices}{Integer vector of 0-based finite-stream indices.}

\item{next_tile_fn}{R function returning the next tile or NULL/0-row matrix.}

\item{reset_fn}{R function (no args) that rewinds the underlying stream.}

\item{preserved_rast}{Optional SpatRaster object to preserve during read.}
}
\value{
Numeric matrix with one row per requested occurrence index.
}
\description{
Drains a callback background stream once and returns the rows at the
requested 0-based finite-stream indices.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/maxent.R
\name{grid_info}
\alias{grid_info}
\title{Get Grid Information}
\usage{
grid_info(grid)
}
\arguments{
\item{grid}{Grid object}
}
\value{
List with grid properties
}
\description{
Retrieves properties of a grid object.
}
\keyword{internal}
